unit AnimationLib;

interface

uses Windows,Types,Classes,SysUtils,Graphics,BitmapLib,GIFImage,CursorLib;



type TAnimationType = (atUnknown,atStaticCursor,atAnimatedCursor);


type TGIFProgress = procedure(Percent: integer);

type TGIFSaveOptions = record
 Delay: integer;
 OptimizeColor,Transparent: boolean;
 OptimizeOptions: TGIFOptimizeOptions;
 end;


type TAnimation = class(TBitmapList)
 private
  FAnimationType: TAnimationType;
  function AddBitmapToGIF(GIF: TGIFImage; Source: TGraphic; DelayMS: integer; Transparent: boolean): integer;
  function TransparentIndex(GIF: TGIFSubImage): byte;
 public
  Title,Author: string;
  constructor Create;
  property AnimationType: TAnimationType read FAnimationType;
  procedure LoadAnimatedCursor(Stream: TStream);  overload;
  procedure LoadAnimatedCursor(Filename: string);  overload;
  procedure LoadStaticCursor(Filename: string);
  procedure LoadCursor(Filename: string);
  procedure SaveAsGIF(Filename: string; SaveOptions: TGIFSaveOptions; Progress: TGIFProgress);
 end;


implementation


constructor TAnimation.Create;
begin
inherited;
FAnimationType:=atUnknown;
end;



procedure TAnimation.LoadAnimatedCursor(Stream: TStream);
var AniCursor: TAniCursor;
var I,R,Inx,Count: integer;
var HasSequence: boolean;
var BM: TBitmap;
begin
AniCursor:=TAniCursor.Create;
try
Clear;
AniCursor.LoadFromStream(Stream);
HasSequence:=AniCursor.SeqCount>0;
if HasSequence then Count:=AniCursor.SeqCount else Count:=AniCursor.FrameCount;
for I:=0 to Count-1 do
	begin
	if HasSequence then Inx:=AniCursor.Sequences[I] else Inx:=I;
	R:=AniCursor.Rates[Inx];
	BM:=IconToBitmap(TIcon(AniCursor.Frames[Inx].AsIcon));
	try
	AddBitmap(BM);
	finally BM.Free; end;
	AuxInfo[I]:=R;
	end;
FAnimationType:=atAnimatedCursor;
Title:=AniCursor.Title;
Author:=AniCursor.Author;
finally AniCursor.Free; end;
end;



procedure TAnimation.LoadStaticCursor(Filename: string);
var BM: TBitmap;
begin
Clear;
BM:=LoadCursorToBitmap(FileName);
try
AddBitmap(BM);
FAnimationType:=atStaticCursor;
Title:='';
Author:='';
finally BM.Free; end;
end;




procedure TAnimation.LoadAnimatedCursor(Filename: string);
var Stream: TStream;
begin
Stream:= TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
try
LoadAnimatedCursor(Stream);
finally Stream.Free; end;
end;




procedure TAnimation.LoadCursor(Filename: string);
var Ext: string;
begin
Ext:=ExtractFileExt(Filename);
if LowerCase(Ext)='.cur' then LoadStaticCursor(Filename)
else LoadAnimatedCursor(Filename);
end;





{---------------------------------------------------------------------}



type TProgressHandler = class(TObject)
 private
 StageCnt: integer;
  StageStep: double;
  FOnProgress: TGIFProgress;
 public
  procedure HandleProgress(Sender: TObject; Stage: TProgressStage; PercentDone: Byte;  RedrawNow: Boolean; const R: TRect; const Msg: string);
  procedure InitializeProgress(Stages: integer);
  property OnProgress: TGIFProgress read FOnProgress write FOnProgress;
 end;


procedure TProgressHandler.HandleProgress(Sender: TObject; Stage: TProgressStage; PercentDone: Byte;  RedrawNow: Boolean; const R: TRect; const Msg: string);
var Progress: integer;
begin
if Stage=psStarting then Inc(StageCnt);
Progress:=trunc((StageStep * (StageCnt-1)) + (StageStep * (PercentDone/100)));
if Assigned(FOnProgress) then FOnProgress(Progress);
end;


procedure TProgressHandler.InitializeProgress(Stages: integer);
begin
StageCnt:=0;
StageStep:=100 / (Stages+1);
end;


{---------------------------------------------------------------------}



function TAnimation.TransparentIndex(GIF: TGIFSubImage): byte;
begin
Result := GIF.Pixels[0, GIF.Height-1];
end;



function TAnimation.AddBitmapToGIF(GIF: TGIFImage; Source: TGraphic; DelayMS: integer; Transparent: boolean): integer;
{Add one images to animated GIF}
var Ext: TGIFGraphicControlExtension;
var LoopExt: TGIFAppExtNSLoop;
begin
Result := GIF.Add(Source);
{ Netscape Loop extension must be the first extension in the first frame! }
if Result = 0 then
	begin
	LoopExt := TGIFAppExtNSLoop.Create(GIF.Images[Result]);
	LoopExt.Loops := 0;
	GIF.Images[Result].Extensions.Add(LoopExt);
	end;
Ext := TGIFGraphicControlExtension.Create(GIF.Images[Result]);
{ Convert dealy to 1/100's of second }
Ext.Delay := DelayMS div 10;
if Transparent then
	begin
	Ext.Transparent:= True;
	Ext.TransparentColorIndex:= TransparentIndex(GIF.Images[Result]);
	Ext.Disposal:=dmBackground;
	end;
GIF.Images[Result].Extensions.Add(Ext);
end;



procedure TAnimation.SaveAsGIF(Filename: string; SaveOptions: TGIFSaveOptions; Progress: TGIFProgress);
var GIF	: TGIFImage;
var ProgressHandler: TProgressHandler;
var I: integer;
begin
GIFImageDefaultColorReduction:=rmQuantize;
ProgressHandler:=TProgressHandler.Create;
GIF := TGIFImage.Create;
try
ProgressHandler.InitializeProgress(Count);
ProgressHandler.OnProgress:=Progress;
GIF.OnProgress:=ProgressHandler.HandleProgress;
for I := 0 to Count-1 do
	begin
	AddBitmapToGIF(GIF, Bitmaps[I],AuxInfo[I]*17,SaveOptions.Transparent);
	end;
if SaveOptions.OptimizeColor then GIF.OptimizeColorMap;

if SaveOptions.OptimizeOptions <> [] then GIF.Optimize(SaveOptions.OptimizeOptions, rmQuantize, dmNearest, 0);
GIF.SaveToFile(Filename);
finally
 ProgressHandler.Free;
 GIF.Free;
 end;
end;




end.
