unit D6Fix;


interface

uses Classes;


implementation

uses HelpIntfs, WinHelpViewer, SysUtils, Windows, Forms, Controls;



var ViewerName : String = 'D6HelpHooker';


{======= This section handles hooking the Delphi Help Manager ===============}



{ implements the interfaces supported by WinHelp ---
  ICustomHelpViewer (required of all Help Viewers),
  IExtendedHelpViewer (Topic and Context),
  ISpecialWinHelpViewer (Winhelp-specific messages),
  IHelpSelector interface to handle Keyword and Table Of Contents requests
  }

type THTMLHelpViewer = class(TInterfacedObject, ICustomHelpViewer, IExtendedHelpViewer{, ISpecialWinHelpViewer})
 private
  FViewerID: Integer;
 public
  FHelpManager: IHelpManager;
  constructor Create;
  destructor Destroy; override;
  function HelpFile(const Name: String) : String;
  procedure InternalShutDown;
  procedure HelpCommand_HELP_SETPOPUP_POS;

  { ICustomHelpViewer }
  function GetViewerName : String;
  procedure NotifyID(const ViewerID: Integer);
  procedure SoftShutDown;
  procedure ShutDown;
  
  function UnderstandsKeyword(const HelpString: String): Integer;
  function GetHelpStrings(const HelpString: String): TStringList;
  function CanShowTableOfContents: Boolean;
  procedure ShowTableOfContents;
  procedure ShowHelp(const HelpString: String);
  
  { IExtendedHelpViewer }
  function UnderstandsTopic(const Topic: String): Boolean;
  procedure DisplayTopic(const Topic: String);
  function UnderstandsContext(const ContextID: Integer; const HelpFileName: String): Boolean;
  procedure DisplayHelpByContext(const ContextID: Integer; const HelpFileName: String);
  property ViewerID : Integer read FViewerID;
  property HelpManager : IHelpManager read FHelpManager write FHelpManager;
 end;

{ global instance of THTMLHelpViewer which HelpIntfs can talk to. }

var HelpViewer : THTMLHelpViewer;


{----------------------------------------------------------------------------}
{ THelpSelector                                                              }
{----------------------------------------------------------------------------}

 { IHelpSelector. IHelpSelector is used by the HelpSystem to ask the
   application to decide which keyword, out of multiple matches returned
   by multiple different Help Viewers, it wishes to support. If an application
   wishes to support this, it passes an IHelpSelector interface into
   IHelpSystem.AssignHelpSelector. }

type THelpSelector = class (TInterfacedObject, IHelpSelector)
 public
  function SelectKeyword(Keywords: TStrings) : Integer;
  function TableOfContents(Contents: TStrings): Integer;
  end;



{Note: Never called - Since we are the only Keyword player in town}

function THelpSelector.SelectKeyword(Keywords: TStrings) : Integer;
begin
Result := 0;      //return index of first item in supplied keyword list
end;


function THelpSelector.TableOfContents(Contents: TStrings): Integer;
{Returning our name poistion in the provided list will ensure that we are used to display the TOC}
var I: Integer;
begin
Result := 0;
for I := 0 to Contents.count-1 do
 if Contents[I] = ViewerName then Result:= I;   //Found ourselves in the list
end;


{----------------------------------------------------------------------------}
{ TWinHelpTester                                                             }
{----------------------------------------------------------------------------}

{ Look though the standard viewer code, WinHelpViewer.pas which this
  module is based on. Typically viewer 1 is first cab off the rank and
  wants to handle the help call. Except... there is this global thing
  "WinHelpTester" which if implemented can override the decisions made
  in Viewer 1. That's what this section is. We implement WinHelpTester
  and can manipulate the decisions made by Viewer 1. Viewer 2 now gets
  a chance to handle all help calls.

}

type TWinHelpTester = class (TInterfacedObject, IWinHelpTester)
 public
  function CanShowALink(const ALink, FileName: string): Boolean;
  function CanShowTopic(const Topic, FileName: string): Boolean;
  function CanShowContext(const Context: Integer; const FileName: string): Boolean;
  function GetHelpStrings(const ALink: string): TStringList;
  function GetHelpPath: string;
  function GetDefaultHelpFile: string;
  end;


function TWinHelpTester.CanShowALink(const ALink, FileName: string): Boolean;
begin
Result := FALSE;
end;


function TWinHelpTester.CanShowTopic(const Topic, FileName: string): Boolean;
begin
Result := False;
end;


function TWinHelpTester.CanShowContext(const Context: Integer; const FileName: string): Boolean;
begin
Result := False;
end;


function TWinHelpTester.GetHelpStrings(const ALink: string): TStringList;
begin
Result := TStringList.Create;
end;


function TWinHelpTester.GetHelpPath: string;
{Used by HelpeViewer 1 for Linux - ignored for now}
begin
Result := '';
end;



function TWinHelpTester.GetDefaultHelpFile: string;
{This will do for now - Not too important if using OnHelp}
begin
Result := '';
if Assigned(HelpViewer) then Result := HelpViewer.HelpFile('');
end;


{----------------------------------------------------------------------------}
{ THTMLHelpViewer                                                            }
{----------------------------------------------------------------------------}

constructor THTMLHelpViewer.Create;
begin
inherited Create;
end;

destructor THTMLHelpViewer.Destroy;
begin
inherited Destroy;
end;


function THTMLHelpViewer.HelpFile(const Name: String): String;
var
FileName : String;
begin
if (Name = '') and Assigned(FHelpManager) then FileName := HelpManager.GetHelpFile
else FileName := Name;
Result := FileName;
end;




procedure THTMLHelpViewer.InternalShutDown;
{ InternalShut Down is called from unit finalization if the unit is exiting
  and the Help Manager needs to be informed. }
begin
SoftShutDown;
if Assigned(FHelpManager) then
	begin
	HelpManager.Release(ViewerID);
	if Assigned(FHelpManager) then HelpManager := nil;
	end;
end;



procedure THTMLHelpViewer.HelpCommand_HELP_SETPOPUP_POS;
{ Send the HELP_SETPOPUP_POS command just before a Context help call. }
var Control: TWinControl;
var Pt: TSmallPoint;


	function ControlHasHelp(const Control: TWinControl) : Boolean;
	begin
	Result := false;
	if (Control.HelpType = htContext) and (Control.HelpContext <> 0) then Result := true
	else if (Control.HelpType = htKeyword) and (Control.HelpKeyword <> '') then Result := true;
	end;

begin
{This is not the best - since F1 press could have come from a memu -- no way of telling}
Control := Screen.ActiveControl;
while (Control <> nil) and (not ControlHasHelp(Control)) do Control:= Control.Parent;
if Control <> nil then
	begin
	Pt := PointToSmallPoint(Control.ClientToScreen(Point(0, 0)));
	Application.HelpCommand(HELP_SETPOPUP_POS, Longint(Pt));
	end;
end;


{----------------------------------------------------------------------------}
{ THTMLHelpViewer - ICustomHelpViewer                                        }
{----------------------------------------------------------------------------}

function THTMLHelpViewer.GetViewerName : String;
begin
Result := ViewerName;
end;


function THTMLHelpViewer.UnderstandsKeyword(const HelpString: String): Integer;
{ UnderstandsKeyword is a querying function that the Help Manager calls to
  determine if the Viewer provide helps on a particular keyword string. }
begin
Result := 1;
end;


function THTMLHelpViewer.GetHelpStrings(const HelpString: String): TStringList;
{ GetHelpStrings is used by the Help Manager to display a list of keyword
  matches from which an application's user can select one. It assumes
  that the String List is properly allocated, so this function should
  never return nil. }
begin
Result := TStringList.Create;
Result.Add(GetViewerName + ': ' + HelpString);
end;



function THTMLHelpViewer.CanShowTableOfContents : Boolean;
{ CanShowTableOfContents is a querying function that the Help Manager
  calls to determine if the Viewer supports tables of contents. WinHelp
  and HyperHelp both do. }
begin
Result := true;
end;



procedure THTMLHelpViewer.ShowTableOfContents;
{ ShowTableOfContents is a command function that the Help Manager uses
  to direct the Viewer to display a table of contents. It is never
  called without being preceded by a call to CanShowTableOfContents. }
begin
{ The Fix!! - data ignored we set to zero}
Application.HelpCommand(HELP_CONTENTS, 0);
end;


procedure THTMLHelpViewer.ShowHelp(const HelpString: String);
{ ShowHelp is the function that the Help Manager calls to request that
  a Help Viewer display help for a given keyword. }
var HelpCommand: array[0..255] of Char;
begin
StrLFmt(HelpCommand, SizeOf(HelpCommand) -1, '%s', [HelpString]);

{ The Fix!! }
Self.HelpCommand_HELP_SETPOPUP_POS;
Application.HelpCommand(HELP_KEY, Longint(@HelpCommand));
end;


procedure THTMLHelpViewer.NotifyID(const ViewerID: Integer);
{ NotifyID is called by the Help Manager after a successful registration
  to provide the Help Viewer with a cookie which uniquely identifies the
  Viewer to the Manager, and can be used in communications between the two. }
begin
FViewerID := ViewerID;
end;


procedure THTMLHelpViewer.SoftShutDown;
{ SoftShutDown is called by the help manager to ask the viewer to
  terminate any externally spawned subsystem without shutting itself down. }
begin
{ The Fix!! }
{rob: Commented this out - caused an error on a user in Win98}
//Application.HelpCommand(HELP_QUIT, 0);
end;


procedure THTMLHelpViewer.ShutDown;
begin
SoftShutDown;
if Assigned(FHelpManager) then HelpManager := nil;
end;

{----------------------------------------------------------------------------}
{ THTMLHelpViewer --- IExtendedHelpViewer                                     }
{----------------------------------------------------------------------------}


function THTMLHelpViewer.UnderstandsTopic(const Topic: String): Boolean;
{ UnderstandsTopic is called by the Help Manager to ask if the Viewer
  is capable of displaying a topic-based help query for a given topic. }
begin
Result := true;
end;




procedure THTMLHelpViewer.DisplayTopic(const Topic: String);
{ DisplayTopic is called by the Help Manager if a Help Viewer claims
  in its response to UnderstandsTopic to be able to provide Topic-based
  help for a particular keyword. }
var HelpCommand: array[0..255] of Char;
begin
StrLFmt(HelpCommand, SizeOf(HelpCommand) -1, '%s', [Topic + 'zzz']);
Self.HelpCommand_HELP_SETPOPUP_POS;
Application.HelpCommand(HELP_KEY, Longint(@HelpCommand));
end;



function THTMLHelpViewer.UnderstandsContext(const ContextID: Integer; const HelpFileName: String): Boolean;
{Tell Help Manager we are capable of hanlding comment }
begin
Result := true;
end;




procedure THTMLHelpViewer.DisplayHelpByContext(const ContextID: Integer; const HelpFileName: String);
{ Help Manager is requesting the Viewer display help by Context ID }
var SaveWinHelpTester: IWinHelpTester;
begin
{ The Fix!! }
Self.HelpCommand_HELP_SETPOPUP_POS;
{ 9-Jan-2003: Stop recursion by saving state}
SaveWinHelpTester := WinHelpViewer.WinHelpTester;
WinHelpViewer.WinHelpTester := nil;
Application.HelpCommand(HELP_CONTEXT, ContextID);
WinHelpViewer.WinHelpTester := SaveWinHelpTester;
end;




{----------------------------------------------------------------------------}
{ THTMLHelpViewer --- ISpecialWinHelpViewer                                   }
{----------------------------------------------------------------------------}



procedure EnableWinHelpTester(Enable: Boolean);
{Enable and disable the WinHelp test as second viewer}
begin
if Enable then
	begin
        {Enable WinHelpTester - ie. Make HelpManager use this viewer}
	if not Assigned(WinHelpViewer.WinHelpTester) then WinHelpViewer.WinHelpTester:= TWinHelpTester.Create;
	end
else
	begin
        {Disable WinHelpTester - ie. Make HelpManager use WinHelpViewer again}
	if Assigned(WinHelpViewer.WinHelpTester) then WinHelpViewer.WinHelpTester:= Nil;
	end;
end;


{============================================================================}


const
    {commands to pass to HtmlHelp(), see HTML Help API Reference}
    HH_DISPLAY_TOPIC        = $0000; {open help topic}
    HH_HELP_FINDER          = $0000; {compatibility, use HH_DISPLAY_TOPIC}
    HH_DISPLAY_TOC          = $0001; {select Contents tab in nav pane}
    HH_DISPLAY_INDEX        = $0002; {select Index + search for keyword}
    HH_DISPLAY_SEARCH       = $0003; {select Search tab in nav pane}
    HH_SET_WIN_TYPE         = $0004;
    HH_GET_WIN_TYPE         = $0005;
    HH_GET_WIN_HANDLE       = $0006;
    HH_ENUM_INFO_TYPE       = $0007;
    HH_SET_INFO_TYPE        = $0008;
    HH_SYNC                 = $0009;
    HH_RESERVED1            = $000A; {not currently implemented}
    HH_RESERVED2            = $000B; {not currently implemented}
    HH_RESERVED3            = $000C; {not currently implemented}
    HH_KEYWORD_LOOKUP       = $000D;
    HH_DISPLAY_TEXT_POPUP   = $000E; {display string resource/text popup}
    HH_HELP_CONTEXT         = $000F; {display topic for context number}
    HH_TP_HELP_CONTEXTMENU  = $0010; {text popup, like HELP_CONTEXTMENU}
    HH_TP_HELP_WM_HELP      = $0011; {text popup, like HELP_WM_HELP}
    HH_CLOSE_ALL            = $0012; {close all windows opened by caller}
    HH_ALINK_LOOKUP         = $0013; {ALink version of HH_KEYWORD_LOOKUP}
    HH_GET_LAST_ERROR       = $0014; {not currently implemented}
    HH_ENUM_CATEGORY        = $0015;
    HH_ENUM_CATEGORY_IT     = $0016;
    HH_RESET_IT_FILTER      = $0017;
    HH_SET_INCLUSIVE_FILTER = $0018;
    HH_SET_EXCLUSIVE_FILTER = $0019;
    HH_INITIALIZE           = $001C;
    HH_UNINITIALIZE         = $001D;
    HH_PRETRANSLATEMESSAGE  = $00FD;
    HH_SET_GLOBAL_PROPERTY  = $00FC;


function HtmlHelp(hwndCaller: THandle; pszFile: PChar;
    uCommand: cardinal; dwData: longint): THandle; stdcall;
    external 'hhctrl.ocx' name 'HtmlHelpA'; {external API call}


type TconHTMLHelp = class(TObject) {encapsulates function}
  function ApplicationHelp(Command: Word; Data: Longint; var CallHelp: Boolean): Boolean;
  end; {TconHTMLHelp..}



function TconHTMLHelp.ApplicationHelp(Command: Word; Data: Longint; var CallHelp: Boolean): Boolean;
{translates WinHelp commands to HTMLHelp commands + calls API}
var HCommand: word;
begin
{make sure VCL doesn't activate WinHelp + function succeeds}
CallHelp := false;
result := true;
{translate WinHelp > HTMLHelp}
case Command of
 HELP_CONTENTS :
	begin
	HCommand :=  HH_DISPLAY_TOC;
	Data := 0;
	end; {HELP_CONTENTS..}
 HELP_CONTEXT: HCommand:= HH_HELP_CONTEXT;
 HELP_CONTEXTPOPUP: HCommand:= HH_HELP_CONTEXT;
 HELP_FINDER: HCommand:= HH_DISPLAY_TOPIC;
 HELP_KEY: HCommand:= HH_DISPLAY_INDEX;
 HELP_QUIT:
	begin
	HCommand:=  HH_CLOSE_ALL;
	Data:= 0;
	end; {HELP_QUIT..}
 else
	begin {default}
	HCommand:= HH_DISPLAY_TOPIC;
	Data:= 0;
	end; {default..}
 end; {case Command..}
{call HTML Help API}
HtmlHelp(Application.MainForm.Handle, PChar(Application.HelpFile), HCommand, Data);
end; {function TconHTMLHelp.ApplicationHelp}


{============================================================================}


var HelpSelector : IHelpSelector;

var hhHelper: TconHTMLHelp;




initialization
EnableWinHelpTester(true);
HelpSelector := THelpSelector.Create;                  //set to Nil by HelpIntfs cleanup
HelpViewer := THTMLHelpViewer.Create;
Application.HelpSystem.AssignHelpSelector(HelpSelector);
HelpIntfs.RegisterViewer(HelpViewer , HelpViewer.FHelpManager);
{create object + assign event handler}
hhHelper:= TconHTMLHelp.Create;
Application.OnHelp:= hhHelper.ApplicationHelp;
finalization
HelpSelector._Release;   //RWC_RELEASE
if Assigned(HelpViewer.FHelpManager) then HelpViewer.InternalShutDown;
{free event handler + object}
Application.OnHelp:= nil;
hhHelper.Free;
end.

