unit CircularBuffer;

interface

uses Windows,Classes,SysUtils,Graphics,Contnrs,SyncObjs,HugeArry,DSupport;

{Object to maintain indexes for LIFOs or FIFOs}

type TOverWriteEvent = procedure(Sender: TObject; Index: integer) of object;

type
 TCircularIndex = class(TObject)
 private
  FMaxSize: integer;
  FOnOverWrite: TOverWriteEvent;
  EnumIndex: integer;
  procedure IncStartIndex;
  procedure IncEndIndex;
  function GetCount: integer;
 protected
  procedure DoOverWrite(Inx: integer);
  procedure SetMaxSize(V: integer); virtual;
 public
  EndInx: integer;
  StartInx : Integer;
  constructor Create(MaxSize: integer); virtual;
  destructor Destroy; override;
  procedure Reset;
  function Add: integer;
  function GetFIFOIndex: integer;
  function GetLIFOIndex: integer;
  function EnumerateIndices(Start: boolean): integer;
  property MaxSize: integer read FMaxSize write SetMaxSize;
  property OnOverWrite: TOverWriteEvent read FOnOverWrite write FOnOverWrite;
  property Count: integer read GetCount;
 end;



type TCircularString = class(TCircularIndex)
 private
 protected
  StringBusy: TCriticalSection;
  procedure SetMaxSize(V: integer); override;
 public
  Buffer: array of char;
  UseSemaphore: boolean;
  constructor Create(MaxSize: integer); virtual;
  destructor Destroy; override;
  procedure AddChar(C: char);
  procedure AddString(S: String);

  function GetFIFOChar(var C: char): boolean;
  function GetFIFOString(var S: string): boolean;
  function GetLIFOChar(var C: char): boolean;
  function GetLIFOString(var S: string): boolean;
 end;



{File based undo system for large objects like bitmaps}

type
 TFileUndo = class(TObject)
 private
  FFilenames: THugeString;
  FCircularIndex: TCircularIndex;
  FFileBasename: string;
  procedure HandleFileDelete(Sender: TObject; Index: integer);
  function GetCount: integer;
  function GetMaxUndos: integer;
  procedure SetMaxUndos(const Value: integer);
    function GetFilename(Index: integer): string;
 protected
 public
  constructor Create(MaxSize: integer);
  destructor Destroy; override;
  procedure Reset;
  procedure Clear;
  function AddUndoItem: string;
  function UndoItem: string;
  property MaxUndos: integer read GetMaxUndos write SetMaxUndos;
  property FileBasename: string read FFileBasename write FFileBasename;
  property Count: integer read GetCount;
  property CircularIndex: TCircularIndex read FCircularIndex;
  property Filename[Index: integer]: string read GetFilename;
 end;



type TBitmapUndo = class(TObject)
 private
  FFileUndo: TFileUndo;
  function GetCount: integer;
 protected
 public
  constructor Create(MaxSize: integer);
  destructor Destroy; override;
  procedure Clear;
  procedure DeleteLastUndo;
  procedure AddUndoBitmap(BM: TBitmap);
  procedure GetUndoBitmap(var BM: TBitmap);
  property Count: integer read GetCount;
  property Fileundo: TFileUndo read FFileUndo;
 end;


type TIntStack = class(TStack)
 private
  function GetDepth: integer;
 public
  procedure Clear;
  function Push(I: integer): boolean;
  function Pop(var I: integer): boolean;
  function Peek(var I: integer): boolean;
  property Depth: integer read GetDepth;
end;


type TStackEx = class(TStack)
 private
  function GetDepth: integer;
 public
  procedure Clear;
  property Depth: integer read GetDepth;
  function Pop(var P: Pointer): boolean; overload;
  function Pop: pointer; overload;
 end;



implementation



procedure TCircularIndex.Reset;
var Inx: integer;
begin
while true do
	begin
	Inx:=GetLIFOIndex;
	if Inx<0 then break;
	DoOverWrite(Inx);
	end;
EndInx:=0; StartInx:=0;
end;



procedure TCircularIndex.SetMaxSize(V: integer);
{Note: Due to the way the buffer wraps around
inorder to get a specific number of useable
items in the list MaxValue has to be Count+1 }
begin
if V<>FMaxSize then
	begin
	Reset;
	FMaxSize:=V;
	end;
end;



constructor TCircularIndex.Create(MaxSize: integer);
{Note: Due to the way the buffer wraps around
inorder to get a specific number of useable
items in the list MaxValue has to be Count+1 }
begin
inherited Create;
Self.MaxSize:=MaxSize;
EndInx:=0; StartInx:=0;
end;



destructor TCircularIndex.Destroy;
begin
Reset;
inherited Destroy;
end;


procedure TCircularIndex.DoOverWrite(Inx: integer);
begin
if Assigned(FOnOverWrite) then FOnOverWrite(Self,Inx);
end;



procedure TCircularIndex.IncStartIndex;
begin
Inc(StartInx);
if  StartInx>=FMaxSize then StartInx:=0;
end;



procedure TCircularIndex.IncEndIndex;
begin
Inc(EndInx);
if EndInx>=FMaxSize then EndInx:=0;
end;



function TCircularIndex.Add: integer;
{ Get index for newly added item and increment index }
begin
Result:=EndInx;
Inc(EndInx);
if EndInx>=FMaxSize then EndInx:=0;
if EndInx=StartInx then
	begin
	DoOverWrite(StartInx);
	IncStartIndex;
	end;
end;



function TCircularIndex.GetLIFOIndex: integer;
{ Get index of next LIFO entry and -1 if none left }
{You must destroy any object connected to the index when you are done}
begin
Result:=-1;
if EndInx=StartInx then exit;
if EndInx=0 then EndInx:=FMaxSize-1 else Dec(EndInx);
Result:=EndInx;
end;



function TCircularIndex.GetFIFOIndex: integer;
{ Get index of next FIFO entry and -1 if none left }
{You must destroy any object connected to the index  when you are done}
begin
Result:=-1;
if StartInx=EndInx then exit;
Result:=StartInx;
Inc(StartInx);
if StartInx>=FMaxSize then StartInx:=0;
end;



function TCircularIndex.GetCount: integer;
begin
if EndInx<StartInx then Result:=EndInx + (FMaxSize-StartInx)
else Result:=EndInx-StartInx;
end;




function TCircularIndex.EnumerateIndices(Start: boolean): integer;
begin
if GetCount<1 then EnumIndex:=-1
else if Start then EnumIndex:=StartInx
else
	begin
	Inc(EnumIndex);
	if EnumIndex>=FMaxSize then EnumIndex:=0;
	if EnumIndex=EndInx then EnumIndex:=-1;
	end;
Result:=EnumIndex;
end;


{----------------------------------------------------------------------------}


function CreateTempFile(const BaseType: string): string;
var Buffer: array[0..MAX_PATH] of char;
begin
repeat
	begin
	GetTempPath(SizeOf(Buffer) - 1, Buffer);
	GetTempFileName(Buffer, PChar(BaseType), GetTickCount, Buffer);
	SetString(Result, Buffer, StrLen(Buffer));
	end
until not FileExists(Result);
end;



procedure TFileUndo.Reset;
begin
FCircularIndex.Reset;
end;



procedure TFileUndo.HandleFileDelete(Sender: TObject; Index: integer);
begin
DeleteFile(FFileNames[Index]);
end;



constructor TFileUndo.Create(MaxSize: integer);
begin
inherited Create;
FFileNames:=THugeString.Create(MaxSize,255);
FCircularIndex:=TCircularIndex.Create(MaxSize);
FCircularIndex.OnOverWrite:=HandleFileDelete;
FileBasename:='Undo';
end;



destructor TFileUndo.Destroy;
begin
FCircularIndex.Free;
FFileNames.Free;
inherited Destroy;
end;



function TFileUndo.AddUndoItem: string;
var Inx: integer;
begin
Inx:=FCircularIndex.Add;
Result:=CreateTempFile(FileBasename);
FFileNames[Inx]:=Result;
end;



function TFileUndo.UndoItem: string;
var Inx: integer;
begin
Result:='';
Inx:=FCircularIndex.GetLIFOIndex;
if Inx>-1 then Result:=FFileNames[Inx];
end;


procedure TFileUndo.Clear;
var S: string;
begin
while true do
	begin
	S:=UndoItem;
	if S='' then break;
	DeleteFile(S);
	end;
Reset;
end;



function TFileUndo.GetCount: integer;
begin
Result:=FCircularIndex.Count;
end;




function TFileUndo.GetMaxUndos: integer;
begin
Result:=FCircularIndex.MaxSize;
end;

procedure TFileUndo.SetMaxUndos(const Value: integer);
begin
FCircularIndex.MaxSize:=Value;
end;



function TFileUndo.GetFilename(Index: integer): string;
begin
Result:=FFilenames[Index];
end;




{-----------------------------------------------------------------------------}



procedure LoadUndoBMP(BM: TBitmap; Filename: string);
{Unscramble BMP header for security}
var MS: TMemoryStream;
var L: int64;
var I: integer;
begin
MS:=TMemoryStream.Create;
try
MS.LoadFromFile(Filename);
L:=1024;
if L>MS.Size then L:=MS.Size;
for I:=0 to L-1 do
 PByteArray(MS.Memory)[I]:=PByteArray(MS.Memory)[I] xor $55;
BM.LoadFromStream(MS);
finally MS.Free; end;
end;



procedure SaveUndoBMP(BM: TBitmap; Filename: string);
{Scramble BMP header for security}
var MS: TMemoryStream;
var L: int64;
var I: integer;
begin
MS:=TMemoryStream.Create;
try
BM.SaveToStream(MS);
MS.Position:=0;
L:=1024;
if L>MS.Size then L:=MS.Size;
for I:=0 to L-1 do
 PByteArray(MS.Memory)[I]:=PByteArray(MS.Memory)[I] xor $55;
MS.SaveToFile(Filename);
finally MS.Free; end;
end;



{-----------------------------------------------------------------------------}


constructor TBitmapUndo.Create(MaxSize: integer);
begin
inherited Create;
FFileUndo:=TFileUndo.Create(MaxSize);
FFileUndo.FileBasename:='BXF';
end;



destructor TBitmapUndo.Destroy;
begin
FFileUndo.Free;
inherited Destroy;
end;





procedure TBitmapUndo.AddUndoBitmap(BM: TBitmap);
var S: string;
begin
S:=FFileUndo.AddUndoItem;
SaveUndoBMP(BM,S);
end;


procedure TBitmapUndo.GetUndoBitmap(var BM: TBitmap);
var S: string;
begin
S:=FFileUndo.UndoItem;
if S<>'' then
	begin
	LoadUndoBMP(BM,S);
	DeleteFile(S);
	end;
end;



procedure TBitmapUndo.DeleteLastUndo;
var S: string;
begin
S:=FFileUndo.UndoItem;
if S<>'' then DeleteFile(S);
end;



procedure TBitmapUndo.Clear;
begin
FFileUndo.Clear;
end;


function TBitmapUndo.GetCount: integer;
begin
Result:=FFileUndo.Count;
end;





{------------------------------------------------------------------------------}

{ TIntStack }

function TIntStack.Peek(var I: integer): boolean;
begin
Result:=List.Count>0;
if Result then I:=Integer(List[List.Count-1]);
end;

function TIntStack.Pop(var I: integer): boolean;
begin
Result:=Peek(I);
if Result then  List.Delete(List.Count-1);
end;

function TIntStack.Push(I: integer): boolean;
begin
Result:=False;
List.Add(Pointer(I));
Result:=True;
end;

procedure TIntStack.Clear;
begin
List.Clear;
end;

function TIntStack.GetDepth: integer;
begin
Result:=List.Count;
end;

{ TStackEx }

procedure TStackEx.Clear;
begin
List.Clear;
end;


function TStackEx.GetDepth: integer;
begin
Result:=List.Count;
end;


function TStackEx.Pop(var P: Pointer): boolean;
begin
Result:=List.Count>0;
if Result then P:= PopItem;
end;

function TStackEx.Pop: pointer;
begin
Pop(Result);
end;



{ TCircularString }

procedure TCircularString.AddChar(C: char);
var Inx: integer;
begin
if UseSemaphore then StringBusy.Acquire;
try
Inx:=Add;
Buffer[Inx]:=C;
finally
 if UseSemaphore then StringBusy.Release;
 end;
end;


procedure TCircularString.AddString(S: String);
var I: integer;
begin
for I:=1 to Length(S) do AddChar(S[I]);
end;


constructor TCircularString.Create(MaxSize: integer);
{Note: Due to the way the buffer wraps around
inorder to get a specific number of useable
items in the list MaxValue has to be Count+1 }
begin
inherited ;
SetLength(Buffer,MaxSize);
UseSemaphore:=False;
StringBusy:=TCriticalSection.Create;
end;


destructor TCircularString.Destroy;
begin
StringBusy.Free;
inherited;
end;



function TCircularString.GetFIFOChar(var C: char): boolean;
var Inx: integer;
begin
if UseSemaphore then StringBusy.Acquire;
try
Inx:=GetFIFOIndex;
Result:=Inx>-1;
if Result then C:=Buffer[Inx];
finally
 if UseSemaphore then StringBusy.Release;
 end;
end;



function TCircularString.GetFIFOString(var S: string): boolean;
var C: char;
begin
S:='';
while GetFIFOChar(C) do S:=S+C;
end;



function TCircularString.GetLIFOChar(var C: char): boolean;
var Inx: integer;
begin
if UseSemaphore then StringBusy.Acquire;
try
Inx:=GetLIFOIndex;
Result:=Inx>-1;
if Result then C:=Buffer[Inx];
finally
 if UseSemaphore then StringBusy.Release;
 end;
end;



function TCircularString.GetLIFOString(var S: string): boolean;
var C: char;
begin
S:='';
while GetLIFOChar(C) do S:=S+C;
end;



procedure TCircularString.SetMaxSize(V: integer);
{Note: Due to the way the buffer wraps around
inorder to get a specific number of useable
items in the list MaxValue has to be Count+1 }
begin
if UseSemaphore then StringBusy.Acquire;
try
inherited;
SetLength(Buffer,MaxSize);
finally
 if UseSemaphore then StringBusy.Release;
 end;
end;



end.
