unit CurveFitLib;

interface

uses Classes,Math,Vector,DSupport;


type TMatrix = class(TObject)
 private
  FMatrix: array of array of double;
  function GetValue(X, Y: integer): double;
  procedure SetValue(X, Y: integer; const Value: double);
  procedure SetRows(const Value: integer);
  function GetRows: integer;
  function GetColumns: integer;
  procedure SetColumns(const Value: integer);
  procedure SetDimensions(Cols, Row: integer);
 protected
 public
  constructor Create;
  procedure Clear(V: double); overload;
  procedure Clear; overload;
  property Cols: integer read GetColumns write SetColumns;
  property Rows: integer read GetRows write SetRows;
  property Value[X,Y: integer]: double read GetValue write SetValue; default;
 end;



type TCurveFitter = class(TObject)
 private
  function GetPower(X: integer; P: double): double;
  function Select(X, Y: Integer): double;
 protected
 public
  Degree: integer;
  Matrix: TMatrix;
  FitData: array of T2DVector;
  constructor Create;
  destructor Destroy; override;
  procedure BuildMatrix;
  procedure DoGaussJordan;
  procedure DoDataFit;
  procedure Add(V: T2DVector); overload;
  procedure Add(X,Y: double); overload;
  procedure Clear;
  procedure Delete(Inx: integer);
  procedure Insert(Inx: integer; V: T2DVector);
  procedure InsertStringList(SL: TStringList; Col: integer);
  procedure InsertString(S: string; Col: integer);
  function GetPolynomialValue(X: double): double;
  procedure Sort;
 end;



implementation


{ TMatrix }



procedure TMatrix.SetDimensions(Cols,Row: integer);
{Set dimensions without regard for the contents}
begin
SetLength(FMatrix,Cols,Row);
end;
                            


function TMatrix.GetColumns: integer;
begin
Result:=Length(FMatrix);
end;



function TMatrix.GetRows: integer;
begin
if Length(FMatrix)>0 then Result:=Length(FMatrix[0])
else Result:=0;
end;



procedure TMatrix.SetRows(const Value: integer);
begin
SetDimensions(Cols,Value);
end;



procedure TMatrix.SetColumns(const Value: integer);
begin
SetDimensions(Value,Rows);
end;




function TMatrix.GetValue(X, Y: integer): double;
begin
Result:=FMatrix[X,Y];
end;



procedure TMatrix.SetValue(X, Y: integer; const Value: double);
begin
FMatrix[X,Y]:=Value;
end;

            

procedure TMatrix.Clear(V: double);
var X,Y: integer;
begin
for Y:=0 to Rows-1 do
 for X:=0 to Cols-1 do
	FMatrix[X,Y]:=V;
end;


procedure TMatrix.Clear;
begin
Clear(0);
end;


constructor TMatrix.Create;
begin
SetDimensions(0,0);
end;



{ TCurveFitter }


constructor TCurveFitter.Create;
begin
Matrix:=TMatrix.Create;
Degree:=1;
end;



destructor TCurveFitter.Destroy;
begin
Matrix.Free;
inherited;
end;



function TCurveFitter.GetPower(X: integer; P: double): double;
{Get the power of the specified X data}
var B: double;
begin
B:=FitData[X].X;
Result:=Power(B,P);
end;




function TCurveFitter.Select(X,Y : Integer): double;
{ Select The Proper Data Item }
begin
if Y=0 then Result:=1
else if Y=Matrix.Rows-1 then Result:=FitData[X].Y
else Result:=GetPower(X,Y);
end;



procedure TCurveFitter.BuildMatrix;
{ Build Determinate Matrix }
var X,Y,S : Integer;
begin
Matrix.Clear;
Matrix.Cols:=Degree+1;
Matrix.Rows:=Degree+2;
for S:=0 to Length(FitData)-1 do
 for X:=0 to Matrix.Cols-1 do
  for Y:=0 to Matrix.Rows-1 do
	Matrix[X,Y]:=Matrix[X,Y] + (SELECT(S,X)*SELECT(S,Y));
end;


procedure TCurveFitter.DoGaussJordan;
{ Do Gauss Jordan Elimination }
var I,K,J : Integer;
var S : double;
begin
K:=0;
while True do
	begin
	S:=Matrix[K,K];
	{Don't use larger number or else you'll get NAN or INF values}
	if S=0.0 then S:=5.0e-162;
	for J:=K to Matrix.Rows-1 do Matrix[K,J]:=Matrix[K,J]/S;
	if K>=Matrix.Cols-1 then break;
	for I:=K+1 to Matrix.Cols-1 do
		begin
		S:=Matrix[I,K];
		for J:=K to Matrix.Rows-1 do
			Matrix[I,J]:=Matrix[I,J]-S*Matrix[K,J];
		end;
	K:=K+1;
	end;

while K>0 do
	begin
	I:=K-1;
	repeat
		begin
		S:=Matrix[I,K];
		for J:=K to Matrix.Rows-1 do
			Matrix[I,J]:=Matrix[I,J]-S*Matrix[K,J];
		I:=I-1;
		end
	until I<0;
	K:=K-1;
	end;
end;


procedure TCurveFitter.DoDataFit;
{Curve the current data}
begin
BuildMatrix;
DoGaussJordan;
end;



function TCurveFitter.GetPolynomialValue(X: double): double;
{Return Polynomial value at a specified point}
var P : double;
var T: integer;
begin
Result:=Matrix[0,Matrix.Rows-1];
P:=1;
for T:=1 to Matrix.Cols-1 do
	begin
	P:=P * X;
	Result:=Result+(Matrix[T,Matrix.Rows-1] * P);
	end;
end;


procedure TCurveFitter.Add(V: T2DVector);
{Add point to Fit Data set}
begin
SetLength(FitData,Length(FitData)+1);
FitData[High(FitData)]:=V;
end;


procedure TCurveFitter.Add(X,Y: double);
{Add point to Fit Data set}
begin
Add(MakeVector2D(X,Y));
end;



procedure TCurveFitter.Clear;
{Set all Fit Data to zero}
begin
SetLength(FitData,0);
end;


procedure TCurveFitter.Delete(Inx: integer);
{Delete item at Inx position}
var R: integer;
begin
if (Inx<0) or (Inx>=Length(FitData)) or (Length(FitData)<0) then exit;
for R:=Inx to Length(FitData)-2 do FitData[R]:=FitData[R+1];
SetLength(FitData,Length(FitData)-1);
end;



procedure TCurveFitter.Insert(Inx: integer; V: T2DVector);
{Insert X,Y value at Inx location}
var R: integer;
begin
if (Inx<0) or (Inx>=Length(FitData)) then exit;
SetLength(FitData,Length(FitData)+1);
for R:=Length(FitData)-2 downto Inx do FitData[R+1]:=FitData[R];
FitData[Inx]:=V;
end;



procedure TCurveFitter.InsertStringList(SL: TStringList; Col: integer);
{Insert the data from string list in Fit Data array}
var I,P,Cnt: integer;
var X,Y: double;
var S: string;
begin
if Length(FitData)<SL.Count then SetLength(FitData,SL.Count);
for I:=0 to SL.Count-1 do
	begin
	S:=SL[I];
	P:=1; Cnt:=0;
	if ExtractNumber(S,P,X) then Inc(Cnt);
	if ExtractNumber(S,P,Y) then Inc(Cnt);
	if Cnt=0 then continue;
	if Col=2 then FitData[I].Y:=X
	else
		begin
		FitData[I].X:=X;
		if Cnt>1 then FitData[I].Y:=Y;
		end;
	if I>Length(FitData) then
		begin
		SetLength(FitData,I+1);
		if Col=2 then FitData[I].X:=0
		else if Cnt=1 then FitData[I].Y:=0;
		end;
	end;
SetLength(FitData,SL.Count);
end;



procedure TCurveFitter.InsertString(S: string; Col: integer);
{Insert one or two column string of values}
var SL: TStringList;
begin
SL:=TStringList.Create;
try
SL.Text:=S;
InsertStringList(SL,Col);
finally SL.Free; end;
end;



procedure TCurveFitter.Sort;
{Bubble sort data by X values}
var I, J: integer;
var T: double;
begin
for I := Length(FitData)-1 downto 0 do
 for J := 0 to Length(FitData) - 2 do
  if FitData[J].X > FitData[J + 1].X then
  	begin
  	T := FitData[J].X; FitData[J].X := FitData[J + 1].X; FitData[J + 1].X := T;
  	T := FitData[J].Y; FitData[J].Y := FitData[J + 1].Y; FitData[J + 1].Y := T;
  	end;
end;


end.
