unit DEMDisplay;

interface

uses Windows,Forms,SysUtils,ExtCtrls,Graphics,Math,DEMLib,BitmapLib,DepthColors,
     GLMesh,DSupport,Vector,Hash;

type TDisplayModes = (dmColor,dmContour,dmShade,dmWater);
type TDisplayMode = set of TDisplayModes;
type TSunDirection=(sdNorth,sdSouth,sdEast,sdWest,sdNorthEast,sdNorthWest,sdSouthEast,sdSouthWest);
type TGLProgress= procedure(Sender: TObject; Action: string; PercentDone: integer);


type TDEMDisplayInfo = record
 Mode: TDisplayMode;
 SunDir: TSunDirection;
 Azimuth,Inclination: double;
 Contrast: double;
 Contour: integer;
 end;

procedure DisplayDEMData(DEMArray: TDEMArray; Image: TImage; DisplayInfo: TDEMDisplayInfo);
procedure LoadDEMToMesh(DEMArray: TDEMArray; Mesh: TGLMesh; Mag: double; Progress: TGLProgress);

implementation

uses DEMLoader;




function IsBetween(L,H,S: double): boolean;
{ Does a contour line fall between L and H?}
var T: double;
begin
{Make sure L<H - swap if they are not}
if L>H then begin T:=L; L:=H; H:=T; end;
{Find nearest topo line below L}
T:=trunc(L / S)*S;
{Check if the next topo line is below H}
Result:=(T+S)<=H;
end;


function GetGrayScaleColor(Percent: double): TColor;
{Get a gray value based on percentage white}
var B: byte;
begin
B:=Round(255 * Percent);
Result:=RGB(B,B,B);
end;


procedure DisplayDEMData(DEMArray: TDEMArray; Image: TImage; DisplayInfo: TDEMDisplayInfo);
{Display Dem array to Image}
var X,Y,CInx: integer;
var Elev: double;
var Step,Depth: double;
var Color24: pRGBTRipleArray;
var SlopeDist: double;


	function GetGrayColor(Slope: double): TColor;
	{Get gray color based on slope and contrast}
	var Scale: double;
	begin
	Scale:=ArcTan(Slope/(SlopeDist*(1/Power(SlopeDist/2,DisplayInfo.Contrast))))/HalfPI;
	Scale:=(Scale/2) + 0.5;
	Scale:=EnsureRange(Scale,0,1);
	Result:=GetGrayScaleColor(Scale);
	end;


	function IsWater: boolean;
	var Limit: double;
	begin
	Limit:=0.00005;
	Result:=(X>0) and (Y>0) and
		(((DEMArray[X,Y-1] - DEMArray[X,Y])/SlopeDist)<Limit) and
		(((DEMArray[X-1,Y] - DEMArray[X,Y])/SlopeDist)<Limit) and
		(((DEMArray[X,Y]   - DEMArray[X,Y-1])/SlopeDist)<Limit) and
		(((DEMArray[X,Y] - DEMArray[X-1,Y])/SlopeDist)<Limit) and
		(((DEMArray[X-1,Y-1] - DEMArray[X,Y])/SlopeDist)<Limit) and
		(((DEMArray[X,Y] - DEMArray[X-1,Y-1])/SlopeDist)<Limit) and
		(((DEMArray[X-1,Y] - DEMArray[X,Y-1])/SlopeDist)<Limit) and
		(((DEMArray[X,Y-1] - DEMArray[X-1,Y])/SlopeDist)<Limit);
	end;


	function GetSlope(X,Y: integer): double;
	begin
	if (X<1) or (Y<1) then Result:=1
	else
		begin
		case DisplayInfo.SunDir of
		 sdNorth    : Result:=DEMArray[X,Y-1] - DEMArray[X,Y];
		 sdEast     : Result:=DEMArray[X-1,Y] - DEMArray[X,Y];
		 sdSouth     : Result:=DEMArray[X,Y]   - DEMArray[X,Y-1];
		 sdWest     : Result:=DEMArray[X,Y] - DEMArray[X-1,Y];
		 sdNorthEast: Result:=DEMArray[X-1,Y-1] - DEMArray[X,Y];
		 sdSouthWest: Result:=DEMArray[X,Y] - DEMArray[X-1,Y-1];
		 sdSouthEast: Result:=DEMArray[X-1,Y] - DEMArray[X,Y-1];
 		 sdNorthWest: Result:=DEMArray[X,Y-1] - DEMArray[X-1,Y];
		 end;
		end;
	end;


	function GetShadeColor(X,Y: integer): TColor;
	{Get shading value based on slope and sun direction}
	var Slope: double;
	begin
	Slope:=GetSlope(X,Y);
	Result:=GetGrayColor(Slope);
	end;


	function IsIntervalPoint(X,Y: integer): boolean;
	{Return true if point is on contour interval}
	begin
	Result:=((X<>0) and IsBetween(DEMArray[X-1,Y],Elev,DisplayInfo.Contour)) or
		((Y<>0) and IsBetween(DEMArray[X,Y-1],Elev,DisplayInfo.Contour));
	end;



	function GetElevationColor: TColor;
	{Get color based on current elevation}
	begin
	if (dmWater in DisplayInfo.Mode) and (Elev<0) then Result:=RGB(0,0,200)
	else if (dmWater in DisplayInfo.Mode) and IsWater then Result:=RGB(0,0,255)
	else
		begin
		CInx:=Round((DEMArray.DEMInfo.ActualMaxVert - Elev) / Step) + 1;
		CInx:=EnsureRange(CInx,0,High(ColorMap47));
		Result:=ColorMap47[CInx];
		end;
	end;


	function AdjustShade(C,Shader: TColor): TColor;
	var S: double;
	var R,G,B: byte;
	begin
	S:=GetRValue(Shader)/255;
	ColorToRGBBytes(C,R,G,B);
	Result:=RGB(round(R*S),round(G*S),round(B*S));
	end;


	function GetPointColor(X,Y: integer): TColor;
	{Get the color for point based mode and elevtion}
	begin
	Result:=clBlack;
	Elev:=DEMArray[X,Y];
	if Elev = EmptyCell then exit;
	if (dmContour in DisplayInfo.Mode) and IsIntervalPoint(X,Y) then Result:=clBlack
	else if (dmColor in DisplayInfo.Mode) and (dmShade in DisplayInfo.Mode) then
		begin
		Result:=AdjustShade(getElevationColor,GetshadeColor(X,Y));
		end
	else if (dmColor in DisplayInfo.Mode) then Result:=getElevationColor
	else if (dmShade in DisplayInfo.Mode) then Result:=GetshadeColor(X,Y)
	else Result:=clWhite;
	end;



begin
Image.Width:=DEMArray.Cols;
Image.Height:=DEMArray.Rows;
Image.Picture.Bitmap.Canvas.Brush.Color:=clBlack;
Image.Picture.Bitmap.Canvas.Rectangle(Image.ClientRect);
Image.Picture.Bitmap.PixelFormat:=pf24bit;
Image.Picture.Bitmap.Width:=DEMArray.Cols;
Image.Picture.Bitmap.Height:=DEMArray.Rows;
SlopeDist:=Sqrt(Sqr(DEMArray.DEMInfo.Resolution[0])+Sqr(DEMArray.DEMInfo.Resolution[1]));
Depth:=DEMArray.DEMInfo.ActualMaxVert - DEMArray.DEMInfo.ActualMinVert;
Step:=Depth / (High(ColorMap47)-1);
for Y:=0 to DEMArray.Rows-1 do
	begin
	Color24:=Image.Picture.Bitmap.ScanLine[DEMArray.Rows-Y-1];
	for X:=0 to DEMArray.Cols-1 do
		begin
		Color24[X]:=ColorToTriple(GetPointColor(X,Y));
		end;
	end;
Image.Invalidate;
Application.ProcessMessages;
end;



{------------------------------------------------------------------------------}


procedure LoadDEMToMesh(DEMArray: TDEMArray; Mesh: TGLMesh; Mag: double; Progress: TGLProgress);
var X,Y: integer;
var P1,P2,P3,P4: T3DVector;
var CX,CY: double;
var MaxZ,MinZ: double;
var Res: T2DVector;


	function GetArrayPoint(X,Y: integer): T3DVector;
	var Z: double;
	begin
	Z:=DEMArray[X,Y];
	if Z=EmptyCell then Z:=0 else Z:=Z-MinZ;
	Result:=MakeVector3D((X-CX)*Res.X,Z*Mag,(CY-Y)*Res.Y);
	end;


	function GetFixedZPoint(X,Y: integer; Z: double): T3DVector;
	begin
	Result:=GetArrayPoint(X,Y);
	Result.Y:=Z * Mag;
	end;


	procedure DoWestSide(Mesh: TGLMesh; C: TColor; Off: double);
	var Y: integer;
	var P1,P2,P3,P4: T3DVector;
	begin
	for Y:=0 to DEMArray.Rows-2 do
		begin
		P1:=GetArrayPoint(0,Y);
		P2:=GetArrayPoint(0,Y+1);
		P3:=GetFixedZPoint(0,Y+1,Off);
		P4:=GetFixedZPoint(0,Y,Off);
		Mesh.AddRectangle(P1,P2,P3,P4,C);
		end;
	end;
	
	
	
	procedure DoEastSide(Mesh: TGLMesh; C: TColor; Off: double);
	var Y: integer;
	var P1,P2,P3,P4: T3DVector;
	begin
	for Y:=0 to DEMArray.Rows-2 do
		begin
		P1:=GetArrayPoint(DEMArray.Cols-1,Y+1);
		P2:=GetArrayPoint(DEMArray.Cols-1,Y);
		P3:=GetFixedZPoint(DEMArray.Cols-1,Y,Off);
		P4:=GetFixedZPoint(DEMArray.Cols-1,Y+1,Off);
		Mesh.AddRectangle(P1,P2,P3,P4,C);
		end;
	end;
	
	
	
	procedure DoNorthSide(Mesh: TGLMesh; C: TColor; Off: double);
	var X: integer;
	var P1,P2,P3,P4: T3DVector;
	begin
	for X:=0 to DemArray.Cols-2 do
		begin
		P1:=GetArrayPoint(X,0);
		P2:=GetFixedZPoint(X,0,Off);
		P3:=GetFixedZPoint(X+1,0,Off);
		P4:=GetArrayPoint(X+1,0);
		Mesh.AddRectangle(P1,P2,P3,P4,C);
		end;
	end;
	
	
	procedure DoSouthSide(Mesh: TGLMesh; C: TColor; Off: double);
	var X: integer;
	var P1,P2,P3,P4: T3DVector;
	begin
	for X:=0 to DEMArray.Cols-2 do
		begin
		P1:=GetArrayPoint(X,DEMArray.Rows-1);
		P2:=GetArrayPoint(X+1,DEMArray.Rows-1);
		P3:=GetFixedZPoint(X+1,DEMArray.Rows-1,Off);
		P4:=GetFixedZPoint(X,DEMArray.Rows-1,Off);
		Mesh.AddRectangle(P1,P2,P3,P4,C);
		end;
	end;
	
	
	
	
	procedure DoBottom(Mesh: TGLMesh; C: TColor; Off: double);
	var P1,P2,P3,P4: T3DVector;
	begin
	P4:=GetFixedZPoint(0,0,Off);
	P3:=GetFixedZPoint(DEMArray.Cols-1,0,Off);
	P2:=GetFixedZPoint(DEMArray.Cols-1,DEMArray.Rows-1,Off);
	P1:=GetFixedZPoint(0,DEMArray.Rows-1,Off);
	Mesh.AddRectangle(P1,P2,P3,P4,C);
	end;

	
	procedure DrawBase(Mesh: TGLMesh; C: TColor);
	var Off: double;
	begin
	Off:=(MaxZ-MinZ)*(50/100);
	DoBottom(Mesh,clBlue,-Off);
	DoNorthSide(Mesh,clRed,-Off);
	DoSouthSide(Mesh,clGreen,-Off);
	DoWestSide(Mesh,clYellow,-Off);
	DoEastSide(Mesh,clWhite,-Off);
	end;
	

begin
GetDEMResolution(DEMArray.DEMInfo,Res);
MaxZ:=DEMArray.DEMInfo.ActualMaxVert;
MinZ:=DEMArray.DEMInfo.ActualMinVert;
CX:=DEMArray.Cols/2;
CY:=DEMArray.Rows/2;

Mesh.Clear;
for Y:=0 to DEMArray.Rows-2 do
	begin
	Progress(Mesh,'Converting', MulDiv(100,Y,DEMArray.Rows-2));
	for X:=0 to DEMArray.Cols-2 do
		begin
		P1:=GetArrayPoint(X,Y);
		P2:=GetArrayPoint(X+1,Y);
		P3:=GetArrayPoint(X+1,Y+1);
		P4:=GetArrayPoint(X,Y+1);
		Mesh.AddRectangle(P1,P2,P3,P4,clGray);
		end;
	end;
DrawBase(Mesh,clRed);
Mesh.OnNormalProgress:=Progress;
Mesh.SetAllNormals;
Mesh.OnNormalProgress:=nil;
end;



end.
