unit DEMLib;

interface

uses Windows,Types,SysUtils,DSupport,Vector,IntVector,Huge2D,Math,CurveFitLib;

type TProgressState = (psStart,psWorking,psDone);

type TDEMProgress = procedure(Percent: integer; State: TProgressState; Status: string);


{SDTS specified number types}

type TSDTSNumberType =(ntBI8,ntBI16,ntBI24,ntBI32,ntBUI,ntBUI8,ntBUI16,ntBUI24,ntBUI32,ntBFP32,ntBFP64);

{Defines empty data point}

const EmptyCell = Low(integer);

{horizontal data point}

type Point2D = record
 case integer of
 0: (East,North: double);
 1: (Vector: T2DVector);
 end;

{Reference Units}

type TDEMUnits = (euRadians=0,euFeet=1,euMeters=2,euArcSec=3);

{Datum values}

type THorizontalDatum = (hdNAD27= 1, hdWGS72= 2, hdWGS84= 3, hdNAD83= 4);

{Map types}

type TMapType=(mtOneDegree,mt7Minute);

{DEM information mostly from the header}

type TDEMInfo = record
  MapName: string[144];
  LevelCode: integer;
  Pattern: integer;
  GroundRef: integer;
  Zone: integer;
  GroundUnits: TDEMUnits;
  ElevationUnits: TDEMUnits;
  PolySides: integer;
  Accuracy: integer;
  DEMRows, DEMCols: integer;
  ActualRows,ActualCols: integer;
  Angle: double;
  BoxAngle: double;
  Projection: array[0..14] of double;
  BoundBoxNW: Point2D;
  SWCorner,NWCorner,NECorner,SECorner: Point2D;
  Resolution: Array [0..2] of double;
  MinElevation, MaxElevation: double;
  ActualMinVert,ActualMaxVert: double;
  MinVertIndex,MaxVertIndex: TPoint;
  SouthMax: double;
  MapHeight: double;
  HorzDatum: THorizontalDatum;
  MapType: TMapType;
  Scale: integer;
  DataID: string;
  Date: string;
  DeRotated: boolean;
  ErrorCount: integer;
  ErrorString: string;
 end;


type TProfileInfo = record
  RowID,ColID: integer;
  ProfRows: integer;
  ProfCols: integer;
  ProfLocation : Point2D;
  ProfElevation: double;
  ProfMinElevation: double;
  ProfMaxElevation: double;
  ProfileLength: integer;
  PixelOffset: integer;
 end;

type TProfInfoArray = array of TProfileInfo;


{Array object to hold information}

type TMergeType = (mtInsert,mtLeft,mtRight,mtTop,mtBottom);

type TDEMArray = class(T2DDoubleArray)
 private
  procedure MergeDEMInfo(Source: TDEMInfo; MergeType: TMergeType);
  procedure PatchMissingData;
  procedure SmoothHorizonSeam(SeamY: integer);
  procedure SmoothVerticalSeam(SeamX: integer);
  procedure AssignExtraData(DEMArray: TDEMArray);
 public
  DEMInfo: TDEMInfo;
  Loaded: boolean;
  CurveDegree,SmoothWidth,AveWidth: integer;
  SmoothSeams,AverageEdges: boolean;
  ProfInfoArray: TProfInfoArray;
  procedure ClearInfo;
  constructor Create;
  procedure SetNewDim(NewSize: integer; ET: TExpandMode); override;
  procedure Assign(DEMArray: TDEMArray);
  procedure OffsetAssign(DEMArray: TDEMArray; Off: TPoint);
  procedure MergeArrays(Source: TDEMArray; MergeType: TMergeType);
  function InRange(X,Y: integer): boolean;
 end;

function GetDatumStr(HD: THorizontalDatum): string;
function GetUnitsStr(Units: TDEMUnits): string;
function GetDEMInfoString(DEMArray: TDEMArray): string;
function GetNumberType(Format: string): TSDTSNumberType;
procedure GetDEMResolution(DEMInfo: TDEMInfo; var Res: T2DVector);

procedure FindDEMAngle(var DEM: TDEMArray);


implementation

uses DEMLoader;



type TAverage = class(TObject)
 private
  Cnt: integer;
  Sum: double;
  DEM: TDEMArray;
 protected
 public
  constructor Create(DEM: TDEMArray);
  procedure Clear;
  procedure Add(X,Y: integer);
  procedure AddAdjacent(X,Y: integer);
  function Average: double;
 end;


{ TAverage }



procedure TAverage.Add(X, Y: integer);
begin
if not DEM.InRange(X,Y) then exit;
Sum:=SUm+DEM[X,Y];
Inc(Cnt);
end;


procedure TAverage.AddAdjacent(X, Y: integer);
var X1,Y1: integer;
begin
for X1:=X-1 to X+1 do
 for Y1:=Y-1 to Y+1 do Add(X1,Y1);
end;



function TAverage.Average: double;
begin
Result:=Sum/Cnt;
end;


procedure TAverage.Clear;
begin
Sum:=0;
Cnt:=0;
end;


constructor TAverage.Create(DEM: TDEMArray);
begin
Self.DEM:=DEM;
end;


{TDEMArray}


procedure CopyProfArrays(var Dest,Source: TProfInfoArray);
var I: integer;
begin
SetLength(Dest,Length(Source));
for I:=0 to High(Dest) do Dest[I]:=Source[I];
end;



procedure TDEMArray.ClearInfo;
begin
DEMInfo.MinElevation:=32767;
DEMInfo.MaxElevation:=-32767;
DEMInfo.ActualMinVert:=32767;
DEMInfo.ActualMaxVert:=-32767;
DEMInfo.MinVertIndex:=Point(-1,-1);
DEMInfo.MaxVertIndex:=Point(-1,-1);
end;


procedure TDEMArray.AssignExtraData(DEMArray: TDEMArray);
begin
DEMInfo:=DEMArray.DEMInfo;
CopyProfArrays(ProfInfoArray,DEMArray.ProfInfoArray);
Loaded:=DEMArray.Loaded
end;



procedure TDEMArray.Assign(DEMArray: TDEMArray);
begin
inherited Assign(DEMArray);
AssignExtraData(DEMArray);
end;





procedure TDEMArray.OffsetAssign(DEMArray: TDEMArray; Off: TPoint);
var I: integer;
begin
inherited OffsetAssign(DEMArray, Off);
AssignExtraData(DEMArray);
{

DEMInfo:=DEMArray.DEMInfo;
DEMINfo.MapName

CopyProfArrays(ProfInfoArray,DEMArray.ProfInfoArray);
Loaded:=Loaded or DEMArray.Loaded
}
end;



procedure TDEMArray.SetNewDim(NewSize: integer; ET: TExpandMode);
var OldDim,Delta: TPoint;

	procedure ReDimLeft;
	var X,Y: integer;
	var Delta: integer;
	begin
	SetLength(ProfInfoArray,NewSize);
	Delta:=NewSize - OldDim.X;
	{shift old data right}
	for X:=0 to OldDim.X-1 do
	 ProfInfoArray[X+Delta]:=ProfInfoArray[X];
	end;

begin
OldDim.X:=Cols;
OldDim.Y:=Rows;
{Expand array of Left or Right}
inherited SetNewDim(NewSize,ET);
case ET of
 emLeft: ReDimLeft;
 emRight: SetLength(ProfInfoArray,NewSize);
 emTop, emBottom: ;
 end;
end;




constructor TDEMArray.Create;
begin
inherited Create(5,5);
Loaded:=False;
DEMInfo.DeRotated:=False;
CurveDegree:=5;
SmoothWidth:=5;
AveWidth:=3;
AverageEdges:=True;
SmoothSeams:=True;
ClearInfo;
end;



function TDEMArray.InRange(X,Y: integer): boolean;
begin
Result:=not ((X<0) or (Y<0) or (X>=Cols) or (Y>=Rows) or (Self[X,Y]=EmptyCell));
end;



procedure TDEMArray.MergeDEMInfo(Source: TDEMInfo; MergeType: TMergeType);
begin
if MergeType=mtInsert then
	begin
	DEMInfo:=Source;
	exit;
	end;

if MergeType=mtLeft then
	begin
	DEMInfo.DEMCols:=DEMInfo.DEMCols+Source.DEMCols;
	DEMInfo.ActualCols:=DEMInfo.DEMCols;
	DEMInfo.SWCorner.East:=Source.SWCorner.East;
	DEMInfo.NWCorner.East:=Source.SWCorner.East;
	end
else if MergeType=mtRight then
	begin
	DEMInfo.DEMCols:=DEMInfo.DEMCols+Source.DEMCols;
	DEMInfo.ActualCols:=DEMInfo.DEMCols;
	DEMInfo.SECorner.East:=Source.SECorner.East;
	DEMInfo.NECorner.East:=Source.SECorner.East;
	end
else if MergeType=mtTop then
	begin
	DEMInfo.ActualRows:=DEMInfo.ActualRows+Source.ActualRows;
	DEMInfo.NWCorner.North:=Source.NWCorner.North;
	DEMInfo.NECorner.North:=Source.SECorner.North;
	end
else if MergeType=mtBottom then
	begin
	DEMInfo.ActualRows:=DEMInfo.ActualRows+Source.ActualRows;
	DEMInfo.SWCorner.North:=Source.SWCorner.North;
	DEMInfo.SECorner.North:=Source.SECorner.North;
	end;


DEMInfo.MapName:='Merge';
DEMInfo.MinElevation:=Min(DEMInfo.MinElevation,Source.MinElevation);
DEMInfo.MaxElevation:=Max(DEMInfo.MaxElevation,Source.MaxElevation);
if Source.ActualMinVert<DEMInfo.ActualMinVert then
	begin
	DEMInfo.ActualMinVert:=Source.ActualMinVert;
	DEMInfo.MinVertIndex:=Source.MinVertIndex;
	end;
if Source.ActualMaxVert<DEMInfo.ActualMaxVert then
	begin
	DEMInfo.ActualMaxVert:=Source.ActualMaxVert;
	DEMInfo.MaxVertIndex:=Source.MaxVertIndex;
	end;
end;




procedure TDEMArray.SmoothHorizonSeam(SeamY: integer);
{Smooth out vertical seam anomalies by curve fitting}
{ and average the edges between the fitted and regular data}
var X,Y,Y1,Y2: integer;
var Fitter: TCurveFitter;
var Average: TAverage;
begin
Fitter:=TCurveFitter.Create;
try
Fitter.Degree:=CurveDegree;
Average:=TAverage.Create(Self);
try
for X:=0 to Cols-1 do
	begin
	Fitter.Clear;
	{Sample line on either side of the seam}
	Y1:=SeamY-SmoothWidth; Y2:=SeamY+SmoothWidth;
	for Y:=Y1 to SeamY-1 do Fitter.Add(Y,Self[X,Y]);
	for Y:=SeamY+1 to Y2 do Fitter.Add(Y,Self[X,Y]);
	{Generate polynomial that approximates terrain}
	Fitter.DoDataFit;
	{Replace distorted terrain with fitted terrain}
	for Y:=Y1+1 to Y2-1 do Self[X,Y]:=Fitter.GetPolynomialValue(Y);
	{Smooth/Average out the edges of the fitted data}
	if AverageEdges then
		begin
		for Y:=-AveWidth to AveWidth do
			begin
			{Average the top edge}
			Average.Clear;
			Average.AddAdjacent(X,Y1+Y);
			Self[X,Y1+Y]:=Average.Average;
			{Average the bottom edge}
			Average.Clear;
			Average.AddAdjacent(X,Y2+Y);
			Self[X,Y2+Y]:=Average.Average;
			end;
		end;
	end;
finally Average.Free; end;
finally Fitter.Free; end;
end;



procedure TDEMArray.SmoothVerticalSeam(SeamX: integer);
{Smooth out vertical seam anomalies by curve fitting}
{ and average the edges between the fitted and regular data}
var X,Y,X1,X2: integer;
var Fitter: TCurveFitter;
var Average: TAverage;
begin
Fitter:=TCurveFitter.Create;
try
Fitter.Degree:=CurveDegree;
Average:=TAverage.Create(Self);
try
for Y:=0 to Rows-1 do
	begin
	Fitter.Clear;
	{Sample line on either side of the seam}
	X1:=SeamX-SmoothWidth; X2:=SeamX+SmoothWidth;
	for X:=X1 to SeamX-1 do Fitter.Add(X,Self[X,Y]);
	for X:=SeamX+1 to X2 do Fitter.Add(X,Self[X,Y]);
	{Generate polynomial that approximates terrain}
	Fitter.DoDataFit;
	{Replace distorted terrain with fitted terrain}
	for X:=X1 to X2 do Self[X,Y]:=Fitter.GetPolynomialValue(X);
	{Smooth/Average out the edges of the fitted data}
	if AverageEdges then
		begin
		for X:=-AveWidth to AveWidth do
			begin
			{Average the left side}
			Average.Clear;
			Average.AddAdjacent(X1+X,Y);
			Self[X1+X,Y]:=Average.Average;
			{Average the rigth edge}
			Average.Clear;
			Average.AddAdjacent(X2+X,Y);
			Self[X2+X,Y]:=Average.Average;
			end;
		end;
	end;
finally Average.Free; end;
finally Fitter.Free; end;
end;



procedure TDEMArray.PatchMissingData;
{Assumes bands missing data will be no more than one pixel wide}
var X,Y,C: integer;
var A: double;

	procedure Average(X,Y: integer);
	begin
	if not InRange(X,Y) then exit;
	A:=A+Self[X,Y];
	Inc(C);
	end;

begin
for Y:=0 to Rows-1 do
 for X:=0 to Cols-1 do
  if Self[X,Y]=EmptyCell then
	begin
	C:=0; A:=0;
	Average(X-1,Y);
	Average(X,Y-1);
	Average(X+1,Y);
	Average(X,Y+1);
	if C>0 then Self[X,Y]:=A/C;
	end;
end;


procedure TDEMArray.MergeArrays(Source: TDEMArray; MergeType: TMergeType);
var SSize,DSize: TPoint;
begin
SSize:=Point(Source.Cols,Source.Rows);
DSize:=Point(Self.Cols,Self.Rows);
Self.FillExpansion:=True;
Self.FillValue:=EmptyCell;
case MergeType of
 mtInsert: Self.Assign(Source);
 mtLeft:
	begin
	if Source.Rows>Self.Rows then Self.SetNewDim(Source.Rows,emBottom);
	Self.SetNewDim(DSize.X+SSize.X,emLeft);
	Self.OffsetAssign(Source,Point(0,0));
	PatchMissingData;
	if SmoothSeams then SmoothVerticalSeam(SSize.X);
	end;
 mtRight:
	begin
	if Source.Rows>Self.Rows then Self.SetNewDim(Source.Rows,emBottom);
	Self.SetNewDim(DSize.X+SSize.X,emRight);
	Self.OffsetAssign(Source,Point(DSize.X,0));
	PatchMissingData;
	if SmoothSeams then SmoothVerticalSeam(DSize.X);
	end;
 mtTop:
	begin
	{These are inverted because DEM profiles are South to North, bottom to top}
	if Source.Cols>Self.Cols then Self.SetNewDim(Source.Cols,emRight);
	Self.SetNewDim(DSize.Y+SSize.Y,emBottom);
	Self.OffsetAssign(Source,Point(0,DSize.Y));
	PatchMissingData;
	if SmoothSeams then SmoothHorizonSeam(DSize.Y-1);
	end;
 mtBottom:
	begin
	{These are inverted because DEM profiles are South to North, bottom to top}
	if Source.Cols>Self.Cols then Self.SetNewDim(Source.Cols,emRight);
	Self.SetNewDim(DSize.Y+SSize.Y,emTop);
	Self.OffsetAssign(Source,Point(0,0));
	PatchMissingData;
	if SmoothSeams then SmoothHorizonSeam(SSize.Y-1);
	end;
 end;
MergeDEMInfo(Source.DEMInfo,MergeType);
Self.Loaded:=Source.Loaded;
end;





{-------------- Support Routines ---------------------------------------------}


function GetNumberType(Format: string): TSDTSNumberType;
{Convert string to SDTS Number type}
begin
if Format='BI8' then  Result:=ntBI8
else if Format='BI16' then  Result:=ntBI16
else if Format='BI24' then  Result:=ntBI24
else if Format='BI32' then  Result:=ntBI32
else if Format='BUI' then  Result:=ntBUI
else if Format='BUI8' then  Result:=ntBUI8
else if Format='BUI16' then  Result:=ntBUI16
else if Format='BUI24' then  Result:=ntBUI24
else if Format='BUI32' then  Result:=ntBUI32
else if Format='BFP32' then  Result:=ntBFP32
else if Format='BFP64' then  Result:=ntBFP64;
end;



function GetDatumStr(HD: THorizontalDatum): string;
{Get string corresponding to horizontal datum type}
begin
case HD of
 hdNAD27: Result:='NAD27';
 hdWGS72: Result:='WGS72';
 hdWGS84: Result:='WGS84';
 hdNAD83: Result:='NAD83';
 end;
end;



function GetUnitsStr(Units: TDEMUnits): string;
{Convert DEM units to string description}
begin
case Units of
 euRadians: Result:='Radians';
 euFeet: Result:='Feet';
 euMeters: Result:='Meters';
 euArcSec: Result:='Arc-Seconds';
 end;
end;




function GetDEMInfoString(DEMArray: TDEMArray): string;
{Return string displaying DEM information}
var I: integer;
var S: string;
begin
with DEMArray do
 begin
 Result:='';
 Result:=Result+'MapName: '+ DEMInfo.MapName+CRLF_Char;
 Result:=Result+'LevelCode: '+ IntToStr(DEMInfo.LevelCode)+CRLF_Char;
 Result:=Result+'Pattern: '+ IntToStr(DEMInfo.Pattern)+CRLF_Char;
 Result:=Result+'GroundRef: '+ IntToStr(DEMInfo.GroundRef)+CRLF_Char;
 Result:=Result+'Zone: '+ IntToStr(DEMInfo.Zone)+CRLF_Char;
 Result:=Result+'GroundUnits: '+ GetUnitsStr(DEMInfo.GroundUnits)+CRLF_Char;
 Result:=Result+'ElevationlUnits: '+ GetUnitsStr(DEMInfo.ElevationUnits)+CRLF_Char;
 Result:=Result+'PolySides: '+ IntToStr(DEMInfo.PolySides)+CRLF_Char;
 Result:=Result+'Accuracy: '+ IntToStr(DEMInfo.Accuracy)+CRLF_Char;
 Result:=Result+'DEMCols: '+ IntToStr(DEMInfo.DEMCols)+CRLF_Char;
 Result:=Result+'DEMRows: '+ IntToStr(DEMInfo.DEMRows)+CRLF_Char;
 Result:=Result+'ActualCols: '+ IntToStr(DEMInfo.ActualCols)+CRLF_Char;
 Result:=Result+'ActualRows: '+ IntToStr(DEMInfo.ActualRows)+CRLF_Char;
 Result:=Result+'Array Size: '+ IntToStr(DEMArray.Cols)+','+IntToStr(DEMArray.Rows)+CRLF_Char;
 Result:=Result+'MinElevation: '+ FloatToStrF(DEMInfo.MinElevation,ffFixed,18,1)+CRLF_Char;
 Result:=Result+'MaxElevation: '+ FloatToStrF(DEMInfo.MaxElevation,ffFixed,18,1)+CRLF_Char;
 Result:=Result+'Angle: '+ FloatToStrF(DEMInfo.Angle,ffFixed,18,1)+CRLF_Char;
 Result:=Result+'BoxAngle: '+ FloatToStrF(DEMInfo.BoxAngle,ffFixed,18,3)+CRLF_Char;
 Result:=Result+'Resolution: '+ FloatToStrF(DEMInfo.Resolution[0],ffFixed,18,1) + ' ' +
 FloatToStrF(DEMInfo.Resolution[1],ffFixed,18,1) + ' ' +
                                      FloatToStrF(DEMInfo.Resolution[2],ffFixed,18,1)+CRLF_Char;
 Result:=Result+'ActualMinVert: ' + FloatToStrF(DEMInfo.ActualMinVert,ffFixed,18,1)+CRLF_Char;
 Result:=Result+'ActualMaxVert: '+ FloatToStrF(DEMInfo.ActualMaxVert,ffFixed,18,1)+CRLF_Char;
 Result:=Result+'MinVertIndex: ' + IntToStr(DEMInfo.MinVertIndex.X)+','+IntToStr(DEMInfo.MinVertIndex.Y)+CRLF_Char;
 Result:=Result+'MaxVertIndex: ' + IntToStr(DEMInfo.MaxVertIndex.X)+','+IntToStr(DEMInfo.MaxVertIndex.Y)+CRLF_Char;

 Result:=Result+'MapType: '+ PickString(DEMInfo.MapType=mtOneDegree,'One Degree','7.5 Minute')+CRLF_Char;

 Result:=Result+'Datum: '+GetDatumStr(DEMInfo.HorzDatum)+CRLF_Char;
 Result:=Result+'Scale: '+IntToStr(DEMInfo.Scale)+CRLF_Char;
 Result:=Result+'DataID: '+DEMInfo.DataID+CRLF_Char;
 Result:=Result+'Date: '+DEMInfo.Date+CRLF_Char;
 Result:=Result+'BoundBoxNW: '+FloatToStrF(DEMInfo.BoundBoxNW.East,ffFixed,18,1)+', '+FloatToStrF(DEMInfo.BoundBoxNW.North,ffFixed,18,1)+CRLF_Char;

 S:='';
 for I:=0 to 14 do S:=S+' '+FloatToStrF(DEMInfo.Projection[I],ffFixed,18,1);
 Result:=Result+'Projections: '+S+CRLF_Char;
 S:=' NW: '+FloatToStrF(DEMInfo.NWCorner.East,ffFixed,18,1)+', '+FloatToStrF(DEMInfo.NWCorner.North,ffFixed,18,1)+CRLF_Char;
 S:=S+' NE: '+FloatToStrF(DEMInfo.NECorner.East,ffFixed,18,1)+', '+FloatToStrF(DEMInfo.NECorner.North,ffFixed,18,1)+CRLF_Char;
 S:=S+' SW: '+FloatToStrF(DEMInfo.SWCorner.East,ffFixed,18,1)+', '+FloatToStrF(DEMInfo.SWCorner.North,ffFixed,18,1)+CRLF_Char;
 S:=S+' SE: '+FloatToStrF(DEMInfo.SECorner.East,ffFixed,18,1)+', '+FloatToStrF(DEMInfo.SECorner.North,ffFixed,18,1)+CRLF_Char;
 Result:=Result+S+CRLF_Char;
 Result:=Result+'SouthMax: '+ FloatToStrF(DEMInfo.SouthMax,ffFixed,18,1)+CRLF_Char;
 Result:=Result+'MapHeight: '+ FloatToStrF(DEMInfo.MapHeight,ffFixed,18,1)+CRLF_Char;
 Result:=Result+'DeRotated: '+ PickString(DEMInfo.DeRotated,'Yes','No')+CRLF_Char;

 Result:=Result+'ErrorCount: '+ IntToStr(DEMInfo.ErrorCount)+CRLF_Char;
 Result:=Result+'Errors: '+CRLF_Char;
 Result:=Result+DEMInfo.ErrorString+CRLF_Char;
 for I:=0 to High(ProfInfoArray) do
	begin
	Result:=Result+'================================'+CRLF_Char;
	Result:=Result+'Profile #: '+IntToStr(I)+CRLF_Char;
	Result:=Result+'================================'+CRLF_Char;
	Result:=Result+'ColID, RowID: '+IntToStr(ProfInfoArray[I].ColID)+','+IntToStr(ProfInfoArray[I].RowID)+CRLF_Char;
	Result:=Result+'Cols, Rows: '+IntToStr(ProfInfoArray[I].ProfCols)+','+IntToStr(ProfInfoArray[I].ProfRows)+CRLF_Char;
	Result:=Result+'Location: '+Vector2DToString(ProfInfoArray[I].ProfLocation.Vector,1,False)+CRLF_Char;
	Result:=Result+'Elevation: '+FloatToStrF(ProfInfoArray[I].ProfElevation,ffNumber,18,1)+CRLF_Char;
	Result:=Result+'Min Elev: '+FloatToStrF(ProfInfoArray[I].ProfMinElevation,ffNumber,18,1)+CRLF_Char;
	Result:=Result+'Max Elev: '+FloatToStrF(ProfInfoArray[I].ProfMaxElevation,ffNumber,18,1)+CRLF_Char;
	Result:=Result+'Length: '+IntToStr(ProfInfoArray[I].ProfileLength)+CRLF_Char;
	Result:=Result+'Pixel Offset: '+IntToStr(ProfInfoArray[I].PixelOffset)+CRLF_Char;
	end;
 end;
end;



{-------------------------------------------------------------------------------}

const GeoOffset : array [0..91-1,0..1] of double = (
 (30.95,30.74),(30.98,30.72),(30.99,30.69),(30.99,30.66),(30.99,30.63),(30.98,30.60),
 (30.95,30.58),(30.92,30.55),(30.88,30.52),(30.82,30.50),(30.76,30.47),(30.69,30.45),
 (30.61,30.42),(30.52,30.40),(30.42,30.38),(30.32,30.35),(30.20,30.33),(30.07,30.31),
 (29.94,30.29),(29.79,30.27),(29.64,30.25),(29.48,30.23),(29.31,30.21),(29.13,30.19),
 (28.94,30.18),(28.74,30.16),(28.53,30.15),(28.32,30.13),(28.09,30.12),(27.86,30.11),
 (27.62,30.10),(27.37,30.09),(27.11,30.08),(26.84,30.07),(26.57,30.06),(26.29,30.06),
 (26.00,30.05),(25.70,30.05),(25.39,30.04),(25.08,30.04),(24.76,30.04),(24.43,30.04),
 (24.09,30.04),(23.75,30.04),(23.40,30.04),(23.04,30.05),(22.67,30.05),(22.30,30.06),
 (21.92,30.06),(21.54,30.07),(21.14,30.08),(20.74,30.09),(20.34,30.10),(19.93,30.11),
 (19.51,30.12),(19.09,30.13),(18.66,30.15),(18.22,30.16),(17.78,30.18),(17.34,30.19),
 (16.89,30.21),(16.43,30.23),(15.97,30.25),(15.50,30.27),(15.03,30.29),(14.56,30.31),
 (14.08,30.33),(13.59,30.35),(13.10,30.38),(12.61,30.40),(12.12,30.42),(11.62,30.45),
 (11.11,30.47),(10.61,30.50),(10.10,30.53),(9.58,30.55),(9.07,30.58),(8.55,30.61),
 (8.03,30.63),(7.50,30.66),(6.98,30.69),(6.45,30.72),(5.92,30.74),(5.39,30.77),
 (4.85,30.80),(4.32,30.83),(3.78,30.86),(3.24,30.89),(2.70,30.92),(2.16,30.94),
 (1.62,30.97));


procedure GetGeoOffset(LatInx: integer; ArcSec: double; var EO,NO: double);
begin
EO:=30; NO:=30;
if (LatInx<0) or (LatInx>90) then exit;
EO:=GeoOffset[LatInx,0]*ArcSec;
NO:=GeoOffset[LatInx,1]*ArcSec;
end;


procedure GetDEMResolution(DEMInfo: TDEMInfo; var Res: T2DVector);
{Set spacing from DEM value}
var Lat,EO,NO: double;
begin
Lat:=Abs(DEMInfo.SWCorner.North/3600);
if DEMInfo.GroundRef=0 then
	begin
	GetGeoOffset(Trunc(Lat),DEMInfo.Resolution[0],EO,NO);
	Res.X:=EO;
	Res.Y:=NO;
	end
else
	begin
	Res.X:=DEMInfo.Resolution[0];
	Res.Y:=DEMInfo.Resolution[1];
	end;
if Res.X=0 then Res.X:=10;
if Res.Y=0 then Res.Y:=10;
end;



procedure FindDEMSides(DEM: TDEMArray; var TopLeft,TopRight,BottomLeft,BottomRight: TPoint);
var Left,Top,Right,Bottom: TPoint;
var MinX,MinY,MaxX,MaxY,CX,CY: integer;
var RotRight: boolean;

	procedure FindTop;
	var X,Y: integer;
	var H,C: integer;
	begin
	C:=DEM.Cols-1;
	H:=C div 2;
	for Y:=0 to DEM.Rows-1  do
	 for X:=0 to H do
		begin
		if DEM[X,Y]<>EmptyCell then begin Top:=Point(X,Y); exit; end;
		if DEM[C-X,Y]<>EmptyCell then begin Top:=Point(C-X,Y); exit; end;
		end;
	end;

	procedure FindBottom;
	var X,Y: integer;
	var H,C: integer;
	begin
	C:=DEM.Cols-1;
	H:=C div 2;
	for Y:=DEM.Rows-1 downto 0  do
	 for X:=0 to H do
		begin
		if DEM[X,Y]<>EmptyCell then begin Bottom:=Point(X,Y); exit; end;
		if DEM[C-X,Y]<>EmptyCell then begin Bottom:=Point(C-X,Y); exit; end;
		end;
	end;

	procedure FindLeft;
	var X,Y: integer;
	var H,C: integer;
	begin
	C:=DEM.Rows-1;
	H:=C div 2;
	for X:=0 to DEM.Cols-1 do
	 for Y:=0 to H do
		begin
		if DEM[X,Y]<>EmptyCell then begin Left:=Point(X,Y); exit; end;
		if DEM[X,C-Y]<>EmptyCell then begin Left:=Point(X,C-Y); exit; end;
		end;
	end;

	procedure FindRight;
	var X,Y: integer;
	var H,C: integer;
	begin
	C:=DEM.Rows-1;
	H:=C div 2;
	for X:=DEM.Cols-1 downto 0 do
	 for Y:=0 to H do
		begin
		if DEM[X,Y]<>EmptyCell then begin Right:=Point(X,Y); exit; end;
		if DEM[X,C-Y]<>EmptyCell then begin Right:=Point(X,C-Y); exit; end;
		end;
	end;

	type TBoxCorner = (bcUpperLeft,bcUpperRight,bcLowerLeft,bcLowerRight);

	function FindCorner(P: TPoint): TBoxCorner;
	begin
	if P.X<CX then
		begin
		if P.Y<CY then Result:=bcUpperLeft
		else Result:=bcLowerLeft
		end
	else
		begin
		if P.Y<CY then Result:=bcUpperRight
		else Result:=bcLowerRight;
		end;
	end;


	function IsSameCorner(P1,P2: TPoint): boolean;
	begin
	Result:=FindCorner(P1)=FindCorner(P2);
	end;


begin
FindTop;
FindBottom;
FindLeft;
FindRight;

MinX:=Min(Min(Top.X,Right.X),Min(Left.X,Bottom.X));
MaxX:=Max(Max(Top.X,Right.X),Max(Left.X,Bottom.X));
MinY:=Min(Min(Top.Y,Right.Y),Min(Left.Y,Bottom.Y));
MaxY:=Max(Max(Top.Y,Right.Y),Max(Left.Y,Bottom.Y));

CX:=(MaxX+MinX) div 2;
CY:=(MaxY+MinX) div 2;

RotRight:=Top.X<CX;
if IsSameCorner(Left,Top) or
   IsSameCorner(Left,Right) or
   IsSameCorner(Left,Bottom) or
   IsSameCorner(Top,Right) or
   IsSameCorner(Top,Bottom) or
   IsSameCorner(Right,Bottom) then
	begin
	TopLeft:=Point(MinX,MinY);
	TopRight:=Point(MaxX,MinY);
	BottomRight:=Point(MaxX,MaxY);
	BottomLeft:=Point(MinX,MaxY);
	end
else if RotRight then
	begin
	TopLeft:=Top;
	TopRight:=Right;
	BottomLeft:=Left;
	BottomRight:=Bottom;
	end
else
	begin                  
	TopLeft:=Left;          
	TopRight:=Top;
	BottomLeft:=Bottom;
	BottomRight:=Right;
	end;
end;



procedure FindDEMAngle(var DEM: TDEMArray);
var TopLeft,TopRight,BottomLeft,BottomRight: TPoint;
var D: TPoint;
var A: double;
begin
FindDEMSides(DEM,TopLeft,TopRight,BottomLeft,BottomRight);
D:=IntVectorSubtract(BottomLeft,TopLeft);
A:=DSupport.ArcTan2(D.X,D.Y) * Radian2Degree;
DEM.DEMInfo.BoxAngle:=-A;
end;


end.