unit FITSReadWrite;

interface

uses Windows,SysUtils,Classes,Graphics,StreamLib,BitmapLib,DSupport;

{FITS data types }

{Array of integers to hold image data}

type T2DFITSArray = array of array of integer;

type TNameString = string[8];
type TKeyString = string[80];

{Holds one key word item}

type TFITSKeyword = record
 Name: TNameString;
 HasValue: boolean;
 Value: string;
 Comment: string;
 end;

{Array to hold all keywords in FITS file}

const KeysPerBlock = 36;

type TFITSHeader = array  of TFITSKeyword;


{Data type to hold all data from a FITS file}

type TFITSFile = record
 Header: TFITSHeader;
 Dimensions: integer;
 Width,Height: integer;
 BitsPerPixel: integer;
 Data: T2DFITSArray;
 Min,Max: integer;
 end;


procedure ReadFITSFile(FileName: string; var FITSFile: TFITSFile; var BMP: TBitmap); overload;
procedure ReadFITSFile(FileName: string; var FITSFile: TFITSFile); overload;
procedure ReadFITSStream(Stream: TStream; var FITSFile: TFITSFile; var BMP: TBitmap); overload;
procedure ReadFITSStream(Stream: TStream; var FITSFile: TFITSFile); overload;


procedure SaveFITSStream(Stream: TStream; var FITSFile: TFITSFile); overload;
procedure SaveFITSFile(FileName: string; var FITSFile: TFITSFile); overload;

procedure SaveFITSStream(Stream: TStream; var Bitmap: TBitmap); overload;
procedure SaveFITSFile(FileName: String; Bitmap: TBitmap); overload;

procedure SaveFITSFile(FileName: String; var Data: T2DFITSArray); overload;
procedure SaveFITSStream(Stream: TStream; var Data: T2DFITSArray); overload;

procedure ClearOptionList;
procedure AddOptionalKeyword(Name: TNameString; HasValue: boolean; Value: string; Comment: string);

implementation



procedure ClearHeader(var Header: TFITSHeader);
var I: integer;
begin
SetLength(Header,0);
end;



function IncrementHeader(var Header: TFITSHeader): integer;
begin
SetLength(Header,Length(Header)+1);
Result:=High(Header);
end;


{------------------------------------------------------------------------------}


procedure ExtractKeyword(S: string; var Keyword: TFITSKeyword);
{Extract the data from a string and put in a Keyword structure}
var I: integer;
var C: char;
var ComFlg,Quote: boolean;
const SQ = #$27;
begin
Keyword.Value:='';
Keyword.Comment:='';
Keyword.Name:=Copy(S,1,8);
Keyword.Name:=Trim(Keyword.Name);
Keyword.HasValue:=(S[9]='=') and (S[10]=' ');
if not Keyword.HasValue then
	begin
	Keyword.Comment:=Copy(S,9,80-8);
	exit;
	end;
ComFlg:=False;
Quote:=False;
for I:=11 to 80 do
	begin
	C:=S[I];
	{quotes are a meta character in Values - ignore slashes inside quotes}
	if not ComFlg and (C=SQ) then Quote:=not Quote;
	if not Quote and (C='/') then ComFlg:=True;
	if ComFlg then Keyword.Comment:=Keyword.Comment+C
	else Keyword.Value:=Keyword.Value+C;
	end;
end;



procedure ExtractHeaderArguments(Stream: TStream; var FITSFile: TFITSFile);
{Extract specific keyword values from header}
{These are arguments that specify the size and depth of image}
var I: integer;
var Name,Value: string;
begin
for I:=0 to High(FITSFile.Header) do
	begin
	Name:=Trim(FITSFile.Header[I].Name);
	Value:=Trim(FITSFile.Header[I].Value);
	{Extract bits per pixel argument}
	if Name='BITPIX' then FITSFile.BitsPerPixel:=StrToIntDef(Value,0);
	{Extract number of dimension in the image - should be two}
	if Name='NAXIS' then FITSFile.Dimensions:=StrToIntDef(Value,0);
	{Extract first dimension of the image}
	if Name='NAXIS1' then FITSFile.Width:=StrToIntDef(Value,0);
	{Extract Second dimension of the image}
	if Name='NAXIS2' then FITSFile.Height:=StrToIntDef(Value,0);
	end;
end;




procedure ReadHeaderBlock(Stream: TStream; var FITSFile: TFITSFile);
{Extract all Keywords from the FITS file header }
{Each keyword is 80-characters long - Each header}
{Header consists of X-number of 36 80-character blocks = 2880 bytes}
var Block: string;
var I,Inx: integer;
var EndFound: boolean;
begin
EndFound:=False;
ClearHeader(FITSFile.Header);
repeat
	begin
	{Read one block}
	for I:=0 to KeysPerBlock-1 do
		begin
		Block:=StreamReadString(Stream,80);
		Inx:=IncrementHeader(FITSFile.Header);
		ExtractKeyword(Block,FITSFile.Header[Inx]);
		if Trim(FITSFile.Header[Inx].Name)='END' then EndFound:=True;
		end;
	{Continue if we haven't reached the end}
	end
until EndFound;
ExtractHeaderArguments(Stream, FITSFile);
end;



procedure ReadDataBlock(Stream: TStream; var FITSFile: TFITSFile);
{Read actual data from the file - at this point only supports 16-bit images}
var X,Y: integer;
var W: word;
begin
FITSFile.Min:=High(word);
FITSFile.Max:=0;
StreamEndian:=BigEnd;
if FITSFile.Dimensions<>2 then raise Exception.Create('Expected 2-Dim File. Found: '+IntToStr(FITSFile.Dimensions));
if FITSFile.BitsPerPixel<>16 then raise Exception.Create('Expected 16-bits/pixel. Fount: '+IntToStr(FITSFile.BitsPerPixel));
SetLength(FITSFile.Data,FITSFile.Width,FITSFile.Height);
for Y:=0 to FITSFile.Height-1 do
 for X:=0 to FITSFile.Width-1 do
	begin
	W:=GetStreamWord(Stream);
        FITSFile.Data[X,Y]:=W;
	if W>FITSFile.Max then FITSFile.Max:=W;
	if W<FITSFile.Min then FITSFile.Min:=W;
	end;
end;



procedure FitsToBMP(var FITSFile: TFITSFile; var BMP: TBitmap);
{Convert the FITS image data to a Windows Bitmap}
{Note: this reduces the precision by 1/256 }
{The maximum gray-scale depth of a BMP is 256 levels}
var X,Y: integer;
var Line: PRGBTripleArray;
var Pix,D: word;
var F: double;
begin
BMP.Width:=FITSFile.Width;
BMP.Height:=FITSFile.Height;
BMP.PixelFormat:=pf24bit;
D:=FITSFile.Max-FITSFile.Min;
for Y:=0 to FITSFile.Height-1 do
	begin
	Line:=BMP.ScanLine[Y];
	for X:=0 to FITSFile.Width-1 do
		begin
		Pix:=MulDiv(FITSFile.Data[X,Y] - FITSFile.Min,255,D);
		Line[X].rgbtRed:=Pix;
		Line[X].rgbtGreen:=Pix;
		Line[X].rgbtBlue:=Pix;
		end;
	end;
end;



procedure ReadFITSStream(Stream: TStream; var FITSFile: TFITSFile); overload;
{Read FITS file from a stream and put data in FITSFile}
begin
ReadHeaderBlock(Stream, FITSFile);
ReadDataBlock(Stream, FITSFile);
end;



procedure ReadFITSStream(Stream: TStream; var FITSFile: TFITSFile; var BMP: TBitmap); overload;
{Read FITS file from a stream and put data in FITSFile}
{Also generate the corresponding bitmap image}
begin
ReadHeaderBlock(Stream, FITSFile);
ReadDataBlock(Stream, FITSFile);
FitsToBMP(FITSFile,BMP);
end;



procedure ReadFITSFile(FileName: string; var FITSFile: TFITSFile); overload;
{Read FITS from file and return data in FITSFile}
var Stream: TMemoryStream;
begin
Stream := TMemoryStream.Create;
Stream.LoadFromFile(Filename);
try
ReadFITSStream(Stream,FITSFile);
finally Stream.Free; end;
end;



procedure ReadFITSFile(FileName: string; var FITSFile: TFITSFile; var BMP: TBitmap); overload;
{Read FITS from file and return data in FITSFile}
{Also return a copy of the image as Windows BMP file}
{Note that windows bitmaps don't have the full resolution of the FITS data}
var Stream: TMemoryStream;
begin
Stream := TMemoryStream.Create;
Stream.LoadFromFile(Filename);
try
ReadFITSStream(Stream,FITSFile,BMP);
finally Stream.Free; end;
end;



{=============== Routines to write FITS files =================================}




procedure SetKeyword(var Keyword: TFITSKeyword; var KeyStr: TKeyString);
var S: string;
const SQ = #$27;
begin
KeyStr:=FillString(' ',80);
Insert(Keyword.Name,KeyStr,1);
if Keyword.HasValue then
	begin
	Insert('= ',KeyStr,9);
	if (Keyword.Value<>'') and (Keyword.Value[1]=SQ) then S:=GetTabStrLeft(Keyword.Value,21)
	else S:=GetTabStrRight(Keyword.Value,20)+' ';
	if (Keyword.Comment<>'') and (Keyword.Comment[1]<>'/') then S:=S+'/';
	S:=S+Keyword.Comment;
	Insert(S,KeyStr,11);
	end
else Insert(Keyword.Comment,KeyStr,8);
end;






procedure WriteHeaderBlock(Stream: TStream; var FITSFile: TFITSFile);
{Write all Keywords to the FITS file }
{Header consists of at least 36 80-character blocks = 2880 bytes}
var Block: TKeyString;
var I,R: integer;
begin
for I:=0 to High(FITSFile.Header) do
	begin
	SetKeyword(FITSFile.Header[I],Block);
	WriteStreamString(Stream,Block);
	end;
{Fill rest of the block}
R:=Length(FITSFile.Header) mod 36;
if R>0 then
	begin
	for I:=0 to (36-R)-1 do WriteStreamNChars(Stream,' ',80);
	end;
end;



procedure WriteDataBlock(Stream: TStream; var FITSFile: TFITSFile);
{Read actual data from the file - at this point only supports 16-bit images}
var X,Y: integer;
begin
StreamEndian:=BigEnd;
for Y:=0 to FITSFile.Height-1 do
 for X:=0 to FITSFile.Width-1 do
	begin
	WriteStreamWord(Stream,FITSFile.Data[X,Y]);
	end;
end;


{-----------------------------------------------------------------------------}


procedure BMPToFits(var BMP: TBitmap; var FITSFile: TFITSFile);
{Convert the FITS image data to a Windows Bitmap}
{Note: this reduces the precision by 1/256 }
{The maximum gray-scale depth of a BMP is 256 levels}
var X,Y: integer;
var Line: PRGBTripleArray;
var Pix: integer;
begin
FITSFile.Width:=BMP.Width;
FITSFile.Height:=BMP.Height;
BMP.PixelFormat:=pf24bit;
SetLength(FITSFile.Data,BMP.Width,BMP.Height);
for Y:=0 to FITSFile.Height-1 do
	begin
	Line:=BMP.ScanLine[Y];
	for X:=0 to FITSFile.Width-1 do
		begin
                Pix:=round((Line[X].rgbtRed+Line[X].rgbtGreen+Line[X].rgbtBlue)/3);
		{Shift 7-bit FITS expects signed numbers}
		FITSFile.Data[X,Y]:=Pix shl 7;
		end;
	end;
end;




procedure AssignData(var FITSFile: TFITSFile; var Data: T2DFITSArray);
{Assign (copy) External FITS Data to FITS file}
var X,Y: integer;
var Pix: integer;
begin
FITSFile.Width:=Length(Data);
FITSFile.Height:=Length(Data[0]);
SetLength(FITSFile.Data,FITSFile.Width,FITSFile.Height);
for Y:=0 to FITSFile.Height-1 do
	begin
	for X:=0 to FITSFile.Width-1 do
		begin
		FITSFile.Data[X,Y]:=Data[X,Y];
		end;
	end;
end;




procedure InsertKeyItem(var Header: TFITSHeader; Key: TFITSKeyword); overload;
var Inx: integer;
begin
Inx:=IncrementHeader(Header);
Header[Inx]:=Key;
end;



procedure InsertKeyItem(var Header: TFITSHeader; Name: TNameString; HasValue: boolean; Value,Comment: string); overload;
var Key: TFITSKeyword;
begin
Key.Name:=Name;
Key.HasValue:=HasValue;
Key.Value:=Value;
Key.Comment:=Comment;
InsertKeyItem(Header,Key);
end;



procedure AddManditoryKeys(var Header: TFITSHeader; Width,Height,BitsPerPixel: integer);
begin
InsertKeyItem(Header,'SIMPLE',True,'T','/ Fits standard');
InsertKeyItem(Header,'BITPIX',True,IntToStr(BitsPerPixel),'/ Bits per pixel');
InsertKeyItem(Header,'NAXIS',True,'2','/ Number of axes');
InsertKeyItem(Header,'NAXIS1',True,IntToStr(Width),'/ Axis length');
InsertKeyItem(Header,'NAXIS2',True,IntToStr(Height),'/ Axis length');
end;



procedure CloseHeader(var Header: TFITSHeader);
var Inx: integer;
begin
Inx:=IncrementHeader(Header);
Header[Inx].Name:='END'; Header[Inx].HasValue:=False; Header[Inx].Value:=''; Header[Inx].Comment:='';
end;


{------------------ Routines to handle Optional Keywords ----------------------}


{List of optional user-supplied keywords}

var OptionList: TFITSHeader;




procedure AppendKeys(var Destination: TFITSHeader; Source: TFITSHeader);
{Used to combine user optional keyword with the manditory keywords}
var I: integer;
begin
for I:=0 to High(Source) do InsertKeyItem(Destination,Source[I]);
end;


procedure ClearOptionList;
{Clear user option list}
begin
ClearHeader(OptionList);
end;



procedure AddOptionalKeyword(Name: TNameString; HasValue: boolean; Value: string; Comment: string);
{Allows user to add additional keywords to the list}
begin
InsertKeyItem(OptionList,Name,HasValue,Value,Comment);
end;



{------------------ Main I/O Routines  ---------------------------------------}



procedure SaveFITSStream(Stream: TStream; var Bitmap: TBitmap); overload;
{Save Bitmap as FITS file}
var FITSFile: TFITSFile;
begin
ClearHeader(FITSFile.Header);
BMPToFITS(Bitmap,FITSFile);
AddManditoryKeys(FITSFile.Header,Bitmap.Width,Bitmap.Height,16);
AppendKeys(FITSFile.Header,OptionList);
CloseHeader(FITSFile.Header);
WriteHeaderBlock(Stream, FITSFile);
WriteDataBlock(Stream, FITSFile);
ClearOptionList;
end;




procedure SaveFITSFile(FileName: String; Bitmap: TBitmap); overload;
var Stream: TMemoryStream;
begin
Stream := TMemoryStream.Create;
SaveFITSStream(Stream,Bitmap);
Stream.SaveToFile(filename);
Stream.Free;
end;


{-----------------------------------------------------------------------------}


procedure SaveFITSStream(Stream: TStream; var Data: T2DFITSArray); overload;
{Save Raw integer array as FITS file}
var FITSFile: TFITSFile;
begin
ClearHeader(FITSFile.Header);
AssignData(FITSFile,Data);
AddManditoryKeys(FITSFile.Header,FITSFile.Width,FITSFile.Height,16);
AppendKeys(FITSFile.Header,OptionList);
CloseHeader(FITSFile.Header);
WriteHeaderBlock(Stream, FITSFile);
WriteDataBlock(Stream, FITSFile);
ClearOptionList;
end;





procedure SaveFITSFile(FileName: String; var Data: T2DFITSArray); overload;
var Stream: TMemoryStream;
begin
Stream := TMemoryStream.Create;
SaveFITSStream(Stream,Data);
Stream.SaveToFile(filename);
Stream.Free;
end;



{-----------------------------------------------------------------------------}


procedure SaveFITSStream(Stream: TStream; var FITSFile: TFITSFile); overload;
{Save a pre-existing FITS data block to FITS stream}
begin
WriteHeaderBlock(Stream, FITSFile);
WriteDataBlock(Stream, FITSFile);
end;



procedure SaveFITSFile(FileName: string; var FITSFile: TFITSFile); overload;
{Save a pre-existing FITS data block to FITS Files}
var Stream: TMemoryStream;
begin
Stream := TMemoryStream.Create;
SaveFITSStream(Stream,FITSFile);
Stream.SaveToFile(filename);
Stream.Free;
end;



initialization
ClearOptionList;
end.
