unit FilterLib;

interface

uses Types, Math, BitmapLib, HugeArry;

type TRealArray = TDoubleDynArray;

procedure CopyArray(var Output: TRealArray; Input: TRealArray);

procedure GaussianFilter1D(Radius: double; var RA: TRealArray);
procedure MedianFilter1D(Radius: integer; var RA: TRealArray);
procedure BilateralFilter1D(Radius: integer; IntSigma, DistSIgma: double; var RA: TRealArray);
procedure TotalVariationFilter1D(var Input, Output: TRealArray; const lambda: double);

implementation


procedure CopyArray(var Output: TRealArray; Input: TRealArray);
var I: integer;
begin
SetLength(Output,Length(Input));
for I:=0 to High(Input) do Output[I]:=Input[I];
end;


procedure GaussianFilter1D(Radius: double; var RA: TRealArray);
var I,J,Inx: integer;
var W,S: double;
var GK: TKernel;
var TH: THistoArray;
begin
SetLength(TH,Length(RA));

MakeGaussianKernel(GK, Radius, 1, 1);
for I:=0 to High(RA) do
	begin
	S:=0;
	for J := -GK.Size to GK.Size do
		begin
		W := GK.Weights[J];
		Inx:=EnsureRange(I - J,0,High(RA));
		S:=S + W * RA[Inx];
		end;
	TH[I]:=S;
	end;
for I:=0 to High(RA) do RA[I]:=TH[I];
end;




procedure MedianFilter1D(Radius: integer; var RA: TRealArray);
var I,J,Inx: integer;
var W,S: double;
var DA: THugeDouble;
begin
DA:=THugeDouble.Create(5);
try
for I:=0 to High(RA) do
	begin
	DA.Clear;
	for J := -Radius to Radius do
		begin
		Inx:=EnsureRange(I + J,0,High(RA));
		DA.Add(RA[Inx]);
		end;
	DA.Sort;
	RA[I]:=DA[DA.Count div 2];
	end;
finally DA.Free; end;
end;


procedure BilateralFilter1D(Radius: integer; IntSigma, DistSIgma: double; var RA: TRealArray);
var I,J,Inx: integer;
var Gi,GS,W,Wp,Filtered: double;
var H1: TRealArray;


	function Gaussian(x, sigma: double): double;
	begin
	Result:=exp(-(sqr(x))/(2 * sqr(sigma)));
	end;


begin
SetLength(H1, Length(RA));
for I:=0 to High(RA) do
	begin
	Filtered:=0;
	wP:=0;
	for J := -Radius to Radius do
		begin
		Inx:=I + J;
		if (Inx<0) or (Inx>High(RA)) then continue;
		{Get the gaussian weight of the difference of RA value}
		Gi:=Gaussian( RA[Inx] - RA[I], IntSigma);
		{Get the gaussian weight of the distance between items}
		GS:=Gaussian(Inx-I, DistSIgma);
		{Combine the two weights}
		W:=GI * GS;

		{Weight and sum the value}
		Filtered:=Filtered + RA[Inx] *  W;
		{Sum the weights}
		wP:=wP + w;
		end;

	{Normalize the pixel}
	Filtered:=Filtered/wP;
	H1[I]:=Filtered;
	end;
CopyArray(RA, H1);
end;




procedure TotalVariationFilter1D(var Input, Output: TRealArray; const lambda: double);
{function implements the "Total Variation Denoising" filter}
{lambda must be zero or positive. lambda=0 just copies input to output}
{Based on the paper "A Direct Algorithm for 1D Total Variation Denoising" by L. Condat}
var k, k0: integer;		{k: current sample location, k0: beginning of current segment}
var umin,umax: double; 		{u is the dual variable}
var vmin, vmax: double;		{bounds for the segment's value}
var kplus,kminus: integer; 	{last positions where umax:=-lambda, umin:=lambda, respectively}
var twolambda: double;		{auxiliary variable}
var minlambda: double;		{auxiliary variable}
begin
{to avoid invalid memory access to input[0]}
if Length(Input)<=0 then exit;
SetLength(Output,Length(Input));
k:=0; k0:=0;			{k: current sample location, k0: beginning of current segment}
umin:=lambda; umax:=-lambda;	{u is the dual variable}
vmin:=input[0]-lambda; vmax:=input[0]+lambda;	{bounds for the segment's value}
kplus:=0; kminus:=0; 	{last positions where umax:=-lambda, umin:=lambda, respectively}
twolambda:=2.0*lambda;	{auxiliary variable}
minlambda:=-lambda;	{auxiliary variable}
while true do
	begin
	{simple loop, the exit test is inside}
	while (k=Length(Input)-1) do
		begin
		{we use the right boundary condition}
		if (umin<0.0) then
			begin
			{vmin is too high -> negative jump necessary}
			repeat
				begin
				output[k0]:=vmin;
				Inc(k0);
				end
			until (k0>kminus);
                                umin:=lambda;
                                k:=k0;
                                kminus:=k;
                                vmin:=input[kminus];
			umax:=vmin+umin-vmax;
			end
		else if (umax>0.0) then
			begin
			{vmax is too low -> positive jump necessary}
			repeat
				begin
				output[k0]:=vmax;
				Inc(K0);
				end
			until k0>kplus;
			umax:=minlambda;
			k:=k0;
			kplus:=k0;
			vmax:=input[k0];
			umin:=vmax+umax-vmin;
			end
		else
			begin
			vmin:=vmin+umin/(k-k0+1);
			repeat
				begin
				output[k0]:=vmin;
				Inc(K0);
				end
			until k0>k;
			exit;
			end;
		end;
	umin:=umin+input[k+1]-vmin;
	if umin<minlambda then
		begin
		{negative jump necessary}
		repeat
			begin
			output[k0]:=vmin;
			Inc(K0);
			end
		until k0>kminus;
		k:=k0;
		kminus:=k0;
		kplus:=k0;
		vmin:=input[k0];
		vmax:=vmin+twolambda;
		umin:=lambda; umax:=minlambda;
		end
	else
		begin
		umax:=umax+input[k+1]-vmax;
		if umax>lambda then
			begin
			{positive jump necessary}
			repeat
				begin
				output[k0]:=vmax;
				Inc(K0);
				end
			until (k0>kplus);
			k:=k0;
			kminus:=k0;
			kplus:=k0;
			vmax:=input[k0];
			vmin:=vmax-twolambda;
			umin:=lambda; umax:=minlambda;
			end
		else
			begin
			{no jump necessary, we continue}
			Inc(k);
			if (umin>=lambda) then
				begin
				{update of vmin}
                                        kminus:=k;
				vmin:=vmin+(umin-lambda)/((K)-k0+1);
				umin:=lambda;
				end;
			if (umax<=minlambda) then
				begin
				{update of vmax}
                                        kplus:=k;
				vmax:=vmax+(umax+lambda)/((k)-k0+1);
				umax:=minlambda;
				end;

			end;
		end;
	end;
end;


end.
