unit GraphicsLib;

interface

uses Windows,Types,Graphics,Contnrs,Vector,IntVector,DSupport,HugeArry;


type TLineParam = record
 Color: TColor;
 Style: TPenStyle;
 end;

type TLineParams = record
 Line1,Line2,Line3: TLineParam;
 end;

{Pattern tag for dotted/dash lines.}

type TLinePattern = (lpSolid,lpDotted,lpDashed,lpDashDot,lpDashDotDot,lpFineDot,lpUserPattern);

{Patterns for common dotted/dashed line}
{These values change if the end cap parameters in ExtCreatePen change}

type TPatternArray = array of cardinal;

var DottedLine:  array [0..1] of cardinal = (3, 3);
var DashLine:  array [0..1] of cardinal = (18, 6);
var DashDotLine:  array [0..3] of cardinal = (9, 6, 3, 6);
var DashDotDotLine:  array [0..5] of cardinal = (9, 3, 3, 3, 3, 3);
var FineDotLine:  array [0..1] of cardinal = (1, 1);
var UserLine: TPatternArray;

type TVertAlign = (vaBottom,vaTop);
type THorizAlign = (haLeft,haRight);

function IsTrueTypeFont(Canvas: TCanvas): boolean;

procedure DrawMultiColorLine(Canvas: TCanvas; P1,P2: TPoint; Params: TLineParams);
procedure DrawArrowLine(Canvas: TCanvas; P1,P2: TPoint; StartArrow,EndArrow: boolean; Length,Width: integer); overload;
procedure DrawArrowLine(Canvas: TCanvas; P1,P2: TPoint; StartArrow,EndArrow: boolean; Color1,Color2: TColor; Length,Width: integer); overload;
procedure DrawPolarSlice(Canvas: TCanvas; A1,A2: double; Radius: integer; Center: TPoint; BorderColor,FillColor: TColor);
function PolarToCartesean(Radius,Angle: double; Center: TPoint): TPoint;
procedure OutputSuperScriptText(Canvas: TCanvas; S: string; X,Y: integer);
procedure DrawOutlineText(const Canvas: TCanvas; const X, Y: Integer; const Text: string);
procedure RightJustifyText(Canvas: TCanvas; X,Y: integer; Text: string);
procedure CenterText(Canvas: TCanvas; Width,Y: integer; Text: string);
procedure DrawTextBox(Canvas: TCanvas; Text: string; Pos: TPoint; HorizAlign: THorizAlign; VertAlign: TVertAlign);

procedure DrawSquareHandle(Canvas: TCanvas; P: TPoint; Color: TColor);
procedure DrawRoundHandle(Canvas: TCanvas; P: TPoint; Color: TColor);
procedure DrawCircle(Canvas: TCanvas; Pos: TPoint; Radius: integer; Fill: boolean; RadiusColor,FillColor: TColor);

procedure ClearCanvas(Canvas: TCanvas; C: TColor);
procedure Draw3DRect(Canvas: TCanvas; Rect: TRect; UL,BR: TColor);

procedure SaveCanvasProperties(Canvas: TCanvas);
procedure RestoreCanvasProperties(Canvas: TCanvas);

procedure SetDashedPen(Canvas: TCanvas; Scale: double; Pattern: array of cardinal); overload;
procedure SetDashedPen(Canvas: TCanvas; Scale: double; Pattern: TLinePattern); overload;


{There are two versions of this algorith, one for object-methods and one for procecedures}


type TPointMethod = procedure(X,Y: integer; UserData: pointer) of object;
type TPointProc = procedure(X,Y: integer; UserData: pointer);

procedure Drawline(StartX, StartY, EndX, EndY : integer; PointProc: TPointMethod; UserData: Pointer); overload;
procedure Drawline(StartX, StartY, EndX, EndY : integer; PointProc: TPointProc; UserData: Pointer); overload;



type TCanvasSaver = class(TObject)
 private
  BrushColor: TColor;
  BrushStyle: TBrushStyle;  PenStyle: TPenStyle;
  PenColor: TColor;
  PenMode: TPenMode;
 public
  procedure Save(Canvas: TCanvas);
  procedure Restore(Canvas: TCanvas);
 end;


var HandleSize: integer = 5;	{Global node Handle size}



{------------------------------------------------------------------------------}

type TPenInfo = class(TObject)
 private
  FWidth: Integer;
  FColor: TColor;
  FMode: TPenMode;
  FStyle: TPenStyle;
 protected
 public
  procedure SetPenParms(Color: TColor; Mode: TPenMode; Style: TPenStyle; Width: integer);
  procedure GetCanvasPen(Canvas: TCanvas);
  procedure SetCanvasPen(var Canvas: TCanvas);
  constructor Create;
  property Color: TColor read FColor write FColor default clBlack;
  property Mode: TPenMode read FMode write FMode default pmCopy;
  property Style: TPenStyle read FStyle write FStyle default psSolid;
  property Width: Integer read FWidth write FWidth default 1;
 end;


{------------------------------------------------------------------------------}

type TBrushInfo = class(TObject)
 private
  FColor: TColor;
  FStyle: TBrushStyle;
 protected
 public
  procedure SetBrushParms(Color: TColor; Style: TBrushStyle);
  constructor Create;
  procedure GetCanvasBrush(Canvas: TCanvas);
  procedure SetCanvasBrush(var Canvas: TCanvas);
  procedure Assign(BrushInfo: TBrushInfo);
  property Color: TColor read FColor write FColor;
  property Style: TBrushStyle read FStyle write FStyle;
 end;

{------------------------------------------------------------------------------}


type TBrushArray = class(TObjectList)
 private
  function GetInfoItem(Index: integer): TBrushInfo;
 protected
 public
  constructor Create;
  function Add: integer; overload;
  function Add(Color: TColor; Style: TBrushStyle): integer; overload;
  property Items[Index: integer]: TBrushInfo read GetInfoItem; default;
 end;


{-------------------------------------------------------------------------------}

type TPenArray = class(TObjectList)
 private
  function GetInfoItem(Index: integer): TPenInfo;
 protected
 public
  constructor Create;
  function Add: integer; overload;
  function Add(Color: TColor; Mode: TPenMode; Style: TPenStyle; Width: integer): integer; overload;
  property Items[Index: integer]: TPenInfo read GetInfoItem; default;
 end;


implementation

{Saved canvas properties}

var BrushColor,PenColor: TColor;
var BrushStyle: TBrushStyle;
var PenMode: TPenMode;
var PenStyle: TPenStyle;



procedure SaveCanvasProperties(Canvas: TCanvas);
{Save colors and attributes of canvas}
begin
BrushColor:=Canvas.Brush.Color;
PenColor:=Canvas.Pen.Color;
BrushStyle:=Canvas.Brush.Style;
PenMode:=Canvas.Pen.Mode;
PenStyle:=Canvas.Pen.Style;
end;



procedure RestoreCanvasProperties(Canvas: TCanvas);
{Restore colors and attributes of canvas}
begin
Canvas.Brush.Color:=BrushColor;
Canvas.Pen.Color:=PenColor;
Canvas.Brush.Style:=BrushStyle;
Canvas.Pen.Mode:=PenMode;
Canvas.Pen.Style:=PenStyle;
end;

function PolarToCartesean(Radius,Angle: double; Center: TPoint): TPoint;
{Convert Polar to cartesian center around point}
begin
Angle:=(Angle-90) * Degree2Radian;
Result.X:=round(Radius*cos(Angle)+Center.X);
Result.Y:=round(Radius*sin(Angle)+Center.Y);
end;



procedure DrawPolarSlice(Canvas: TCanvas; A1,A2: double; Radius: integer; Center: TPoint; BorderColor,FillColor: TColor);
{Draw pie-shaped slice between Angle A1 and A2 of specified Radius around center point}
var P1,P2: TPoint;
var Rect: TRect;
begin
Rect.TopLeft:=Center;
Rect.BottomRight:=Center;
InflateRect(Rect,Radius,Radius);
Canvas.Pen.Color:=BorderColor;
Canvas.Brush.Color:=FillColor;
P1:=PolarToCartesean(Radius,A2,Center);
P2:=PolarToCartesean(Radius,A1,Center);
if (P1.X=P2.X) and (P1.Y=P2.Y) then
	begin
	Canvas.MoveTo(Center.X,Center.Y);
	Canvas.LineTo(P2.X,P2.Y);
	end
else Canvas.Pie(Rect.Left,Rect.Top,Rect.Right,Rect.Bottom, P1.X, P1.Y, P2.X, P2.Y);
end;


{------------------------------------------------------------------------------}



procedure DrawSquareHandle(Canvas: TCanvas; P: TPoint; Color: TColor);
{Draw one square node handles}
var R: TRect;
begin
R:=Rect(P.X-HandleSize,P.Y-HandleSize,P.X+HandleSize-1,P.Y+HandleSize-1);
Canvas.Brush.Style:=bsSolid;
Canvas.Brush.Color:=Color;
Canvas.Pen.Style:=psSolid;
Canvas.Pen.Color:=clWhite;
Canvas.Pen.Mode:=pmCopy;
Canvas.Rectangle(R.Left,R.Top,R.Right,R.Bottom);
end;




procedure DrawRoundHandle(Canvas: TCanvas; P: TPoint; Color: TColor);
{Draw one round node handles}
var R: TRect;
begin
R:=Rect(P.X-HandleSize,P.Y-HandleSize,P.X+HandleSize,P.Y+HandleSize);
Canvas.Brush.Style:=bsSolid;
Canvas.Brush.Color:=Color;
Canvas.Pen.Style:=psSolid;
Canvas.Pen.Color:=clWhite;
Canvas.Pen.Mode:=pmCopy;
Canvas.Ellipse(R);
end;





procedure DrawCircle(Canvas: TCanvas; Pos: TPoint; Radius: integer; Fill: boolean; RadiusColor,FillColor: TColor);
{Draw circle of specified Pos, with specified Radius, and Color}
var R: TRect;
begin
if Radius<1 then Canvas.Pixels[Pos.X,Pos.Y]:=RadiusColor
else
	begin
	Canvas.Pen.Color:=RadiusColor;
	Canvas.Brush.Color:=FillColor;
	if Fill then Canvas.Brush.Style:=bsSolid
	else Canvas.Brush.Style:=bsClear;
	R:=Rect(Pos.X,Pos.Y,Pos.X,Pos.Y);
	InflateRect(R,Radius,Radius);
	Canvas.Ellipse(R);
	end;
end;


{------------------------------------------------------------------------------}


function IsTrueTypeFont(Canvas: TCanvas): boolean;
var TM: TTextMetric;
begin
GetTextMetrics(Canvas.Handle,TM);
Result:=(TM.tmPitchAndFamily and TMPF_TRUETYPE)<>0;
end;


procedure DrawArrowHead(Canvas: TCanvas; P1,P2: T2DVector; Len,Width: double);
{Draw arrow of the specified Length and width}
var UnitVector,Base,VLen,VWidth: T2DVector;
var PG1,PG2,PG3: TPoint;
begin
UnitVector:=UnitVector2D(VectorSubtract2D(P2,P1));
VLen:=ScalarProduct2D(UnitVector,Len);
VWidth:=ScalarProduct2D(UnitVector,Width);
Base:=VectorSubtract2D(P2,VLen);
PG1:=Vector2DtoPoint(P2);
PG2:=Point(trunc(Base.X + VWidth.Y), trunc(Base.Y - VWidth.X));
PG3:=Point(trunc(Base.X - VWidth.Y), trunc(Base.Y + VWidth.X));
Canvas.Polygon([PG1,PG2,PG3]);
end;



procedure DrawArrowLine(Canvas: TCanvas; P1,P2: TPoint; StartArrow,EndArrow: boolean; Color1,Color2: TColor; Length,Width: integer);
{Draw a line with arrow on the end from P1 to P2. StartArrow and EndArrow control which ends get arrows.}
{Shaft Color = Color1, Arrow Color = Color2. Length and Width control size of arrow.}
var PV1,PV2: T2DVector;
begin
Canvas.Pen.Color:=Color1;
Canvas.MoveTo(P1.X,P1.Y);
Canvas.LineTo(P2.X,P2.Y);
PV1:=MakeVector2D(P1.X,P1.Y);
PV2:=MakeVector2D(P2.X,P2.Y);
Canvas.Pen.Color:=Color2;
if EndArrow then DrawArrowHead(Canvas,PV1,PV2,Length,Width);
if StartArrow then DrawArrowHead(Canvas,PV2,PV1,Length,Width);
end;


procedure DrawArrowLine(Canvas: TCanvas; P1,P2: TPoint; StartArrow,EndArrow: boolean; Length,Width: integer);
{Draw a line with arrow on the end from P1 to P2, StartArrow and EndArrow control which ends get arrows}
{Length and Width control size of arrow. Color comes from current canvas's pen Color}
begin
DrawArrowLine(Canvas,P1,P2,StartArrow,EndArrow,Canvas.Pen.Color,Canvas.Pen.Color,Length,Width);
end;


procedure OutputSuperScriptText(Canvas: TCanvas; S: string; X,Y: integer);
{Output text after "^" as super script - "^" is ignored}
var I,H: integer;
var SuperScript: boolean;
begin
H:=Canvas.TextHeight(S);
SuperScript:=False;
for I:=1 to Length(S) do
	begin
	if S[I]='^' then SuperScript:=True
	else
		begin
		if SuperScript then Canvas.TextOut(X,Y-(H div 2),S[I])
		else Canvas.TextOut(X,Y,S[I]);
		X:=Canvas.PenPos.X;
		end;
	end;
end;




procedure DrawOutlineText(const Canvas: TCanvas; const X, Y: Integer; const Text: string);
{ Draw Outlined text. Color scheme is}
{ Outer Text = Pen.Color }
{ Inner Text = Font.Color }
{ Background = Brush.Color }
var OldBkMode: Integer;
var OB: TColor;
begin
{Draw the background of the text using current brush color}
BeginPath(Canvas.Handle);
Canvas.TextOut(X, Y, Text);
EndPath(Canvas.Handle);
FillPath(Canvas.Handle);

{Draw the body of the text and outline}
OB:=Canvas.Brush.Color;
{Use the font color for the body}
Canvas.Brush.Color:=Canvas.Font.Color;
{Make back ground transparent because the text drawn on top of the background}
OldBkMode := SetBkMode(Canvas.Handle, TRANSPARENT);
BeginPath(Canvas.Handle);
Canvas.TextOut(X, Y, Text);
EndPath(Canvas.Handle);
{The pen color strokes the border }
StrokeAndFillPath(Canvas.Handle);
{Restore modes and colors}
SetBkMode(Canvas.Handle, OldBkMode);
Canvas.Brush.Color:=OB;
end;


procedure RightJustifyText(Canvas: TCanvas; X,Y: integer; Text: string);
{Right justify text, starting from X position}
var X1: integer;
begin
X1:=X-Canvas.TextWidth(Text);
Canvas.TextOut(X1,Y,Text);
end;


procedure CenterText(Canvas: TCanvas; Width,Y: integer; Text: string);
{Center justify text on a line with a specified width}
var X1: integer;
begin
X1:=(Width-Canvas.TextWidth(Text)) div 2;
Canvas.TextOut(X1,Y,Text);
end;



procedure DrawTextBox(Canvas: TCanvas; Text: string; Pos: TPoint; HorizAlign: THorizAlign; VertAlign: TVertAlign);
{Draw Text in box at specified position, using specified alignment }
{You must specify the box color, line color and transparency. e.g.: }
{   Canvas.Brush.Color := clWhite; }
{   Canvas.Pen.Color:=clBlack; }
{   Canvas.Brush.Style:=bsClear; }
var R: TRect;
var Size,Off: TPoint;
const Border = 2;
begin
{Create Base Rectangle}
Size:=TPoint(Canvas.TextExtent(Text));
R:=Rect(0,0,Size.X+Border*2,Size.Y+Border*2);
{Offset according to horizontal alignment}
case HorizAlign of
 haLeft: OffsetRect(R,0,0);
 haright: OffsetRect(R,-(Size.X+Border+Canvas.Pen.Width),0);
 end;
{Offset according to Vertical alignment}
case VertAlign of
 vaTop: OffsetRect(R,0,0);
 vaBottom: OffsetRect(R,0,-(Size.Y+Border+Canvas.Pen.Width));
end;
{Moved to specified position}
OffsetRect(R,Pos.X,Pos.Y);
{Draw using current pen color and brush color}
Canvas.Rectangle(R);
{This is base alignment - It doesn't change}
SetTextAlign(Canvas.Handle, TA_Top or TA_Left);
{Draw text transparent on top of rectangle}
Canvas.Brush.Style:=bsClear;
Canvas.TextOut(R.Left+Border,R.Top+Border,Text);
end;




{------------------------------------------------------------------------------}



procedure DrawMultiColorLine(Canvas: TCanvas; P1,P2: TPoint; Params: TLineParams);
var D: T2DVector;
var R: double;
var I: TPoint;
begin
D:=VectorSubtract2D(PointToVector2D(P2),PointToVector2D(P1));
R:=D.Y/D.X;
if abs(R)>1 then I:=Point(1,0) else I:=Point(0,1);
Canvas.Pen.Color:=Params.Line1.Color;
Canvas.Pen.Style:=Params.Line1.Style;
Canvas.MoveTo(P1.X-I.X,P1.Y-I.Y);
Canvas.LineTo(P2.X-I.X,P2.Y-I.Y);
Canvas.Pen.Color:=Params.Line2.Color;
Canvas.Pen.Style:=Params.Line2.Style;
Canvas.MoveTo(P1.X-I.X,P1.Y-I.Y);
Canvas.MoveTo(P1.X,P1.Y);
Canvas.LineTo(P2.X,P2.Y);
Canvas.Pen.Color:=Params.Line3.Color;
Canvas.Pen.Style:=Params.Line3.Style;
Canvas.MoveTo(P1.X-I.X,P1.Y-I.Y);
Canvas.MoveTo(P1.X+I.X,P1.Y+I.Y);
Canvas.LineTo(P2.X+I.X,P2.Y+I.Y);
end;


{------------------------------------------------------------------------------}


procedure TCanvasSaver.Save(Canvas: TCanvas);
begin
BrushStyle:=Canvas.Brush.Style;
BrushColor:=Canvas.Brush.Color;
PenStyle:=Canvas.Pen.Style;
PenColor:=Canvas.Pen.Color;
PenMode:=Canvas.Pen.Mode;
end;



procedure TCanvasSaver.Restore(Canvas: TCanvas);
begin
Canvas.Brush.Style:=BrushStyle;
Canvas.Brush.Color:=BrushColor;
Canvas.Pen.Style:=PenStyle;
Canvas.Pen.Color:=PenColor;
Canvas.Pen.Mode:=PenMode;
end;




procedure ClearCanvas(Canvas: TCanvas; C: TColor);
{Clear bitmap image }
begin
Canvas.Brush.Color:=C;
Canvas.FillRect(Canvas.ClipRect);
end;


procedure Draw3DRect(Canvas: TCanvas; Rect: TRect; UL,BR: TColor);
{Draw rectangle to colors to simulate 3D appearance}
begin
Canvas.Pen.Color:=UL;
Canvas.Rectangle(Rect);
Canvas.Pen.Color:=BR;
Canvas.MoveTo(Rect.Left,Rect.Bottom-1);
Canvas.LineTo(Rect.Right-1,Rect.Bottom-1);
Canvas.MoveTo(Rect.Right-1,Rect.Bottom-1);
Canvas.LineTo(Rect.Right-1,Rect.Top-1);
end;

{------------------------------------------------------------------------------}

{
Mapping the pattern to a particular device is tricky and not well documented.
According to Microsoft, the the pattern is in "Style Units," which is about 25
pixels per inch. When I designed the patterns, I did it on a 96 DPI screen.
Programs may put the output into different mapping modes. For example, Compass
uses MM_HIENGLISH mode, which puts the units into 1/1000 of an inch.
Also, when the dotted lines are sent to a printer, the printer may have
a different number of DPIs. For this reason, the program has to take into
account both issues. That's what MapPttern attempts to do.
}

procedure MapPattern(Canvas: TCanvas; Scale: double; var Pattern: array of cardinal);
{Scale pattern to match map mode}
var Pts: array of TPoint;
var ZP: TPoint;
var I: integer;
var PPI: integer;
begin
{Get origin of canvas according to current mapping mode}
ZP:=Point(0,0);
DPtoLP(Canvas.Handle,ZP,1);
{Convert Pattern points into mapping mode units}
SetLength(Pts,Length(Pattern));
for I:=0 to High(Pts) do Pts[I]:=Point(Pattern[I],Pattern[I]);
DPtoLP(Canvas.Handle,Pts[0],Length(Pattern));
for I:=0 to High(Pts) do Pattern[I]:=Round(IntVectorAbs(IntVectorSubtract(Pts[I],ZP))*Scale);
{Scale according pixels per inch of device}
PPI:=GetDeviceCaps(Canvas.Handle,LOGPIXELSX);
for I:=0 to High(Pattern) do Pattern[I]:=MulDiv(Pattern[I],PPI,96);
end;





procedure SetDashedPen(Canvas: TCanvas; Scale: double; Pattern: array of cardinal); overload;
{Set canvas to use numerical do/dash pattern}
var lb: TLogBrush;
begin
lb.lbStyle := BS_SOLID;
lb.lbColor := Canvas.Pen.Color;
MapPattern(Canvas, Scale, Pattern);
Canvas.Pen.Handle:=ExtCreatePen(PS_GEOMETRIC or PS_USERSTYLE or PS_ENDCAP_FLAT, Canvas.Pen.Width, lb, length(pattern), @pattern);
end;




procedure SetDashedPen(Canvas: TCanvas; Scale: double; Pattern: TLinePattern); overload;
{Set canvas to use specified dot/dash pattern}
begin
case Pattern of
 lpSolid: Canvas.Pen.Style:=psSolid;
 lpDotted:
	begin
	if Canvas.Pen.Width=1 then Canvas.Pen.Style:=psDot
	else SetDashedPen(Canvas,Scale,DottedLine);
	end;
 lpDashed:
	begin
	if Canvas.Pen.Width=1 then Canvas.Pen.Style:=psDash
	else SetDashedPen(Canvas,Scale,DashLine);
	end;
 lpDashDot:
	begin
	if Canvas.Pen.Width=1 then Canvas.Pen.Style:=psDashDot
	else SetDashedPen(Canvas,Scale,DashDotLine);
	end;
 lpDashDotDot:
	begin
	if Canvas.Pen.Width=1 then Canvas.Pen.Style:=psDashDotDot
	else SetDashedPen(Canvas,Scale,DashDotDotLine);
	end;
 lpFineDot: SetDashedPen(Canvas,Scale,FineDotLine);
 lpUserPattern: SetDashedPen(Canvas,Scale,UserLine);
 end;
end;


{------------------------------------------------------------------------------}


procedure Drawline(StartX, StartY, EndX, EndY : integer; PointProc: TPointMethod; UserData: Pointer);
{ Bresenham's Line Algorithm. This one is for object metheds}
var i:  integer;		{ loop index}
var x,y: integer;		{ current x and y coordinates}
var DX, DY: integer;		{ displacements in x and y}
var d: integer;			{ decision variable}
var DiagInc: integer;		{ d's increment for diagonal steps}
var DiagX: integer;		{ diagonal x step for next pixel}
var DiagY: integer;		{ diagonal y step for next pixel}
var NonDiagDX: integer;		{ nondiagonal x step for next pixel}
var NonDiagDY: integer;		{ nondiagonal y step for next pixel}
var NonDiagInc: integer;	{ d's increment for nondiagonal steps}
var SwapTmp: integer;		{ temporary variable for swap}
begin
x := StartX;			{ line starting point}
y := StartY;
{ Determine drawing direction and step to the next pixel.}
DX := EndX - StartX;		{ difference in x dimension}
DY := EndY - StartY;		{ difference in y dimension}
{ Determine whether end point lies to right or left of start point.}
if DX < 0 then			{ drawing towards smaller x values?}begin
	DX := -DX;		{ make 'a' positive}
	DiagX := -1
	end
else DiagX := 1;

{ Determine whether end point lies above or below start point.}
if DY < 0 then			{ drawing towards smaller x values?}
	begin
	DY := -DY;		{ make 'a' positive}
	DiagY := -1
	end
else DiagY := 1;
{ Identify octant containing end point.}
if DX < DY then
	begin
	SwapTmp := DX;
	DX := DY;
	DY := SwapTmp;
	NonDiagDX := 0;
	NonDiagDY := DiagY
	end
else
	begin
	NonDiagDX := DiagX;
	NonDiagDY := 0
	end;
d := DY + DY - DX;		{ initial value for d is 2*DY - a}
NonDiagInc := DY + DY;		{ set initial d increment values}
DiagInc    := DY + DY - DX - DX;
for i := 0 to DX do
	begin			{ draw the a+1 pixels}
	PointProc(x, y, UserData);
	if d < 0 then		{ is midpoint above the line?}
		begin		{ step nondiagonally}
		x := x + NonDiagDX;
		y := y + NonDiagDY;
		d := d + NonDiagInc	{ update decision variable}
		end
	else
		begin			{ midpoint is above the line; step diagonally}
		x := x + DiagX;
		y := y + DiagY;
		d := d + DiagInc;
		end;
	end;
end;



procedure Drawline(StartX, StartY, EndX, EndY : integer; PointProc: TPointProc; UserData: Pointer);
{ Bresenham's Line Algorithm. This is version is for ordinary procedures}
var i:  integer;		{ loop index}
var x,y: integer;		{ current x and y coordinates}
var DX, DY: integer;		{ displacements in x and y}
var d: integer;			{ decision variable}
var DiagInc: integer;		{ d's increment for diagonal steps}
var DiagX: integer;		{ diagonal x step for next pixel}
var DiagY: integer;		{ diagonal y step for next pixel}
var NonDiagDX: integer;		{ nondiagonal x step for next pixel}
var NonDiagDY: integer;		{ nondiagonal y step for next pixel}
var NonDiagInc: integer;	{ d's increment for nondiagonal steps}
var SwapTmp: integer;		{ temporary variable for swap}
begin
x := StartX;			{ line starting point}
y := StartY;
{ Determine drawing direction and step to the next pixel.}
DX := EndX - StartX;		{ difference in x dimension}
DY := EndY - StartY;		{ difference in y dimension}
{ Determine whether end point lies to right or left of start point.}
if DX < 0 then			{ drawing towards smaller x values?}begin
	DX := -DX;		{ make 'a' positive}
	DiagX := -1
	end
else DiagX := 1;

{ Determine whether end point lies above or below start point.}
if DY < 0 then			{ drawing towards smaller x values?}
	begin
	DY := -DY;		{ make 'a' positive}
	DiagY := -1
	end
else DiagY := 1;
{ Identify octant containing end point.}
if DX < DY then
	begin
	SwapTmp := DX;
	DX := DY;
	DY := SwapTmp;
	NonDiagDX := 0;
	NonDiagDY := DiagY
	end
else
	begin
	NonDiagDX := DiagX;
	NonDiagDY := 0
	end;
d := DY + DY - DX;		{ initial value for d is 2*DY - a}
NonDiagInc := DY + DY;		{ set initial d increment values}
DiagInc    := DY + DY - DX - DX;

for i := 0 to DX do
	begin			{ draw the a+1 pixels}
	PointProc(x, y, UserData);
	if d < 0 then		{ is midpoint above the line?}
		begin		{ step nondiagonally}
		x := x + NonDiagDX;
		y := y + NonDiagDY;
		d := d + NonDiagInc	{ update decision variable}
		end
	else
		begin			{ midpoint is above the line; step diagonally}
		x := x + DiagX;
		y := y + DiagY;
		d := d + DiagInc;
		end;
	end;
end;


procedure IntializeGraphics;
const Default: array [0..1] of cardinal = (3, 4);
var I: integer;
begin
SetLength(UserLine,Length(Default));
for I:=0 to High(UserLine) do UserLine[I]:=Default[I];
end;



{ TPenData }

constructor TPenInfo.Create;
begin
FColor:=clBlack;
FMode:=pmCopy;
FStyle:=psSolid;
FWidth:=1;
end;

procedure TPenInfo.GetCanvasPen(Canvas: TCanvas);
begin
FColor:=Canvas.Pen.Color;
FMode:=Canvas.Pen.Mode;
FStyle:=Canvas.Pen.Style;
FWidth:=Canvas.Pen.Width;
end;


procedure TPenInfo.SetCanvasPen(var Canvas: TCanvas);
begin
Canvas.Pen.Color:=FColor;
Canvas.Pen.Mode:=FMode;
Canvas.Pen.Style:=FStyle;
Canvas.Pen.Width:=FWidth;
end;



procedure TPenInfo.SetPenParms(Color: TColor; Mode: TPenMode; Style: TPenStyle; Width: integer);
begin
FColor:=Color;
FMode:=Mode;
FStyle:=Style;
FWidth:=Width;
end;

{ TBrushInfo }

constructor TBrushInfo.Create;
begin
FColor:=clWhite;
FStyle:=bsSolid;
end;

procedure TBrushInfo.GetCanvasBrush(Canvas: TCanvas);
begin
FColor:=Canvas.Brush.Color;
FStyle:=Canvas.Brush.Style;
end;


procedure TBrushInfo.SetCanvasBrush(var Canvas: TCanvas);
begin
Canvas.Brush.Color:=FColor;
Canvas.Brush.Style:=FStyle;
end;



procedure TBrushInfo.SetBrushParms(Color: TColor; Style: TBrushStyle);
begin
FColor:=Color;
FStyle:=Style;
end;



procedure TBrushInfo.Assign(BrushInfo: TBrushInfo);
begin
Self.SetBrushParms(BrushInfo.Color,BrushInfo.Style);
end;

{ TBrushArray }

function TBrushArray.Add: integer;
begin
Result:=inherited Add(TBrushInfo.Create);
end;



function TBrushArray.Add(Color: TColor; Style: TBrushStyle): integer;
begin
Result:=Add;
TBrushInfo(Self[Result]).SetBrushParms(Color,Style);
end;

constructor TBrushArray.Create;
begin
inherited Create(True);
end;


function TBrushArray.GetInfoItem(Index: integer): TBrushInfo;
begin
Result:=TBrushInfo(inherited Items[Index]);
end;



{ TPenArray }

function TPenArray.Add: integer;
begin
Result:=inherited Add(TPenInfo.Create);
end;



function TPenArray.Add(Color: TColor; Mode: TPenMode; Style: TPenStyle; Width: integer): integer;
begin
Result:=Add;
TPenInfo(Self[Result]).SetPenParms(Color,Mode,Style,Width);
end;

constructor TPenArray.Create;
begin
inherited Create(True);
end;


function TPenArray.GetInfoItem(Index: integer): TPenInfo;
begin
Result:=TPenInfo(inherited Items[Index]);
end;

initialization
IntializeGraphics;
end.
