unit GridLibrary;

interface

uses Math,DSupport,Vector,HugeArry;



type TGridPoint = record
  P: T2DVector;
  Value: double;
  end;
type PGridPoint = ^TGridPoint;


type TCorners = record
 TopLeft,TopRight: PGridPoint;
 BottomLeft,BottomRight: PGridPoint;
 end;


type TGridCell = record
  Corners: TCorners;
  Points: array of integer;
  end;

type TGridArray = array of array of TGridPoint;
type TCellArray = array of array of TGridCell;

{Object to concvert randomly scattered data to regular grid}

type TGridder = class(TObject)
 private
  FGridMin: T2DVector;
  FGridMax: T2DVector;
  FValueMax: double;
  FValueMin: double;
  FDistPower: double;
  FEqualizeXY: boolean;
    FWrapY: boolean;
    FWrapX: boolean;
    FWrapRangeY: double;
    FWrapRangeX: double;
    FInterpolateMax: double;
    FInterpolateMin: double;
  procedure SetDistPower(const Value: double);
  function EqualizePoint(Pos: T2DVector): T2DVector;
  function GetDistance(P1, P2: T2DVector): double;
 protected
 public
  GridArray: TGridArray;
  CellArray: TCellArray;
  ScatterPoints: THuge3DVector;
  constructor Create;
  destructor Destroy; override;
  procedure Clear;
  procedure SetupGrid;
  function GridIterpolate(PX, PY: double): double;
  procedure AddPoint(X,Y,Z: double); overload;
  procedure AddPoint(V: T3DVector); overload;
  procedure SetGridSize(X,Y: integer);
  function GetGridPointValue(X,Y: integer): double;
  procedure GridData;
  property GridMin: T2DVector read FGridMin write FGridMin;
  property GridMax: T2DVector read FGridMax write FGridMax;
  property ValueMin: double read FValueMin;
  property ValueMax: double read FValueMax;
  property InterpolateMin: double read FInterpolateMin;
  property InterpolateMax: double read FInterpolateMax;
  property DistPower: double read FDistPower write SetDistPower;
  property EqualizeXY: boolean read FEqualizeXY write FEqualizeXY default False;
  property WrapRangeX: double read FWrapRangeX write FWrapRangeX;
  property WrapRangeY: double read FWrapRangeY write FWrapRangeY;
  property WrapX: boolean read FWrapX write FWrapX default False;
  property WrapY: boolean read FWrapY write FWrapY default False;
 end;

implementation



{ TGridder }

procedure TGridder.SetupGrid;
var Range,Offset,Step: T2DVector;
var CX, CY,CntX,CntY: Integer;
var I: integer;
var P: T3DVector;
begin
{Calculate range}
CntX:=High(GridArray);
CntY:=High(GridArray[0]);
Range:=VectorSubtract2D(FGridMax, FGridMin);
{Range should never be zero}
if abs(Range.X)<1E-9 then Range.X:=1E-9;
if abs(Range.Y)<1E-9 then Range.Y:=1E-9;
Offset:=FGridMin;
Step:=VectorDivide2D(Range,MakeVector2D(CntX,CntY));

{ Create the grid positions}
for cx:= 0 to High(GridArray) do
 for cy:= 0 to High(GridArray[0]) do
	begin
	GridArray[cx, cy].P.X:= cx*Step.X + Offset.X;		{ grid from -1 to 1}
	GridArray[cx, cy].P.Y:= cy*Step.Y + Offset.Y;	{ grid from -1 to 1}
	end;

{ Create the cells. Each cell points to 4 grid points}
for cx:= 0 to High(CellArray) do
 for cy:= 0 to High(CellArray[0]) do
	begin
	SetLength(CellArray[cx,cy].Points,0);
	CellArray[cx,cy].Corners.TopLeft:=     @GridArray[CX,  CY  ];
	CellArray[cx,cy].Corners.TopRight:=    @GridArray[CX+1,CY  ];
	CellArray[cx,cy].Corners.BottomLeft:=  @GridArray[CX,  CY+1];
	CellArray[cx,cy].Corners.BottomRight:= @GridArray[CX+1,CY+1];
	end;
{Use this step size to make the range inclusive}
Step:=VectorDivide2D(ScalarProduct2D(Range,1.00001),MakeVector2D(CntX,CntY));
{Put all scatter points in the appropriate cell}
for I:=0 to ScatterPoints.Count-1 do
	begin
	P:=ScatterPoints[I];
	CX:=trunc((P.X-Offset.X)/Step.X);
	CY:=trunc((P.Y-Offset.Y)/Step.Y);
	SetLength(CellArray[cx,cy].Points,Length(CellArray[cx,cy].Points)+1);
	CellArray[cx,cy].Points[High(CellArray[cx,cy].Points)]:=I;
	end;
end;



function TGridder.EqualizePoint(Pos: T2DVector): T2DVector;
{Scale X and Y so they are in the 0 to 1 range}
{This equalizes the influence of X and Y coordinates}
var Range: T2DVector;
begin
Range:=VectorSubtract2D(FGridMax, FGridMin);
Result:=VectorSubtract2D(Pos,FGridMin);
Result:=VectorDivide2D(Result,Range);
end;


function TGridder.GetDistance(P1,P2: T2DVector): double;
{Get the distance between the two points }
{Handling wrapping around if enabled}
var D: T2DVector;
var WX,WY: double;
begin
WX:=PickFloat(EqualizeXY,1,WrapRangeX);
WY:=PickFloat(EqualizeXY,1,WrapRangeY);
D.X:=PickFloat(WrapX,WrappedDifference(P1.X,P2.X,WX), P1.X-P2.X);
D.Y:=PickFloat(WrapY,WrappedDifference(P1.Y,P2.Y,WY), P1.Y-P2.Y);
Result:=VectorAbs2D(D);
end;




function TGridder.GetGridPointValue(X,Y: integer): double;
{Get the Z value for a specific grid point}
{Value is interpolated from the randomly scattered points}
var I: integer;
var Dist,Sum1,Sum2,D,Z: double;
var Pos,Pos2: T2DVector;
begin
Pos:=GridArray[X, Y].P;
if EqualizeXY then Pos:=EqualizePoint(Pos);
Sum1:=0; Sum2:=0;
{Step through all randomly scattered points}
for I:=0 to ScatterPoints.Count-1 do
	begin
	Pos2:=Vector3DTo2D(ScatterPoints[I]);
	Z:=ScatterPoints[I].Z;
	if EqualizeXY then Pos2:=EqualizePoint(Pos2);
	{Calculate the inflence of each point based on distance}
	Dist:=GetDistance(Pos2,Pos);
	{If distance=0 then it is right on top of scatter point}
	{so use the use the unweighted scatter value}
	if Dist=0 then begin Result:=Z; exit; end;
	{Calculate the weight and apply to sums}
	D:=Power(Dist,FDistPower);
	Sum1:=Sum1+Z/D;
	Sum2:=Sum2+1/D;
	end;
{Calculate final result}
if Sum2=0 then Result:=0
else Result:=Sum1/Sum2;
end;




procedure TGridder.GridData;
{Interpolate the scattered data to assign a value to each position}
var CX, CY,CntX,CntY: Integer;
var P: T2DVector;
begin
FInterpolateMin:=9e99;
FInterpolateMax:=-9e99;

{ Step through all grid positions}
for cy:= 0 to High(GridArray[0]) do
 for cx:= 0 to High(GridArray) do
	begin
	{Get interpolated value for a specific grid position}
	GridArray[CX,CY].Value:=GetGridPointValue(CX,CY);
	FInterpolateMin:=Min(FInterpolateMin,GridArray[CX,CY].Value);
	FInterpolateMax:=Max(FInterpolateMax,GridArray[CX,CY].Value);
	end;
end;




procedure TGridder.SetGridSize(X, Y: integer);
{Set grid size}
begin
SetLength(GridArray,X+1,Y+1);
SetLength(CellArray,X,Y);
SetupGrid;
end;



procedure TGridder.SetDistPower(const Value: double);
{Set interpolation distance power}
begin
if Value<0.0001 then exit;
FDistPower := Value;
end;




constructor TGridder.Create;
begin
ScatterPoints:=THuge3DVector.Create(5);
SetGridSize(10,10);
Clear;
FDistPower:=2;
FEqualizeXY:=False;
FWrapX:=False;
FWrapY:=False;
FWrapRangeX:=360;
FWrapRangeY:=360;
end;



destructor TGridder.Destroy;
begin
ScatterPoints.Free;
inherited;
end;



procedure TGridder.AddPoint(X, Y, Z: double);
{Add scatter point value}
begin
FGridMin:=GetMinimumVector2D(FGridMin,MakeVector2D(X,Y));
FGridMax:=GetMaximumVector2D(FGridMax,MakeVector2D(X,Y));
ScatterPoints.Add(MakeVector3D(X, Y, Z));
FValueMin:=Min(FValueMin,Z);
FValueMax:=Max(FValueMax,Z);
end;



procedure TGridder.AddPoint(V: T3DVector);
{Add scatter point value}
begin
AddPoint(V.X,V.Y,V.Z);
end;


procedure TGridder.Clear;
{Clear all scatter point data }
begin
ScatterPoints.Clear;
FGridMin:=MakeVector2D(9e99,9e99);
FGridMax:=MakeVector2D(-9e99,-9e99);
FValueMin:=9e99;
FValueMax:=-9e99;
FInterpolateMin:=9e99;
FInterpolateMax:=-9e99;
SetupGrid;
end;



function TGridder.GridIterpolate(PX,PY: double): double;
{Using gridded data find the value of a point somewhere in the grid}
{Requires that the subroutine - GridData has been called previously }
var Range,Offset: T2DVector;
var X2,Y2,FX,FY,P1,P2: double;
var InxX,InxY: integer;
var Cell: TGridCell;
begin
X2:=Length(CellArray)*PX;
Y2:=Length(CellArray[0])*PY;
InxX:=Trunc(X2);
InxY:=Trunc(Y2);
FX:=Frac(X2);
FY:=Frac(Y2);
Cell:=CellArray[InxX,InxY];
P1:=LinInterpolate(Cell.Corners.TopLeft.Value,Cell.Corners.TopRight.Value,FX);
P2:=LinInterpolate(Cell.Corners.BottomLeft.Value,Cell.Corners.BottomRight.Value,FX);
Result:=LinInterpolate(P1,P2,FY);
end;


end.
