unit LBMLib;

interface

uses Windows,Classes,SysUtils,Graphics,StreamLib,BitmapLib;

{Data structure used to define the details of a LBM file}

type TFormType = (ftUnknown,ftILBM,ftPBM);

type TLBMInfo = record
 HaveBmhd: boolean;
 FormSize: integer;
 Width, Height, Depth: integer;
 XOffset, YOffset: integer;
 Masking: integer;
 Compression: integer;
 Pad1: byte;
 Transparent: integer;
 XAspect, YAspect: integer;
 PageWidth, PageHeight: integer;
 ColorArray: array of TRGBTriple;
 FormType: TFormType;
 end;

function ReadLBMFile(BM: TBitmap; FileName: string): TLBMInfo;
function LoadLBMStream(Stream: TStream; BM: TBitmap): TLBMInfo;
procedure SaveLBMToStream(Stream: TStream; Bitmap: TBitmap; Compress: boolean);
procedure SaveLBMToFile(FileName: String; Bitmap: TBitmap; Compress: boolean);

{Defines a progress event. The user can create a subroutine and
{ assign it to FOnProgress and the subroutine will be called  }
{ periodically during the file loading to indicate progress. }

type TLBMProgress= procedure(Sender: TObject; Stage: TProgressStage; PercentDone: Byte; RedrawNow: Boolean; const R: TRect; const Msg: string);

var OnLBMProgress: TLBMProgress;



implementation


{================== Routines To Write LBM Files ===============================}


{------------------- Handle Progress Events --------------------------}

var ProgressValue: double;	{Current progress value }
var ProgressInc: double;	{Amount to increment progress }



procedure AdvanceProgress(Stage: TProgressStage; Steps: integer);
{Increment progress value and calls Progress event }
{if a procedure is assigned to FOnProgress, it is
{called periodically during file loading process }
var DrawNow: boolean;
var Msg: string;
begin
DrawNow:=False;
case Stage of
 psStarting:
	begin
	ProgressValue:=0;
	ProgressInc:=100/Steps;
	Msg:='Starting';
	end;
 psRunning:
	begin
	ProgressValue:=ProgressValue+ProgressInc;
	Msg:='Running';
	end;
 psEnding:
	begin
	DrawNow:=True;
	Msg:='Ending';
	end;
 end;

if Assigned(OnLBMProgress) then
    OnLBMProgress(nil,Stage,trunc(ProgressValue),DrawNow,Rect(0,0,0,0),Msg);
end;



{------------------ Support Routines ------------------------------------------}

procedure WriteChunkStart(Stream: TStream; var  SizeLoc: integer);
{Call at the beginning of writing to a chunk after the Label is written}
begin
SizeLoc:=Stream.Position;
WriteStreamDWord(Stream,0);
end;



function WriteChunkEnd(Stream: TStream; var  SizeLoc: integer): integer;
{Call at the beginning of writing to a chunk after the Label is written}
var Size: integer;
var OldPos: int64;
begin
OldPos:=Stream.Position;
Size:=OldPos-(SizeLoc+SizeOf(DWord));
Stream.Position:=SizeLoc;
WriteStreamDWord(Stream,Size);
Stream.Position:=OldPos;
Result:=Size;
end;


{------------------ Writes "BODY" part of LBM File ----------------------------}

procedure WritePackBitsLine(Stream: TStream; P: PByteArray; N: integer);
{Do packbits compression for one line of image data  }
var Literals: array [0..127] of byte;
var LitInx,I,OutCnt,InIndex: word;


	procedure WriteLiterals;
	{Write accumulated literals}
	begin
	WriteStreamByte(Stream,(LitInx-1) and $7F);
	Inc(OutCnt);
	Stream.Write(Literals,LitInx);
	Inc(OutCnt,LitInx);
	LitInx:=0;
	end;

	procedure WriteByteCode;
	{Write code count and value}
	begin
	WriteStreamByte(Stream,not I+1);
	WriteStreamByte(Stream,P[InIndex+I]);
	Inc(OutCnt,2);
	Inc(InIndex,I+1);
	end;


begin
LitInx:=0; I:=0;
OutCnt:=0; InIndex:=0;
repeat
	begin
	I:=0;
	while (I < 127) and (I < (N-1)) and ((InIndex+I+1) < N) and
	      (P[InIndex+I] = P[InIndex+I+1]) do Inc(I);

	if (I > 1) or (LitInx >= 127) then
		begin
		if LitInx<>0 then WriteLiterals;
		if I<>0 then WriteByteCode;
		end
	else
		begin
		Literals[LitInx]:=P[InIndex];
		Inc(LitInx); Inc(InIndex);
		end;
	end
until InIndex >= N;

if LitInx<>0 then WriteLiterals;
if (OutCnt and $01)<>0  then WriteStreamByte(Stream,$80);
end;



procedure OutputScanLine(Stream: TStream; P: PByteArray; N: integer; Compress: boolean);
{Output compressed or uncompressed scan line}
begin
if Compress then WritePackBitsLine(Stream, P, N)
else Stream.Write(P^,N);
end;


procedure WriteBODY(Stream: TStream; var Info: TLBMInfo; Bitmap: TBitmap; Compress: boolean);
{ Write "BODY" part of LBM. Contains the actual image data}
var SizeLoc: integer;
var Y,N: integer;
var ScanLine: PByteArray;
begin
WriteStreamString(Stream,'BODY');
WriteChunkStart(Stream,SizeLoc);
for Y:=0 to Bitmap.Height-1 do
	begin
	ScanLine:=Bitmap.ScanLine[Y];
	N:=Bitmap.Width;
	OutputScanline(Stream, ScanLine,N,Compress);
	end;
WriteChunkEnd(Stream,SizeLoc);
AdvanceProgress(psRunning,0);
end;


{----------------- Write the "BMHD" "Bitmap Header part of file -------------- }


type TPaletteEntries = array [0..255] of TPaletteEntry;	{Define Palette array data format}


procedure BuildHeader(var Info: TLBMInfo; BM: TBitmap; Compress: boolean);
{Extract Header information from BMP and store in LBM info structure}
var I:integer;
var BitmapPalette: TPaletteEntries;
begin
Info.HaveBmhd:=True;
Info.FormSize:=0;
Info.Width:=BM.Width;
Info.Height:=BM.Height;
Info.Depth:=8;
Info.XOffset:=0;
Info.YOffset:=0;
Info.Masking:=0;
if Compress then Info.Compression:=1 else Info.Compression:=0;
Info.Pad1:=0;
Info.Transparent:=255;
Info.XAspect:=1;
Info.YAspect:=1;
Info.PageWidth:=BM.Width;
Info.PageHeight:=BM.Height;
Info.FormType:=ftPBM;
SetLength(Info.ColorArray,256);
GetPaletteEntries(BM.Palette,0,256,BitmapPalette);
for I:=0 to 256-1 do
	begin
	Info.ColorArray[I].rgbtRed:=BitmapPalette[I].peRed;
	Info.ColorArray[I].rgbtGreen:=BitmapPalette[I].peGreen;
	Info.ColorArray[I].rgbtBlue:=BitmapPalette[I].peBlue;
	end;
end;



procedure WriteBMHD(Stream: TStream;var Info: TLBMInfo);
{Write the BMHD part of the file which is the "Bitmap Header"}
var SizeLoc: integer;
begin
WriteStreamString(Stream,'BMHD');
WriteChunkStart(Stream,SizeLoc);
WriteStreamWord(Stream,Info.Width);
WriteStreamWord(Stream,Info.Height);
WriteStreamWord(Stream,Info.XOffset);
WriteStreamWord(Stream,Info.YOffset);
WriteStreamByte(Stream,Info.Depth);
WriteStreamByte(Stream,Info.Masking);
WriteStreamByte(Stream,Info.Compression);
WriteStreamByte(Stream,Info.Pad1);
WriteStreamWord(Stream,Info.Transparent);
WriteStreamByte(Stream,Info.XAspect);
WriteStreamByte(Stream,Info.YAspect);
WriteStreamWord(Stream,Info.PageWidth);
WriteStreamWord(Stream,Info.PageHeight);
WriteChunkEnd(Stream,SizeLoc);
AdvanceProgress(psRunning,0);
end;



procedure WriteCMAP(Stream: TStream; var Info: TLBMInfo);
{Write the "CMAP" part of file which is the LBM color palette }
var SizeLoc: integer;
var I: integer;
begin
WriteStreamString(Stream,'CMAP');
WriteChunkStart(Stream,SizeLoc);

for I:=0 to High(Info.ColorArray) do
	begin
	WriteStreamByte(Stream,Info.ColorArray[I].rgbtRed);
	WriteStreamByte(Stream,Info.ColorArray[I].rgbtGreen);
	WriteStreamByte(Stream,Info.ColorArray[I].rgbtBlue);
	end;
WriteChunkEnd(Stream,SizeLoc);
AdvanceProgress(psRunning,0);
end;



procedure WriteForm(Stream: TStream; var Info: TLBMInfo; Bitmap: TBitmap; Compress: boolean);
{Write "FORM" part of LBM. "FORM" is the top level and holds everything}
var SizeLoc: integer;
begin
WriteStreamString(Stream,'FORM');
WriteChunkStart(Stream,SizeLoc);
WriteStreamString(Stream,'PBM ');
WriteBMHD(Stream,Info);
WriteCMAP(Stream,Info);
WriteBODY(Stream,Info,Bitmap,Compress);
Info.FormSize:=WriteChunkEnd(Stream,SizeLoc);
end;



procedure SaveLBMToStream(Stream: TStream; Bitmap: TBitmap; Compress: boolean);
{Save BMP to specified Stream as LBM}
var Info: TLBMInfo;
var WorkBitmap: TBitmap;
begin
AdvanceProgress(psStarting,4);
WorkBitmap:=TBitmap.Create;
try
WorkBitmap.Assign(Bitmap);
WorkBitmap.PixelFormat:=pf8Bit;
StreamEndian:=BigEnd;
FillMemory(@Info,SizeOf(TLBMInfo),0);
BuildHeader(Info,WorkBitmap,Compress);
WriteForm(Stream,Info,WorkBitmap,Compress);
finally WorkBitmap.Free; end;
AdvanceProgress(psEnding,0);
end;



procedure SaveLBMToFile(FileName: String; Bitmap: TBitmap; Compress: boolean);
{Save BMP to specified File as LBM}
var Stream: TMemoryStream;
begin
Stream:=TMemoryStream.Create;
SaveLBMToStream(Stream,Bitmap,Compress);
Stream.SaveToFile(filename);
Stream.Free;
end;



{================== Routines to Read LBM files ================================}


{------------------- Chunk Handling Routines-----------------------------------}

const EvenMask = High(Cardinal)-1;

var ChunkSize,ChunkStart: integer;


function StartChunk(Stream: TStream): integer;
begin
{Every chunk is padded to an even length}
Result:=GetStreamDWord(Stream);
ChunkSize:=(Result+1) and EvenMask;
ChunkStart:=Stream.Position;
end;


procedure CleanupChunk(Stream: TStream);
{Point to the beginning of the next chunk}
begin
Stream.Position:=ChunkStart+ChunkSize;
end;


{------------------------------------------------------------------------------}


procedure EatChunk(Stream: TStream);
{Read in and dispose of a chunk}
begin
StartChunk(Stream);
CleanupChunk(Stream);
end;


{------------------------------------------------------------------------------}



procedure DoBMHD(Stream: TStream; var Info: TLBMInfo; BM: TBitmap);
{Read in BitMap Header chunk}
begin
StartChunk(Stream);
Info.Width:= GetStreamWord(Stream);		{Image width and height in pixels}
Info.Height:= GetStreamWord(Stream);
Info.XOffset:= GetStreamWord(Stream);		{Offset (in pixels) for this image }
Info.YOffset:= GetStreamWord(Stream);
Info.Depth:= GetStreamByte(Stream);		{Number of bit planes (can be 0)}
Info.Masking:= GetStreamByte(Stream);		{Masking technique ('no mask' is assumed here)}
Info.Compression:=GetStreamByte(Stream);	{Compression algorithm (0=none, 1=byte run)}
Info.Pad1:=GetStreamByte(Stream);		{Defined as an unused byte (note you cannot read word)}
Info.Transparent:= GetStreamWord(Stream);	{Transparent color (only used with masking)}
Info.XAspect:= GetStreamByte(Stream);		{Aspect ratio (1:1 is assumed here)}
Info.YAspect:= GetStreamByte(Stream);
Info.PageWidth:= GetStreamWord(Stream);		{Source page size }
Info.PageHeight:= GetStreamWord(Stream);
Info.HaveBMHD:= true;				{Indicate that essential BMHD chunk has been read in}
CleanupChunk(Stream);
BM.Width:=Info.Width;
BM.Height:=Info.Height;
AdvanceProgress(psRunning,0);
end;	{DoBMHD}



procedure DoCMAP(Stream: TStream; var Info: TLBMInfo);
{Color Map, set up color registers}
var ColRegs: integer;	{Number of color registers to use}
var N: integer;
var R, G, B: integer;
var CkSize: integer;
begin
CkSize:=StartChunk(Stream);
ColRegs:= CkSize div 3;
if ColRegs >256 then ColRegs:= 256;	{Limit maximum number of color registers}
SetLength(Info.ColorArray,ColRegs);
for N:=0 to ColRegs-1 do
	begin
	R:= GetStreamByte(Stream);
	G:= GetStreamByte(Stream);
	B:= GetStreamByte(Stream);
	Info.ColorArray[N]:=MakeRBGTriple(R,G,B);
	end;

CleanupChunk(Stream);
AdvanceProgress(psRunning,0);
end;	{DoCMAP}

{----------------------------------------------------------------------}

type TPlaneLine = array of byte;	{Single line on plane}
type TLBMLine = array of TPlaneLine;	{Combined line from all planes}


procedure ReadCompressedLine(Stream: TStream; Width: integer; var Line: TPlaneLine);
{Read a single line for one bit plane (compressed data)}
var N: ShortInt;
var M: integer;
var J,I: integer;
begin
J:=0;
repeat
	begin
	N:= GetStreamByte(Stream);
	if N >=0 then
		begin
		{Copy the next N+1 bytes literally}
		for I:=1 to N+1 do
			begin
			Line[J]:=GetStreamByte(Stream);
			Inc(J);
			end;
		end
	else if N = -128 then begin end		{Ignore}
	else	begin
		{N is in the range: $81..$FF, so Replicate the next byte -Extend(N)+1 times}
		{Read the byte to Replicate}
		M:= GetStreamByte(Stream);
		for I:=1 to -N+1 do
			begin
			Line[J]:=M;
			Inc(J);
			end;
		end;
	end
until J >= Width;
end;



procedure ReadLine(Stream: TStream; Width: integer; var Line: TPlaneLine);
{Read one line from one plane Uncompressed}
var X: integer;
begin
for X:=0 to Width-1 do Line[X]:=GetStreamByte(Stream);
end;



procedure GetCombinedLine(Stream: TStream; var Info: TLBMInfo; Width: integer; var Line: TLBMLine);
{Read one lines on  each plane }
var D: integer;
begin
for D:=0 to Info.Depth-1 do
	begin
	if Info.Compression<>0 then ReadCompressedLine(Stream,Width,Line[D])
	else  ReadLine(Stream,Width,Line[D]);
	end;
end;



function GetPixelValue(Line: TLBMLine; Info: TLBMInfo; Mask: byte; Inx: integer): integer;
{Combine each line from each plane and combine them to the associated pixel value}
var D: integer;
begin
Result:=0;
{Iterrate through all associated planes}
for D:=Info.Depth-1 downto 0 do
	begin
	Result:=Result shl 1;
	if (Mask and Line[D,Inx])<>0 then Result:=Result or 1;
	end;
end;



procedure SetColorValues(Line: TLBMLine; Info: TLBMInfo; Y: integer; BM: TBitmap);
{Insert the colors into the bitmap for one line}
var Mask: byte;
var X,B: integer;
var PInx: integer;
var Scan: pRGBTripleArray;
begin
Scan:=BM.ScanLine[Y];
for X:=0 to (Info.Width div 8)-1 do
	begin
	Mask:=$80;
	for B:=0 to 7 do
		begin
		PInx:=GetPixelValue(Line,Info,Mask,X);
		Scan[X * 8 + B]:=Info.ColorArray[Pinx];
		Mask:=Mask shr 1;
		end;
	end;
end;



procedure HandleILBMLine(Stream: TStream; var Info: TLBMInfo; BM: TBitmap; Y: integer);
{Handle a line from a ILBM type file}
var Line: TLBMLine;
var ByteWidth: integer;
begin
ByteWidth:=Info.Width div 8;		{Width in bytes}
SetLength(Line,Info.Depth,ByteWidth);
GetCombinedLine(Stream,Info,ByteWidth,Line);
SetColorValues(Line,Info,Y,BM);
end;



procedure HandlePBMLine(Stream: TStream; var Info: TLBMInfo; BM: TBitmap; Y: integer);
{Handle a line from a PBM type file}
var Line: TPlaneLine;
var X: integer;
var Scan: pRGBTripleArray;
begin
Scan:=BM.ScanLine[Y];
SetLength(Line,Info.Width);
if Info.Compression<>0 then ReadCompressedLine(Stream,Info.Width,Line)
else ReadLine(Stream,Info.Width,Line);
for X:=0 to Info.Width-1 do
	Scan[X]:=Info.ColorArray[Line[X]];
end;



procedure DoBODY(Stream: TStream; var Info: TLBMInfo; BM: TBitmap);
{Handle the HODY Chunk which holds the pixel data}
var Y: integer;		{Vertical scan line in pixels (0..HEIGHT-1)}
begin	{DoBODY}
if not Info.HaveBMHD then raise Exception.Create('"BMHD" is missing.');
StartChunk(Stream);

for Y:= 0 to Info.Height-1 do
	begin
	case Info.FormType of
	 ftILBM: HandleILBMLine(Stream,Info,BM,Y);
	 ftPBM: HandlePBMLine(Stream,Info,BM,Y);
	 else HandleILBMLine(Stream,Info,BM,Y);
	 end;
	end;
CleanupChunk(Stream);
AdvanceProgress(psRunning,0);
end;	{DoBODY}




function LoadLBMStream(Stream: TStream; BM: TBitmap): TLBMInfo;
{Load LBM stream into specified bitmap }
{Reference: Amiga ROM Kernel Reference Manual, p I-27.}
var S: string;
var Info: TLBMInfo;
begin
AdvanceProgress(psStarting,4);
FillMemory(@Info,SizeOf(TLBMInfo),0);
Info.HaveBMHD:= false;
Info.FormType:=ftUnknown;
try
StreamEndian:=BigEnd;
S:=GetStreamString(Stream,4);
if S<>'FORM' then raise Exception.Create('file did not start with "FORM".');
Info.FormSize:=GetStreamDWord(Stream);
S:=GetStreamString(Stream,4);
if S='ILBM' then Info.FormType:=ftILBM
else if S='PBM ' then Info.FormType:=ftPBM
else raise Exception.Create('Unknown Form Type.');
BM.PixelFormat:=pf24Bit;

while true do
	begin
	{Read chunk header}
	S:=GetStreamString(Stream,4);
	if S[1]=#0 then break;
	if S='BMHD' then DoBMHD(Stream,Info,BM)
	else if S='CMAP' then DoCMAP(Stream,Info)
	else if S='BODY' then begin DoBODY(Stream,Info,BM); break; end
	else EatChunk(Stream);		{Ignore any unrecognized chunks}
	end;
finally Result:=Info; end;
AdvanceProgress(psEnding,0);
end;	{LoadLBM}





function ReadLBMFile(BM: TBitmap; FileName: string): TLBMInfo;
{Read bmp from file }
var Stream: TMemoryStream;
begin
Stream := TMemoryStream.Create;
Stream.LoadFromFile(Filename);
try
Result:=LoadLBMStream(Stream,BM);
finally Stream.Free; end;
end;





end.
