unit MeasureTool;

interface

uses Windows,Classes,Types,Graphics,DSupport,Math,GraphicsLib,ImageEx,Vector,IntVector;


type
 TImageCoords = record
 BaseX,BaseY,XScale,YScale: double;
 Zone: integer;
 North: boolean;
 end;

const LabelOffset: TPoint = (X: 4; Y: 4);	{Note: must be positive}

type TPath = array of TPoint;
type TNames = array of string;



type TUserLengthProc = function(Inx1,Inx2: integer): double;
type TUserPosProc = function(Inx: integer): T2DVector;


type TMeasureTool = class(TObject)
 private
  FOnUserLength: TUserLengthProc;
  FOnUserPosition: TUserPosProc;
  FOnChange: TNotifyEvent;
  FShowLabels: boolean;
  procedure Changed;
  procedure DeletePoint;
  function GetPathRect(Path: TPath; Canvas: TCanvas): TRect;
    function GetPoint(Inx: integer; Pixels: boolean): T2DVector;
 public
  CurrentPoint: integer;
  Points: TPath;
  Names: TNames;
  Coords: TImageCoords;
  DrawScale: double;
  constructor Create;
  function StorePoint(X,Y: integer): integer;
  function PathHitTest(Pos: TPoint): integer;
  procedure DeleteLastPoint(Image: TImageEx);
  procedure Reset;
  procedure DrawPath(Canvas: TCanvas); overload;
  procedure DrawPath(Image: TImageEx); overload;
  procedure ErasePath(Image: TImageEx);
  function GetLength(P1,P2: integer; Pixels: boolean): double; overload;
  function GetLength(Pixels: boolean): double; overload;
  function GetArea(Pixels: boolean): double;
  function GetAzimuth: double;  overload;
  function GetAzimuth(P1,P2: integer): double; overload;
  procedure HandleMouseDown(X, Y: integer; Shift: TShiftState; Image: TImageEx);
  procedure HandleMouseMove(X, Y: integer; Shift: TShiftState; Image: TImageEx);
  procedure HandleMouseUp(Image: TImageEx);
  property OnChange: TNotifyEvent read FOnChange write FOnChange;
  property ShowLabels: boolean read FShowLabels write FShowLabels;
  property OnUserLength: TUserLengthProc read FOnUserLength write FOnUserLength;
  property OnUserPosition: TUserPosProc read FOnUserPosition write FOnUserPosition;
 end;




implementation


{-----------------------------------------------------------------------------}


procedure TMeasureTool.Changed;
begin
if Assigned(FOnChange) then FOnChange(Self);
end;



function TMeasureTool.PathHitTest(Pos: TPoint): integer;
var I: integer;
var R: TRect;
var Pnt: TPoint;
begin
Result:=-1;
for I:=0 to Length(Points)-1 do
	begin
	Pnt:=Points[I];
	R:=Rect(Pnt.X-HandleSize,Pnt.Y-HandleSize,Pnt.X+HandleSize,Pnt.Y+HandleSize);
	if PtInRect(R,Pos) then
		begin
		Result:=I;
		exit;
		end;
	end;
end;




procedure TMeasureTool.DrawPath(Canvas: TCanvas);
var I: integer;
var Name1,Name2: shortstring;
var P1,P2: TPoint;
var Params: TLineParams;

	procedure DisplayStation(Inx: integer; Name: string; P: TPoint);
	var C: TColor;
	begin
	if Inx=CurrentPoint then C:=clAqua else C:=clRed;
	DrawRoundHandle(Canvas,P,C);
	Canvas.Brush.Style:=bsClear;
	if ShowLabels then Canvas.TextOut(P.X+LabelOffset.X,P.Y+LabelOffset.Y,Names[Inx]);
	end;

begin
if Length(Points)<1 then exit;
HandleSize:=4;
Params.Line1.Color:=clRed;
Params.Line1.Style:=psInsideFrame;
Params.Line2.Color:=clBlack;
Params.Line2.Style:=psInsideFrame;
Params.Line3.Color:=clBlue;
Params.Line3.Style:=psClear;
Canvas.Pen.Width:=1;
DisplayStation(0,Name1,IntScalarProduct(Points[0],DrawScale));
for I:=0 to Length(Points)-2 do
	begin
	P1:=IntScalarProduct(Points[I],DrawScale);
	P2:=IntScalarProduct(Points[I+1],DrawScale);
	DrawMultiColorLine(Canvas,P1,P2,Params);
	DisplayStation(I+1,Name2,P2);
	end;

end;



procedure TMeasureTool.DrawPath(Image: TImageEx);
begin
DrawPath(Image.Canvas);
end;



function TMeasureTool.GetPathRect(Path: TPath; Canvas: TCanvas): TRect;
var I: integer;
var Size,BigText: TSize;
begin
Result.Left:=High(Integer);
Result.Right:=Low(Integer);
Result.Top:=High(Integer);
Result.Bottom:=Low(Integer);
BigText.cx:=0;
BigText.cy:=0;
for I:=0 to Length(Path)-1 do
	begin
	Size:=Canvas.TextExtent(Names[I]);
	if Size.cx>BigText.cx then BigText.cx:=Size.cx;
	if Size.cy>BigText.cy then BigText.cy:=Size.cy;
	if Path[I].X<Result.Left then Result.Left:=Path[I].X;
	if Path[I].Y<Result.Top then Result.Top:=Path[I].Y;
	if Path[I].X>Result.Right then Result.Right:=Path[I].X;
	if Path[I].Y>Result.Bottom then Result.Bottom:=Path[I].Y;
	end;
InflateRect(Result,HandleSize,HandleSize);
if ShowLabels then
	begin
	Inc(Result.Right,BigText.cx*2);
	Inc(Result.Bottom,BigText.cy*2);
	end;
end;



procedure TMeasureTool.HandleMouseDown(X,Y: integer; Shift: TShiftState; Image: TImageEx);
var Rect: TRect;
begin
if (ssLeft in Shift) or (ssRight in Shift) then
	begin
	Rect:=GetPathRect(Points,Image.Canvas);
	if (ssLeft in Shift) then
		begin
		CurrentPoint:=PathHitTest(Point(X,Y));
		if CurrentPoint<0 then
			begin
			CurrentPoint:=StorePoint(X,Y);
			end;
		end
	else DeletePoint;
	Image.RestoreRectangle(Rect);
	DrawPath(Image);
	end;
end;


procedure TMeasureTool.ErasePath(Image: TImageEx);
var Rect: TRect;
begin
Rect:=GetPathRect(Points,Image.Canvas);
Image.RestoreRectangle(Rect);
end;



procedure TMeasureTool.HandleMouseMove(X,Y: integer; Shift: TShiftState;  Image: TImageEx);
var P,D: TPoint;
var I: integer;
var Rect: TRect;
begin
if CurrentPoint>-1 then
	begin
	P:=Point(X,Y);
	Rect:=GetPathRect(Points,Image.Canvas);
	if ssCtrl in Shift then
		begin
		D:=IntVectorSubtract(P,Points[CurrentPoint]);
		for I:=0 to High(Points) do Points[I]:=IntVectorAdd(Points[I],D);
		end
	else Points[CurrentPoint]:=P;
	Image.RestoreRectangle(Rect);
	DrawPath(Image);
	Changed;
	end;
end;





procedure TMeasureTool.HandleMouseUp(Image: TImageEx);
begin
if CurrentPoint>-1 then
	begin
	CurrentPoint:=-1;
	Image.RestoreImage;
	DrawPath(Image);
	end;
end;


{------------------Measurement Tool Object ------------------------------}


procedure TMeasureTool.DeletePoint;
begin
if Length(Points)>0 then
	begin
	SetLength(Points,Length(Points)-1);
	SetLength(Names,Length(Names)-1);
	Changed;
	end;
end;



procedure TMeasureTool.DeleteLastPoint(Image: TImageEx);
var Rect: TRect;
begin
Rect:=GetPathRect(Points,Image.Canvas);
DeletePoint;
Image.RestoreRectangle(Rect);
DrawPath(Image);
end;


function TMeasureTool.StorePoint(X,Y: integer): integer;
begin
SetLength(Points,Length(Points)+1);
SetLength(Names,Length(Names)+1);
Result:=High(Points);
Points[Result]:=Point(X,Y);
Names[Result]:='';
Changed;
end;


procedure TMeasureTool.Reset;
begin
CurrentPoint:=-1;
SetLength(Points,0);
SetLength(Names,0);
Changed;
end;



constructor TMeasureTool.Create;
begin
Inherited Create;
DrawScale:=1;
Reset;
end;

{------------------------------------------------------------------------------}


function TMeasureTool.GetPoint(Inx: integer; Pixels: boolean): T2DVector;
var XScale,YScale: double;
begin
Result:=MakeVector2D(0,0);
if (Inx<0) or (Inx>High(Points))  then exit;
if Pixels or not Assigned(FOnUserPosition) then
	begin
	if Pixels then begin XScale:=1; YScale:=1; end
	else begin XScale:=Coords.XScale; YScale:=Coords.YScale; end;
	Result.X:=Points[Inx].X * XScale;
	Result.Y:=Points[Inx].Y * YScale;
	end
else Result:=FOnUserPosition(Inx);
end;




function TMeasureTool.GetLength(P1,P2: integer; Pixels: boolean): double;
{Get distance between two points P1 and P2}
var XScale,YScale,DX,DY: double;
begin
Result:=0;
if (P1<0) or (P2<0) or (P1>High(Points)) or (P2>High(Points)) then exit;
if Pixels or not Assigned(FOnUserLength) then
	begin
	if Pixels then begin XScale:=1; YScale:=1; end
	else begin XScale:=Coords.XScale; YScale:=Coords.YScale; end;
	DX:=(Points[P2].X - Points[P1].X) * XScale;
	DY:=(Points[P2].Y - Points[P1].Y) * YScale;
	Result:=sqrt(sqr(DX) + sqr(DY));
	end
else Result:=FOnUserLength(P1,P2)
end;




function TMeasureTool.GetLength(Pixels: boolean): double;
var I: integer;
begin
Result:=0;
for I:=0 to Length(Points)-2 do
	begin
	Result:=Result+GetLength(I,I+1,Pixels);
	end;
end;



function TMeasureTool.GetAzimuth(P1,P2: integer): double;
{Return azimuth of two specified points}
var DX,DY: double;
begin
Result:=0;
if (P1<0) or (P2<0) or (P1>High(Points)) or (P2>High(Points)) then exit;
DX:=Points[P2].X - Points[P1].X;
DY:=Points[P2].Y - Points[P1].Y;
Result:=(ArcTan2(DX,-DY) * Radian2Degree);
if Result<0 then Result:=360 + Result;
end;



function TMeasureTool.GetAzimuth: double;
{Return azimuth of last two points}
begin
Result:=GetAzimuth(Length(Points)-2,Length(Points)-1);
end;




function TMeasureTool.GetArea(Pixels: boolean): double;
{K = [(x1y2 + x2y3 + x3y4 + ... + xny1) - (x2y1 + x3y2 + x4y3 + ... + x1yn)]/2}
var I,J: integer;
var Sum1,Sum2: double;
var P1,P2: T2DVector;
begin
Result:=0;
Sum1:=0; Sum2:=0;
for I:=0 to Length(Points)-1 do
	begin
	if I >= (Length(Points)-1) then J:=0 else J:=I+1;
	P1:=GetPoint(I,Pixels); P2:=GetPoint(J,Pixels);
	Sum1:=Sum1 + (P1.X * P2.Y);
	Sum2:=Sum2 + (P2.X * P1.Y);
	end;

Result:=abs((Sum1 - Sum2)/2);
end;




end.
