unit MorphLibrary;

interface

uses Windows,Types,Graphics,SysUtils,Math,Vector;

type TMorphFactors = record
 A,B,P: double;
 end;

type TMorphProgress = procedure(Percent: integer);


function DoMultiLineMorph(Source,Dest: T2DLineArray; DestPoint: T2DVector; MF: TMorphFactors): T2DVector;
function GetTransformedPoint(P1,Q1: T2DVector; U,V: double): T2DVector;
procedure MorphBitmap(FromBitmap,ToBitmap: TBitmap; FromLine1,ToLine2: T2DLineArray;
		      MF: TMorphFactors; BackColor: TColor; Progress: TMorphProgress);



implementation


{ Beier-Neely morphing Algorithm from Siggraph '92 paper }


function CalculateU(P,Q,X: T2DVector): double;
{Calculate "U" value = Percent along P-Q the closest intercept falls}
{Values less than 0 mean the intercept beyond the start of the Line}
{Values Greater than 1 mean the intercept beyond the End of the Line}
var XP,QP: T2DVector;
begin
XP:=VectorSubtract2D(X,P);
QP:=VectorSubtract2D(Q,P);
Result:=DotProduct2D(XP,QP)/sqr(VectorABS2D(QP));
end;


function CalculateV(P,Q,X: T2DVector): double;
{Calculate "V" value = signed distance from Point X to Line P-Q}
var XP,QP: T2DVector;
begin
XP:=VectorSubtract2D(X,P);
QP:=VectorSubtract2D(Q,P);
QP:=GetNormal2D(QP);
Result:=DotProduct2D(XP,QP)/VectorABS2D(QP);
end;



function GetTransformedPoint(P1,Q1: T2DVector; U,V: double): T2DVector;
{Given U/V find equivalent postion for line P1,Q1}
var QP: T2DVector;
begin
{Use V to find right-angle distance from line Q1-P1 }
QP:=VectorSubtract2D(Q1,P1);
Result:=GetNormal2D(QP);
Result:=ScalarProduct2D(Result,V);
{Use U to find length along line Q1-P1 }
Result:=ScalarDivide2D(Result,VectorABS2D(QP));
Result:=VectorAdd2D(ScalarProduct2D(QP,U),Result);
Result:=VectorAdd2D(Result,P1);
end;



function MorphTransform(SourceLine,DestLine: T2DLine; DestPoint: T2DVector): T2DVector;
{Find U/V position on Destination Line set and move point to same location on Source}
var U,V: double;
begin
U:=CalculateU(DestLine.P1,DestLine.P2,DestPoint);
V:=CalculateV(DestLine.P1,DestLine.P2,DestPoint);
Result:=GetTransformedPoint(SourceLine.P1,SourceLine.P2,U,V);
end;


function DoMultiLineMorph(Source,Dest: T2DLineArray; DestPoint: T2DVector; MF: TMorphFactors): T2DVector;
{ Returns new destination bitmap position based on DestPoint Input}
{ The difference between Source and Dest Line sets indicates how morphing is done}
{NF contains morphing factors that control how the lines are weighted }
var I: integer;
var DSum,Xi,Di: T2DVector;
var WeightSum,Dist,Weight,LengthP: double;
var U,V: double;
begin
if (Length(Source)<1) or (Length(Dest)<1) then raise Exception.Create('Empty Arrays.');
if Length(Source)<>Length(Dest) then
 raise Exception.Create('Arrays not same size: '+IntToStr(Length(Source))+'/'+IntToStr(Length(Dest)));

{Calculate weighted sum of the effects of all individaul lines on the point}
DSum:=MakeVector2D(0,0);
WeightSum:=0;
{Iterate averaging the effect of all lines on point}
for I:=0 to Length(Dest)-1 do
	begin
	{Get U/V coordinates of Point relative to Destination Line Set}
	U:=CalculateU(Dest[I].P1,Dest[I].P2,DestPoint);
	V:=CalculateV(Dest[I].P1,Dest[I].P2,DestPoint);
	{Transform point to same U/V position relative to Source line set}
	Xi:=GetTransformedPoint(Source[I].P1,Source[I].P2,U,V);
	Di:=VectorSubtract2D(Xi,DestPoint);
	{Calculate normaized distance}
	if U<0 then Dist:=VectorAbs2D(VectorSubtract2D(Dest[I].P1,DestPoint))
	else if U>1 then Dist:=VectorAbs2D(VectorSubtract2D(Dest[I].P2,DestPoint))
	else Dist:=abs(V);
	{Calculate and sum all weighting factors}
	LengthP:=Power(VectorABS2D(VectorSubtract2D(Dest[I].P1,Dest[I].P2)),MF.P);
	Weight:=Power(LengthP/(MF.A + Dist),MF.B);
	DSUM:=VectorAdd2D(DSUM,ScalarProduct2D(Di,Weight));
	WeightSum:=WeightSum + Weight;
	end;
{Apply Sum of Weights to Destination Point}
Result:=VectorAdd2D(DestPoint, ScalarDivide2D(DSUM,WeightSum));
end;




type TIntegerArray = array[0..32767] of Integer;
type PIntegerArray = ^TIntegerArray;


function SwapRedBlue(Color: TColor): TColor;
begin
PByteArray(@Result)[0]:=PByteArray(@Color)[2];
PByteArray(@Result)[1]:=PByteArray(@Color)[1];
PByteArray(@Result)[2]:=PByteArray(@Color)[0];
PByteArray(@Result)[3]:=PByteArray(@Color)[3];
end;


procedure MorphBitmap(FromBitmap,ToBitmap: TBitmap; FromLine1,ToLine2: T2DLineArray;
		      MF: TMorphFactors; BackColor: TColor; Progress: TMorphProgress);
{Morph bitmap based FromLine1, ToLine2 which control the how the image is morphed}
{MF controls how the lines are weighted, BackColor controls the background color}
var X,Y: integer;
var P: TPoint;
var ScanLine,SL2: PIntegerArray;
begin
BackColor:=SwapRedBlue(BackColor);
ToBitmap.Width:=FromBitmap.Width;
ToBitmap.Height:=FromBitmap.Height;

FromBitmap.PixelFormat:=pf32Bit;
ToBitmap.PixelFormat:=pf32Bit;
for Y:=0 to ToBitmap.Height-1 do
	begin
	ScanLine:=ToBitmap.ScanLine[Y];
	if Assigned(Progress) then Progress(MulDiv(100,Y,ToBitmap.Height));
	for X:=0 to ToBitmap.Width-1 do
		begin
		P:=Vector2DToPoint(DoMultiLineMorph(FromLine1,ToLine2,MakeVector2D(X,Y),MF));
		if (P.X<0) or (P.Y<0) or (P.X>=FromBitmap.Width) or (P.Y>=FromBitmap.Height) then ScanLine[X]:=BackColor
		else
			begin
			SL2:=FromBitmap.ScanLine[P.Y];
			ScanLine[X]:=SL2[P.X];
			end;
		end;
	end;
if Assigned(Progress) then Progress(100);
end;



end.
