unit GLFonts;

interface

uses Windows,Classes,SysUtils,Graphics,Contnrs,Vector,OpenGL15,GLRenderer,GLLibrary;




type TGL3DFont = class(TObject)
 private
   FFontBuilt: boolean;
   FFontDisplayList: GLuint;
   GlyphMetricsArray: array[0..255] of GLYPHMETRICSFLOAT;
   FDelphiFont: TFont;
   FFontDepth: double;
   FOnChange: TNotifyEvent;
   procedure BuildFont(DC: HDC; Font: TFont); overload;
   procedure BuildFont; overload;
   procedure SetFont(F: TFont);
   procedure SetFontDepth(V: double);
   procedure Change;
 public
   constructor Create;
   constructor CreateFrom(Font: TGL3DFont);
   destructor Destroy; override;
   procedure Assign(AObject: TObject);
   property Font: TFont read FDelphiFont write SetFont;
   property FontDepth: double read FFontDepth write SetFontDepth;
   property OnChange: TNotifyEvent read FOnChange write FOnChange;
 end;



type TTextJustify = (tjLeft,tjRight,tjCenter);


type TGL3DString = class(TGLRenderer)
 private
   FText: string;
   FJustify: TTextJustify;
   FSize: T3DVector;
   FWidth,FHeight,FCharHeight: double;
   FShowLineBreaks: boolean;
   FLineSpacing: double;
    FExtentMin: T3DVector;
    FExtentMax: T3DVector;
   function GetLineLength(S: string): double;
   procedure GetTextExtents;
   procedure SetTextPosition(Line: integer; S: string);
   procedure HandleFontChange(Sender: TObject);
   procedure SetText(V: string);
   procedure OutputSeparateLines(S: string);
   procedure SetShowLineBreaks(const Value: boolean);
   procedure SetLineSpacing(const Value: double);
    function ReadExtentMax: T3DVector;
    function ReadExtentMin: T3DVector;
    function GetHeight: double;
    function GetWidth: double;
 public
   GL3DFont: TGL3DFont;
   constructor Create; override;
   constructor CreateFrom(GL3DString: TGL3DString);
   destructor Destroy; override;
   procedure Assign(GLObject: TGLObject); override;
   property Text: string read FText write SetText;
   property Justify: TTextJustify read FJustify write FJustify;
   procedure TextOut(S: string);
   procedure Render; override;
   procedure Clear; override;
   procedure ClearData;
   procedure SetColor(const Color: TColor); override;
   function GetColor: TColor; override;
   property Size: T3DVector read FSize;
   property Width: double read GetWidth;
   property Height: double read GetHeight;
   property LineSpacing: double read FLineSpacing write SetLineSpacing;
   property CharHeight: double read FCharHeight;
   property ShowLineBreaks: boolean read FShowLineBreaks write SetShowLineBreaks;
   property Color: TColor read GetColor write SetColor;
   property ExtentMin: T3DVector read ReadExtentMin write FExtentMin;
   property ExtentMax: T3DVector read ReadExtentMax write FExtentMax;
 end;


{ TGLStringList }


type  TGLStringList3D = class(TGLObjectList)
  private
    procedure HandleStringChange(Sender: TObject);
  protected
    function GetItem(Index: Integer): TGL3DString;
    procedure SetItem(Index: Integer; AObject: TGL3DString);
  public
    function Add(var Str: TGL3DString): integer; overload;
    function Add: integer; overload;
    function AddFrom(Str: TGL3DString): integer;
    function Insert(Inx: integer; var Str: TGL3DString): integer; overload;
    function Insert(Inx: integer): integer; overload;
    property Items[Index: Integer]: TGL3DString read GetItem write SetItem; default;
    procedure Assign(ObjectList: TObjectList);
  end;


implementation



{ TGL3DFont }

function GetFontHandle(Font: TFont): HFONT;
{Get font handle this way because we need to force True Type fonts}
var LogFont: TLogFont;
begin
{Note: size is ignored by GL routines}
logfont.lfHeight:=-27;
logfont.lfWidth:=27;
logfont.lfEscapement:=0;
logfont.lfOrientation:=0;
 if fsBold in Font.Style then LogFont.lfWeight := FW_BOLD
 else LogFont.lfWeight:=FW_NORMAL;
LogFont.lfItalic:=Byte(fsItalic in Font.Style);
LogFont.lfUnderline:=Byte(fsUnderline in Font.Style);
LogFont.lfStrikeOut:=Byte(fsStrikeOut in Font.Style);
LogFont.lfCharSet:=Byte(Font.Charset);
logfont.lfOutPrecision:=OUT_TT_ONLY_PRECIS;
logfont.lfClipPrecision:=CLIP_DEFAULT_PRECIS;
logfont.lfQuality:=PROOF_QUALITY;
case Font.Pitch of
 fpVariable: LogFont.lfPitchAndFamily:=VARIABLE_PITCH;
 fpFixed: LogFont.lfPitchAndFamily:=FIXED_PITCH;
 else LogFont.lfPitchAndFamily:=DEFAULT_PITCH;
 end;
if AnsiCompareText(Font.Name, 'Default') = 0 then StrPCopy(LogFont.lfFaceName, DefFontData.Name)
else StrPCopy(LogFont.lfFaceName, Font.Name);
Result:=CreateFontIndirect(LogFont);
end;


procedure TGL3DFont.Change;
begin
if Assigned(FOnChange) then FOnChange(Self);
end;


procedure TGL3DFont.BuildFont(DC: HDC; Font: TFont);
var FontHandle,Old: HFONT;
begin
FFontDisplayList:=glGenLists(256);
FontHandle:=GetFontHandle(Font);
Old:=SelectObject(DC,FontHandle);
FFontBuilt:=wglUseFontOutlines(DC,0,256,FFontDisplayList,0,FFontDepth,WGL_FONT_POLYGONS,@GlyphMetricsArray);
SelectObject(DC,Old);
DeleteObject(FontHandle);
Change;
end;




procedure TGL3DFont.BuildFont;
begin
BuildFont(wglGetCurrentDC,FDelphiFont);
end;



constructor TGL3DFont.Create;
begin
inherited;
FFontBuilt:=False;
FDelphiFont:=TFont.Create;
FDelphiFont.Height:=16;
FDelphiFont.Style:=[fsBold];
FDelphiFont.Charset:=ANSI_CHARSET;
FDelphiFont.Name:='Arial';
FDelphiFont.Color:=clBlue;
FFontDepth:=0.2;
end;



constructor TGL3DFont.CreateFrom(Font: TGL3DFont);
begin
inherited;
Create;
Self.Assign(Font);
end;



destructor TGL3DFont.Destroy;
begin
FDelphiFont.Free;
inherited;
end;




procedure TGL3DFont.SetFont(F: TFont);
begin
FDelphiFont.Assign(F);
FFontBuilt:=False;
end;




procedure TGL3DFont.SetFontDepth(V: double);
begin
if V<>FFontDepth then
	begin
	FFontDepth:=V;
	FFontBuilt:=False;
	end;
end;



procedure TGL3DFont.Assign(AObject: TObject);
begin
inherited;
if AObject is TGL3DFont then
	begin
	FDelphiFont.Assign(TGL3DFont(AObject).FDelphiFont);
	FFontDepth:=TGL3DFont(AObject).FontDepth;
	FFontBuilt:=False;
	end;
end;

{----------------------- GL3DString ---------------------------------}



procedure RandomStippling;
var P: array [0..31,0..31] of byte;
var X,Y: integer;
begin
for Y:=0 to 31 do
 for X:=0 to 31 do P[X,Y]:=Random(255);
glPolygonStipple(@P);
glEnable(GL_POLYGON_STIPPLE);
end;



procedure TGL3DString.Assign(GLObject: TGLObject);
begin
Clear;
inherited Assign(GLObject);
if GLObject is TGL3DString then
	begin
	GL3DFont.Assign(TGL3Dstring(GLObject).GL3DFont);
	Text:=TGL3DString(GLObject).Text;
	Justify:=TGL3DString(GLObject).Justify;
	LineSpacing:=TGL3DString(GLObject).LineSpacing;
	ShowLineBreaks:=TGL3DString(GLObject).ShowLineBreaks;
	RenderMode:=TGL3DString(GLObject).RenderMode;
	SetColor(TGL3DString(GLObject).GetColor);
	end;
end;



procedure TGL3DString.Clear;
begin
inherited;
FText:='';
end;



procedure TGL3DString.ClearData;
begin
FText:='';
end;

constructor TGL3DString.Create;
begin
inherited;
GL3DFont:=TGL3DFont.Create;
GL3DFont.OnChange:=HandleFontChange;
FJustify:=tjCenter;
FLineSpacing:=0.1;
Self.FShowLineBreaks:=True;
end;



constructor TGL3DString.CreateFrom(GL3DString: TGL3DString);
begin
inherited Create;
GL3DFont:=TGL3DFont.CreateFrom(GL3DString.GL3DFont);
GL3DFont.OnChange:=HandleFontChange;

Text:=GL3DString.Text;
Justify:=GL3DString.Justify;
FSize:=GL3DString.Size;
//FHeight:=GL3DString.Height;
LineSpacing:=GL3DString.LineSpacing;
FCharHeight:=GL3DString.CharHeight;
ShowLineBreaks:=GL3DString.ShowLineBreaks;
RenderMode:=GL3DString.RenderMode;
end;



destructor TGL3DString.Destroy;
begin
GL3DFont.Free;
inherited;
end;



function TGL3DString.GetHeight: double;
begin
GetTextExtents;
Result:=FHeight;
end;


function TGL3DString.GetLineLength(S: string): double;
var I: integer;
begin
Result:=0;
for I:=1 to length(S) do
        begin
        Result:=Result + GL3DFont.GlyphMetricsArray[byte(S[I])].gmfCellIncX;
        end;
end;



procedure TGL3DString.GetTextExtents;
var Z: double;

	procedure MultiLineExtents(S: string; CharHeight: double; var X,Y: double);
	var I: integer;
	var SL: TStringList;
	var L: double;
	begin
	SL:=TStringList.Create;
	try
	X:=0;
	SL.Text:=S;
	for I:=0 to SL.Count-1 do
		begin
		L:=GetLineLength(SL[I]);
		if L>X then X:=L;
		end;
	Y:=CharHeight * SL.Count;
	Y:=Y + (Self.FLineSpacing * (SL.Count-1));
	finally SL.Free; end;
	end;


begin
FExtentMin:=MakeVector3D(0,0,0);
if not GL3DFont.FFontBuilt then GL3DFont.BuildFont;
FCharHeight:=GL3DFont.GlyphMetricsArray[$41].gmfBlackBoxY;
Z:=-GL3DFont.FontDepth;
FExtentMax:=MakeVector3D(0,FCharHeight,Z);
if Self.FShowLineBreaks then
	begin
	MultiLineExtents(Text,FCharHeight,FExtentMax.X,FExtentMax.Y);
	end
else
	begin
	FExtentMax.X:=GetLineLength(Text);
	end;
FSize:=VectorSubtract3D(FExtentMax,FExtentMin);
FWidth:=FSize.X;
FHeight:=FSize.Y;
end;



function TGL3DString.GetWidth: double;
begin
GetTextExtents;
Result:=FWidth;
end;

procedure TGL3DString.HandleFontChange(Sender: TObject);
begin
end;



procedure TGL3DString.OutputSeparateLines(S: string);
var SL: TStringList;
var I: integer;
var S1: string;
begin
SL:=TStringList.Create;
try
SL.Clear;
SL.Text:=S;
for I:=0 to SL.Count-1 do
	begin
	glPushMatrix;
	S1:=SL[I];
	SetTextPosition(I,S1);
	glListBase(GL3DFont.FFontDisplayList);
	glCallLists(length(S1),GL_UNSIGNED_BYTE,Pchar(S1));
	glPopMatrix;
	end;
finally SL.Free; end;
end;



function TGL3DString.ReadExtentMax: T3DVector;
begin
GetTextExtents;
Result:=FExtentMin;
end;


function TGL3DString.ReadExtentMin: T3DVector;
begin
GetTextExtents;
Result:=FExtentMax;
end;




procedure TGL3DString.Render;
begin
glPushAttrib(GL_ALL_ATTRIB_BITS);
try
inherited Render;
TextOut(FText);
finally glPopAttrib; end;
end;




procedure TGL3DString.SetColor(const Color: TColor);
begin
inherited;
GL3DFont.FDelphiFont.Color:=Color;
end;



function TGL3DString.GetColor: TColor;
begin
inherited GetColor;
Result:=GL3DFont.FDelphiFont.Color;
end;



procedure TGL3DString.SetTextPosition(Line: integer; S: string);
var Y,L,SP: double;
begin
SP:=(FHeight/2)-FCharHeight;
Y:=SP-((FCharHeight+FLineSpacing)* Line);
GetTextExtents;
L:=GetLineLength(S);
case Justify of
 tjLeft: glTranslatef(0,Y,0);
 tjCenter: glTranslatef(-L/2,Y,0);
 tjRight: glTranslatef(-L,Y,0);
 end;
end;



procedure TGL3DString.SetText(V: string);
begin
if FText<>V then
	begin
	FText:=V;
	Change;
	end;
end;




procedure TGL3DString.SetShowLineBreaks(const Value: boolean);
begin
if FShowLineBreaks<>Value then
	begin
	FShowLineBreaks:=Value;
	Change;
	end;
end;


procedure TGL3DString.SetLineSpacing(const Value: double);
begin
if FLineSpacing<>Value then
	begin
	FLineSpacing:=Value;
	Change;
	end;
end;





procedure TGL3DString.TextOut(S: string);
var R,G,B,A: single;
const S2 ='Frog';
begin
if S='' then exit;
GetTextExtents;
glPushAttrib(GL_LIST_BIT);
GetRGB(GL3DFont.FDelphiFont.Color,R,G,B);
A:=GetCurrentAlpha;
glColor4f(R,G,B,A);
if FRenderMode=rmWireFrame then glPolygonMode(GL_FRONT, GL_LINE);
if Self.FShowLineBreaks then
	begin
	OutputSeparateLines(S);
	end
else
	begin
	SetTextPosition(0,S);
	glListBase(GL3DFont.FFontDisplayList);
	glCallLists(length(S),GL_UNSIGNED_BYTE,Pchar(S));
	end;
glPopAttrib;
end;



{ TGLStringList }


procedure TGLStringList3D.HandleStringChange(Sender: TObject);
begin
Change;
end;



function TGLStringList3D.Add(var Str: TGL3DString): integer;
begin
Str:=TGL3DString.Create;
Str.OnChange:=HandleStringChange;
Str.Name:='String3D-'+IntToStr(Self.Count+1);
Result:=inherited Add(Str);
Self[Result].ResetTransforms;
Self[Result].ResetMaterials;
end;



function TGLStringList3D.AddFrom(Str: TGL3DString): integer;
var NewStr: TGL3DString;
begin
NewStr:=TGL3DString.CreateFrom(Str);
Result:=inherited Add(NewStr);
NewStr.Assign(Str);
end;


function TGLStringList3D.Add: integer;
var Str: TGL3DString;
begin
Result:=Add(Str);
end;



function TGLStringList3D.Insert(Inx: integer; var Str: TGL3DString): integer;
begin
Str:=TGL3DString.Create;
Str.Name:='String3D-'+IntToStr(Count+1);
inherited Insert(Inx,Str);
Result:=Inx;
end;



function TGLStringList3D.Insert(Inx: integer): integer;
var Str: TGL3DString;
begin
Result:=Insert(Inx,Str);
end;



function TGLStringList3D.GetItem(Index: Integer): TGL3DString;
begin
Result:=TGL3DString(inherited GetItem(Index));
end;


procedure TGLStringList3D.SetItem(Index: Integer; AObject: TGL3DString);
begin
inherited SetItem(Index, AObject);
end;



procedure TGLStringList3D.Assign(ObjectList: TObjectList);
var I,Inx: integer;
begin
if ObjectList is TGLStringList3D then
	begin
	Clear;
	for I:=0 to ObjectList.Count-1 do
		begin
		Inx:=AddFrom(TGL3DString(ObjectList[I]));
		Self[Inx].Assign(TGL3DString(ObjectList[I]));
		end;
	end
else inherited Assign(ObjectList);
end;




end.
