unit GLFrustumLib;

interface

type TFrustum = array [0..5,0..3] of double;


{Constants specifying frustum sides}

const FrustRight= 0; FrustLeft = 1;
const FrustBottom = 2; FrustTop = 3;
const FrustBack = 4; FrustFront = 5;

{Constants specifying the A,B,C,D elements of plane equation}
{Corresponds to X,Y,Z value of plane's normal and the distance from the orgin}

const PlaneA = 0; PlaneB = 1;
const PlaneC = 2; PlaneD = 3;

function CalculateFrustum: TFrustum;
function PointInFrustum(Frustum: TFrustum; X,Y,Z: double): boolean;

implementation

uses OpenGL15, GLMatrix;



procedure  NormalizePlane(Frustum: TFrustum;  Side : integer);
{Normalize specified plane which corresponds a side of the frustum}
var Magnitude: double;
begin 
{Calculate magnitude of normal}
Magnitude:=sqrt(Sqr(Frustum[Side,PlaneA]) + Sqr(Frustum[Side,PlaneB]) + Sqr(Frustum[Side,PlaneC]));

{ Then we divide the plane's values by it's magnitude. }
Frustum[Side,PlaneA]:= Frustum[Side,PlaneA] / Magnitude;
Frustum[Side,PlaneB]:= Frustum[Side,PlaneB] / Magnitude;
Frustum[Side,PlaneC]:= Frustum[Side,PlaneC] / Magnitude;
Frustum[Side,PlaneD]:= Frustum[Side,PlaneD] / Magnitude; 
end;



function CalculateFrustum: TFrustum;
{Extract the clipping planes for the current Frustum}
var ProjectMatrix,ModelMatrix,ClipMatrix: TGLMatrix;
begin
{Get projection matrix}
glGetDoublev(GL_PROJECTION_MATRIX, @ProjectMatrix);

{ Get model matrix }
glGetDoublev( GL_MODELVIEW_MATRIX, @ModelMatrix);

{Multiply the two matrices to transform the }
ClipMatrix:=MatrixMultiply(ModelMatrix,ProjectMatrix);

{ Extract the Right side of the frustum }
Result[FrustRight,PlaneA]:= ClipMatrix[ 3] - ClipMatrix[ 0];
Result[FrustRight,PlaneB]:= ClipMatrix[ 7] - ClipMatrix[ 4];
Result[FrustRight,PlaneC]:= ClipMatrix[11] - ClipMatrix[ 8];
Result[FrustRight,PlaneD]:= ClipMatrix[15] - ClipMatrix[12];

{ Normalize the Right side }
NormalizePlane(Result, FrustRight);

{ This will extract the Left side of the frustum }
Result[FrustLeft,PlaneA]:= ClipMatrix[ 3] + ClipMatrix[ 0];
Result[FrustLeft,PlaneB]:= ClipMatrix[ 7] + ClipMatrix[ 4];
Result[FrustLeft,PlaneC]:= ClipMatrix[11] + ClipMatrix[ 8];
Result[FrustLeft,PlaneD]:= ClipMatrix[15] + ClipMatrix[12];

{ Normalize the Left side }
NormalizePlane(Result, FrustLeft);

{ This will extract the Bottom side of the frustum }
Result[FrustBottom,PlaneA]:= ClipMatrix[ 3] + ClipMatrix[ 1];
Result[FrustBottom,PlaneB]:= ClipMatrix[ 7] + ClipMatrix[ 5];
Result[FrustBottom,PlaneC]:= ClipMatrix[11] + ClipMatrix[ 9];
Result[FrustBottom,PlaneD]:= ClipMatrix[15] + ClipMatrix[13];

{ Normalize the Bottom side }
NormalizePlane(Result, FrustBottom);

{ This will extract the Top side of the frustum }
Result[FrustTop,PlaneA]:= ClipMatrix[ 3] - ClipMatrix[ 1];
Result[FrustTop,PlaneB]:= ClipMatrix[ 7] - ClipMatrix[ 5];
Result[FrustTop,PlaneC]:= ClipMatrix[11] - ClipMatrix[ 9];
Result[FrustTop,PlaneD]:= ClipMatrix[15] - ClipMatrix[13];

{ Normalize the Top side }
NormalizePlane(Result, FrustTop);

{ This will extract the Back side of the frustum }
Result[FrustBack,PlaneA]:= ClipMatrix[ 3] - ClipMatrix[ 2];
Result[FrustBack,PlaneB]:= ClipMatrix[ 7] - ClipMatrix[ 6];
Result[FrustBack,PlaneC]:= ClipMatrix[11] - ClipMatrix[10];
Result[FrustBack,PlaneD]:= ClipMatrix[15] - ClipMatrix[14];

{ Normalize the Back side }
NormalizePlane(Result, FrustBack);

{ This will extract the Front side of the frustum }
Result[FrustFront,PlaneA]:= ClipMatrix[ 3] + ClipMatrix[ 2];
Result[FrustFront,PlaneB]:= ClipMatrix[ 7] + ClipMatrix[ 6];
Result[FrustFront,PlaneC]:= ClipMatrix[11] + ClipMatrix[10];
Result[FrustFront,PlaneD]:= ClipMatrix[15] + ClipMatrix[14];

{ Normalize the Front side }
NormalizePlane(Result, FrustFront);
end;



function PointInFrustum(Frustum: TFrustum; X,Y,Z: double): boolean;
{Calculate if point is inside frustrum}
begin
Result:=False;
{  Go through all six sides of the frustum }
{  Calculate the plane equation and check if the point is behind a side of the frustum }
if (Frustum[FrustRight, PlaneA]  * x + Frustum[FrustRight, PlaneB]  * y + Frustum[FrustRight, PlaneC]  * z + Frustum[FrustRight, PlaneD]) <= 0 then exit;
if (Frustum[FrustLeft,  PlaneA]  * x + Frustum[FrustLeft,  PlaneB]  * y + Frustum[FrustLeft,  PlaneC]  * z + Frustum[FrustLeft,  PlaneD]) <= 0 then exit;
if (Frustum[FrustBottom,PlaneA]  * x + Frustum[FrustBottom,PlaneB]  * y + Frustum[FrustBottom,PlaneC]  * z + Frustum[FrustBottom,PlaneD]) <= 0 then exit;
if (Frustum[FrustTop,   PlaneA]  * x + Frustum[FrustTop,   PlaneB]  * y + Frustum[FrustTop,   PlaneC]  * z + Frustum[FrustTop,   PlaneD]) <= 0 then exit;
if (Frustum[FrustBack,  PlaneA]  * x + Frustum[FrustBack,  PlaneB]  * y + Frustum[FrustBack,  PlaneC]  * z + Frustum[FrustBack,  PlaneD]) <= 0 then exit;
if (Frustum[FrustFront, PlaneA]  * x + Frustum[FrustFront, PlaneB]  * y + Frustum[FrustFront, PlaneC]  * z + Frustum[FrustFront, PlaneD]) <= 0 then exit;
Result:=True;
end;



end.
