unit GLLights;

interface

uses Graphics,OpenGL15,GLLibrary;


type TLightType = (ltPoint,ltSpot,ltDirectional);

type TGLLightSource = class(TObject)
 private
  FEnabled: boolean;
 public
  Ambient,Diffuse,Specular: TGLArrayf4;
  SpotCutoff: TGLfloat;
  Position: TGLArrayf4;
  SpotDirection: TGLArrayf3;
  DirectionalDir: TGLArrayf4;
  LightNumber: TGLenum;
  LightType: TLightType;
  ConstantAttenuation: TGLfloat;
  LinearAttenuation: TGLfloat;
  QuadraticAttenuation: TGLfloat;
  constructor Create;
  procedure Reset;
  procedure RenderLight;
  property Enabled: boolean read FEnabled write FEnabled;
 end;






const MaxLights = 8;

type TLightArray = array [0..MaxLights-1] of TGLLightSource;

type TGLLightArray = class(TObject)
 private
  FLights: TLightArray;
  function GetLights(index: integer): TGLLightSource;
 public
  constructor Create;
  destructor Destroy; override;
  procedure Reset;
  procedure RenderLights;
  property Lights[index: integer]: TGLLightSource read GetLights; default;
 end;



implementation


{------------------ Light Source ----------------------------------------------}



procedure TGLLightSource.Reset;
begin
Position:=MakeGLArrayf4(8,8,16,1);
Ambient:=MakeGLArrayf4(0,0,0,1);
Diffuse:=MakeGLArrayf4(1,1,1,1);
Specular:=MakeGLArrayf4(1,1,1,1);
SpotDirection:=MakeGLArrayf3(0,0,-1);
DirectionalDir:=MakeGLArrayf4(0,0,1,0);
SpotCutoff:=15;
ConstantAttenuation:= 1;
LinearAttenuation:= 0;
QuadraticAttenuation:= 0;
LightType:=ltPoint;
end;



constructor TGLLightSource.Create;
begin
Reset;
LightNumber:=GL_LIGHT0;
end;



procedure TGLLightSource.RenderLight;

	procedure HandlePointLight;
	begin
	glLightf(LightNumber, GL_SPOT_CUTOFF, 180);
	Position[3]:=1;
	glLightfv(LightNumber,GL_POSITION,@Position);
	end;

	procedure HandleSpotLight;
	begin
	glLightf(LightNumber, GL_SPOT_CUTOFF, SpotCutoff);
	glLightfv(LightNumber, GL_SPOT_DIRECTION, @SpotDirection);
	glLightfv(LightNumber,GL_POSITION,@Position);
	end;

	procedure HandleDirectionalLight;
	begin
	glLightf(LightNumber, GL_SPOT_CUTOFF, 180);
	DirectionalDir[3]:=0;
	glLightfv(LightNumber,GL_POSITION,@DirectionalDir);
	end;



begin
if FEnabled then
	begin
	glEnable(LightNumber);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity;
	glLightfv(LightNumber,GL_AMBIENT,@Ambient);
	glLightfv(LightNumber,GL_DIFFUSE,@Diffuse);
	glLightfv(LightNumber,GL_SPECULAR,@Specular);

	glLightf(LightNumber, GL_CONSTANT_ATTENUATION, ConstantAttenuation);
	glLightf(LightNumber, GL_LINEAR_ATTENUATION, LinearAttenuation);
	glLightf(LightNumber, GL_QUADRATIC_ATTENUATION, QuadraticAttenuation);

	case LightType of
	 ltPoint: HandlePointLight;
	 ltSpot: HandleSpotLight;
	 ltDirectional: HandleDirectionalLight;
	 end;
	end
else glDisable(LightNumber);
end;



{------------------ TGLLightArray ---------------------------------------------}




constructor TGLLightArray.Create;
var I: integer;
begin
for I:=0 to MaxLights-1 do
	begin
	FLights[I]:=TGLLightSource.Create;
	FLights[I].LightNumber:=GL_LIGHT0 + I;
	end;
end;



destructor TGLLightArray.Destroy;
var I: integer;
begin
for I:=0 to MaxLights-1 do FLights[I].Free;
inherited;
end;


function TGLLightArray.GetLights(index: integer): TGLLightSource;
begin
Result:=FLights[Index];
end;

procedure TGLLightArray.RenderLights;
var I: integer;
begin
for I:=0 to MaxLights-1 do Lights[I].RenderLight;
end;




procedure TGLLightArray.Reset;
var I: integer;
begin
for I:=0 to MaxLights-1 do Lights[I].Reset;
end;



end.
