unit GLLines;

{This unit hold the TGLLines object which handles the creation, memory
allocation, normal generation, and rendering of an array of lines.}

interface

uses Windows, Classes, SysUtils, Graphics, Contnrs, OpenGL15, Vector, Math, HugeArry,
     GLQuatLib, GLLibrary, GLOctTree, GLRenderer,GLMatrix,StrokeFontData,StrokeFontLib;


{ Define line properties }

type TLineProperty = record
 Width: integer;
 Stipple: boolean;
 Pattern: integer;
 Factor: integer;
 end;

{ Define an array of line properties }

type TLinePropertyArray = class(THugeArray)
 private
  procedure Put(Index: integer; Item: TLineProperty); virtual;
  function Get(Index: integer): TLineProperty;
 protected
 public
  constructor Create(Number: integer);
  function Add(Item: TLineProperty): integer; reintroduce; overload; virtual;
  function AddPointer(Item: TLineProperty; Ptr: integer): Integer; virtual;
  procedure Insert(Value: TLineProperty; Index: integer); virtual;
  procedure Delete(Index: integer);
  procedure Move(CurIndex, NewIndex: Integer);
  property Value[Index: integer]: TLineProperty read Get write Put; default;
 end;

{Defines call back progress routine for mesh building process}

type TGLProgress= procedure(Sender: TObject; Action: string; PercentDone: integer);

{Mesh object declaration}

type TGLLines = class(TGLRenderer)
 private
  Sorter: TList;
  FRenderMode: TRenderMode;
  FOnNormalProgress: TGLProgress;
  procedure DoLineColor(I: integer);
  procedure DoLine(I: integer);
  procedure DrawLine(Inx,Cell: integer);
  procedure SetRenderMode(V: TRenderMode); reintroduce; virtual;
  procedure SetMinMax(T: T3DLine);
  procedure SetupOctaltree;
  procedure RenderLines;
  procedure DoProperties(I: integer);
  procedure DrawFontStroke(P1, P2, Pos: T3DVector; Height: integer; Color: TColor);
 protected
 public
  Vertices: THuge3DVectorEx;
  LineColor: THugeLongInt;
  Properties: TLinePropertyArray;
  TreeBuilt: boolean;
  UseDisplayList: boolean;
  OctalTree: TOctTree;
  StrokeFont: TStrokeFont;
  constructor Create; override;
  destructor Destroy; override;
  procedure Append(GLObject: TGLObject); override;
  procedure AddLine(T: T3DLine; C: TColor; P: TLineProperty); overload;
  procedure AddLine(T: T3DLine; C: TColor); overload;
  procedure AddLine(P1,P2: T3DVector; C: TColor); overload;
  procedure AddLine(P1,P2: T3DVector; C: TColor; Size: integer); overload;
  procedure AddLine(T: T3DLine; C: TColor; Width: integer); overload;
  procedure SetLine(Inx: integer; C: TColor; P: TLineProperty);
  procedure SetColor(const Color: TColor); override;
  procedure SetWidth(const Width: integer);
  procedure LinesClear;
  procedure Clear; override;
  procedure ClearData;
  procedure Render; override;
  procedure DisplayProgress(Status: string; Progress,Total: integer);
  property RenderMode: TRenderMode read FRenderMode write SetRenderMode;
  property OnNormalProgress: TGLProgress read FOnNormalProgress write FOnNormalProgress;
  function GetColor: TColor; override;
  function GetWidth: integer;
 end;


{ TGLLineList }


type  TGLLineList = class(TGLObjectList)
  protected
    function GetItem(Index: Integer): TGLLines;
    procedure SetItem(Index: Integer; AObject: TGLLines);
  public
    function Add(var Lines: TGLLines): integer; overload;
    function Add: integer; overload;
    property Items[Index: Integer]: TGLLines read GetItem write SetItem; default;
    procedure Assign(ObjectList: TObjectList);
  end;

implementation


{--------------------------------------------------------------------------------}
{  Implementation for huge Line Property array object based on huge array object }
{--------------------------------------------------------------------------------}

constructor TLinePropertyArray.Create(Number: Integer);
{Create a huge property array}
begin
inherited Create(Number,SizeOf(TLineProperty));
end;


function TLinePropertyArray.Add(Item: TLineProperty): integer;
{Add item to property array}
begin
Result:=inherited Add(Item);
end;


function TLinePropertyArray.AddPointer(Item: TLineProperty; Ptr: integer): Integer;
begin
Result:=inherited AddPointer(Item,Ptr);
end;



procedure TLinePropertyArray.Put(Index: integer; Item: TLineProperty);
{Insert property at index}
var PTR: ^TLineProperty;
begin
PTR:=GetItemPtr(Index);
PTR^:=Item;
end;



function TLinePropertyArray.Get(Index: integer): TLineProperty;
{Get property at the index}
var PTR: ^TLineProperty;
begin
PTR:=GetItemPtr(Index);
Result:=PTR^;
end;


procedure TLinePropertyArray.Insert(Value: TLineProperty; Index: integer);
begin
InsertItem(Value,Index);
end;


procedure TLinePropertyArray.Delete(Index: integer);
begin
DeleteItem(Index);
end;



procedure TLinePropertyArray.Move(CurIndex, NewIndex: Integer);
var T: TLineProperty;
begin
T:=Value[CurIndex];
Delete(CurIndex);
Insert(T,NewIndex);
end;



{------------------ Property Setting Routines ---------------------------------}



procedure TGLLines.SetRenderMode(V: TRenderMode);
{Set current render mode}
begin
if RenderMode<>V then
	begin
	FRenderMode:=V;
	TreeBuilt:=False;
	end;
end;





procedure TGLLines.DisplayProgress(Status: string; Progress,Total: integer);
{This routine calls 'progress" callback routine if the user has set one up}
begin
if not Assigned(FOnNormalProgress) then exit;
if (Progress mod (Total div 100 + 1))<>0 then exit;
FOnNormalProgress(Self,Status,MulDiv(100,Progress,Total));
end;



{----------------- Stroke Font Routines ----------------------------------}



procedure TGLLines.DrawFontStroke(P1,P2,Pos: T3DVector; Height: integer; Color: TColor);
begin
AddLine(P1,P2,Color)
end;



{---------------- Object Event handlers ----------------------------------}


constructor TGLLines.Create;
begin
inherited;
Vertices:=THuge3DVectorEx.CreateAsLines(5);
LineColor:=THugeLongInt.Create(5);
Properties:=TLinePropertyArray.Create(5);
Sorter:=TList.Create;
OctalTree:=TOctTree.Create;
StrokeFont:=TStrokeFont.Create;
StrokeFont.LineDrawProc:=DrawFontStroke;
FRenderMode:=rmFlatFace;
TreeBuilt:=False;
Position:=DefaultPos;
ResetPos:=DefaultPos;
UseDisplayList:=False;
end;



destructor TGLLines.Destroy;
begin
StrokeFont.Free;
Vertices.Free;
LineColor.Free;
Properties.Free;
Sorter.Free;
OctalTree.Free;
inherited Destroy;
end;


{---------------- Routines to add Lines to the list ---------------------------}


procedure TGLLines.SetMinMax(T: T3DLine);
{Set min/max from of mesh}
begin
if Vertices.ItemIndex=0 then
	begin
	ExtentMin:=T.P1;
	ExtentMax:=T.P1;
	ExtentMin:=GetMinimumVector(ExtentMin,T.P2);
	ExtentMax:=GetMaximumVector(ExtentMax,T.P2);
	end
else
	begin
	ExtentMin:=GetMinimumVector(ExtentMin,T.P1);
	ExtentMin:=GetMinimumVector(ExtentMin,T.P2);
	ExtentMax:=GetMaximumVector(ExtentMax,T.P1);
	ExtentMax:=GetMaximumVector(ExtentMax,T.P2);
	end;
end;


procedure TGLLines.SetLine(Inx: integer; C: TColor; P: TLineProperty);
begin
LineColor[Inx]:=C;
Properties[Inx]:=P;
end;


procedure TGLLines.AddLine(T: T3DLine; C: TColor; P: TLineProperty);
{Add a Line to the array}
begin
SetMinMax(T);
Vertices.AddLine(T);
LineColor.Add(C);
Properties.Add(P);
TreeBuilt:=false;
end;


procedure TGLLines.AddLine(T: T3DLine; C: TColor; Width: integer);
{Add a Line to the array}
var P: TLineProperty;
begin
P.Width:=Width;
P.Stipple:=False;
P.Pattern:=0;
P.Factor:=1;
AddLine(T,C,P);
end;


procedure TGLLines.AddLine(T: T3DLine; C: TColor);
{Add a Line to the array}
begin
AddLine(T,C,1);
end;


procedure TGLLines.AddLine(P1,P2: T3DVector; C: TColor);
{Add a Line to the array}
var L: T3DLine;
begin
L.P1:=P1; L.P2:=P2;
AddLine(L,C);
end;



procedure TGLLines.AddLine(P1,P2: T3DVector; C: TColor; Size: integer);
{Add a Line to the array}
var L: T3DLine;
begin
L.P1:=P1; L.P2:=P2;
AddLine(L,C,Size);
end;


procedure TGLLines.LinesClear;
{Clear Lines only leaving position, rotation and scaling}
begin
Vertices.Clear;
LineColor.Clear;
Properties.Clear;
Sorter.Clear;
TreeBuilt:=false;
ExtentMin:=MakeVector3D(0,0,0);
ExtentMax:=MakeVector3D(0,0,0);
end;



procedure TGLLines.Clear;
{Clear the Lines}
begin
inherited Clear;
LinesClear;
end;



procedure TGLLines.ClearData;
{Clear the Lines without reseting transforms}
begin
LinesClear;
end;


procedure TGLLines.Append(GLObject: TGLObject);
begin
Clear;
inherited Assign(GLObject);
if GLObject is TGLLines then
	begin
	Vertices.Assign(TGLLines(GLObject).Vertices);
	LineColor.Assign(TGLLines(GLObject).LineColor);
	Properties.Assign(TGLLines(GLObject).Properties);
	end;
end;


{-------------- Routines To Render Lines --------------------------------------}



procedure TGLLines.DoLineColor(I: integer);
{Tell OpenGL about the color of specified line}
var R,G,B,A: single;
begin
GetRGB(LineColor.Value[I],R,G,B);
A:=GetCurrentAlpha;
glColor4f(R,G,B,A);
end;




procedure TGLLines.DoProperties(I: integer);
{Tell OpenGL about the properties of specified line}
begin
glLineWidth(Properties[I].Width);
glLineStipple(Properties[I].Factor, Properties[I].Pattern);
if Properties[I].Stipple then glEnable(GL_LINE_STIPPLE)
else glDisable(GL_LINE_STIPPLE);
end;



procedure TGLLines.DoLine(I: integer);
{Tell OpenGL about specified line }
var T: T3DLine;
begin
T:=Vertices.Line[I];
glBegin(GL_LINES);
glVertex3dv(@T.P1); glVertex3dv(@T.P2);
glEnd;
end;



procedure TGLLines.DrawLine(Inx,Cell: integer);
begin
DoLineColor(Inx);
DoProperties(Inx);
DoLine(Inx);
end;




procedure TGLLines.RenderLines;
{Handle displaying normal GL mode}
begin
glDisable(GL_LIGHTING);
OctalTree.IterateLines(DrawLine);
{glEnd not needed here}
end;




procedure TGLLines.SetupOctalTree;
{Put all Lines in the boxes}
{This is only done when the list are changed}
var I: integer;
begin
OctalTree.ResetBoxes;
OctalTree.MeshDimensions:=MakeBox(ExtentMin,ExtentMax);
for I:=0 to Vertices.LineCount-1 do
	begin
	{If Lines span boxes, we'll use the box for the first point}
	OctalTree.PutLineInBox(Vertices.Line[I],I);
	end;
TreeBuilt:=True;
end;



procedure TGLLines.Render;
{Render whole mesh}
begin
glPushAttrib(GL_ALL_ATTRIB_BITS);
try
if not TreeBuilt then SetupOctalTree;
inherited Render;
OctalTree.FlagBoxesInFrustum;
RenderLines;
finally glPopAttrib; end;
end;



procedure TGLLines.SetColor(const Color: TColor);
var I: integer;
begin
inherited;
for I:=0 to LineColor.Count-1 do LineColor[I]:=Color;
end;



procedure TGLLines.SetWidth(const Width: integer);
var I: integer;
begin
for I:=0 to Properties.Count-1 do
	begin
	TLineProperty(Properties.GetItemPtr(I)^).Width:=Width;
	end;
end;



function TGLLines.GetColor: TColor;
begin
inherited GetColor;
if LineColor.Count>0 then Result:=LineColor[0]
else Result:=clWhite;
end;



function TGLLines.GetWidth: integer;
begin
if Properties.Count>0 then Result:=Properties[0].Width
else Result:=1;
end;




{ TGLLineList }

function TGLLineList.Add(var Lines: TGLLines): integer;
begin
Lines:=TGLLines.Create;
Result:=inherited Add(Lines);
Lines.Name:='Lines-'+IntToStr(Result+1);
Change;
end;




function TGLLineList.Add: integer;
var Lines: TGLLines;
begin
Result:=Add(Lines);
end;



procedure TGLLineList.Assign(ObjectList: TObjectList);
var I: integer;
var O: TGLLines;
begin
if ObjectList is TGLLineList then
	begin
	Clear;
	for I:=0 to ObjectList.Count-1 do
		begin
		Add(O);
		O.Assign(TGLLines(ObjectList[I]));
		end;
	end
else inherited Assign(ObjectList);
end;



function TGLLineList.GetItem(Index: Integer): TGLLines;
begin
Result := TGLLines(inherited GetItem(Index));
end;


procedure TGLLineList.SetItem(Index: Integer; AObject: TGLLines);
begin
inherited SetItem(Index, AObject);
end;





end.
