unit GLMatrix;

{Libary for manipulating OpenGL-style homongenous coordinate matrices}

interface

uses Types,Graphics,SysUtils,Grids,Math,OpenGL15,Vector,DSupport,VCLLibrary;


{

===================================================
Notes on matrix layout and display
===================================================
In the 3D graphics, matrices can be stored in memory
in two different ways:

Storage Option 1: A,B,C,D,E,F,G,H,J,K,L,M,N,O,P,Q
Storage Option 2: A,E,J,N,B,F,K,O,C,G,L,P,D,H,M,Q

OpenGL and DirectX both use option #2

===================================================
Matrices can be type set and displayed two different
ways:

----------------------------------------------------
Typesetting Option 1  - The columns are vectors
The translation vector is in parenthesis
---------------
A   B   C   (D)
E   F   G   (H)
J   K   L   (M)
N   O   P   (Q)
---------------
OpenGL uses the column major format in docs

----------------------------------------------------
Typesetting Option 2 - The rows are vectors
The translation vector is in parenthesis
---------------
 A   E   J   N
 B   F   K   O
 C   G   L   P
(D) (H) (M) (Q)
---------------

DirectX uses the column major format in docs

Translation from 1-D to 2-D arrays

0  [0,0]              [0,0] 0
1  [0,1]              [1,0] 4
2  [0,2]              [2,0] 8
3  [0,3]              [3,0] 12
4  [1,0]              [0,1] 1
5  [1,1]              [1,1] 5
6  [1,2]              [2,1] 9
7  [1,3]              [3,1] 13
8  [2,0]              [0,2] 2
9  [2,1]              [1,2] 6
10 [2,2]              [2,2] 10
11 [2,3]              [3,2] 14
12 [3,0]              [0,3] 3
13 [3,1]              [1,3] 7
14 [3,2]              [2,3] 11
15 [3,3]              [3,3] 15

}

type TGLMatrix = array [0..15] of double;
type TStandMatrix = array [1..4,1..4] of double;
type TProgMatrix = array [0..3,0..3] of double;
type TColVector = array [0..3] of double;


const IdentityMatrix: TGLMatrix = (1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1);
const GLIndentityMatrix: TGLMatrixd4 = ((1,0,0,0),(0,1,0,0),(0,0,1,0),(0,0,0,1));


procedure MatrixToEulerAngles(M: TGLMatrix; var Pitch,Yaw,Roll: double);
function MatrixVectorMult(M: TGLMatrix; V: T3DVector): T3DVector;
function MatrixVectorMultRow(M: TGLMatrix; V: T3DVector): T3DVector;
function MatrixTriangleMult(M: TGLMatrix; T: TTriangle): TTriangle;
function MatrixVectorMultHom(M: TGLMatrix; V: T3DVector): T3DVector;

function MatrixVectorMultOld(M: TGLMatrix; V: T3DVector): T3DVector;
function MatrixMultiply(const M1, M2: TGLMatrix): TGLMatrix;
function TransposeMatrix(M: TGLMatrix): TGLMatrix;
function GetMatrixStr(M: TGLMatrix; Digits: integer): string;
procedure MatrixToGrid(Grid: TDrawGrid; Matrix: TGLMatrix; ACol, ARow, Digits: Integer; Rect: TRect);

function RemoveGLMatrixScaling(M: TGLMatrix): TGLMatrix;
function SetGLMatrixScale(Scale: T3DVector): TGLMatrix;
function GetGLMatrixScale(M: TGLMatrix): T3DVector;
function GetGLMatrixPosition(M: TGLMatrix): T3DVector;
function RemoveGLMatrixPosition(M: TGLMatrix): TGLMatrix;
function InvertGLMatrix(M: TGLMatrix): TGLMatrix;
function GetGLMatrixDeterminant(M: TGLMatrix): double;
function GetMatrixRotation(M: TGLMatrix): TGLMatrix;
function DirectionVectorToMatrix(Up,Right,Front: T3DVector): TGLMatrix; overload;
function DirectionVectorToMatrix(Front: T3DVector; Swap: boolean): TGLMatrix; overload;
function AxisAngleToMatrix(Axis: T3DVector; Angle: double): TGLMatrix;
function GetAlignMatrix(V1,V2: T3DVector): TGLMatrix;
function GetDirAlignMatrix(P1,P2,Dir: T3DVector): TGLMatrix;



function GetTranslationMatrix(V: T3DVector): TGLMatrix;
function GetTranslationVector(M: TGLMatrix): T3DVector;
function AddTranslation(M: TGLMatrix; V: T3DVector): TGLMatrix;
function SubtractTranslation(M: TGLMatrix; V: T3DVector): TGLMatrix;
function GetScaleMatrix(V: T3DVector): TGLMatrix;
function AddScale(M: TGLMatrix; V: T3DVector): TGLMatrix;
function AddRotation(M: TGLMatrix; X,Y,Z: double): TGLMatrix;
function AddOrthoProjection(M: TGLMatrix; Min,Max: T3DVector): TGLMatrix;
function GetOrthoProjMatrix(Min,Max: T3DVector): TGLMatrix;
function GetPerspectiveMatrix(Min,Max: T3DVector): TGLMatrix;
function AddPerspective(M: TGLMatrix; Min,Max: T3DVector): TGLMatrix;

function MatrixScalarMultiply(M: TGLMatrix; S: double): TGLMatrix;
function MatrixAdd(M1,M2: TGLMatrix): TGLMatrix;



implementation

uses GLQuatLib;




function TransposeMatrix(M: TGLMatrix): TGLMatrix;
begin
Result[0]:=M[0]; Result[1]:=M[4]; Result[2]:=M[8]; Result[3]:=M[12];
Result[4]:=M[1]; Result[5]:=M[5]; Result[6]:=M[9]; Result[7]:=M[13];
Result[8]:=M[2]; Result[9]:=M[6]; Result[10]:=M[10]; Result[11]:=M[14];
Result[12]:=M[3]; Result[13]:=M[7]; Result[14]:=M[11]; Result[15]:=M[15];
end;




function GetGLMatrixDeterminant(M: TGLMatrix): double;
{Get the determinate of a specified matrix}
begin
Result:=M[12]*M[9]*M[6]*M[3] - M[8]*M[13]*M[6]*M[3] - M[12]*M[5]*M[10]*M[3] +
	M[4]*M[13]*M[10]*M[3] + M[8]*M[5]*M[14]*M[3] - M[4]*M[9]*M[14]*M[3] -
	M[12]*M[9]*M[2]*M[7] + M[8]*M[13]*M[2]*M[7] + M[12]*M[1]*M[10]*M[7] -
	M[0]*M[13]*M[10]*M[7] - M[8]*M[1]*M[14]*M[7] + M[0]*M[9]*M[14]*M[7] +
	M[12]*M[5]*M[2]*M[11] - M[4]*M[13]*M[2]*M[11] - M[12]*M[1]*M[6]*M[11] +
	M[0]*M[13]*M[6]*M[11] + M[4]*M[1]*M[14]*M[11] - M[0]*M[5]*M[14]*M[11] -
	M[8]*M[5]*M[2]*M[15] + M[4]*M[9]*M[2]*M[15] + M[8]*M[1]*M[6]*M[15] -
	M[0]*M[9]*M[6]*M[15] -  M[4]*M[1]*M[10]*M[15] +  M[0]*M[5]*M[10]*M[15];
end;



function InvertGLMatrix(M: TGLMatrix): TGLMatrix;
{Invert the specified matrix- This reverses the action of every operation in the matrix}
{This is why a Maxtrix times its inverse yields an Identity Matrix}
var X: double;
begin
X:=GetGLMatrixDeterminant(M);
if X=0 then raise Exception.Create('Invert Error: Singular Matrix');
Result[0]:= (-M[13]*M[10]*M[7] +M[9]*M[14]*M[7] +M[13]*M[6]*M[11]-M[5]*M[14]*M[11] -M[9]*M[6]*M[15] +M[5]*M[10]*M[15])/x;
Result[4]:= ( M[12]*M[10]*M[7] -M[8]*M[14]*M[7] -M[12]*M[6]*M[11]+M[4]*M[14]*M[11] +M[8]*M[6]*M[15] -M[4]*M[10]*M[15])/x;
Result[8]:= (-M[12]*M[9]* M[7] +M[8]*M[13]*M[7] +M[12]*M[5]*M[11]-M[4]*M[13]*M[11] -M[8]*M[5]*M[15] +M[4]*M[9]* M[15])/x;
Result[12]:=( M[12]*M[9]* M[6] -M[8]*M[13]*M[6] -M[12]*M[5]*M[10]+M[4]*M[13]*M[10] +M[8]*M[5]*M[14] -M[4]*M[9]* M[14])/x;
Result[1]:= ( M[13]*M[10]*M[3] -M[9]*M[14]*M[3] -M[13]*M[2]*M[11]+M[1]*M[14]*M[11] +M[9]*M[2]*M[15] -M[1]*M[10]*M[15])/x;
Result[5]:= (-M[12]*M[10]*M[3] +M[8]*M[14]*M[3] +M[12]*M[2]*M[11]-M[0]*M[14]*M[11] -M[8]*M[2]*M[15] +M[0]*M[10]*M[15])/x;
Result[9]:= ( M[12]*M[9]* M[3] -M[8]*M[13]*M[3] -M[12]*M[1]*M[11]+M[0]*M[13]*M[11] +M[8]*M[1]*M[15] -M[0]*M[9]* M[15])/x;
Result[13]:=(-M[12]*M[9]* M[2] +M[8]*M[13]*M[2] +M[12]*M[1]*M[10]-M[0]*M[13]*M[10] -M[8]*M[1]*M[14] +M[0]*M[9]* M[14])/x;
Result[2]:= (-M[13]*M[6]* M[3] +M[5]*M[14]*M[3] +M[13]*M[2]*M[7]-M[1]*M[14]*M[7] -M[5]*M[2]*M[15] +M[1]*M[6]* M[15])/x;
Result[6]:= ( M[12]*M[6]* M[3] -M[4]*M[14]*M[3] -M[12]*M[2]*M[7]+M[0]*M[14]*M[7] +M[4]*M[2]*M[15] -M[0]*M[6]* M[15])/x;
Result[10]:=(-M[12]*M[5]* M[3] +M[4]*M[13]*M[3] +M[12]*M[1]*M[7]-M[0]*M[13]*M[7] -M[4]*M[1]*M[15] +M[0]*M[5]* M[15])/x;
Result[14]:=( M[12]*M[5]* M[2] -M[4]*M[13]*M[2] -M[12]*M[1]*M[6]+M[0]*M[13]*M[6] +M[4]*M[1]*M[14] -M[0]*M[5]* M[14])/x;
Result[3]:= ( M[9]* M[6]* M[3] -M[5]*M[10]*M[3] -M[9]* M[2]*M[7]+M[1]*M[10]*M[7] +M[5]*M[2]*M[11] -M[1]*M[6]* M[11])/x;
Result[7]:= (-M[8]* M[6]* M[3] +M[4]*M[10]*M[3] +M[8]* M[2]*M[7]-M[0]*M[10]*M[7] -M[4]*M[2]*M[11] +M[0]*M[6]* M[11])/x;
Result[11]:=( M[8]* M[5]* M[3] -M[4]*M[9]* M[3] -M[8]* M[1]*M[7]+M[0]*M[9]* M[7] +M[4]*M[1]*M[11] -M[0]*M[5]* M[11])/x;
Result[15]:=(-M[8]* M[5]* M[2] +M[4]*M[9]* M[2] +M[8]* M[1]*M[6]-M[0]*M[9]* M[6] -M[4]*M[1]*M[10] +M[0]*M[5]* M[10])/x;
end;



function GetGLMatrixScale(M: TGLMatrix): T3DVector;
{Get the scale vector from Matrix}
var Up,Right,Frwrd: T3DVector;
var PM: TProgMatrix;
begin
PM:=TProgMatrix(M);
{Scale is just the magnitude of the three column vectors}
Up:=MakeVector3D(PM[0,0],PM[0,1],PM[0,2]);
Right:=MakeVector3D(PM[1,0],PM[1,1],PM[1,2]);
Frwrd:=MakeVector3D(PM[2,0],PM[2,1],PM[2,2]);
Result:=MakeVector3D(VectorAbs3D(Up), VectorAbs3D(Right), VectorAbs3D(Frwrd))
end;




function SetGLMatrixScale(Scale: T3DVector): TGLMatrix;
{Make matrix inserting scale factors}
begin
Result:=IdentityMatrix;
TProgMatrix(Result)[0,0]:=Scale.X;
TProgMatrix(Result)[1,1]:=Scale.Y;
TProgMatrix(Result)[2,2]:=Scale.Z;
end;



function RemoveGLMatrixScaling(M: TGLMatrix): TGLMatrix;
{Remove scaling from matrix}
var Scale: T3DVector;
var PM: TProgMatrix;
begin
PM:=TProgMatrix(M);
Scale:=GetGLMatrixScale(M);
PM[0,0]:=PM[0,0]/Scale.X; PM[1,0]:=PM[1,0]/Scale.Y; PM[2,0]:=PM[2,0]/Scale.Z;
PM[0,1]:=PM[0,1]/Scale.X; PM[1,1]:=PM[1,1]/Scale.Y; PM[2,1]:=PM[2,1]/Scale.Z;
PM[0,2]:=PM[0,2]/Scale.X; PM[1,2]:=PM[1,2]/Scale.Y; PM[2,2]:=PM[2,2]/Scale.Z;
Result:=TGLMatrix(PM);
end;




function MatrixMultiply(const M1, M2: TGLMatrix): TGLMatrix;
var I, J: Integer;
begin
for I := 0 to 3 do
 for J := 0 to 3 do
   TProgMatrix(Result)[I, J] := TProgMatrix(M1)[I, 0] * TProgMatrix(M2)[0, J] +
				TProgMatrix(M1)[I, 1] * TProgMatrix(M2)[1, J] +
				TProgMatrix(M1)[I, 2] * TProgMatrix(M2)[2, J] +
				TProgMatrix(M1)[I, 3] * TProgMatrix(M2)[3, J];
end;





function GetMatrixStr(M: TGLMatrix; Digits: integer): string;
{Displays matrix as column vectors which is the Opengl text format}
{Note: this display is oposite of the way it appears in memory. See notes at the top}
var X,Y: integer;
begin
Result:='';
for Y:=0 to 3 do
	begin
	for X:=0 to 3 do
		begin
		Result:=Result+FloatToStrF(TProgMatrix(M)[X,Y],ffFixed,18,Digits);
		if X=3 then Result:=Result+CRLF_Char
		else Result:=Result+', ';
		end;
	end;
end;




procedure MatrixToGrid(Grid: TDrawGrid; Matrix: TGLMatrix; ACol, ARow, Digits: Integer; Rect: TRect);
{Draw matrix into grid}
var X,Y: integer;
begin
X:=ACol-1;
Y:=ARow-1;
if ARow=0 then
	begin
	if ACol = 0 then OwnerDrawItem(Grid.Canvas,'#',Rect)
	else OwnerDrawItem(Grid.Canvas,IntToStr(X),Rect);
	end
else
	begin
	if ACol = 0 then OwnerDrawItem(Grid.Canvas,IntToStr(Y),Rect)
	else OwnerDrawItem(Grid.Canvas,FloatToStrF(TProgMatrix(Matrix)[X,Y],ffFixed,18,Digits),Rect);
	end;
end;


{-----------------------------------------------------------------------------}

function MatrixVectorMultOld(M: TGLMatrix; V: T3DVector): T3DVector;
{Multiplye GLMatrix by vector}
var X,Y: integer;
var M1: TProgMatrix;
var Vect,Final: TColVector;
var Sum: double;
begin
Vect[0]:=V.X; Vect[1]:= V.Y; Vect[2]:= V.Z; Vect[3]:= 1;
M1:=TProgMatrix(M);

for X:= 0 to 3 do
	begin
	Sum:=0;
	for Y:=0 to 3 do Sum:=Sum + M1[X,Y] * Vect[Y];
	Final[X]:=Sum;
	end;
Result:=MakeVector3D(Final[0],Final[1],Final[2])
end;


function MatrixVectorMultRow(M: TGLMatrix; V: T3DVector): T3DVector;
{Matrix-Vector multiply for row vector memory storage order }
var W: double;
begin
W:=1;
Result.X:= V.X * M[0] + V.Y * M[1] + V.Z * M[2] + W * M[3];
Result.Y:= V.X * M[4] + V.Y * M[5] + V.Z * M[6] + W * M[7];
Result.Z:= V.X * M[8] + V.Y * M[9] + V.Z * M[10] + W * M[11];
end;



function MatrixVectorMult(M: TGLMatrix; V: T3DVector): T3DVector;
{Matrix-Vector multiply for column vector memory storage }
{The Format by DirectX and OpenGL  }
begin
Result.X:= V.X * M[0] + V.Y * M[4] + V.Z * M[8] +  M[12];
Result.Y:= V.X * M[1] + V.Y * M[5] + V.Z * M[9] +  M[13];
Result.Z:= V.X * M[2] + V.Y * M[6] + V.Z * M[10] + M[14];
end;




function MatrixVectorMultHom(M: TGLMatrix; V: T3DVector): T3DVector;
{Matrix-Vector multiply for column vector memory storage }
{Coordinates are homogenized by dividing through by W}
{Homogenization used for perspective transforms }
{Normally this would be a 4x4 matrix times a 4D vector }
{The W would come from the Vector. In this case W=1 }
var W: double;
begin
Result.X:= V.X * M[0] + V.Y * M[4] + V.Z * M[8] +  M[12];
Result.Y:= V.X * M[1] + V.Y * M[5] + V.Z * M[9] +  M[13];
Result.Z:= V.X * M[2] + V.Y * M[6] + V.Z * M[10] + M[14];
       W:= V.X * M[3] + V.Y * M[7] + V.Z * M[11] + M[15];
Result:=ScalarDivide3D(Result,W);
end;


function MatrixTriangleMult(M: TGLMatrix; T: TTriangle): TTriangle;
begin
Result.P1:=MatrixVectorMult(M,T.P1);
Result.P2:=MatrixVectorMult(M,T.P2);
Result.P3:=MatrixVectorMult(M,T.P3);
end;



procedure MatrixToEulerAngles(M: TGLMatrix; var Pitch,Yaw,Roll: double);
{Find euler angles from 4D matrix }
var Q: TQuaternion;
begin
Q:=GLMatrixToQuaternion(M);
QuatToEulerAngles(Q,Pitch,Yaw,Roll);
end;





function GetGLMatrixPosition(M: TGLMatrix): T3DVector;
var PM: TProgMatrix;
begin
PM:=TProgMatrix(M);
Result:=MakeVector3D(PM[3,0], PM[3,1],  PM[3,2]);
end;




function RemoveGLMatrixPosition(M: TGLMatrix): TGLMatrix;
{Remove Position from matrix}
begin
Result:=M;
TProgMatrix(Result)[3,0]:=0;
TProgMatrix(Result)[3,1]:=0;
TProgMatrix(Result)[3,2]:=0;
end;



function GetMatrixRotation(M: TGLMatrix): TGLMatrix;
{Return rotation portion of matrix}
begin
Result:=RemoveGLMatrixScaling(RemoveGLMatrixPosition(M));
end;




function DirectionVectorToMatrix(Up,Right,Front: T3DVector): TGLMatrix;
{Create rotation matrix from direction vector}
var PM: TProgMatrix;
begin
{Start with identity matrix}
PM:=TProgMatrix(IdentityMatrix);
{Normalize vectors}
Up:=UnitVector3D(Up);
Right:=UnitVector3D(Right);
Front:=UnitVector3D(Front);
{Copy vectors into columns}
PM[0,0]:=Up.X; PM[1,0]:=Right.X; PM[2,0]:=Front.X;
PM[0,1]:=Up.Y; PM[1,1]:=Right.Y; PM[2,1]:=Front.Y;
PM[0,2]:=Up.Z; PM[1,2]:=Right.Z; PM[2,2]:=Front.Z;
Result:=TGLMatrix(PM);
end;



function DirectionVectorToMatrix(Front: T3DVector; Swap: boolean): TGLMatrix;
{Create rotation matrix from direction vector}
var Up,Right: T3DVector;
begin
{Intersect two planes}
if Swap then Right:=CrossProduct3D(Front,MakeVector3D(0, 0, -1))
else Right:=CrossProduct3D(Front,MakeVector3D(0, 0, 1));
Up:=CrossProduct3D(Front,Right);
Result:=DirectionVectorToMatrix(Up,Right,Front);
end;


{

Uses OpenGL's description of glRotate command


( XX(1-C)+C	XY(1-C)-ZS  XZ(1-C)+YS	 0  )
(					    )
( YX(1-C)+ZS	YY(1-C)+C   YZ(1-C)-XS	 0  (
(					    )
( XZ(1-C)-YS	YZ(1-C)+XS  ZZ(1-C)+C	 0  )
(					    )
(	 0	     0		 0	 1  )
}

function AxisAngleToMatrix(Axis: T3DVector; Angle: double): TGLMatrix;
{Convert Direction Vector and Angle to matrix}
var RCos,RSin: double;
begin
Axis:=UnitVector3D(Axis);
RCos:=cos(Angle);
RSin:=sin(Angle);
Result:=IdentityMatrix;

{First row}
TProgMatrix(Result)[0,0]:= Axis.X*Axis.X*(1-RCos) + RCos;
TProgMatrix(Result)[1,0]:= Axis.X*Axis.Y*(1-RCos) - Axis.Z * RSin;
TProgMatrix(Result)[2,0]:= Axis.X*Axis.Z*(1-RCos) + Axis.Y * RSin;

{Second row}
TProgMatrix(Result)[0,1]:= Axis.Y*Axis.X*(1-RCos) + Axis.Z * RSin;
TProgMatrix(Result)[1,1]:= Axis.Y*Axis.Y*(1-RCos) + RCos;
TProgMatrix(Result)[2,1]:= Axis.Y*Axis.Z*(1-RCos) - Axis.X * RSin;

{Third row}
TProgMatrix(Result)[0,2]:=Axis.X*Axis.Z*(1-RCos) - Axis.Y * RSin;
TProgMatrix(Result)[1,2]:=Axis.Y*Axis.Z*(1-RCos) + Axis.X * RSin;
TProgMatrix(Result)[2,2]:=Axis.Z*Axis.Z*(1-RCos) + RCos;
end;



function GetAlignMatrix(V1,V2: T3DVector): TGLMatrix;
{Return the rotation the matrix that aligns V2 to V2 }
var VC: T3DVector;
var Angle: double;
begin
{Rotation axis is at right angles to the V1-V2 plane}
VC:=CrossProduct3D(V1,V2);
{Rotation angle is just the angle between vectors}
Angle:=VectorAngle3D(V1,V2);
{Build rotation matrix}
Result:=AxisAngleToMatrix(VC,Angle);
end;



function GetDirAlignMatrix(P1,P2,Dir: T3DVector): TGLMatrix;
{Get matrix to align points to Dir}
var V1,V2,Int: T3DVector;
begin
V1:=VectorSubtract3D(P2,P1);
V2:=Dir;
Result:=GetAlignMatrix(V1,V2);
Result:=SubtractTranslation(Result,P1);
end;


{------------------------------------------------------------------------------}

function GetTranslationMatrix(V: T3DVector): TGLMatrix;
{Get Specified Translation matrix}
begin
Result:=IdentityMatrix;
TProgMatrix(Result)[3,0]:=V.X;
TProgMatrix(Result)[3,1]:=V.Y;
TProgMatrix(Result)[3,2]:=V.Z;
end;



function GetTranslationVector(M: TGLMatrix): T3DVector;
{Get translation from matrix as a Vector}
begin
Result.X:=TProgMatrix(M)[3,0];
Result.Y:=TProgMatrix(M)[3,1];
Result.Z:=TProgMatrix(M)[3,2];
end;




function AddTranslation(M: TGLMatrix; V: T3DVector): TGLMatrix;
{Add specified Translation to matrix}
var M1: TGLMatrix;
begin
M1:=GetTranslationMatrix(V);
Result:=MatrixMultiply(M,M1);
end;



function SubtractTranslation(M: TGLMatrix; V: T3DVector): TGLMatrix;
{Subtract specified Translation with matrix}
var M1: TGLMatrix;
begin
M1:=GetTranslationMatrix(ScalarProduct3D(V,-1));
Result:=MatrixMultiply(M1,M);
end;




{------------------------------------------------------------------------------}


function GetScaleMatrix(V: T3DVector): TGLMatrix;
{Get Specified scale matrix}
begin
Result:=IdentityMatrix;
TProgMatrix(Result)[0,0]:=V.X;
TProgMatrix(Result)[1,1]:=V.Y;
TProgMatrix(Result)[2,2]:=V.Z;
end;



function AddScale(M: TGLMatrix; V: T3DVector): TGLMatrix;
{Combine specified scale with matrix}
var M1: TGLMatrix;
begin
M1:=GetScaleMatrix(V);
Result:=MatrixMultiply(M,M1);
end;

{------------------------------------------------------------------------------}


function AddRotation(M: TGLMatrix; X,Y,Z: double): TGLMatrix;
{Add rotation to matrix around specified axis}
var M1: TGLMatrix;
begin
M1:=EulerToMatrix(X,Y,Z);
Result:=MatrixMultiply(M,M1);
end;

{------------------------------------------------------------------------------}


function GetOrthoProjMatrix(Min,Max: T3DVector): TGLMatrix;
{Get an Orthogonal Projection Matrix}
{Used to transform vectors Min,Max volume to -1,-1,-1 +1,+1,+1}
begin
Result:=IdentityMatrix;
TProgMatrix(Result)[0,0]:=2/(Max.X-Min.X);
TProgMatrix(Result)[1,1]:=2/(Max.Y-Min.Y);
TProgMatrix(Result)[2,2]:=1/(Max.Z-Min.Z);

TProgMatrix(Result)[3,0]:=-((Max.X+Min.X)/(Max.X-Min.X));
TProgMatrix(Result)[3,1]:=-((Max.Y+Min.Y)/(Max.Y-Min.Y));
TProgMatrix(Result)[3,2]:=-(Min.Z/(Max.Z-Min.Z));
end;



function AddOrthoProjection(M: TGLMatrix; Min,Max: T3DVector): TGLMatrix;
{Project to 2D and normalize to a range of -1  +1}
var M1: TGLMatrix;
begin
M1:=GetOrthoProjMatrix(Min,Max);
Result:=MatrixMultiply(M,M1);
end;

{------------------------------------------------------------------------------}


function GetPerspectiveMatrix(Min,Max: T3DVector): TGLMatrix;
{Get an Orthogonal Projection Matrix}
begin
Result:=IdentityMatrix;
TProgMatrix(Result)[2,2]:=0;
TProgMatrix(Result)[2,3]:=-1/Min.Z/2;
TProgMatrix(Result)[3,3]:=1;
end;



function AddPerspective(M: TGLMatrix; Min,Max: T3DVector): TGLMatrix;
{Project to 2D and normalize to a range of -1  +1}
var M1: TGLMatrix;
begin
M1:=GetPerspectiveMatrix(Min,Max);
Result:=MatrixMultiply(M,M1);
end;



function MatrixScalarMultiply(M: TGLMatrix; S: double): TGLMatrix;
{Multiply all elements of Matrix by scalar}
var I: integer;
begin
for I:=0 to High(M) do Result[I]:=M[I]*S;
end;





function MatrixAdd(M1,M2: TGLMatrix): TGLMatrix;
{Add element of a two matrices together}
var I: integer;
begin
for I:=0 to High(M1) do Result[I]:=M1[I]+M2[I]
end;



end.
