unit GLPoints;

{This unit hold the TGLLines object which handles the creation, memory
allocation, normal generation, and rendering of an array of lines.}

interface

uses Windows, Classes, SysUtils, Graphics, Contnrs, OpenGL15, Vector, Math, HugeArry,
     GLQuatLib, GLLibrary, GLOctTree, GLRenderer,GLMatrix;


{ Define Point properties }

type TPointProperty = record
 Size: integer;
 Round: boolean;
 end;

{ Define an array of Point properties }

type TPointPropertyArray = class(THugeArray)
 private
  procedure Put(Index: integer; Item: TPointProperty); virtual;
  function Get(Index: integer): TPointProperty;
 protected
 public
  constructor Create(Number: integer);
  function Add(Item: TPointProperty): integer; reintroduce; overload; virtual;
  function AddPointer(Item: TPointProperty; Ptr: integer): Integer; virtual;
  procedure Insert(Value: TPointProperty; Index: integer); virtual;
  procedure Delete(Index: integer);
  procedure Move(CurIndex, NewIndex: Integer);
  property Value[Index: integer]: TPointProperty read Get write Put; default;
 end;

{Defines call back progress routine for mesh building process}

type TGLProgress= procedure(Sender: TObject; Action: string; PercentDone: integer);

{Mesh object declaration}

type TGLPoints = class(TGLRenderer)
 private
  Sorter: TList;
  FRenderMode: TRenderMode;
  FOnNormalProgress: TGLProgress;
  procedure DoPointColor(I: integer);
  procedure DoPoint(I: integer);
  procedure DrawPoint(Inx,Cell: integer);
  procedure SetRenderMode(V: TRenderMode); reintroduce; virtual;
  procedure SetMinMax(T: T3DVector);
  procedure SetupOctaltree;
  procedure RenderPoints;
  procedure DoProperties(I: integer);
 protected
 public
  Vertices: THuge3DVectorEx;
  PointColor: THugeLongInt;
  Properties: TPointPropertyArray;
  TreeBuilt: boolean;
  UseDisplayList: boolean;
  OctalTree: TOctTree;
  constructor Create; override;
  destructor Destroy; override;
  procedure Append(GLObject: TGLObject); override;

  procedure AddPoint(T: T3DVector; C: TColor; Size: integer); overload;
  procedure AddPoint(T: T3DVector; C: TColor; P: TPointProperty); overload;
  procedure AddPoint(T: T3DVector; C: TColor); overload;

  procedure SetPoint(Inx: integer; C: TColor; P: TPointProperty);
  procedure SetColor(const Color: TColor); override;
  procedure SetSize(const Size: integer);
  procedure PointsClear;
  procedure Clear; override;
  procedure ClearData;
  procedure Render; override;
  procedure DisplayProgress(Status: string; Progress,Total: integer);
  property RenderMode: TRenderMode read FRenderMode write SetRenderMode;
  property OnNormalProgress: TGLProgress read FOnNormalProgress write FOnNormalProgress;
  function GetColor: TColor; override;
  function GetSize: integer;
 end;


{ TGLPointList }


type  TGLPointList = class(TGLObjectList)
  protected
    function GetItem(Index: Integer): TGLPoints;
    procedure SetItem(Index: Integer; AObject: TGLPoints);
  public
    function Add(var Points: TGLPoints): integer; overload;
    function Add: integer; overload;
    property Items[Index: Integer]: TGLPoints read GetItem write SetItem; default;
    procedure Assign(ObjectList: TObjectList);
  end;

implementation


{--------------------------------------------------------------------------------}
{  Implementation for huge Point Property array object based on huge array object }
{--------------------------------------------------------------------------------}

constructor TPointPropertyArray.Create(Number: Integer);
{Create a huge property array}
begin
inherited Create(Number,SizeOf(TPointProperty));
end;


function TPointPropertyArray.Add(Item: TPointProperty): integer;
{Add item to property array}
begin
Result:=inherited Add(Item);
end;


function TPointPropertyArray.AddPointer(Item: TPointProperty; Ptr: integer): Integer;
begin
Result:=inherited AddPointer(Item,Ptr);
end;



procedure TPointPropertyArray.Put(Index: integer; Item: TPointProperty);
{Insert property at index}
var PTR: ^TPointProperty;
begin
PTR:=GetItemPtr(Index);
PTR^:=Item;
end;



function TPointPropertyArray.Get(Index: integer): TPointProperty;
{Get property at the index}
var PTR: ^TPointProperty;
begin
PTR:=GetItemPtr(Index);
Result:=PTR^;
end;


procedure TPointPropertyArray.Insert(Value: TPointProperty; Index: integer);
begin
InsertItem(Value,Index);
end;


procedure TPointPropertyArray.Delete(Index: integer);
begin
DeleteItem(Index);
end;



procedure TPointPropertyArray.Move(CurIndex, NewIndex: Integer);
var T: TPointProperty;
begin
T:=Value[CurIndex];
Delete(CurIndex);
Insert(T,NewIndex);
end;



{------------------ Property Setting Routines ---------------------------------}



procedure TGLPoints.SetRenderMode(V: TRenderMode);
{Set current render mode}
begin
if RenderMode<>V then
	begin
	FRenderMode:=V;
	TreeBuilt:=False;
	end;
end;





procedure TGLPoints.DisplayProgress(Status: string; Progress,Total: integer);
{This routine calls 'progress" callback routine if the user has set one up}
begin
if not Assigned(FOnNormalProgress) then exit;
if (Progress mod (Total div 100 + 1))<>0 then exit;
FOnNormalProgress(Self,Status,MulDiv(100,Progress,Total));
end;





{---------------- Object Event handlers ----------------------------------}


constructor TGLPoints.Create;
begin
inherited;
Vertices:=THuge3DVectorEx.Create(5);
PointColor:=THugeLongInt.Create(5);
Properties:=TPointPropertyArray.Create(5);
Sorter:=TList.Create;
OctalTree:=TOctTree.Create;
FRenderMode:=rmFlatFace;
TreeBuilt:=False;
Position:=DefaultPos;
ResetPos:=DefaultPos;
UseDisplayList:=False;
end;



destructor TGLPoints.Destroy;
begin
Vertices.Free;
PointColor.Free;
Properties.Free;
Sorter.Free;
OctalTree.Free;
inherited Destroy;
end;


{---------------- Routines to add Points to the list ---------------------------}


procedure TGLPoints.SetMinMax(T: T3DVector);
{Set min/max from of mesh}
begin
if Vertices.ItemIndex=0 then
	begin
	ExtentMin:=T;
	ExtentMax:=T;
	end
else
	begin
	ExtentMin:=GetMinimumVector(ExtentMin,T);
	ExtentMax:=GetMaximumVector(ExtentMax,T);
	end;
end;


procedure TGLPoints.SetPoint(Inx: integer; C: TColor; P: TPointProperty);
begin
PointColor[Inx]:=C;
Properties[Inx]:=P;
end;


procedure TGLPoints.AddPoint(T: T3DVector; C: TColor; P: TPointProperty);
{Add a Point to the array}
begin
SetMinMax(T);
Vertices.Add(T);
PointColor.Add(C);
Properties.Add(P);
TreeBuilt:=false;
end;


procedure TGLPoints.AddPoint(T: T3DVector; C: TColor; Size: integer);
{Add a Point to the array}
var P: TPointProperty;
begin
P.Size:=Size;
P.Round:=True;
AddPoint(T,C,P);
end;


procedure TGLPoints.AddPoint(T: T3DVector; C: TColor);
{Add a Point to the array}
begin
AddPoint(T,C,2);
end;


procedure TGLPoints.PointsClear;
{Clear Points only leaving position, rotation and scaling}
begin
Vertices.Clear;
PointColor.Clear;
Properties.Clear;
Sorter.Clear;
TreeBuilt:=false;
ExtentMin:=MakeVector3D(0,0,0);
ExtentMax:=MakeVector3D(0,0,0);
end;



procedure TGLPoints.Clear;
{Clear the Points}
begin
inherited Clear;
PointsClear;
end;



procedure TGLPoints.ClearData;
{Clear the Points without reseting transforms}
begin
PointsClear;
end;


procedure TGLPoints.Append(GLObject: TGLObject);
begin
Clear;
inherited Assign(GLObject);
if GLObject is TGLPoints then
	begin
	Vertices.Assign(TGLPoints(GLObject).Vertices);
	PointColor.Assign(TGLPoints(GLObject).PointColor);
	Properties.Assign(TGLPoints(GLObject).Properties);
	end;
end;


{-------------- Routines To Render Points --------------------------------------}



procedure TGLPoints.DoPointColor(I: integer);
{Tell OpenGL about the color of specified Point}
var R,G,B,A: single;
begin
GetRGB(PointColor.Value[I],R,G,B);
A:=GetCurrentAlpha;
glColor4f(R,G,B,A);
end;




procedure TGLPoints.DoProperties(I: integer);
{Tell OpenGL about the properties of specified Point}
begin
glPointSize(Properties[I].Size);
if Properties[I].Round then
	begin
	glEnable( GL_POINT_SMOOTH );
	glEnable( GL_BLEND );
	end;
end;



procedure TGLPoints.DoPoint(I: integer);
{Tell OpenGL about specified Point }
var T: T3DVector;
begin
T:=Vertices[I];
glBegin(GL_Points);
glVertex3dv(@T);
glEnd;
end;



procedure TGLPoints.DrawPoint(Inx,Cell: integer);
begin
DoPointColor(Inx);
DoProperties(Inx);
DoPoint(Inx);
end;




procedure TGLPoints.RenderPoints;
{Handle displaying normal GL mode}
begin
glDisable(GL_LIGHTING);
OctalTree.IterateLines(DrawPoint);
{glEnd not needed here}
end;




procedure TGLPoints.SetupOctalTree;
{Put all Points in the boxes}
{This is only done when the list are changed}
var I: integer;
begin
OctalTree.ResetBoxes;
OctalTree.MeshDimensions:=MakeBox(ExtentMin,ExtentMax);
for I:=0 to Vertices.Count-1 do
	begin
	{If Points span boxes, we'll use the box for the first point}
	OctalTree.PutPointInBox(Vertices[I],I);
	end;
TreeBuilt:=True;
end;



procedure TGLPoints.Render;
{Render whole mesh}
begin
glPushAttrib(GL_ALL_ATTRIB_BITS);
try
if not TreeBuilt then SetupOctalTree;
inherited Render;
OctalTree.FlagBoxesInFrustum;
RenderPoints;
finally glPopAttrib; end;
end;



procedure TGLPoints.SetColor(const Color: TColor);
var I: integer;
begin
inherited;
for I:=0 to PointColor.Count-1 do PointColor[I]:=Color;
end;



procedure TGLPoints.SetSize(const Size: integer);
var I: integer;
begin
for I:=0 to Properties.Count-1 do
	begin
	TPointProperty(Properties.GetItemPtr(I)^).Size:=Size;
	end;
end;



function TGLPoints.GetColor: TColor;
begin
inherited GetColor;
if PointColor.Count>0 then Result:=PointColor[0]
else Result:=clWhite;
end;



function TGLPoints.GetSize: integer;
begin
if Properties.Count>0 then Result:=Properties[0].Size
else Result:=1;
end;




{ TGLPointList }

function TGLPointList.Add(var Points: TGLPoints): integer;
begin
Points:=TGLPoints.Create;
Result:=inherited Add(Points);
Points.Name:='Points-'+IntToStr(Result+1);
Change;
end;




function TGLPointList.Add: integer;
var Points: TGLPoints;
begin
Result:=Add(Points);
end;



procedure TGLPointList.Assign(ObjectList: TObjectList);
var I: integer;
var O: TGLPoints;
begin
if ObjectList is TGLPointList then
	begin
	Clear;
	for I:=0 to ObjectList.Count-1 do
		begin
		Add(O);
		O.Assign(TGLPoints(ObjectList[I]));
		end;
	end
else inherited Assign(ObjectList);
end;



function TGLPointList.GetItem(Index: Integer): TGLPoints;
begin
Result := TGLPoints(inherited GetItem(Index));
end;


procedure TGLPointList.SetItem(Index: Integer; AObject: TGLPoints);
begin
inherited SetItem(Index, AObject);
end;





end.
