unit GLRenderer;


interface

uses Windows,SysUtils, Classes,Controls, Graphics,Contnrs, OpenGL15, Vector, Math, HugeArry,
     GLQuatLib, GLLibrary, GLMatrix, GLOctTree, GLTexture, GLAnimator, DSupport;




type TGLObject = class(TObject)
 private
  MoveX,MoveY: integer;
  FEnabled: boolean;
  FName: string;
  FParentObject: TGLObject;
  FScale3D: T3DVector;
  FSelected: boolean;
  FHitTestEnabled: boolean;
  FExtentMin, FExtentMax: T3DVector;
  FFreeRotate: boolean;
  FMouseDown: boolean;
  procedure SetName(const Value: string);
  procedure SetParentObject(const Value: TGLObject);

  procedure SetPitch(const Value: double);
  procedure SetRoll(const Value: double);
  procedure SetYaw(const Value: double);
  procedure SetScale(const Value: double);
  function GetPitch: double;
  function GetRoll: double;
  function GetYaw: double;
  function GetScale: double;
  procedure SetSelected(const Value: boolean);
  function GetSize: T3DVector;
 protected
  FOnChange: TNotifyEvent;
  procedure Change;
  function GetExtentMax: T3DVector; virtual;
  function GetExtentMin: T3DVector; virtual;
  procedure SetExtentMax(const Value: T3DVector); virtual;
  procedure SetExtentMin(const Value: T3DVector); virtual;
 public
  ConYaw,ConPitch: double;
  LastMatrix: TGLMatrix;
  ParentMatrix: TGLMatrix;
  PitchInc,YawInc,RollInc: double;
  Quaternion: TQuaternion;
  Position, ResetPos: T3DVector;
  ResetRot: T3DVector;
  RotationCenter: T3DVector;
  FlyThrough: boolean;
  OwnerList: TObjectList;
  Animator: TGLAnimator;
  constructor Create; virtual;
  destructor Destroy; override;
  procedure Assign(GLObject: TGLObject); virtual;
  procedure Append(GLObject: TGLObject); virtual;
  procedure Render; virtual;
  procedure UpdateMatrices;
  procedure ResetTransforms;
  procedure ResetRotIncrements;
  procedure ResetAngles;
  procedure ResetPosition;
  procedure SetDefaultTransform;
  procedure Reset; virtual;
  procedure AddRotation(Pitch,Yaw,Roll: double);
  procedure SetRotation(Pitch,Yaw,Roll: double);
  procedure SetDirection(Dir: T3DVector); overload;
  procedure SetDirection(Up,Dir: T3DVector); overload;
  procedure SetYawPitchOrient(Yaw, Pitch: double);
  procedure IncFreeRotation(Pitch, Yaw, Roll: double);
  procedure IncConstRotation(Pitch, Yaw: double);
  procedure IncRotations(Pitch, Yaw, Roll: double);
  procedure RepeatLastRotation;
  procedure SetPosition(X,Y,Z: double);
  procedure AddPosition(X,Y,Z: double); overload;
  procedure AddPosition(P: T3DVector); overload;
  procedure DoRotations;
  procedure DoPosition;
  procedure HandleMouseMove(Shift: TShiftState; X, Y: Integer; BaseObject: TGLObject); overload;
  procedure HandleMouseMove(Shift: TShiftState; X, Y: Integer); overload;
  procedure HandleMouseDown(X, Y: Integer);
  procedure HandleMouseUp;
  property ExtentMin: T3DVector read GetExtentMin write SetExtentMin;
  property ExtentMax: T3DVector read GetExtentMax write SetExtentMax;
  property Roll: double read GetRoll write SetRoll;
  property Pitch: double read GetPitch write SetPitch;
  property Yaw: double read GetYaw write SetYaw;
  property Scale: double read GetScale write SetScale;
  property Scale3D: T3DVector read FScale3D write FScale3D;
  function GetMatrix: TGLMatrix;
  procedure SetMatrix(M: TGLMatrix);
  function GetSceneScaleFactor(GLO: TGLObject): double; overload;
  function GetSceneScaleFactor: double; overload;
  property Selected: boolean read FSelected write SetSelected;
  property Enabled: boolean read FEnabled write FEnabled;
  property HitTestEnabled: boolean read FHitTestEnabled write FHitTestEnabled;
  property Name: string read FName write SetName;
  property ParentObject: TGLObject read FParentObject write SetParentObject;
  property OnChange: TNotifyEvent read FOnChange write FOnChange;
  property Matrix: TGLMatrix read GetMatrix write SetMatrix;
  property Size: T3DVector read GetSize;
  property FreeRotate: boolean read FFreeRotate write FFreeRotate;
 end;

{Define rending modes}

type TRenderMode = (rmWireFrame,rmFlatFace,rmSmoothFace);

const DefaultPos: T3DVector = (X:0;Y:0;Z:0);
const DefaultRot: T3DVector = (X:0;Y:0;Z:0);

{Render object declaration}

type
 TGLRenderer = class(TGLObject)
 private
  FBackColor,FFrontColor: TColor;
  FDoTwoSidedLighting: boolean;
  FDoFrontBackColor: boolean;
    FTransparency: double;
  procedure SetAlphaBlend;
  procedure SetMaterials;
  procedure SetLighting;
  procedure SetDoTwoSidedLighting(const Value: boolean);
  procedure SetDoFrontBackColor(const Value: boolean);
  procedure SetBackColor(const Value: TColor);
  procedure SetFrontColor(const Value: TColor);
 protected
  FRenderMode: TRenderMode;
  function GetColor: TColor; virtual;
  function GetCurrentAlpha: double;
  procedure SetRenderMode(const Value: TRenderMode); virtual;
  procedure SetTransparency(const Value: double); virtual;
 public
  DoAlphaBlend: boolean;
  UserBlend,TransparentDepth,BlendBackground: boolean;
  SrcUserBlend,DestUserBlend,BlendEquation: TGLenum;
  LightEmission: TGLArrayf4;
  LightSpecular: TGLArrayf4;
  LightShininess: GLFloat;
  Texture: TGLObjectTexture;
  constructor Create; override;
  destructor Destroy; override;
  procedure Assign(GLObject: TGLObject); override;
  procedure Append(GLObject: TGLObject); override;
  procedure Clear; virtual;
  procedure ResetMaterials;
  procedure Reset; reintroduce;
  procedure SetColor(const Color: TColor); virtual;
  procedure Render; override;
  property Alpha: double read GetCurrentAlpha;
  property Color: TColor read GetColor write SetColor;
  property RenderMode: TRenderMode read FRenderMode write SetRenderMode;
  property DoTwoSidedLighting: boolean read FDoTwoSidedLighting write SetDoTwoSidedLighting;
  property DoFrontBackColor: boolean read FDoFrontBackColor write SetDoFrontBackColor;
  property FrontColor: TColor read FFrontColor write SetFrontColor;
  property BackColor: TColor read FBackColor write SetBackColor;
  property Transparency: double read FTransparency write SetTransparency;
 end;


type TListChangeEvent = procedure(GLObject: TGLObject; Action: TListNotification) of object;


type TGLObjectList = class(TObjectList)
 private
  FOnChange: TNotifyEvent;
  FOnListChange: TListChangeEvent;
  function GetExtentMax: T3DVector;
  function GetExtentMin: T3DVector;
 protected
  procedure SetItem(Index: Integer; AObject: TObject);
  procedure Notify(Ptr: Pointer; Action: TListNotification); override;
 public
  procedure Change;
  constructor Create;
  property OnListChange: TListChangeEvent read FOnListChange write FOnLIstChange;
  property OnChange: TNotifyEvent read FOnChange write FOnChange;
  function Add(GLObject: TGLObject; OwnerList: TObjectList): Integer; overload; virtual;
  function Add(GLObject: TGLObject): Integer; overload; virtual;
  function Add: TGLObject; overload; virtual;
  procedure Clear; reintroduce; virtual;
  procedure Delete(Index: Integer);
  procedure Exchange(Index1, Index2: Integer);
  procedure Insert(Index: Integer; Item: Pointer);
  procedure Sort(Compare: TListSortCompare);
  procedure Move(CurIndex, NewIndex: Integer);
  property ExtentMin: T3DVector read GetExtentMin;
  property ExtentMax: T3DVector read GetExtentMax;
  procedure Assign(ObjectList: TObjectList); virtual;
 end;

procedure GetRootRelativePosition(GLO: TGLObject; var P: T3DVector);

procedure GetRootRelativeRotation(GLO: TGLObject; var Pitch,Yaw,Roll: double);
procedure GetMatrixRelativeRotation(Matrix: TGLMatrix; var Pitch,Yaw,Roll: double);

implementation


{ Support routines }


procedure GetRootRelativePosition(GLO: TGLObject; var P: T3DVector);
{Convert position from object relative to base frame relative}
begin
{Do move based on derotated position - i.e. remove scaling and transpose matrix}
P:=MatrixVectorMult(TransposeMatrix(RemoveGLMatrixScaling(GLO.ParentMatrix)),P);
end;



procedure GetMatrixRelativeRotation(Matrix: TGLMatrix; var Pitch,Yaw,Roll: double);
{Convert Yaw, Pitch and Roll for object relative to Specified matrix}
var ParentQuat,NewQuat,Q1: TQuaternion;
begin
{Get matrix-quat from specified maxtrix}
ParentQuat:=GLMatrixToQuaternion(Matrix);
NewQuat:=EulerToQuat(Pitch,Yaw,Roll);

{Unrotate new value to match parent rotations from base frame down}
Q1:=QuatMultiply(QuatConjugate(ParentQuat),NewQuat);
Q1:=QuatMultiply(Q1,ParentQuat);

{Convert back to Euler angles}
QuatToEulerAngles(Q1,Pitch,Yaw,Roll);
end;



procedure GetRootRelativeRotation(GLO: TGLObject; var Pitch,Yaw,Roll: double);
{Convert Yaw, Pitch and Roll for object relative to base frame relative}
begin
{Use matrix from all branches to just above object}
GetMatrixRelativeRotation(GLO.ParentMatrix,Pitch,Yaw,Roll);
end;



{ TGLObject }

procedure TGLObject.Change;
begin
if Assigned(FOnChange) then FOnChange(Self);
if OwnerList<>nil then TGLObjectList(OwnerList).Change;
end;


constructor TGLObject.Create;
begin
inherited Create;
Animator:=TGLAnimator.Create;
Enabled:=True;
HitTestEnabled:=True;
Selected:=True;
Reset;
ParentMatrix:=IdentityMatrix;
LastMatrix:=IdentityMatrix;
FreeRotate:=True;
FMouseDown:=False;
end;


destructor TGLObject.Destroy;
begin
Animator.Free;
inherited;
end;



procedure TGLObject.SetName(const Value: string);
begin
if FName<>Value then
	begin
	FName:=Value;
	Change;
	end;
end;



procedure TGLObject.SetParentObject(const Value: TGLObject);
begin
if FParentObject<>Value then
	begin
	FParentObject:=Value;
	end;
end;


{----------------- Routines to handle position and rotation -------------------}



procedure TGLObject.ResetAngles;
begin
SetRotation(ResetRot.X,ResetRot.Y,ResetRot.Z);
ConYaw:=0; ConPitch:=0;
end;



procedure TGLObject.ResetPosition;
begin
Position:=ResetPos;
end;



procedure TGLObject.ResetTransforms;
{Reset position and rotation to the default location}
begin
Quaternion:=GetIdentityQuaternion;
Position:=ResetPos;
RotationCenter:=MakeVector3D(0,0,0);
FScale3D:=MakeVector3D(1,1,1);
UpdateMatrices;
end;



procedure TGLObject.ResetRotIncrements;
begin
RollInc:=0; YawInc:=0; PitchInc:=0;
end;


procedure TGLObject.SetDefaultTransform;
{Reset position and rotation to the defaul location}
begin
ResetPos:=DefaultPos;
ResetRot:=DefaultRot;
ResetTransforms;
end;



procedure TGLObject.Reset;
{Reset all parameters}
begin
ResetTransforms;
ResetAngles;
ResetPosition;
SetDefaultTransform;
end;





procedure TGLObject.SetYawPitchOrient(Yaw,Pitch: double);
{Set Orientation constrained to Yaw, Pitch value}
var Up,Dir,Right: T3DVector;
var Cent: T3DVector;
var Qt: TQuaternion;
var Mat: TGLMatrix;
begin
ConYaw:=Yaw;
ConPitch:=Pitch;
Pitch:=Pitch+90;
Cent:=MakeVector3D(0,0,0);
Up:=MakeVector3D(1,0,0);
RotatePoint3D(UP,Cent,Yaw,Pitch);
Dir:=MakeVector3D(0,-1,0);
RotatePoint3D(Dir,Cent,Yaw,Pitch);
Right:=CrossProduct3D(Dir,Up);
Mat:=DirectionVectorToMatrix(Up,Right,Dir);
Qt:=GLMatrixToQuaternion(Mat);
Quaternion:=Qt;
end;


procedure TGLObject.SetDirection(Dir: T3DVector);
var Qt: TQuaternion;
var Mat: TGLMatrix;
var Up,Right: T3DVector;
begin
VectorToUpRight(Dir,Up,Right);
Mat:=DirectionVectorToMatrix(Up,Right,Dir);
Qt:=GLMatrixToQuaternion(Mat);
Quaternion:=Qt;
end;




procedure TGLObject.SetDirection(Up,Dir: T3DVector);
var Right: T3DVector;
var Qt: TQuaternion;
var Mat: TGLMatrix;
begin
Right:=CrossProduct3D(Dir,Up);
Mat:=DirectionVectorToMatrix(Up,Right,Dir);
Qt:=GLMatrixToQuaternion(Mat);
Quaternion:=Qt;
end;



procedure TGLObject.SetRotation(Pitch,Yaw,Roll: double);
{Set the current Quaternions to the specified Pitch, Yaw and Roll}
begin
Quaternion:=EulerToQuat(Pitch,Yaw,Roll);
end;



procedure TGLObject.AddRotation(Pitch,Yaw,Roll: double);
{Add the specified Pitch, Yaw and Roll to current Quaternion}
var QT: TQuaternion;
begin
QT:=EulerToQuat(Pitch,Yaw,Roll);
Quaternion:=QuatMultiply(QT,Quaternion);
end;




procedure TGLObject.SetPosition(X,Y,Z: double);
{Set current position to X,Y,Z}
begin
Position:=MakeVector3D(X,Y,Z);
end;


function TGLObject.GetScale: double;
{Only looks at one dimemsion, maybe should average all three}
begin
Result:=FScale3D.X;
end;



procedure TGLObject.SetScale(const Value: double);
begin
if GetScale<>Value then FScale3D:=MakeVector3D(Value,Value,Value);
end;


procedure TGLObject.AddPosition(P: T3DVector);
{Add X,Y,Z to current position}
begin
Position:=VectorAdd3D(Position,P);
end;




procedure TGLObject.AddPosition(X,Y,Z: double);
{Add X,Y,Z to current position}
begin
AddPosition(MakeVector3D(X,Y,Z));
end;





procedure TGLObject.DoRotations;
{Rotate the current matrix using quaternions}
var V: T3DVector;
var Angle: double;
begin
glTranslateD(RotationCenter.X, RotationCenter.Y, RotationCenter.Z);
QuatToAxisAngle(Quaternion,V,Angle);
glRotateD(RadToDeg(Angle),V.X,V.Y,V.Z);
glTranslateD(-RotationCenter.X, -RotationCenter.Y, -RotationCenter.Z);
end;





procedure TGLObject.DoPosition;
{Set matrix to current postion}
var P: T3DVector;
begin
P:=Position;
glTranslateD(P.X, P.Y, P.Z);
end;



function TGLObject.GetSceneScaleFactor(GLO: TGLObject): double;
{Return scaling factor for movements}
var Z,Size,Depth: double;
var C,P: T3DVector;
begin
{Work out distance from camera to center of the object}
{Get the center of the object}
C:=ScalarDivide3D(VectorAdd3D(GLO.ExtentMax,GLO.ExtentMin),2);
{Get frame frame position}
P:=GetGLMatrixPosition(GLO.LastMatrix);
{Combine to find distance from center to camera}
P:=VectorAdd3D(P,C);
Z:=abs(P.Z);

{Work out the general size of the object}
Size:=VectorABS3D(VectorSubtract3D(GLO.ExtentMax,GLO.ExtentMin));
{Work out a depth 1/4 the way into the object}
Depth:=sqrt(sqr(Size/4));
{Compensate for the fact that as you closer and or near zero they get infinitely small }
if Z<Depth then Z:=Depth;
Result:= Abs(PixelsPerUnit/(Z/WorkDistance));
end;






function TGLObject.GetSceneScaleFactor: double;
begin
Result:=GetSceneScaleFactor(Self);
end;

{------------------------------------------------------------------------------}





procedure TGLObject.IncFreeRotation(Pitch,Yaw,Roll: double);

	procedure Normalize;
	{Normalize angles to +/-360 range}
	begin
	while PitchInc>360 do PitchInc:=PitchInc-360;
	while PitchInc<0 do PitchInc:=PitchInc+360;
	while YawInc>360 do YawInc:=YawInc-360;
	while YawInc<0 do YawInc:=YawInc+360;

	while RollInc>360 do RollInc:=RollInc-360;
	while RollInc<0 do RollInc:=RollInc+360;
	end;


begin
RollInc:=Roll; YawInc:=Yaw; PitchInc:=Pitch;
Normalize;
GetRootRelativeRotation(Self,PitchInc,YawInc,RollInc);
AddRotation(PitchInc,YawInc,RollInc);
end;



procedure TGLObject.IncConstRotation(Pitch,Yaw: double);


	procedure NormalAngles;
	begin
	if ConPitch>90 then ConPitch:=90
	else if ConPitch<-90 then ConPitch:=-90;
	while ConYaw>360 do ConYaw:=ConYaw-360;
	while ConYaw<0 do ConYaw:=ConYaw+360;
	end;


begin
RollInc:=0; YawInc:=Yaw; PitchInc:=Pitch;
ConYaw:=ConYaw-Yaw;
ConPitch:=ConPitch-Pitch;
NormalAngles;
SetYawPitchOrient(ConYaw,ConPitch);
end;



procedure TGLObject.IncRotations(Pitch,Yaw,Roll: double);
begin
if FreeRotate then IncFreeRotation(Pitch,Yaw,Roll)
else IncConstRotation(Pitch,Yaw);
end;


{------------------------------------------------------------------------------}




procedure TGLObject.HandleMouseMove(Shift: TShiftState; X, Y: Integer; BaseObject: TGLObject);
var DX,DY,XO,Scale: double;
var P: T3DVector;
begin
if not FMouseDown then exit;
Scale:=BaseObject.GetSceneScaleFactor(BaseObject);
DX:=(X-MoveX); DY:=(Y-MoveY);
if ((ssLeft in Shift) and (ssRight in Shift)) or (Shift=[ssMiddle]) then
	begin
	{Handle mouse zooms}
	if FlyThrough then XO:=-DX/Scale*2 else XO:=0;
	P:=MakeVector3D(XO,0,-DY/Scale*2);
	GetRootRelativePosition(BaseObject,P);
	AddPosition(P);
	MoveX:=X; MoveY:=Y;
	end
else if Shift=[ssLeft] then
	begin
	{Handle mouse rotation}
	DX:=(X-MoveX); DY:=(Y-MoveY);
	{Note reversed DY,DX}
	IncRotations(DY,DX,0);
	MoveX:=X; MoveY:=Y;
	end
else if Shift=[ssRight] then
	begin
	{Handle mouse translation}
	P:=MakeVector3D(DX/Scale,-DY/Scale,0);
	GetRootRelativePosition(BaseObject,P);
	AddPosition(P);
	MoveX:=X; MoveY:=Y;
	end;
end;



procedure TGLObject.HandleMouseMove(Shift: TShiftState; X, Y: Integer);
begin
HandleMouseMove(Shift,X,Y,Self);
end;



procedure TGLObject.HandleMouseDown(X, Y: Integer);
{Set starting location of mouse}
begin
FMouseDown:=True;
MoveX:=X; MoveY:=Y;
ResetRotIncrements;
end;


procedure TGLObject.HandleMouseUp;
begin
FMouseDown:=False;
end;



{-----------------------------------------------------------------------------}


procedure TGLObject.RepeatLastRotation;
begin
IncRotations(PitchInc,YawInc,RollInc);
end;




procedure TGLObject.SetPitch(const Value: double);
var Pitch,Yaw,Roll: double;
begin
QuatToEulerAngles(Self.Quaternion,Pitch,Yaw,Roll);
if Pitch<>Value then
	begin
	Pitch:=Value;
	SetRotation(Pitch,Yaw,Roll);
	end;
end;



procedure TGLObject.SetRoll(const Value: double);
var Pitch,Yaw,Roll: double;
begin
QuatToEulerAngles(Self.Quaternion,Pitch,Yaw,Roll);
if Roll<>Value then
	begin
	Roll:=Value;
	SetRotation(Pitch,Yaw,Roll);
	end;
end;



procedure TGLObject.SetYaw(const Value: double);
var Pitch,Yaw,Roll: double;
begin
QuatToEulerAngles(Self.Quaternion,Pitch,Yaw,Roll);
if Yaw<>Value then
	begin
	Yaw:=Value;
	SetRotation(Pitch,Yaw,Roll);
	end;
end;




function TGLObject.GetPitch: double;
var Pitch,Yaw,Roll: double;
begin
QuatToEulerAngles(Self.Quaternion,Pitch,Yaw,Roll);
Result:=Pitch;
end;



function TGLObject.GetRoll: double;
var Pitch,Yaw,Roll: double;
begin
QuatToEulerAngles(Self.Quaternion,Pitch,Yaw,Roll);
Result:=Roll;
end;



function TGLObject.GetYaw: double;
var Pitch,Yaw,Roll: double;
begin
QuatToEulerAngles(Self.Quaternion,Pitch,Yaw,Roll);
Result:=Yaw;
end;





procedure TGLObject.UpdateMatrices;
begin
{Get combined matrix from just above the level}
glGetDoublev(GL_MODELVIEW_MATRIX, @ParentMatrix);
DoPosition;
DoRotations;
glScaleD(FScale3D.X,FScale3D.Y,FScale3D.Z);
{Get combined matrix at this level}
glGetDoublev(GL_MODELVIEW_MATRIX, @LastMatrix);
end;




procedure TGLObject.Render;
begin
UpdateMatrices;
end;


function TGLObject.GetMatrix: TGLMatrix;
{Get current local matrix to this node}
begin
glMatrixMode(GL_MODELVIEW);
glPushMatrix;
glLoadIdentity;
Render;
glGetDoublev( GL_MODELVIEW_MATRIX, @Result);
glPopMatrix;
end;



procedure TGLObject.SetMatrix(M: TGLMatrix);
{Set current local matrix for this node}
begin
FScale3D:=GetGLMatrixScale(M);
Quaternion:=GLMatrixToQuaternion(M);
Position:=GetGLMatrixPosition(M);
end;



procedure TGLObject.Assign(GLObject: TGLObject);
begin
Animator.Assign(GLObject.Animator);
ExtentMin:=GLObject.ExtentMin;
ExtentMax:=GLObject.ExtentMax;
Quaternion:=GLObject.Quaternion;
Position:=GLObject.Position;
ResetPos:=GLObject.ResetPos;
RotationCenter:=GLObject.RotationCenter;
FlyThrough:=GLObject.FlyThrough;
LastMatrix:=GLObject.LastMatrix;
ParentMatrix:=GLObject.ParentMatrix;

PitchInc:=GLObject.PitchInc;
YawInc:=GLObject.YawInc;
RollInc:=GLObject.RollInc;

Scale:=GLObject.Scale;
Scale3D:=GLObject.Scale3D;
Selected:=GLObject.Selected;
Enabled:=GLObject.Enabled;

HitTestEnabled:=GLObject.HitTestEnabled;
Name:=GLObject.Name;
ParentObject:=GLObject.ParentObject;
{Don't copy owner list, the previous owner doesn't matter}
//OwnerList:=GLObject.OwnerList;
{Don't copy Onchange these or event will trigger in copies that shouldn't }
//OnChange:=GLObject.OnChange;
end;



procedure TGLObject.Append(GLObject: TGLObject);
begin
ExtentMin:=GetMinimumVector(GLObject.ExtentMin,ExtentMin);
ExtentMax:=GetMaximumVector(GLObject.ExtentMax,ExtentMax);
end;



procedure TGLObject.SetSelected(const Value: boolean);
begin
if FSelected <> Value then
	begin
	FSelected := Value;
	Change;
	end;
end;


function TGLObject.GetSize: T3DVector;
begin
Result:=VectorSubtract3D(ExtentMax,ExtentMin);
end;


{-------------- TGLObjectList -------------------------------------------------}


procedure TGLObjectList.Change;
begin
if Assigned(FOnChange) then FOnChange(Self);
end;



procedure TGLObjectList.Notify(Ptr: Pointer; Action: TListNotification);
begin
inherited;
if Assigned(FOnListChange) then FOnListChange(Ptr,Action);
end;



function TGLObjectList.Add(GLObject: TGLObject; OwnerList: TObjectList): Integer;
begin
Result:=inherited Add(GLObject);
TGLObject(GLObject).OwnerList:=OwnerList;
Change;
end;


function TGLObjectList.Add(GLObject: TGLObject): Integer;
begin
Result:=Add(GLObject,Self);
end;


function TGLObjectList.Add: TGLObject;
begin
Result:=TGLObject.Create;
Add(Result);
end;



procedure TGLObjectList.Clear;
begin
inherited Clear;
Change;
end;



procedure TGLObjectList.Delete(Index: Integer);
begin
inherited Delete(Index);
Change;
end;


procedure TGLObjectList.Exchange(Index1, Index2: Integer);
begin
inherited Exchange(Index1,Index2);
Change;
end;


procedure TGLObjectList.Insert(Index: Integer; Item: Pointer);
begin
inherited Insert(Index,Item);
Change;
end;


procedure TGLObjectList.Sort(Compare: TListSortCompare);
begin
inherited Sort(Compare);
Change;
end;


procedure TGLObjectList.Move(CurIndex, NewIndex: Integer);
begin
inherited Move(CurIndex,NewIndex);
Change;
end;



constructor TGLObjectList.Create;
begin
inherited Create;
end;





procedure TGLObjectList.SetItem(Index: Integer; AObject: TObject);
var Old: boolean;
begin
{Make false so user can replace the existing object without
destroying the original object which may be owned by someone else }
Old:=OwnsObjects;
OwnsObjects:=False;
inherited ;
OwnsObjects:=Old;
end;



procedure TGLObjectList.Assign(ObjectList: TObjectList);
var I: integer;
var O: TGLObject;
begin
if ObjectList is TGLObjectList then
	begin
        OnChange:=TGLObjectList(ObjectList).OnChange;
        OnListChange:=TGLObjectList(ObjectList).OnListChange;
	Clear;
	for I:=0 to ObjectList.Count-1 do
		begin
		O:=Add;
		O.Assign(TGLObject(ObjectList[I]));
		end;
	end
else inherited Assign(ObjectList);
end;


{------------------ TRenderer -------------------------------------------------}





procedure TGLRenderer.Reset;
{Reset all parameters}
begin
inherited Reset;
Scale:=1;
ResetMaterials;
end;


{---------------- Object Event handlers ----------------------------------}


constructor TGLRenderer.Create;
{Warning: Textures must be generated only after the GL Window has}
{been created. Other wise "glGenTextures' names will be wrong}
begin
inherited Create;
Texture:=TGLObjectTexture.Create;
ResetPos:=DefaultPos;
Position:=DefaultPos;
DoAlphaBlend:=False;
UserBlend:=False;
TransparentDepth:=False;
BlendBackground:=False;
SrcUserBlend:=GL_SRC_ALPHA;
DestUserBlend:=GL_ONE;
BlendEquation:=GL_FUNC_ADD;
FlyThrough:=True;
Transparency:=0;
Reset;
FRenderMode:=rmSmoothFace;
FDoTwoSidedLighting:=False;
FDoFrontBackColor:=False;
FBackColor:=clBlue;
FFrontColor:=clGreen;
end;



destructor TGLRenderer.Destroy;
begin
Texture.Free;
inherited Destroy;
end;


{---------------- Alpha/Transparency routines --------------------------------}

procedure TGLRenderer.SetAlphaBlend;
begin
if DoAlphaBlend then
	begin
	if UserBlend then
		begin
		glBlendEquation(BlendEquation);
		glBlendFunc(SrcUserBlend, DestUserBlend);
		end
	else if BlendBackground then
		begin
		glBlendEquation(GL_FUNC_ADD);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		end
	else
		begin
		glBlendEquation(GL_FUNC_ADD);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE);
		end;
	glEnable(GL_BLEND);
	if TransparentDepth then glEnable(GL_DEPTH_TEST)
	else glDisable(GL_DEPTH_TEST);
	end
else
	begin
	glDisable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);
	end;
end;


function TransparencyToAlpha(T: double): double;
begin
Result:=(100-T)/100;
end;


function TGLRenderer.GetCurrentAlpha: double;
begin
Result:=TransparencyToAlpha(Transparency);
end;



procedure TGLRenderer.ResetMaterials;
begin
LightEmission:= MakeGLArrayf4(0,0,0,0);
LightSpecular:= MakeGLArrayf4(0.5,0.5,0.5,1.0);
LightShininess:= 10;
end;



procedure TGLRenderer.SetMaterials;
begin
{
GL_Diffuse = Base Color
GL_Specular = Highlight Color
GL_Ambient  = Low Light Color
GL_Shininess = Surface Smoothness
GL_Emission = Glow Color
}
glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION,@LightEmission);
glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR,@LightSpecular);
glMaterialfv(GL_FRONT_AND_BACK,GL_SHININESS,@LightShininess);
end;



procedure TGLRenderer.SetLighting;
var FC,BC: TGLArrayf4;
begin
if FDoFrontBackColor then
	begin
	glDisable(GL_COLOR_MATERIAL);
	FC:=ColorToGLRGBA(FFrontColor);
	BC:=ColorToGLRGBA(FBackColor);
	glMaterialfv(GL_BACK , GL_AMBIENT_AND_DIFFUSE, @BC);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, @FC);
	end
else
	begin
	glEnable(GL_COLOR_MATERIAL);
	end;

if FDoTwoSidedLighting then glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE)
else glEnable(GL_COLOR_MATERIAL);
end;




procedure TGLRenderer.Render;
{Render whole mesh}
begin
{Don't load indentify matrix here}
SetAlphaBlend;
inherited Render;
SetMaterials;
SetLighting;
{Do your rendering here in decendent class}
end;



function TGLObjectList.GetExtentMax: T3DVector;
var I: integer;
begin
Result:=MakeVector3D(0,0,0);
if Count<1 then exit;
Result:=TGLRenderer(Self[0]).ExtentMax;
for I:=1 to Count-1 do
	Result:=GetMaximumVector(TGLRenderer(Self[I]).ExtentMax,Result);
end;





function TGLObjectList.GetExtentMin: T3DVector;
var I: integer;
begin
Result:=MakeVector3D(0,0,0);
if Count<1 then exit;
Result:=TGLRenderer(Self[0]).ExtentMin;
for I:=1 to Count-1 do
	Result:=GetMinimumVector(TGLRenderer(Self[I]).ExtentMin,Result);
end;



procedure TGLRenderer.Clear;
begin
Reset;
end;



procedure TGLRenderer.SetColor(const Color: TColor);
begin
end;



procedure TGLRenderer.SetRenderMode(const Value: TRenderMode);
begin
if FRenderMode<>Value then
	begin
	FRenderMode := Value;
	Change;
	end;
end;



procedure TGLRenderer.Assign(GLObject: TGLObject);
begin
inherited Assign(GLObject);
if not (GLObject is TGLRenderer) then exit;
DoAlphaBlend:=TGLRenderer(GLObject).DoAlphaBlend;
Transparency:=TGLRenderer(GLObject).Transparency;
TransparentDepth:=TGLRenderer(GLObject).TransparentDepth;
BlendBackground:=TGLRenderer(GLObject).BlendBackground;
LightEmission:=TGLRenderer(GLObject).LightEmission;
LightSpecular:=TGLRenderer(GLObject).LightSpecular;
LightShininess:=TGLRenderer(GLObject).LightShininess;
Texture.Assign(TGLRenderer(GLObject).Texture);
Color:=TGLRenderer(GLObject).Color;
RenderMode:=TGLRenderer(GLObject).RenderMode;

DoTwoSidedLighting:=TGLRenderer(GLObject).DoTwoSidedLighting;
DoFrontBackColor:=TGLRenderer(GLObject).DoFrontBackColor;
FrontColor:=TGLRenderer(GLObject).FrontColor;
BackColor:=TGLRenderer(GLObject).BackColor;
end;





procedure TGLRenderer.Append(GLObject: TGLObject);
begin
inherited Append(GLObject);
if not (GLObject is TGLRenderer) then exit;
{We will retain the destination mesh's properties}
end;


function TGLRenderer.GetColor: TColor;
{This is a virtual function filled in by descendents}
begin
end;


procedure TGLRenderer.SetDoTwoSidedLighting(const Value: boolean);
begin
if FDoTwoSidedLighting<> Value then
	begin
	FDoTwoSidedLighting:=Value;
	Change;
	end;
end;




procedure TGLRenderer.SetDoFrontBackColor(const Value: boolean);
begin
if FDoFrontBackColor<>Value then
	begin
	FDoFrontBackColor:=Value;
	Change;
	end;
end;

procedure TGLRenderer.SetBackColor(const Value: TColor);
begin
if FBackColor<>Value then
	begin
	FBackColor:=Value;
	Change;
	end;
end;

procedure TGLRenderer.SetFrontColor(const Value: TColor);
begin
if FFrontColor<>Value then
	begin
	FFrontColor:=Value;
	Change;
	end;
end;




function TGLObject.GetExtentMax: T3DVector;
begin
Result:=FExtentMax;
end;

function TGLObject.GetExtentMin: T3DVector;
begin
Result:=FExtentMin;
end;

procedure TGLObject.SetExtentMax(const Value: T3DVector);
begin
FExtentMax:=Value;
end;

procedure TGLObject.SetExtentMin(const Value: T3DVector);
begin
FExtentMin:=Value;
end;


procedure TGLRenderer.SetTransparency(const Value: double);
begin
if FTransparency<>Value then FTransparency := Value;
end;

end.
