unit GLTesselator;

interface

uses Windows, Forms,Classes,SysUtils,Graphics,OpenGL15,Vector,GLLibrary,GLPolygons,GLMesh;

type PGLArrayd3 = ^TGLArrayd3;

var  ErrorCount: integer;
var  ErrorMessage: string;

type TTesselateProgress = procedure(Percent: integer);


type TGLTesselator = class(TOBject)
 private
  procedure CreateBackSurface(Mesh: TGLMesh; Depth: double);
  procedure CreateSideSurfaces(Mesh: TGLMesh; Polygons: TGLPolygonArray; Depth: double);
  procedure NormalizePolygons(Polygons: TGLPolygonArray); overload;
  procedure NormalizePolygons(Polygons: TRealPolygonArray); overload;
 protected
 public
  procedure ExtrudeBeveled(RPA: TRealPolygonArray; Mesh: TGLMesh; SplineSteps: integer; Depth,Bevel: double; ProgressProc: TTesselateProgress);
  procedure TesselatePolygons(Polygons: TGLPolygonArray; Mesh: TGLMesh; ProgressProc: TTesselateProgress);
  procedure ExtrudePolygons(Polygons: TGLPolygonArray; Mesh: TGLMesh; Depth: double; ProgressProc: TTesselateProgress); overload;
  procedure ExtrudePolygons(RPA: TRealPolygonArray; Mesh: TGLMesh; SplineSteps: integer; Depth: double; ProgressProc: TTesselateProgress); overload;
end;

procedure SmoothArray(RPA,SPA: TRealPolygonArray; Steps: integer);


implementation



{--------- The call back routines have to be regular functions  ---------}


 {Items used by Tesselator  }

type TPointerArray = Array of pointer;

type TTessItems = record
 VertexArray: array [0..2] of T3DVector;
 VertexCount: integer;
 CombineArray: TPointerArray;
 Mesh: TGLMesh;
 end;

procedure TessBeginCallBack(atype: GLenum; var TessItems: TTessItems); stdcall;
begin
with TessItems do
 begin
 VertexCount:=0;
 end;
end;


procedure TessEndCallBack(var TessItems: TTessItems); stdcall;
begin
end;



procedure TessEdgeFlagCallBack(Flag: GLBoolean; var TessItems: TTessItems); stdcall;
{Note: the edge call back is required because it}
{ forces the Tesselator to use only triangles}
begin
end;




procedure CombineCallBack(var GA: TGLArrayd3; var Vertices: TGLArrayi4; var Weight: TGLArrayf4; var OutData: PGLArrayd3; var TessItems: TTessItems); stdcall;
{Becareful with the combine array. If the vertices are
stored in an array and a pointer to each items is used, the
pointer may change when the array reallocates space. This
will cause problems.}
var H: integer;
var GLA: PGLArrayd3;
begin
with TessItems do
 begin
 New(GLA);
 GLA[0]:=GA[0];
 GLA[1]:=GA[1];
 GLA[2]:=GA[2];
 OutData:=GLA;

 SetLength(CombineArray,Length(CombineArray)+1);
 H:=High(CombineArray);
 CombineArray[H]:=GLA;
 end;
end;




procedure TessErrorCallBack(Error: GLenum; var TessItems: TTessItems); stdcall;
var S: string;
begin
S:=gluErrorString(Error);
end;



procedure TessVertexCallBack(var GA: TGLArrayd3; var TessItems: TTessItems); stdcall;
begin
with TessItems do
 begin
 VertexArray[VertexCount]:=MakeVector3D(GA[0],GA[1],GA[2]);
 Inc(VertexCount);
 if VertexCount>2 then
	begin
	Mesh.AddTriangle(VertexArray[2],VertexArray[1],VertexArray[0],clRed);
	VertexCount:=0;
	end;
 end;
end;




procedure TGLTesselator.TesselatePolygons(Polygons: TGLPolygonArray; Mesh: TGLMesh; ProgressProc: TTesselateProgress);
var Tess: PGLUtriangulatorObj;
var I,J: integer;
var TessItems: TTessItems;
begin
TessItems.Mesh:=Mesh;
Tess:=gluNewTess;
if Tess=nil then raise Exception.Create('Unable to create Tesselator');
{Note: the Edge call back is required because it}
{ forces the Tesselator to use only triangles}
gluTessCallback(tess, GLU_TESS_EDGE_FLAG_DATA, @TessEdgeFlagCallBack);
gluTessCallback(tess, GLU_TESS_BEGIN_DATA, @TessBeginCallBack);
gluTessCallback(tess, GLU_TESS_VERTEX_DATA, @TessVertexCallBack);
gluTessCallback(tess, GLU_TESS_END_DATA, @TessEndCallBack);
gluTessCallback(tess, GLU_TESS_COMBINE_DATA, @CombineCallBack);
gluTessCallback(tess, GLU_TESS_ERROR_DATA, @TessErrorCallBack);
gluTessProperty(tess, GLU_TESS_WINDING_RULE,GLU_TESS_WINDING_ODD);

SetLength(TessItems.CombineArray,0);

gluTessBeginPolygon(Tess,@TessItems);
for I:=0 to High(Polygons) do
	begin
	if Assigned(ProgressProc) then ProgressProc(MulDiv(95,I,High(Polygons)));
	gluTessBeginContour(Tess);
        for J:=0 to High(Polygons[I]) do
		begin
		gluTessVertex(Tess,Polygons[I,J],@Polygons[I,J]);
		end;
	gluTessEndContour(Tess);
	end;
gluTessEndPolygon(Tess);
gluDeleteTess(tess);
for I:=0 to High(TessItems.CombineArray) do
	Dispose(TessItems.CombineArray[I]);
end;



procedure TGLTesselator.CreateBackSurface(Mesh: TGLMesh; Depth: double);
var I: integer;
var T: TTriangle;
begin
Depth:=-abs(Depth);
for I:=0 to Mesh.Vertices.TriCount-1 do
	begin
	T:=Mesh.Vertices.Triangle[I];
	T.P1.Z:=T.P1.Z+Depth;
	T.P2.Z:=T.P2.Z+Depth;
	T.P3.Z:=T.P3.Z+Depth;
	T:=ReverseWinding(T);
	Mesh.AddTriangle(T,Mesh.FaceColor.Face[I]);
	end;
end;



procedure TGLTesselator.CreateSideSurfaces(Mesh: TGLMesh; Polygons: TGLPolygonArray; Depth: double);
var I,J: integer;
var R3D: T3DRectangle;
begin
Depth:=-abs(Depth);
for I:=0 to High(Polygons) do
 for J:=0 to High(Polygons[I]) do
	begin
	R3D.P1:=GLArrayToVector3D(Polygons[I,J]);
	R3D.P4:=R3D.P1; R3D.P4.Z:=R3D.P4.Z+Depth;
	if J=High(Polygons[I]) then R3D.P2:=GLArrayToVector3D(Polygons[I,0])
	else R3D.P2:=GLArrayToVector3D(Polygons[I,J+1]);
	R3D.P3:=R3D.P2;  R3D.P3.Z:=R3D.P3.Z+Depth;
	Mesh.AddRectangle(R3D,clBlue);
	end;
end;



procedure TGLTesselator.NormalizePolygons(Polygons: TGLPolygonArray);
{Normalize polygons so interior and exterior polygons are counter clockwise and interiors are clockwise}
{Note this routines assumes OpenGL coordinages where Y is reversed from Screen}
var I: integer;
var S: string;
begin
S:='';
for I:=0 to High(Polygons) do
	begin
	S:=S+IntToStr(I)+':';
	if Length(Polygons[I])<1 then continue;
	if IsExteriorPoint(Polygons,I) then
		begin
		S:=S+' E';
		if IsClockwise(Polygons[I]) then
			begin
			S:=S+' C';
			ReverseLines(Polygons[I]);
			end;
		end
	else
		begin
		S:=S+' I';
		if not IsClockwise(Polygons[I]) then
			begin
			S:=S+' CC';
			ReverseLines(Polygons[I]);
			end;
		end;
	S:=S+', ';
	end;
//TessForm.Caption:=S;
end;


procedure TGLTesselator.NormalizePolygons(Polygons: TRealPolygonArray);
{Normalize polygons so interior and exterior polygons are counter clockwise and interiors are clockwise}
{Note this routines assumes screen coordinate where Y is reversed from opengl}
var I: integer;
begin
for I:=0 to Polygons.Count-1 do
	begin
	if Polygons[I].Count<1 then continue;
	if IsExteriorPoint(Polygons,I) then
		begin
		if not IsClockwise(Polygons[I]) then ReverseLines(Polygons[I]);
		end
	else
		begin
		if IsClockwise(Polygons[I]) then ReverseLines(Polygons[I]);
		end;
	end;
end;



procedure SmoothArray(RPA,SPA: TRealPolygonArray; Steps: integer);
var I,J,K: integer;
var PG: T2DVectorArray;
begin
SPA.Clear;
for I:=0 to RPA.Count-1 do
	begin
	SPA.Add;
	for J:=0 to RPA[I].Count-1 do
		begin
		GetSplineLine(J,RPA[I],Steps,PG);
		for K:=0 to High(PG)-1 do SPA[I].Add(PG[K]);
		end;
	end;
end;



procedure TGLTesselator.ExtrudePolygons(Polygons: TGLPolygonArray; Mesh: TGLMesh; Depth: double; ProgressProc: TTesselateProgress);
begin
NormalizePolygons(Polygons);
TesselatePolygons(Polygons,Mesh,ProgressProc);
CreateBackSurface(Mesh,Depth);
CreateSideSurfaces(Mesh,Polygons,Depth);
end;



procedure OffsetPolygon(var Poly: TGLPolygonArray; Offset: double);
var I,J: integer;
begin
for I:=0 to High(Poly) do
for J:=0 to High(Poly[I]) do
	Poly[I,J,2]:=Poly[I,J,2]+Offset;
end;


procedure CreateEdge(Inner,Outer: TGLPolygonArray; Mesh: TGLMesh; Reverse: boolean);
var I,J,K: integer;
var R3D: T3DRectangle;
var V: T3DVector;
begin
for I:=0 to High(Inner) do
 for J:=0 to High(Inner[I]) do
	begin
	R3D.P1:=GLArrayToVector3D(Inner[I,J]);
	R3D.P4:=GLArrayToVector3D(Outer[I,J]);
	if J=High(Inner[I]) then K:=0 else K:=J+1;
	R3D.P2:=GLArrayToVector3D(Inner[I,K]);
	R3D.P3:=GLArrayToVector3D(Outer[I,K]);
	if Reverse then
		begin
		Vector.ExchangeVectors3D(R3D.P1,R3D.P2);
		Vector.ExchangeVectors3D(R3D.P3,R3D.P4);
		end;
	Mesh.AddRectangle(R3D,clBlue);
	end;
end;




procedure TGLTesselator.ExtrudeBeveled(RPA: TRealPolygonArray; Mesh: TGLMesh; SplineSteps: integer; Depth,Bevel: double; ProgressProc: TTesselateProgress);
var Inner: TRealPolygonArray;
var I,Inx: integer;
var InPoly,OutPoly: TGLPolygonArray;
var SPA: TRealPolygonArray;
begin
Bevel:=Bevel * 100;
SPA:=TRealPolygonArray.Create;
Inner:=TRealPolygonArray.Create;

SmoothArray(RPA,SPA,SplineSteps);
CenterPolygons(SPA);
NormalizePolygons(SPA);
ProgressProc(0);
for I:=0 to SPA.Count-1 do
	begin
	if (I mod 10)=0 then ProgressProc(MulDiv(100,I,SPA.Count));
	Inx:=Inner.Add;
	InflatePolygon(SPA[I],Inner[Inx],False,True,-Bevel);
	end;

RealPolygonArrayToGLPolygons(SPA,OutPoly,100);
RealPolygonArrayToGLPolygons(Inner,InPoly,100);
TesselatePolygons(InPoly,Mesh,ProgressProc);
CreateBackSurface(Mesh,Depth);

{Create front edge}
OffsetPolygon(OutPoly,-Depth/3);
CreateSideSurfaces(Mesh,OutPoly,Depth/3);
CreateEdge(InPoly,OutPoly,Mesh,False);

{Create Back edge}
OffsetPolygon(OutPoly,-Depth/3);
OffsetPolygon(InPoly,-Depth);
CreateEdge(InPoly,OutPoly,Mesh,True);

Inner.Free;
SPA.Free;
end;



procedure TGLTesselator.ExtrudePolygons(RPA: TRealPolygonArray; Mesh: TGLMesh; SplineSteps: integer; Depth: double; ProgressProc: TTesselateProgress);
var Polygons: TGLPolygonArray;
var SPA: TRealPolygonArray;
begin
SPA:=TRealPolygonArray.Create;
SmoothArray(RPA,SPA,SplineSteps);
CenterPolygons(SPA);

RealPolygonArrayToGLPolygons(SPA,Polygons,100);
ExtrudePolygons(Polygons,Mesh,Depth,ProgressProc);
SPA.Free;
end;





end.
