unit GLTexture;

interface

uses Windows,Forms,Classes,ExtCtrls,SysUtils,Math,Graphics,
  ImageLib,HugeArry,OpenGL15,GLLibrary,GLCubeMap,Vector,DSupport,GLQuatLib,GLArrays;



{------ Object to display background texture --------}

type TGLBackTexture = class(TGLBaseBitmap)
 private
 protected
 public
  Position: T3DVector;
  Zoom: T2DVector;
  constructor Create; reintroduce; virtual;
  procedure Render;
 end;


{------ Object to display textures for Graphic objects--------}

type TTextureMode = (tmAuto,tmManual,tmCubeMap);
type TAutoTextureMap = (amObjectLinear,amEyeLinear,amSphereMap,amNormalMap,amReflectionMap);
type TManualTextureMap = (mmLinear,mmCylindrical,mmSpherical,mmToroidal);


type TGLObjectTexture = class(TGLBaseBitmap)
 private
  FManualTextureMap: TManualTextureMap;
  FScale: T2DVector;
  FOffset: T2DVector;
  FRadius: double;
  FYaw: double;
  FPitch: double;
  FTPlane: TGLArrayf4;
  FSPlane: TGLArrayf4;
  FTextureBuilt: boolean;
  FHighlight: boolean;
  FRoll: double;
  procedure SetFlatTexture;
  procedure SetAutoTexture;
  procedure SetManualTextureMap(const Value: TManualTextureMap);
  procedure SetScale(const Value: T2DVector);
  procedure SetOffset(const Value: T2DVector);
  procedure SetYaw(const Value: double);
  procedure SetPitch(const Value: double);
  function GetFlatTextureUV(V,DX,Min: T3DVector; Offset,Scale: T2DVector): T2DVector;
  function GetCylindricalTextureUV(Cross: boolean; V: T3DVector; Offset,Scale: T2DVector): T2DVector;
  function GetSphericalTextureUV(Cross: boolean; V: T3DVector; Offset,Scale: T2DVector): T2DVector;
  procedure SetSPlane(const Value: TGLArrayf4);
  procedure SetTPlane(const Value: TGLArrayf4);
  procedure BuildTexture;
  procedure SetHighlight(const Value: boolean);
  procedure SetRoll(const Value: double);
  function GetToriodalTextureUV(Cross: boolean; V: T3DVector; Radius: double; Offset, Scale: T2DVector): T2DVector;
  procedure SetRadius(const Value: double);
 protected
  procedure Change; override;
 public
  UVCoordinates: THuge2DVector;
  TextureMode: TTextureMode;
  AutoTextureMap: TAutoTextureMap;
  CubeMap: TGLCubeMap;
  procedure SetupTextureVertices(VA: THuge3DVectorEx; Min,Max: T3DVector); overload;
  procedure SetupTextureVertices(Triangles: TTriInterface; Min,Max: T3DVector); overload;
  constructor Create; override;
  destructor Destroy; override;
  procedure Render;
  procedure FinalizeRender;
  procedure Assign(Source: TPersistent);
  property ManualTextureMap: TManualTextureMap read FManualTextureMap write SetManualTextureMap;
  property Scale: T2DVector read FScale write SetScale;
  property Offset: T2DVector read FOffset write SetOffset;
  property Radius: double read FRadius write SetRadius;
  property Pitch: double read FPitch write SetPitch;
  property Roll: double read FRoll write SetRoll;
  property Yaw: double read FYaw write SetYaw;
  property SPlane: TGLArrayf4 read FSPlane write SetSPlane;
  property TPlane: TGLArrayf4 read FTPlane write SetTPlane;
  property Highlight: boolean read FHighlight write SetHighlight;
 end;

            

implementation


{------------------------ Back Texture ----------------------------------------}


constructor TGLBackTexture.Create;
begin
inherited;
Position:=MakeVector3D(0,0,-10);
Zoom:=MakeVector2D(1,1);
end;



procedure TGLBackTexture.Render;
begin
glRasterPos3d(0,0,Position.Z);
OpenGL15.glBitmap(0, 0, 0, 0, Position.X,Position.Y, nil);
glPixelZoom(Zoom.X,Zoom.Y);
glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
glDrawPixels(Width,Height, GL_RGBA,GL_UNSIGNED_BYTE,GLBitmap);
end;


{ =================== TGLObjectTexture ==================================}




function RotateVector3DX(V: T3DVector; Pitch,Yaw,Roll: double): T3DVector;
{ Rotate Coordinates relative to origin  }
var T: T3DVector;
var SinA,CosA,PSinA,PCosA: double;
begin
CosA:=cos(Yaw*Degree2Radian); SinA:=sin(Yaw*Degree2Radian);
PCosA:=cos(Pitch*Degree2Radian); PSinA:=sin(Pitch*Degree2Radian);
T:=V;
Result.Y := (SinA  * -T.X) + (CosA  * T.Y);
Result.X := (CosA  * T.X) + (SinA  * T.Y);
Result.Z := (PSinA  * -V.Y) + (PCosA  * T.Z);
Result.Y := (PCosA  * V.Y) + (PSinA  * T.Z);
end;


function RotateVector3D(V: T3DVector; Pitch,Yaw,Roll: double): T3DVector;
var Q: TQuaternion;
begin
Q:=EulerToQuat(-Pitch,Yaw,Roll);
Result:=QuaternionTransform(Q,V);
end;


function RotateTriangle(T: TTriangle; Yaw,Pitch: double): TTriangle;
{ Rotate Triangle relative to origin  }
begin
Result.P1:=RotateVector3D(T.P1,Pitch,Yaw,0);
Result.P2:=RotateVector3D(T.P2,Pitch,Yaw,0);
Result.P3:=RotateVector3D(T.P3,Pitch,Yaw,0);
end;





function TGLObjectTexture.GetFlatTextureUV(V,DX,Min: T3DVector; Offset,Scale: T2DVector): T2DVector;
begin
V:=RotateVector3D(V,FPitch,FYaw,FRoll);
Result.X:=V.X * Scale.X - Offset.X;
Result.Y:=V.Y * Scale.Y - Offset.Y;
end;


function GetOriginAngle(V: T3DVector): double;
{Calculate angle between origin an point}
{Angle is negative going clockwise so texture is not reversed}
begin
Result:=ArcTan2(V.Y,V.X);
end;



function TGLObjectTexture.GetCylindricalTextureUV(Cross: boolean; V: T3DVector; Offset,Scale: T2DVector): T2DVector;
{Calculate Cylindrical texture coordinates }
begin
{Only pitch is relevant offset already rotates around central axis}
V:=RotateVector3D(V,FPitch,FYaw,0);
Result.X:=((1/(2*Pi)) * ArcTan2(V.Y,V.X));
Result.Y:=V.Z * Scale.Y - Offset.Y;
{Make U-Coordinate negative if we cross Quarant four and U is positive }
if Cross and (Result.X>0) then Result.X:=Result.X - 1;
{Handle scale and offset}
Result.X:=Result.X * Scale.X- Offset.X
end;



function TGLObjectTexture.GetSphericalTextureUV(Cross: boolean; V: T3DVector; Offset,Scale: T2DVector): T2DVector;
begin
{Only pitch is relevant offset already rotates around central axis}
V:=RotateVector3D(V,FPitch,0,0);
Result.X:=(1/(2*Pi) * ArcTan2(V.Y,V.X));
Result.Y:=(Scale.Y/(Pi) * ArcTan2(V.Z,VectorAbs3D(V)))- Offset.Y;
{Make U-Coordinate negative if we cross Quarant four and U is positive }
if Cross and (Result.X>0) then Result.X:=Result.X - 1;
{Handle scale and offset}
Result.X:=Result.X * Scale.X- Offset.X
end;




function TGLObjectTexture.GetToriodalTextureUV(Cross: boolean; V: T3DVector; Radius: double; Offset,Scale: T2DVector): T2DVector;
var A,B,C,D,Angle: double;
begin
{Only pitch is relevant offset already rotates around central axis}
V:=RotateVector3D(V,FPitch,0,0);

Result.X:=(1/(2*Pi) * (ArcTan2(V.Y,V.X) + Pi));
{Make U-Coordinate negative if we cross Quarant four and U is positive }
if Cross and (Result.X>=0.5) then Result.X:=Result.X - 1;
{Handle scale and offset}
Result.X:=Result.X * Scale.X - Offset.X;

{Calculate V or X texture coordinat by Calculating the}
{angle of vertex relative to center radius of torus. }
{ Points: Orgin = P1, Vertex = P4, Radius = P3 and P2 = Vertical from P4}
{ Sides: A = P1-P2, B = P2-P4, C = P1-P3, D = P3-P4}
{ Triangles: A,B,C and B,D,E}
C:=VectorAbs3D(V);		{Hypotenuse}
B:=V.Z;				{Opposite }
A:=sqrt(sqr(C) - sqr(B));	{Adjacent}
D:=Radius - A;			{Secondary adjacent}
Angle:=ArcTan2(B,D);

{Scale and offset and use as V texture coordinnate}
Result.Y:=(Scale.Y/(2*Pi) * (Angle+Pi)) - Offset.Y;
end;



function SpanQuadrants(T: TTriangle): boolean;
{Does specified triangle span quadrants 2 and 3}
var A1,A2,A3,N: double;
var Neg,Pos: boolean;
begin
{Calculate angle of ray from origin to center of face}
N:=GetOriginAngle(GetTriangleCenter(T));

{Calculate angle of the triangle vertices}
A1:=GetOriginAngle(T.P1);
A2:=GetOriginAngle(T.P2);
A3:=GetOriginAngle(T.P3);
{if the angles are both positive and negative it crosses 2/3 or 4/1 }
Pos:=(A1>=0) or (A2>=0) or (A3>=0);
Neg:=(A1<0) or (A2<0) or (A3<0);
{if the Ray angle is greater than 90 it is 2/3}
Result:=Pos and Neg and (abs(N)>(pi/2));
end;



function MaxTriangeLength(T: TTriangle): double;
{Return the length of the longest side of the triangle}
var L1,L2,L3: double;
begin
L1:=VectorAbs3D(VectorSubtract3D(T.P2,T.P1));
L2:=VectorAbs3D(VectorSubtract3D(T.P3,T.P1));
L3:=VectorAbs3D(VectorSubtract3D(T.P3,T.P2));
Result:=Max(Max(L1,L2),L3);
end;


function GetArcUVDistance(Center: T3DVector; P: T3DVector; T: TTriangle): T2DVector;
{Currently unused - routine to see if we could get better rock texture - never tested}
var Azm,Len,Incl: double;
var ArcLen,CurlLen,Angle,TriSize: double;
var Normal: T3DVector;
var S: string;
begin
{First calculate U-values: Get polar values}
VectorToPolar3D(Center,P,Azm,Incl,Len);
{Calculate distance clockwise around circumference}
{ Then add to radial distance - this is U value}
ArcLen:=Len * (360-Azm) * Degree2Radian;

{Caculate V-Value from normal to triangle}
Normal:=GetTriangleNormal(T);
{Get angle of normal relative to horizontal}
Angle:=VectorAngle3D(MakeVector3D(0,0,1),Normal);
{Set the sign of the value}
if Normal.Y<0 then Angle:=-Angle;
{Get length of longest side of triangle}
TriSize:=MaxTriangeLength(T);
{Use this for the V-Length}
CurlLen:=TriSize * Angle;
Result.X:=ArcLen;
Result.Y:=CurlLen;
end;




procedure TGLObjectTexture.SetupTextureVertices(VA: THuge3DVectorEx; Min,Max: T3DVector);
{Apply manual Texture for regular TGLMesh}
var I: integer;
var DX: T3DVector;
var T: TTriangle;
var Cross: boolean;
begin
DX:=VectorSubtract3D(Max,Min);
UVCoordinates.Clear;
case Self.FManualTextureMap of
 mmLinear: for I:=0 to VA.Count-1 do UVCoordinates.Add(GetFlatTextureUV(VA[I],DX,Min,FOffset,FScale));
 mmCylindrical: for I:=0 to VA.TriCount-1 do
	begin
	T:=VA.Triangle[I];
	Cross:=SpanQuadrants(RotateTriangle(T,0,FPitch));
	UVCoordinates.Add(GetCylindricalTextureUV(Cross,T.P1,FOffset,FScale));
	UVCoordinates.Add(GetCylindricalTextureUV(Cross,T.P2,FOffset,FScale));
	UVCoordinates.Add(GetCylindricalTextureUV(Cross,T.P3,FOffset,FScale));
	end;
 mmSpherical: for I:=0 to VA.TriCount-1 do
	begin
	T:=VA.Triangle[I];
	Cross:=SpanQuadrants(RotateTriangle(T,0,FPitch));
	UVCoordinates.Add(GetSphericalTextureUV(Cross,T.P1,FOffset,FScale));
	UVCoordinates.Add(GetSphericalTextureUV(Cross,T.P2,FOffset,FScale));
	UVCoordinates.Add(GetSphericalTextureUV(Cross,T.P3,FOffset,FScale));
	end;
 mmToroidal: for I:=0 to VA.TriCount-1 do
	begin
	T:=VA.Triangle[I];
	Cross:=SpanQuadrants(RotateTriangle(T,0,FPitch));
	UVCoordinates.Add(GetToriodalTextureUV(Cross,T.P1,FRadius,FOffset,FScale));
	UVCoordinates.Add(GetToriodalTextureUV(Cross,T.P2,FRadius,FOffset,FScale));
	UVCoordinates.Add(GetToriodalTextureUV(Cross,T.P3,FRadius,FOffset,FScale));
	end;
 end;
end;







procedure TGLObjectTexture.SetupTextureVertices(Triangles: TTriInterface; Min,Max: T3DVector);
{Apply manual Texture for new TGLFastMesh}
var I: integer;
var DX: T3DVector;
var T: TTriangle;
var Cross: boolean;
var Face: TFace;
begin
DX:=VectorSubtract3D(Max,Min);
UVCoordinates.Clear;

if Triangles.Vertices.Count<1 then UVCoordinates.MaxItems:=1
else UVCoordinates.MaxItems:=Triangles.Vertices.Count;
case Self.FManualTextureMap of
 mmLinear: for I:=0 to Triangles.Vertices.Count-1 do UVCoordinates.Add(GetFlatTextureUV(Triangles.Vertices[I],DX,Min,FOffset,FScale));
 mmCylindrical: for I:=0 to Triangles.Count-1 do
	begin
	T:=Triangles[I];
	Face:=Triangles.Faces[I];
	Cross:=SpanQuadrants(RotateTriangle(T,0,FPitch));
	UVCoordinates[Face.P1]:=GetCylindricalTextureUV(Cross,T.P1,FOffset,FScale);
	UVCoordinates[Face.P2]:=GetCylindricalTextureUV(Cross,T.P2,FOffset,FScale);
	UVCoordinates[Face.P3]:=GetCylindricalTextureUV(Cross,T.P3,FOffset,FScale);
	end;
 mmSpherical: for I:=0 to Triangles.Count-1 do
	begin
	T:=Triangles[I];
	Face:=Triangles.Faces[I];
	Cross:=SpanQuadrants(RotateTriangle(T,0,FPitch));
	UVCoordinates[Face.P1]:=GetSphericalTextureUV(Cross,T.P1,FOffset,FScale);
	UVCoordinates[Face.P2]:=GetSphericalTextureUV(Cross,T.P2,FOffset,FScale);
	UVCoordinates[Face.P3]:=GetSphericalTextureUV(Cross,T.P3,FOffset,FScale);
	end;
 mmToroidal: for I:=0 to Triangles.Count-1 do
	begin
	T:=Triangles[I];
	Face:=Triangles.Faces[I];
	Cross:=SpanQuadrants(RotateTriangle(T,0,FPitch));
	UVCoordinates[Face.P1]:=GetToriodalTextureUV(Cross,T.P1,FRadius,FOffset,FScale);
	UVCoordinates[Face.P2]:=GetToriodalTextureUV(Cross,T.P2,FRadius,FOffset,FScale);
	UVCoordinates[Face.P3]:=GetToriodalTextureUV(Cross,T.P3,FRadius,FOffset,FScale);
	end;
 end;
end;




procedure TGLObjectTexture.BuildTexture;
begin
SetupTextureParameters(TGLBaseBitmap(Self));
FTextureBuilt:=True;
end;



procedure TGLObjectTexture.Change;
begin
inherited;
FTextureBuilt:=False;
end;



constructor TGLObjectTexture.Create;
begin
inherited;
FTextureBuilt:=False;
UVCoordinates:=THuge2DVector.Create(5);
CubeMap:=TGLCubeMap.Create;
FScale:=MakeVector2D(1,1);
FOffset:=MakeVector2D(0,0);
FPitch:=0;
FYaw:=0;
FRoll:=0;
TextureMode:=tmAuto;
AutoTextureMap:=amObjectLinear;
FSPlane:=MakeGLArrayf4(1,0,0,0);
FTPlane:=MakeGLArrayf4(0,1,0,0);
FRadius:=1;
end;



destructor TGLObjectTexture.Destroy;
begin
CubeMap.Free;
UVCoordinates.Free;
inherited;
end;




procedure TGLObjectTexture.SetAutoTexture;
begin
glMatrixMode(GL_TEXTURE);
glLoadIdentity;
glScalef(FScale.X, FScale.y, 1);
glTranslatef(FOffset.X,FOffset.Y,0);
{Apply roll here for simple way to rotate the texture around its center}
glRotated(FRoll,0,0,1);
case AutoTextureMap of
 amObjectLinear:
	begin
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	glTexGenfv(GL_S, GL_OBJECT_PLANE, @FSPlane);
	glTexGenfv(GL_T, GL_OBJECT_PLANE, @FTPlane);
	end;
 amEyeLinear:
	begin
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	glTexGenfv(GL_S, GL_EYE_PLANE, @FSPlane);
	glTexGenfv(GL_T, GL_EYE_PLANE, @FTPlane);
	end;
 amSphereMap:
	begin
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	end;
 amNormalMap: ;
 amReflectionMap: ;
 end;

{Automatically generate coordinates}
glEnable(GL_TEXTURE_GEN_S);
glEnable(GL_TEXTURE_GEN_T);
glMatrixMode(GL_MODELVIEW);
{ Select The Texture }
glBindTexture(GL_TEXTURE_2D, Textname);
{ Enable Texture Mapping }
glEnable(GL_TEXTURE_2D);
end;


procedure TGLObjectTexture.SetFlatTexture;
begin
glMatrixMode(GL_TEXTURE);
glLoadIdentity;
glDisable(GL_TEXTURE_GEN_S);
glDisable(GL_TEXTURE_GEN_T);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
glMatrixMode(GL_MODELVIEW);
{ Select The Texture }
glBindTexture(GL_TEXTURE_2D, Textname);
{ Enable Texture Mapping }
glEnable(GL_TEXTURE_2D);
end;



procedure TGLObjectTexture.Render;
begin
if Self.FHighlight then glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR)
else glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SINGLE_COLOR);
if not FTextureBuilt then BuildTexture;
case TextureMode of
 tmManual: SetFlatTexture;
 tmAuto: SetAutoTexture;
 tmCubeMap: CubeMap.Render;
 else SetAutoTexture;
 end;
end;



procedure TGLObjectTexture.FinalizeRender;
begin
case TextureMode of
 tmManual: glDisable(GL_TEXTURE_2D);
 tmAuto: glDisable(GL_TEXTURE_2D);
 tmCubeMap: CubeMap.FinalizeRender;
 else glDisable(GL_TEXTURE_2D);
 end;
end;



procedure TGLObjectTexture.SetManualTextureMap(const Value: TManualTextureMap);
begin
if FManualTextureMap<>Value then
	begin
	FManualTextureMap := Value;
	Change;
	end;
end;



procedure TGLObjectTexture.SetScale(const Value: T2DVector);
begin
if not AreEqual2D(Value,FScale) then
	begin
	FScale:=Value;
	Change;
	end;
end;



procedure TGLObjectTexture.SetOffset(const Value: T2DVector);
begin
if not AreEqual2D(Value,FOffset) then
	begin
	FOffset:=Value;
	Change;
	end;
end;



procedure TGLObjectTexture.SetYaw(const Value: double);
begin
if Yaw<>Value then
	begin
	FYaw:=Value;
	Change;
	end;
end;



procedure TGLObjectTexture.SetPitch(const Value: double);
begin
if Pitch<>Value then
	begin
	FPitch:=Value;
	Change;
	end;
end;



procedure TGLObjectTexture.SetRoll(const Value: double);
begin
if Roll<>Value then
	begin
	FRoll:=Value;
	Change;
	end;
end;



procedure TGLObjectTexture.SetRadius(const Value: double);
begin
if FRadius<>Value then
	begin
	FRadius := Value;
	Change;
	end;
end;



procedure TGLObjectTexture.SetSPlane(const Value: TGLArrayf4);
begin
if not AreEqualf4(FSPlane,Value) then
	begin
	FSPlane:=Value;
	Change;
	end;;
end;


procedure TGLObjectTexture.SetTPlane(const Value: TGLArrayf4);
begin
if not AreEqualf4(FTPlane,Value) then
	begin
	FTPlane := Value;
	Change;
	end;
end;


procedure TGLObjectTexture.SetHighlight(const Value: boolean);
begin
if FHighlight<>Value then
	begin
	FHighlight := Value;
	Change;
	end;
end;



procedure TGLObjectTexture.Assign(Source: TPersistent);
var ST: TGLObjectTexture;
begin
inherited Assign(Source);
if Source is TGLObjectTexture then
	begin
	ST:=TGLObjectTexture(Source);
	FTextureBuilt:=False;
	TextureMode:=ST.TextureMode;
	AutoTextureMap:=ST.AutoTextureMap;
	ManualTextureMap:=ST.ManualTextureMap;
	Scale:=ST. Scale; Offset:=ST.Offset;
	Yaw:=ST.Yaw; 	Pitch:=ST.Pitch; Roll:=ST.Roll;
	SPlane:=ST.SPlane; TPlane:=ST.TPlane;
        UVCoordinates.Assign(TGLObjectTexture(Source).UVCoordinates);
	end;
end;


end.
