unit GLXFileWriter;

interface

uses Windows,Classes,SysUtils,Vector,Graphics,HugeArry,DSupport,OpenGL15,
     GLViewer,GLCustomMesh,GLMesh,GLFastMesh,GLFrame,GLXFileTextReader,GLXFileGlobal,GLColorArray,Hash,GLLibrary, BitmapLib;

procedure WriteXFile(FileName: string; Viewer: TGLViewer; ProgressProc: TProgressProcedure);
procedure WriteXStream(Stream: TStream; Viewer: TGLViewer; ProgressProc: TProgressProcedure);


implementation

uses StreamLib;

var IndentLevel: integer;


{------------------ Progress handling routines -------------------------------}


var StreamProgress: TStreamProgress;




{----------------- Low level support routines ---------------------------------}


procedure OutputXFileStr(S: string; StringList: TStringList);
{Copy S into XText addin appropriate indentation}
var I: integer;
begin
for I:=1 to IndentLevel do S:=' '+S;
StringList.Add(S);
end;




procedure OutputBeginBlock(Tag,Name: string; StringList: TStringList);
var S: string;
begin
OutputXFileStr('',StringList);
S:=Tag+' ';
if Name<>'' then S:=S+Name+' ';
OutputXFileStr(S+'{',StringList);
Inc(IndentLevel);
end;


procedure OutputEndBlock(StringList: TStringList);
begin
Dec(IndentLevel);
OutputXFileStr('}',StringList);
end;


procedure OutputInteger(I: integer; Term: char; StringList: TStringList);
begin
OutputXFileStr(IntToStr(I)+Term,StringList);
end;


procedure OutputFloat(F: double; Term: char; StringList: TStringList);
begin
OutputXFileStr(FloatToStrF(F,ffFixed,18,6)+Term,StringList);
end;



{--------------- Write Mesh Object -------------------------------------------}



procedure OutputVertexList(VertList: TVertexList; StringList: TStringList; Progress: TProgressProcedure);
{Display whole vertex list}
var I,J,Inx,PntCnt: integer;
var S: string;
begin
OutputXFileStr(IntToStr(VertList.Vertices.Count)+';',StringList);
for I:=0 to VertList.Vertices.Count-1 do
	begin
	S:=Vector3DToString(VertList.Vertices[I],6,';')+';';
	if I=(VertList.Vertices.Count-1) then S:=S+';' else S:=S+',';
	OutputXFileStr(S,StringList);
	if Assigned(Progress) then Progress(MulDiv(50,I,VertList.Vertices.Count));
	end;
OutputXFileStr(IntToStr(VertList.FaceCount)+';',StringList);

Inx:=0;
for I:=0 to VertList.FaceCount-1 do
	begin
	PntCnt:=VertList.Indices[Inx];
	Inc(Inx);
	S:=IntToStr(PntCnt)+';';
	for J:=0 to PntCnt-1 do
		begin
		S:=S+IntToStr(VertList.Indices[Inx]);
		Inc(Inx);
		if J=(PntCnt-1) then S:=S+';' else S:=S+',';
		end;
	if I=(VertList.FaceCount-1) then S:=S+';' else S:=S+',';
	OutputXFileStr(S,StringList);
        if Assigned(Progress) then Progress(50+MulDiv(50,I,VertList.FaceCount));
	end;
end;



procedure OutputVertices(Mesh: TGLCustomMesh; StringList: TStringList; Progress: TProgressProcedure);
{Read generic vertex list - i.e. face vertices or normal vertices}
var I,Inx: integer;
var VertexList: TVertexList;
var V: T3DVector;
begin
VertexList:=TVertexList.Create;
try
VertexList.Clear;
for I:=0 to Mesh.VertexCount-1 do
	begin
	V:=Mesh.Vertex[I];
	V.Z:=-V.Z;
	VertexList.Vertices.Add(V);
        if Assigned(Progress) then Progress(MulDiv(50,I,Mesh.VertexCount));
	end;
VertexList.FaceCount:=Mesh.VertexCount div 3;
for I:=0 to VertexList.FaceCount-1 do
	begin
	Inx:=I * 3;
	VertexList.Indices.Add(3);
	VertexList.Indices.Add(Inx+2);
	VertexList.Indices.Add(Inx+1);
	VertexList.Indices.Add(Inx);
        if Assigned(Progress) then Progress(50+MulDiv(50,I,VertexList.FaceCount));
	end;
OutputVertexList(VertexList,StringList,Progress);
finally VertexList.Free; end;
end;



{----------------- Handle Material Lists ---------------------------------------}


procedure BuildColorArrays(Mesh: TGLCustomMesh; Colors: TIntegerHashTable; Indices: THugeInteger);
var I,Inx: integer;
var C: TColor;
begin
Colors.Clear;
Indices.Clear;
for I:=0 to Mesh.FaceCount-1 do
	begin
	C:=Mesh.WholeFaceColor[I];
	Inx:=Colors.Search(C);
	if Inx<0 then Inx:=Colors.Insert(C);
	Indices.Add(Inx);
	end;
end;



procedure OutputRGBA(RGBA: TGLArrayf4; StringList: TStringList);
var S: string;
begin
S:=FloatToStrF(RGBA[0],ffFixed,18,6);
S:=S+';'+FloatToStrF(RGBA[1],ffFixed,18,6);
S:=S+';'+FloatToStrF(RGBA[2],ffFixed,18,6);
S:=S+';'+FloatToStrF(RGBA[3],ffFixed,18,6)+';;';
OutputXFileStr(S,StringList);
end;



procedure OutputRGB(RGBA: TGLArrayf3; StringList: TStringList);
var S: string;
begin
S:=FloatToStrF(RGBA[0],ffFixed,18,6);
S:=S+';'+FloatToStrF(RGBA[1],ffFixed,18,6);
S:=S+';'+FloatToStrF(RGBA[2],ffFixed,18,6)+';;';
OutputXFileStr(S,StringList);
end;



procedure OutputMaterial(FaceColor: TColor; Alpha,Shininess: double; Emission,Specular: TGLArrayf3; StringList: TStringList);
var GCL: TGLArrayf4;
begin
OutputBeginBlock('Material','',StringList);
GCL:=ColorToGLRGBA(FaceColor,Alpha);
OutputRGBA(GCL,StringList);
OutputFloat(Shininess,';',StringList);
OutputRGB(Specular,StringList);
OutputRGB(Emission,StringList);
OutputEndBlock(StringList);
end;



procedure OutputMaterials(Mesh: TGLCustomMesh; StringList: TStringList);
var Colors: TIntegerHashTable;
var Indices: THugeInteger;
var I: integer;
begin
Colors:=TIntegerHashTable.Create(5);
try
Indices:=THugeInteger.Create(5);
try
OutputBeginBlock('MeshMaterialList','',StringList);
BuildColorArrays(Mesh,Colors,Indices);
OutputInteger(Colors.Count,';',StringList);
OutputInteger(Indices.Count,';',StringList);
for I:=0 to Indices.Count-1 do OutputInteger(Indices[I],';',StringList);

for I:=0 to Colors.Count-1 do
	begin
	OutputMaterial(Colors[I],Mesh.Alpha,Mesh.LightShininess,
		      GLArray4fTo3f(Mesh.LightEmission),GLArray4fTo3f(Mesh.LightSpecular),StringList);
	end;

OutputEndBlock(StringList);
finally  Indices.Free; end;
finally Colors.Free; end;
end;



procedure OutputNormals(Mesh: TGLCustomMesh; StringList: TStringList; Progress: TProgressProcedure);
{Read generic vertex list - i.e. face vertices or normal vertices}
var I,Inx: integer;
var VertexList: TVertexList;
begin
VertexList:=TVertexList.Create;
try
VertexList.Clear;
OutputBeginBlock('MeshNormals','',StringList);
for I:=0 to Mesh.VertexCount-1 do
	begin
	VertexList.Vertices.Add(Mesh.VertexNormal[I]);
	end;

VertexList.FaceCount:=Mesh.VertexCount div 3;
for I:=0 to VertexList.FaceCount-1 do
	begin
	Inx:=I * 3;
	VertexList.Indices.Add(3);
	VertexList.Indices.Add(Inx);
	VertexList.Indices.Add(Inx+1);
	VertexList.Indices.Add(Inx+2);
	end;
OutputVertexList(VertexList,StringList,Progress);
OutputEndBlock(StringList);
finally VertexList.Free; end;
end;



function ColorToRGBAStr(C: TColor; Alpha: double): string;
var R,G,B: byte;
begin
ColorToRGBBytes(C,R,G,B);
Result:=FloatToStrF(R/255,ffFixed,18,6);
Result:=Result+';'+FloatToStrF(G/255,ffFixed,18,6);
Result:=Result+';'+FloatToStrF(B/255,ffFixed,18,6);
Result:=Result+';'+FloatToStrF(Alpha,ffFixed,18,6)+';;';
end;



procedure OutputVertexColors(Mesh: TGLCustomMesh; StringList: TStringList; Progress: TProgressProcedure);
{Write vertex color list for each vertex}
var I: integer;
var GLC: TGLFaceColor;
var S: string;
begin
OutputBeginBlock('MeshVertexColors','',StringList);
OutputInteger(Mesh.VertexCount,';',StringList);
for I:=0 to Mesh.FaceCount-1 do
	begin
	GLC:=Mesh.FaceColors[I];
	S:=IntToStr(I+0)+'; '+ColorToRGBAStr(GLC.C1,1);
	OutputXFileStr(S,StringList);
	S:=IntToStr(I+1)+'; '+ColorToRGBAStr(GLC.C2,1);
	OutputXFileStr(S,StringList);
	S:=IntToStr(I+2)+'; '+ColorToRGBAStr(GLC.C3,1);
	OutputXFileStr(S,StringList);
        if Assigned(Progress) then Progress(MulDiv(100,I,Mesh.FaceCount));
	end;
OutputEndBlock(StringList);
end;




procedure WriteMesh(Mesh: TGLCustomMesh; StringList: TStringList; Progress: TProgressProcedure);
{Read and display Mesh objects}
begin
OutputBeginBlock('Mesh',Mesh.Name,StringList);
OutputVertices(Mesh,StringList,Progress);
OutputMaterials(Mesh,StringList);
OutputNormals(Mesh,StringList,Progress);
OutputVertexColors(Mesh,StringList,Progress);
OutputEndBlock(StringList);
end;



{--------------- Write Header Object -------------------------------------------}





procedure WriteHeader(StringList: TStringList);
{Write X file header object}
begin
OutputBeginBlock('Header','',StringList);
OutputInteger(1,';',StringList);
OutputInteger(0,';',StringList);
OutputInteger(0,';',StringList);
OutputEndBlock(StringList);
end;




{------------------------------------------------------------------------------}


procedure AddTextHeader(Stream: TStream; StringList: TStringList);
{Create and a text header}
begin
StringList.Add('xof 0302txt 0064');
end;



procedure WriteXStream(Stream: TStream; Viewer: TGLViewer; ProgressProc: TProgressProcedure);
var StringList: TStringList;

	procedure WriteFrame(Frame: TGLFrame);
	var F,M: integer;
	begin
	for M:=0 to Frame.MeshList.Count-1 do
		WriteMesh(Frame.MeshList[M],Stringlist,ProgressProc);
	for F:=0 to Frame.FrameList.Count-1 do
		WriteFrame(Frame.FrameList[F]);
	end;

begin
StringList:=TStringList.Create;
try
IndentLevel:=0;
StringList.Clear;
Stream.Seek(0,soFromBeginning);
AddTextHeader(Stream,StringList);
WriteHeader(StringList);
WriteFrame(Viewer.BaseFrame);
StringList.SaveToStream(Stream);
finally StringList.Free; end;
end;



procedure WriteXFile(FileName: string; Viewer: TGLViewer; ProgressProc: TProgressProcedure);
{Read X  file }
var Stream: TMemoryStream;
begin
Stream:= TMemoryStream.Create;
try
WriteXStream(Stream,Viewer,ProgressProc);
Stream.SaveToFile(Filename);
finally Stream.Free; end;
end;



initialization
StreamProgress:=TStreamProgress.Create;
finalization
StreamProgress.Free;
end.
