unit GLLibrary;

interface

uses Windows,Classes,SysUtils,Controls,Graphics, OpenGL15, Vector, HugeArry,ImageLib;

{Define line - basic unit of line array}

type TTLine = record
 P1,P2: T3DVector;
 end;


{Define basic box, orthogonal to X,Y and Z axes}

type TBox = record
 Min,Max: T3DVector;
 end;


type TFace = packed record
 P1,P2,P3: integer;
 end;


type TGLFaceColor = record
 C1,C2,C3: TColor;
 end;


type TGLColor = packed record
 R,G,B,A: byte;
 end;

type PGLColor = ^TGLColor;


const GLWhite: TGLColor = (R: 255; G: 255; B: 255; A: 255);


type TRGBA = record
  R,G,B,A: single;
  end;

type TVertexChangeEvent = procedure(Value: T3DVector; Index: integer) of object;




type THuge3DVectorEx = class(THuge3DVector)
  private
    FOnChange: TVertexChangeEvent;
    FOnAdd: TVertexChangeEvent;
    FOnDelete: TVertexChangeEvent;
    FOnDeleteTriangle: TVertexChangeEvent;
    procedure PutTriangle(Index: integer; Item: TTriangle);
    procedure PutLine(Index: integer; Item: T3DLine);
    function GetTriangle(Index: integer): TTriangle;
    function GetLine(Index: integer): T3DLine;
    function GetTriIndex: integer;
    function GetLineIndex: integer;
    procedure SetTriIndex(Value: integer);
    procedure SetLineIndex(Value: integer);
    procedure DeleteItem(Index: integer); override;
  protected
    procedure Change(Value: T3DVector; Index: integer);
    procedure Put(Index: integer; Item: T3DVector); override;
  public
    constructor CreateAsTriangles(Number: integer);
    constructor CreateAsLines(Number: Integer; dummy : integer = 0);
    procedure Insert(Value: T3DVector; Index: integer); override;
    function AddPointer(Item: T3DVector; Ptr: integer): Integer; override;
    function Add(Item: T3DVector): integer; reintroduce; virtual;
    function AddTriangle(Item: TTriangle): integer;
    function AddLine(Item: T3DLine): integer;
    function GetTrianglePointer(Index: integer): PTriangle;
    procedure InsertTriangle(Value: TTriangle; Index: integer);
    procedure InsertLine(Value: T3DLine; Index: integer);
    procedure DeleteTriangle(Index: integer);
    procedure DeleteLine(Index: integer);
    property Triangle[Index: integer]: TTriangle read GetTriangle write PutTriangle;
    property TriCount: integer read GetTriIndex write SetTriIndex;
    property Line[Index: integer]: T3DLine read GetLine write PutLine;
    property LineCount: integer read GetLineIndex write SetLIneIndex;
    property OnChange: TVertexChangeEvent read FOnChange write FOnChange;
    property OnDelete: TVertexChangeEvent read FOnDelete write FOnDelete;
    property OnAdd: TVertexChangeEvent read FOnAdd write FOnAdd;
    property OnDeleteTriangle: TVertexChangeEvent read FOnDeleteTriangle write FOnDeleteTriangle;
  end;

{ Define clip plane distances. Note, all these value are
relative to the eye or camera. When the camera is positioned,
the actual, real world coordinates of the frustum will be
defined. }


const PixelsPerUnit= 100;	{Scaling factor for screen}
const WorkDistance = 10;	{Distance at which above scaling applies}
const FrustNear = 1;		{Position of the near frustrum}
const FrustFar = 100000;	{Position of the far frustrum}



type TGLWindow = class(TObject)
  private
    WinControl: TWinControl;
    FRealWidth,FRealHeight: double;
  protected
    procedure BuildFormatDescriptor(var PFD: TPixelFormatDescriptor);
    procedure SetupPixelFormat(DC: HDC);
    procedure SetupGLWindow(Handle: HWND);
  public
    GLWindowDC: HDC;
    OpenGLRC: HGLRC;
    constructor Create(Control: TWinControl; Width,Height: integer);
    destructor Destroy; override;
    procedure SetViewPort(Width,Height: integer);
    property RealWidth: double read FRealWidth;
    property RealHeight: double read FRealHeight;
  end;

{---------------- Object to hold bitmpa images  ------------------------------}


type TRGBQuadArray = array [0..65535] of TRGBQuad;
type PRGBQuadArray = ^TRGBQuadArray;

{This object can load BMP style bitmaps and return pointers that OGL can use}

type TGLBaseBitmap = class(TPersistent)
 private
  FTextName: GLuint;
  FByteArray: THugeByte;
  FWidth,FHeight: integer;
  FEnabled: boolean;
  FOnChange: TNotifyEvent;
  function GetTexturePointer: pointer;
 protected
  procedure Change; virtual;
 public
  Name: string;
  constructor Create; virtual;
  destructor Destroy; override;
  procedure Assign(Source: TPersistent); reintroduce;
  procedure LoadBitmap(BM: TBitmap);
  procedure LoadFromStream(Stream: TStream);
  procedure LoadFromFile(Filename: string);
  procedure LoadFromResource(Name: string);
  procedure Clear;
  property Width: integer read FWidth write FWidth;
  property Height: integer read FHeight write FHeight;
  property GLBitmap: Pointer read GetTexturePointer;
  property Enabled: boolean read FEnabled write FEnabled;
  property OnChange: TNotifyEvent read FOnChange write FOnChange;
  property Textname: GLuint read FTextName;
  property ByteArray: THugeByte read FByteArray;
 end;


{Viewer related types and variables}

type TRenderType = (rtMesh,rtString3D,rtLines,rtFrame,rtFastMesh,rtString2D);


procedure GetRGB(Color: TColor; var R,G,B: single);
procedure GetRGBA(Color: TColor; var R,G,B,A: single);
function MakeGLArrayf3(X,Y,Z: double): TGLArrayf3;
function MakeGLArrayf4(X,Y,Z,W: double): TGLArrayf4;
function GLArrayToVector3D(A: TGLArrayf4): T3DVector; overload;
function GLArrayToVector3D(A: TGLArrayf3): T3DVector; overload;
function GLArrayToVector3D(A: TGLArrayd3): T3DVector; overload;
function GLArrayf4ToStr(GA: TGLArrayf4): string;

function Vector3DToGLArray(V: T3DVector): TGLArrayf3;
function Vector3DToGLArrayf4(V: T3DVector): TGLArrayf4; overload;
function Vector3DToGLArrayf4(V: T3DVector; W: double): TGLArrayf4; overload;
function AreEqualf4(V1,V2: TGLArrayf4): boolean;
function GLArray4fTo3f(A: TGLArrayf4): TGLArrayf3;

function MakeBox(Min,Max: T3DVector): TBox;
function InflateBox(Box: TBox; D: double): TBox;

function MakeGLColor(R,G,B: byte): TGLColor; overload;
function MakeGLColor(R,G,B,A: byte): TGLColor; overload;

function ColorToGLColor(C: TColor): TGLColor;
function GLColorToColor(C: TGLColor): TColor;

function ColorToGLRGBA(Color: TColor; Alpha: double): TGLArrayf4; overload;
function ColorToGLRGBA(Color: TColor): TGLArrayf4; overload;
function ColorToGLRGB(Color: TColor): TGLArrayf3;
function ColorToGLArrayf4(Color: TColor): TGLArrayf4;

function GLRGBToColor(GLRGB: TGLArrayf4): TColor;
function GetGLMouseRay(X, Y: Integer): T3DLine;
function ProjectGLScreenRay(X, Y: Integer;  Pitch,Yaw,Roll: double): T3DLine;
function ScreenToWorld(V: T3DVector): T3DVector;
function WorldToScreen(V: T3DVector): T3DVector;

procedure SetupTextureParameters(var GT: TGLBaseBitmap);
function IsGLExtensionSupported(Extension: string): boolean;

implementation

{----------------- Support Routines ---------------------------------}


procedure SetupTextureParameters(var GT: TGLBaseBitmap);
begin
glBindTexture(GL_TEXTURE_2D, GT.TextName);
glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);
gluBuild2DMipmaps(GL_TEXTURE_2D, 3, GT.Width, GT.Height, GL_RGBA, GL_UNSIGNED_BYTE,GT.GLBitmap);
end;



function MakeGLColor(R,G,B,A: byte): TGLColor;
begin
Result.R:=R; Result.G:=G; Result.B:=B; Result.A:=A;
end;


function MakeGLColor(R,G,B: byte): TGLColor;
begin
Result:=MakeGLColor(R,G,B,$FF);
end;



function ColorToGLColor(C: TColor): TGLColor;
begin
Result:=TGLColor(C);
Result.A:=((C shr 24) and $FF);
end;


function GLColorToColor(C: TGLColor): TColor;
begin
Result:=RGB(C.R,C.G,C.B);
end;

procedure GetRGB(Color: TColor; var R,G,B: single);
{Convert Delphi color to separate RBG values in 0 to 1 range}
begin
R:=(Color and $FF)/255;
G:=((Color shr 8) and $FF)/255;
B:=((Color shr 16) and $FF)/255;
end;





procedure GetRGBA(Color: TColor; var R,G,B,A: single);
{Convert Delphi color to separate RBG and Alpha values in 0 to 1 range}
begin
R:=(Color and $FF)/255;
G:=((Color shr 8) and $FF)/255;
B:=((Color shr 16) and $FF)/255;
A:=((Color shr 24) and $FF)/255;
end;



function ColorToGLArrayf4(Color: TColor): TGLArrayf4;
begin
GetRGBA(Color,Result[0],Result[1],Result[2],Result[3]);
end;



function ColorToGLRGB(Color: TColor): TGLArrayf3;
begin
GetRGB(Color,Result[0],Result[1],Result[2]);
end;





function ColorToGLRGBA(Color: TColor; Alpha: double): TGLArrayf4;
begin
GetRGB(Color,Result[0],Result[1],Result[2]);
Result[3]:=Alpha;
end;



function ColorToGLRGBA(Color: TColor): TGLArrayf4;
begin
Result:=ColorToGLRGBA(Color,1);
end;


function GLRGBToColor(GLRGB: TGLArrayf4): TColor;
begin
Result:=RGB(trunc(GLRGB[0] * 255), trunc(GLRGB[1] * 255),trunc(GLRGB[2] * 255));
end;



function GLArray4fTo3f(A: TGLArrayf4): TGLArrayf3;
begin
Result[0]:=A[0];
Result[1]:=A[1];
Result[2]:=A[2];
end;


function MakeGLArrayf3(X,Y,Z: double): TGLArrayf3;
{Convert X,Y,Z values to OpenGL array}
begin
Result[0]:=X; Result[1]:=Y; Result[2]:=Z;
end;


function MakeGLArrayf4(X,Y,Z,W: double): TGLArrayf4;
{Convert X,Y,Z,W values to OpenGL array}
begin
Result[0]:=X; Result[1]:=Y; Result[2]:=Z; Result[3]:=W;
end;


function AreEqualf4(V1,V2: TGLArrayf4): boolean;
begin
Result:=(V1[0]=V2[0]) and (V1[1]=V2[1]) and (V1[2]=V2[2]) and (V1[3]=V2[3]);
end;


function GLArrayToVector3D(A: TGLArrayf3): T3DVector;
{Convert 3D OpenGL array to 3D vector}
begin
Result.X:=A[0]; Result.Y:=A[1]; Result.Z:=A[2];
end;


function GLArrayToVector3D(A: TGLArrayf4): T3DVector;
{Convert 3D OpenGL array to 3D vector}
begin
Result.X:=A[0]; Result.Y:=A[1]; Result.Z:=A[2];
end;


function GLArrayToVector3D(A: TGLArrayd3): T3DVector;
{Convert 3D OpenGL array to 3D vector}
begin
Result.X:=A[0]; Result.Y:=A[1]; Result.Z:=A[2];
end;



function GLArrayf4ToStr(GA: TGLArrayf4): string;
begin
Result:=FloatToStrF(GA[0],ffGeneral,18,4)+','+
	FloatToStrF(GA[1],ffGeneral,18,4)+','+
	FloatToStrF(GA[2],ffGeneral,18,4)+','+
	FloatToStrF(GA[3],ffGeneral,18,4);
end;



function Vector3DToGLArray(V: T3DVector): TGLArrayf3;
{Convert 3D vector to 3D OpenGL array}
begin
Result[0]:=V.X; Result[1]:=V.Y; Result[2]:=V.Z;
end;



function Vector3DToGLArrayf4(V: T3DVector): TGLArrayf4;
begin
Result[0]:=V.X; Result[1]:=V.Y; Result[2]:=V.Z; Result[3]:=0;
end;



function Vector3DToGLArrayf4(V: T3DVector; W: double): TGLArrayf4;
begin
Result[0]:=V.X; Result[1]:=V.Y; Result[2]:=V.Z; Result[3]:=W;
end;


function MakeBox(Min,Max: T3DVector): TBox;
begin
Result.Min:=Min; Result.Max:=Max;
end;


function InflateBox(Box: TBox; D: double): TBox;
begin
Result.Min:=ScalarSubtract3D(Box.Min,D);
Result.Max:=ScalarAdd3D(Box.Max,D);
end;



function GetGLMouseRay(X, Y: Integer): T3DLine;
{Bug: should use the camera position if camera is beyond clipping plane}
var viewport: TVector4i;
var modelview: TGLMatrixd4;
var projection: TGLMatrixd4;
begin
glGetDoublev(GL_MODELVIEW_MATRIX, @modelview);
glGetDoublev(GL_PROJECTION_MATRIX, @projection);
glGetIntegerv(GL_VIEWPORT, @viewport);
if Y = 0 then Y:=1;	{problem reported with Y=0}
{Get near plane position of X,Y position}
gluUnProject(X, viewport[3]-Y, 0, modelview, projection, viewport, @Result.P1.X, @Result.P1.Y, @Result.P1.Z);
{Get far plane position form a ray between the two points}
gluUnProject(X, viewport[3]-Y, 1, modelview, projection, viewport, @Result.P2.X, @Result.P2.Y, @Result.P2.Z);
end;



function GetGLMouseRayX(X, Y: Integer; Pitch,Yaw,Roll: double): T3DLine;
{Bug: should use the camera position if camera is beyond clipping plane}
var viewport: TVector4i;
var modelview: TGLMatrixd4;
var projection: TGLMatrixd4;
begin
glGetDoublev(GL_MODELVIEW_MATRIX, @modelview);
glGetDoublev(GL_PROJECTION_MATRIX, @projection);
glGetIntegerv(GL_VIEWPORT, @viewport);
if Y = 0 then Y:=1;	{problem reported with Y=0}

{Get near plane position of X,Y position}
gluUnProject(X, viewport[3]-Y, 0, modelview, projection, viewport, @Result.P1.X, @Result.P1.Y, @Result.P1.Z);
{Get far plane position form a ray between the two points}
gluUnProject(X, viewport[3]-Y, 1, modelview, projection, viewport, @Result.P2.X, @Result.P2.Y, @Result.P2.Z);
RotatePoint3D(Result.P2,Result.P1,Pitch,Yaw,Roll);
end;



function ProjectGLScreenRay(X, Y: Integer; Pitch,Yaw,Roll: double): T3DLine;
{Project ray starting at X,Y at angle specified by Pitch,Yaw and Roll}
{Bug: should use the camera position if camera is beyond clipping plane}
var viewport: TVector4i;
var modelview: TGLMatrixd4;
var projection: TGLMatrixd4;
var P1,P2: T3DVector;
begin
glGetDoublev(GL_MODELVIEW_MATRIX, @modelview);
glGetDoublev(GL_PROJECTION_MATRIX, @projection);
glGetIntegerv(GL_VIEWPORT, @viewport);
if Y = 0 then Y:=1;	{problem reported with Y=0}

{Setup projection geometry}
P1:=MakeVector3D(X, viewport[3]-Y, 0);
P2:=MakeVector3D(X, viewport[3]-Y, viewport[2]);
RotatePoint3D(P2,P1,Pitch,Yaw,Roll);
P1.Z:=0; P2.Z:=1;

{Get near plane position of X,Y position}
gluUnProject(P1.X,P1.Y,P1.Z, modelview, projection, viewport, @Result.P1.X, @Result.P1.Y, @Result.P1.Z);
{Get far plane position form a ray between the two points}
gluUnProject(P2.X,P2.Y,P2.Z, modelview, projection, viewport, @Result.P2.X, @Result.P2.Y, @Result.P2.Z);
end;


function ScreenToWorld(V: T3DVector): T3DVector;
var viewport: TVector4i;
var modelview: TGLMatrixd4;
var projection: TGLMatrixd4;
begin
{Note: loading these matrices are slow - speed this by only loading them once }
glGetDoublev(GL_MODELVIEW_MATRIX, @modelview);
glGetDoublev(GL_PROJECTION_MATRIX, @projection);
glGetIntegerv(GL_VIEWPORT, @viewport);
gluUnProject(V.X, Viewport[3] - V.Y, V.z, modelview, projection, viewport, @Result.X, @Result.Y, @Result.Z);
end;




function WorldToScreen(V: T3DVector): T3DVector;
var ViewPort: TVector4i;
var ModelView: TGLMatrixd4;
var Projection: TGLMatrixd4;
begin
{Note: loading these matrices are slow - speed this by only loading them once }
glGetDoublev(GL_MODELVIEW_MATRIX, @ModelView);
glGetDoublev(GL_PROJECTION_MATRIX, @Projection);
glGetIntegerv(GL_VIEWPORT, @ViewPort);
gluProject(V.X, V.Y, V.Z, ModelView, Projection, ViewPort, @Result.X, @Result.Y, @Result.Z);
Result.Y:=ViewPort[3]-Result.Y;
end;



{--------------- Extended Vector Array Object --------------------------------}


procedure THuge3DVectorEx.Change(Value: T3DVector; Index: integer);
begin
if Assigned(FOnChange) then FOnChange(Value,Index);
end;

{Override these procedures so they generate events}


procedure THuge3DVectorEx.Put(Index: integer; Item: T3DVector);
begin
inherited Put(Index,Item);
Change(Item,Index);
end;


function THuge3DVectorEx.Add(Item: T3DVector): integer;
begin
Result:=inherited Add(Item);
Change(Item,Result);
if Assigned(FOnAdd) then FOnAdd(Item,Result);
end;



function THuge3DVectorEx.AddPointer(Item: T3DVector; Ptr: integer): Integer;
begin
Result:=inherited AddPointer(Item,Ptr);
Change(Item,Result);
end;


procedure THuge3DVectorEx.Insert(Value: T3DVector; Index: integer);
begin
inherited Insert(Value,Index);
Change(Value,Index);
end;



constructor THuge3DVectorEx.CreateAsTriangles(Number: Integer);
{Create as huge Triangle array}
begin
inherited Create(Number*3);
end;


function THuge3DVectorEx.AddTriangle(Item: TTriangle): integer;
{Add item to Triangle array}
begin
Result:=Add(Item.P1) div 3;
Add(Item.P2);
Add(Item.P3);
end;


procedure THuge3DVectorEx.PutTriangle(Index: integer; Item: TTriangle);
{Insert Triangle at index}
begin
Index:=Index * 3;
Value[Index]:=Item.P1;
Value[Index+1]:=Item.P2;
Value[Index+2]:=Item.P3;
end;


function THuge3DVectorEx.GetTriangle(Index: integer): TTriangle;
{Get Triangle at the index}
begin
Index:=Index * 3;
Result:=TTriangle(GetItemPtr(Index)^);
end;


function THuge3DVectorEx.GetTrianglePointer(Index: integer): PTriangle;
{Get Triangle at the index}
begin
Index:=Index * 3;
Result:=GetItemPtr(Index);
end;


procedure THuge3DVectorEx.InsertTriangle(Value: TTriangle; Index: integer);
{Insert Triangle at the index}
begin
Index:=Index * 3;
InsertItem(Value.P1,Index);
InsertItem(Value.P2,Index+1);
InsertItem(Value.P3,Index+2);
end;


procedure THuge3DVectorEx.DeleteItem(Index: integer);
var V: T3DVector;
begin
V:=Self[Index];
inherited DeleteItem(Index);
Change(Self[Index],Index);
if Assigned(FOnDelete) then FOnDelete(V,Index);
end;


procedure THuge3DVectorEx.DeleteTriangle(Index: integer);
{Delete Triangle at the index}
var V: T3DVector;
var VertInx: integer;
begin
VertInx:=Index * 3;
V:=Self[VertInx];
DeleteItem(VertInx+2);
DeleteItem(VertInx+1);
DeleteItem(VertInx);
if Assigned(FOnDeleteTriangle) then FOnDeleteTriangle(V,Index);
end;


function THuge3DVectorEx.GetTriIndex: integer;
begin
Result:=ItemIndex div 3;
end;



procedure THuge3DVectorEx.SetTriIndex(Value: integer);
begin
ItemIndex:=Value * 3;
end;


{--------------- Line Aspect of Extened Array --------------------------------}


constructor THuge3DVectorEx.CreateAsLines(Number: Integer; dummy : integer = 0);
{Create as huge Lines array}
{ "Dummy" deals with a C++  compatibility issue with two constructors}
begin
inherited Create(Number*2);
end;


function THuge3DVectorEx.AddLine(Item: T3DLine): integer;
{Add item to Line array}
begin
Result:=Add(Item.P1) div 2;
Add(Item.P2);
end;


procedure THuge3DVectorEx.PutLine(Index: integer; Item: T3DLine);
{Insert Line at index}
begin
Index:=Index * 2;
Value[Index]:=Item.P1;
Value[Index+1]:=Item.P2;
end;


function THuge3DVectorEx.GetLine(Index: integer): T3DLine;
{Get Line at the index}
begin
Index:=Index * 2;
Result:=T3DLine(GetItemPtr(Index)^);
end;


procedure THuge3DVectorEx.InsertLine(Value: T3DLine; Index: integer);
{Insert Triangle at the index}
begin
Index:=Index * 2;
InsertItem(Value.P1,Index);
InsertItem(Value.P2,Index+1);
end;


procedure THuge3DVectorEx.DeleteLine(Index: integer);
{Delete Line at the index}
begin
Index:=Index * 2;
DeleteItem(Index);
DeleteItem(Index+1);
end;


function THuge3DVectorEx.GetLineIndex: integer;
begin
Result:=ItemIndex div 2;
end;



procedure THuge3DVectorEx.SetLineIndex(Value: integer);
begin
ItemIndex:=Value * 3;
end;

{---------------- Routines to setup the OpenGL window ------------------------}


procedure TGLWindow.BuildFormatDescriptor(var PFD: TPixelFormatDescriptor);
{Set the pixel format descriptor for our window }
begin
FillMemory(@PFD,Sizeof(TPixelFormatDescriptor),0);
PFD.nSize:=sizeof(TPixelFormatDescriptor);
PFD.nVersion:=1;
PFD.dwFlags:=PFD_SUPPORT_OPENGL or PFD_DRAW_TO_WINDOW or PFD_DOUBLEBUFFER;
PFD.iPixelType:=PFD_TYPE_RGBA;
PFD.cColorBits:=24;
PFD.cDepthBits:=16;
PFD.iLayerType:=PFD_MAIN_PLANE;
end;



procedure TGLWindow.SetupPixelFormat(DC: HDC);
{Setup pixel format for our window}
var PixelFormat:integer;
var PFD: TPixelFormatDescriptor;
begin
BuildFormatDescriptor(PFD);
PixelFormat := ChoosePixelFormat(DC, @pfd);
if PixelFormat = 0 then
	begin
	MessageBox(WindowFromDC(DC), 'ChoosePixelFormat failed.', 'Error', MB_ICONERROR or MB_OK);
	exit;
	end;
if not SetPixelFormat(DC, pixelFormat, @pfd) then
	begin
	MessageBox(WindowFromDC(DC), 'SetPixelFormat failed.', 'Error', MB_ICONERROR or MB_OK);
	exit;
	end;
end;



procedure TGLWindow.SetupGLWindow(Handle: HWND);
{Create the GL handle for our window}
begin
GLWindowDC:=GetDC(Handle);
SetupPixelFormat(GLWindowDC);
OpenGLRC:=wglCreateContext(GLWindowDC);
wglMakeCurrent(GLWindowDC, OpenGLRC);
end;



procedure TGLWindow.SetViewPort(Width,Height: integer);
{Set the view port parameter for our window}
var FrustWidth,FrustHeight: double;
begin
FrustWidth:= Width / PixelsPerUnit / WorkDistance;
FrustHeight:= Height / PixelsPerUnit / WorkDistance;
FRealWidth:=FrustWidth * WorkDistance;
FRealHeight:=FrustHeight * WorkDistance;
glViewport(0, 0, Width, Height);
glMatrixMode(GL_PROJECTION);
glLoadIdentity;
glFrustum(-(FrustWidth/2), FrustWidth/2, -(FrustHeight/2), FrustHeight/2, FrustNear,FrustFar);
glMatrixMode(GL_MODELVIEW);
end;



constructor TGLWindow.Create(Control: TWinControl; Width,Height: integer);
begin
inherited Create;
WinControl:=Control;
SetupGLWindow(Control.Handle);
SetViewPort(Width,Height);
end;


destructor TGLWindow.Destroy;
begin
wglMakeCurrent(0, 0);
wglDeleteContext(OpenGLRC);
ReleaseDC(WinControl.Handle, GLWindowDC);
inherited Destroy;
end;
{============== GLTexture Object ==============================================}



constructor TGLBaseBitmap.Create;
begin
inherited;
FByteArray:=THugeByte.Create(5);
{Warning: Textures must be generated only after the GL Window has}
{been created. Other wise "glGenTextures' names will be wrong}
glGenTextures(1,@FTextName);
end;



destructor TGLBaseBitmap.Destroy;
begin
glDeleteTextures(1,@FTextName);
FByteArray.Free;
inherited;
end;


function TGLBaseBitmap.GetTexturePointer: Pointer;
begin
Result:=FByteArray.GetItemPtr(0);
end;


procedure TGLBaseBitmap.Assign(Source: TPersistent);
var ST: TGLBaseBitmap;
begin
FByteArray.Clear;
if Source is TGLBaseBitmap then
	begin
	FByteArray.Assign(TGLBaseBitmap(Source).FByteArray);
	ST:=TGLBaseBitmap(Source);
	FWidth:=ST.Width;
	FHeight:=ST.Height;
	Enabled:=ST.Enabled;
	Change;
	end
else inherited Assign(Source);
end;




procedure TGLBaseBitmap.LoadBitmap(BM: TBitmap);
{New method about 10 times faster but maybe not as compatible}
var Scanline: PRGBQuadArray;
var X,Y,Inx: integer;
var PB: PByteArray;
begin
FByteArray.Clear;
BM.PixelFormat:=pf32Bit;
FByteArray.ReAllocMem(BM.Height * BM.Width * 4);
PB:=FByteArray.GetItemPtr(0);
for Y:=0 to BM.Height-1 do
	begin
	Scanline:=BM.Scanline[(BM.Height-1)-Y];
	for X:=0 to BM.Width-1 do
		begin
		Inx:=(X + (Y * BM.Width)) * 4;
		PB[Inx]:=Scanline^[X].rgbRed;
		PB[Inx+1]:=Scanline^[X].rgbGreen;
		PB[Inx+2]:=Scanline^[X].rgbBlue;
		PB[Inx+3]:=255;
		end;
	end;
FByteArray.ItemIndex:=BM.Width * BM.Height * 4;
FWidth:=BM.Width;
FHeight:=BM.Height;
Change;
end;



procedure TGLBaseBitmap.LoadFromStream(Stream: TStream);
var BM: TBitmap;
begin
BM:=TBitmap.Create;
try
BM.LoadFromStream(Stream);
LoadBitmap(BM);
finally BM.Free; end;
end;




procedure TGLBaseBitmap.LoadFromFile(Filename: string);
var BM: TBitmap;
begin
BM:=TBitmap.Create;
try
LoadBitmapByExtension(BM,Filename,False);
LoadBitmap(BM);
finally BM.Free; end;
end;




procedure TGLBaseBitmap.LoadFromResource(Name: string);
var BM: TBitmap;
begin
BM := TBitmap.Create;
try
BM.Handle := Windows.LoadBitmap(hInstance, PChar(Name));
LoadBitmap(BM);
finally BM.Free; end;
end;




procedure TGLBaseBitmap.Change;
begin
if Assigned(FOnChange) then FOnChange(Self);
end;




procedure TGLBaseBitmap.Clear;
begin
FByteArray.Clear;
FWidth:=0;
FHeight:=0;
Enabled:=False;
end;


{------------------------------------------------------------------------------}


function IsGLExtensionSupported(Extension: string): boolean;
{Determine if named extension is supported}
{A OpenGL rendering context must be setup with }
{"SetupPixelFormat," "wglCreateContext" and "wglMakeCurrent" }
var ExtStr: string;
var Inx: integer;
begin
Result:=False;
{Validate Extension}
Inx := Pos(' ',Extension);
{Shouldn't have space or be empty}
if (Inx <> 0) or (Extension = '') then exit;
ExtStr := glGetString(GL_EXTENSIONS);
if Pos(Extension,ExtStr) = 0 then exit;
Result := true;
end;



end.
