unit PCXLib;

interface

uses Windows,Graphics,SysUtils,Classes,BitmapLib,DSupport,StreamLib;

type TPCXInfo = record
 Company: byte;
 Version: byte;
 Encoding: byte;
 BitsPerPixel: byte;
 XMin,YMin,XMax,YMax: Word;
 HDPI,VDPI: Word;
 Palette1: array [0..15] of TRGBTriple;
 Pad1: byte;
 Planes: byte;
 BytesPerLine: word;
 PaletteFormat: word;
 Width,Height: word;
 Pal2Tag: byte;
 Palette2: array [0..255] of TRGBTriple;
 PeakValue: integer;
 end;

function LoadPCXStream(Stream: TStream; BM: TBitmap): TPCXInfo;
function ReadPCXFile(BM: TBitmap; FileName: string): TPCXInfo;
procedure SavePCXToStream(Stream: TStream; Bitmap: TBitmap);
procedure SavePCXToFile(FileName: String; Bitmap: TBitmap);

{Defines a progress event. The user can create a subroutine and
{ assign it to FOnProgress and the subroutine will be called  }
{ periodically during the file loading to indicate progress. }

type TPCXProgress= procedure(Sender: TObject; Stage: TProgressStage; PercentDone: Byte; RedrawNow: Boolean; const R: TRect; const Msg: string);

var OnPCXProgress: TPCXProgress;


implementation


{------------------- Handle Progress Events --------------------------}

var ProgressValue: double;	{Current progress value }
var ProgressInc: double;	{Amount to increment progress }



procedure AdvanceProgress(Stage: TProgressStage; Steps: integer);
{Increment progress value and calls Progress event }
{if a procedure is assigned to FOnProgress, it is
{called periodically during file loading process }
var DrawNow: boolean;
var Msg: string;
begin
DrawNow:=False;
case Stage of
 psStarting:
	begin
	ProgressValue:=0;
	ProgressInc:=100/Steps;
	Msg:='Starting';
	end;
 psRunning:
	begin
	ProgressValue:=ProgressValue+ProgressInc;
	Msg:='Running';
	end;
 psEnding:
	begin
	DrawNow:=True;
	Msg:='Ending';
	end;
 end;

if Assigned(OnPCXProgress) then
    OnPCXProgress(nil,Stage,trunc(ProgressValue),DrawNow,Rect(0,0,0,0),Msg);
end;


{================= Routines to Write PCX files ================================}

type TLineArray = array of byte;
type TPlaneIndex = (piBlue=0,piGreen=1,piRed=2);	{This order is manditory}



procedure StoreByte(var Data: TLineArray; B: byte);
begin
SetLength(Data,Length(Data)+1);
Data[High(Data)]:=B;
end;



procedure StoreBytePair(var Data: TLineArray; B1,B2: byte);
{Store two at time for speed}
begin
SetLength(Data,Length(Data)+2);
Data[High(Data)-1]:=B1;
Data[High(Data)]:=B2;
end;



function EncodeBytePair(Byte, Count: DWord; var Data: TLineArray): integer;
{ Subroutine for writing an encoded byte pair (or single byte if it doesn't encode) }
{ It returns the count of bytes written, 0 if error }
begin
if Count<>0 then
	begin
	if (Count = 1) and ($C0 <> ($C0 and Byte)) then
		begin
		StoreByte(Data,Byte);
		Result:=1;
		end
	else
		begin
		StoreBytePair(Data,$C0 or Count,Byte);
		Result:=2;
		end
	end
else Result:=0;
end;


function EncodeScanline(var Scan: pRGBTripleArray; Plane: TPlaneIndex; Width: integer;var Data: TLineArray): integer;
{Encode one BMP scan line and store encoded data in "Data"}
var First, Last: byte;
var X, I: integer;
var Total: integer;
var runCount: integer;
begin
Total:=0;
RunCount:=1;
Last:=PByteArray(@Scan[0])[Integer(Plane)];

for X:=1 to Width-1 do
	begin
	First:=PByteArray(@Scan[X])[Integer(Plane)];
	if First = Last then
		begin
		{ There is at least two duplicates so encode it }
		Inc(RunCount);
		if RunCount = 63 then
			begin
			I:=EncodeBytePair(Last, RunCount, Data);
			Total:=Total+I;
			RunCount:=0;
			end
		end
	else
		begin
                { There are no (more) duplicates}
		if RunCount<>0 then
			begin
                        I:=EncodeBytePair(Last, RunCount, Data);
			Total:=Total + I;
			end;
		Last:=First;
		RunCount:=1;
		end
	end;
if RunCount<>0 then
	begin
        I:=EncodeBytePair(Last,RunCount,Data);
	Total:=Total + I;
	end;
Result:=Total;
end;



procedure WriteImage(Stream: TStream; BM: TBitmap);
{Write compressed (ecoded) Image to specified stream}
var Y: integer;
var Data: TLineArray;
var Scan: pRGBTripleArray;
begin
for Y:=0 to BM.Height-1 do
	begin
	Scan:=BM.Scanline[Y];
	EncodeScanline(Scan,piRed,BM.Width,Data);
	EncodeScanline(Scan,piGreen,BM.Width,Data);
	EncodeScanline(Scan,piBlue,BM.Width,Data);
	end;
Stream.Write(Data[0],Length(Data));
AdvanceProgress(psRunning,0);
end;



procedure WriteHeader(Stream: TStream; Info: TPCXInfo);
{Write info to Stream as header}
var I: integer;
begin
WriteStreamByte(Stream,Info.Company);
WriteStreamByte(Stream,Info.Version);
WriteStreamByte(Stream,Info.Encoding);
WriteStreamByte(Stream,Info.BitsPerPixel);
WriteStreamWord(Stream,Info.XMin);
WriteStreamWord(Stream,Info.YMin);
WriteStreamWord(Stream,Info.XMax);
WriteStreamWord(Stream,Info.YMax);
WriteStreamWord(Stream,Info.HDPI);
WriteStreamWord(Stream,Info.VDPI);
for I:=0 to 16-1 do
	begin
	WriteStreamByte(Stream,0);
	WriteStreamByte(Stream,0);
	WriteStreamByte(Stream,0);
	end;
WriteStreamByte(Stream,Info.Pad1);
WriteStreamByte(Stream,Info.Planes);
WriteStreamWord(Stream,Info.BytesPerLine);
WriteStreamWord(Stream,Info.PaletteFormat);
WriteStreamWord(Stream,Info.Width);
WriteStreamWord(Stream,Info.Height);
for I:=1 to 54 do WriteStreamByte(Stream,0);
AdvanceProgress(psRunning,0);
end;



procedure BuildHeader(var Info: TPCXInfo; BM: TBitmap);
{Extract Header information from BMP and store in PCX info structure}
var I,XPPM,YPPM:integer;
begin
Info.Company:=10;
Info.Version:=5;
Info.Encoding:=1;
Info.BitsPerPixel:=8;
Info.XMin:=0;
Info.YMin:=0;
Info.XMax:=BM.Width-1;
Info.YMax:=BM.Height-1;
BitmapLib.GetBmpPPM(BM,XPPM,YPPM);

Info.HDPI:=round(XPPM * Inches2Meters);
Info.VDPI:=round(YPPM * Inches2Meters);

for I:=0 to 16-1 do Info.Palette1[I]:=MakeRBGTriple(0,0,0);
Info.Pad1:=0;
Info.Planes:=3;
{Doc says this value should be even
but no one else makes it even}
Info.BytesPerLine:=BM.Width;
Info.PaletteFormat:=1;
Info.Width:=BM.Width;
Info.Height:=BM.Height;
AdvanceProgress(psRunning,0);
end;




procedure SavePCXToStream(Stream: TStream; Bitmap: TBitmap);
{Save BMP to specified Stream as PCX}
var Info: TPCXInfo;
begin
StreamEndian:=LittleEnd;
FillMemory(@Info,SizeOf(TPCXInfo),0);
AdvanceProgress(psStarting,3);
BuildHeader(Info,Bitmap);
WriteHeader(Stream,Info);
WriteImage(Stream,Bitmap);
AdvanceProgress(psEnding,0);
end;




procedure SavePCXToFile(FileName: String; Bitmap: TBitmap);
{Save BMP to specified File as PCX}
var Stream: TMemoryStream;
begin
Stream:=TMemoryStream.Create;
SavePCXToStream(Stream,Bitmap);
Stream.SaveToFile(filename);
Stream.Free;
end;



{================== Routines for Read PCX files ===============================}


procedure ReadHeader(Stream: TStream; var Info: TPCXInfo);
var I: integer;
var R, G, B: integer;
begin
Info.Company:=GetStreamByte(Stream);
Info.Version:=GetStreamByte(Stream);
Info.Encoding:=GetStreamByte(Stream);
Info.BitsPerPixel:=GetStreamByte(Stream);
Info.XMin:=GetStreamWord(Stream);
Info.YMin:=GetStreamWord(Stream);
Info.XMax:=GetStreamWord(Stream);
Info.YMax:=GetStreamWord(Stream);
Info.HDPI:=GetStreamWord(Stream);
Info.VDPI:=GetStreamWord(Stream);
for I:=0 to 16-1 do
	begin
	R:= GetStreamByte(Stream);
	G:= GetStreamByte(Stream);
	B:= GetStreamByte(Stream);
	Info.Palette1[I]:=MakeRBGTriple(R,G,B);
	end;
Info.Pad1:=GetStreamByte(Stream);
Info.Planes:=GetStreamByte(Stream);
Info.BytesPerLine:=GetStreamWord(Stream);
Info.PaletteFormat:=GetStreamWord(Stream);
Info.Width:=GetStreamWord(Stream);
Info.Height:=GetStreamWord(Stream);
for I:=1 to 54 do GetStreamByte(Stream);
AdvanceProgress(psRunning,0);
end;



procedure ReadSecondPalette(Stream: TStream; var Info: TPCXInfo);
var I: integer;
var R, G, B: byte;
begin
Info.Pal2Tag:=GetStreamByte(Stream);
if Info.Pal2Tag<>$0C then exit;
for I:=0 to 256-1 do
	begin
	R:= GetStreamByte(Stream);
	G:= GetStreamByte(Stream);
	B:= GetStreamByte(Stream);
	Info.Palette2[I]:=MakeRBGTriple(R,G,B);
	end;
AdvanceProgress(psRunning,0);
end;


type TBodyData = array of byte;

procedure ReadRawImage(Stream: TStream; Info: TPCXInfo; var RawImage: TBodyData);
var J,I: integer; 	{Index into Image}
var Byte: integer;	{Byte read from PCX file}
var Cnt: integer;	{Number of bytes to repeat}
var LineBytes: integer;
var YSize,TotalSize: integer;
begin
LineBytes:=Info.Planes * Info.BytesPerLine;
YSize:=Info.Ymax - Info.Ymin + 1;
TotalSize:=LineBytes * YSize;
SetLength(RawImage,TotalSize);
J:= 0;
repeat
	begin
	Byte:= GetStreamByte(Stream);			{Get byte from .PCX file}
	if Byte < $C0 then
		begin
		RawImage[J]:= Byte;
		J:= J +1;
		end
	else	begin			{Decompress multiple copies}
		Cnt:= Byte and $3F;
		Byte:= GetStreamByte(Stream);
		for I:= 1 to Cnt do
			begin
			RawImage[J]:= Byte;
			J:= J +1;
			end;
		end;
	end
until J>= TotalSize;
AdvanceProgress(psRunning,0);
end;

{-------------------------------------------------------------------------------}

procedure HandlePlane1Bits1(var RawImage: TBodyData; BM: TBitmap; Info: TPCXInfo);
var X,Y, Width,Height,Inx,Bit: integer;
const Bits: array [0..7] of byte = ($80,$40,$20,$10,$08,$04,$02,$01);
var Scan: pRGBTripleArray;
begin
BM.PixelFormat:=pf24Bit;
Width:=Info.Xmax - Info.Xmin + 1;
Height:=Info.Ymax - Info.Ymin + 1;
BM.Width:=Width;
BM.Height:=Height;
for Y:=0 to Height-1 do
	begin
	Scan:=BM.ScanLine[Y];
	for X:=0 to Width-1 do
		begin
		Bit:=Bits[X Mod 8];
		Inx:=(X div 8)+(Y * Info.BytesPerLine);
		if (RawImage[Inx] and Bit)<>0 then Scan[X]:=Info.Palette1[1]
		else Scan[X]:=Info.Palette1[0];
		end;
	end;
end;



procedure HandlePlane1Bits2(var RawImage: TBodyData; BM: TBitmap; Info: TPCXInfo);
var X,Y, Width,Height,Inx,Bit,Shift,V: integer;
const Bits: array [0..3] of byte = ($C0,$30,$0C,$03);
const Shifts: array [0..3] of integer = (6,4,2,0);
var Scan: pRGBTripleArray;
begin
BM.PixelFormat:=pf24Bit;
Width:=Info.Xmax - Info.Xmin + 1;
Height:=Info.Ymax - Info.Ymin + 1;
BM.Width:=Width;
BM.Height:=Height;
for Y:=0 to Height-1 do
	begin
	Scan:=BM.ScanLine[Y];
	for X:=0 to Width-1 do
		begin
		Bit:=Bits[X Mod 4];
		Shift:=Shifts[X Mod 4];
		Inx:=(X div 4)+(Y * Info.BytesPerLine);
		V:=(RawImage[Inx] and Bit) shr Shift;
		Scan[X]:=Info.Palette1[V];
		end;
	end;
end;



procedure HandlePlane3Bits1(var RawImage: TBodyData; BM: TBitmap; Info: TPCXInfo);
var X,Y, Width,Height,Inx,Bit: integer;
var Plane,PlaneOff: integer;
var PalInx: integer;
const Bits: array [0..7] of byte = ($80,$40,$20,$10,$08,$04,$02,$01);
var Scan: pRGBTripleArray;
begin
BM.PixelFormat:=pf24Bit;
Width:=Info.Xmax - Info.Xmin + 1;
Height:=Info.Ymax - Info.Ymin + 1;
BM.Width:=Width;
BM.Height:=Height;
for Y:=0 to Height-1 do
	begin
	Scan:=BM.ScanLine[Y];
	for X:=0 to Width-1 do
		begin
		Bit:=Bits[X Mod 8];
		Inx:=(X div 8)+(Y * 4 * Info.BytesPerLine);
		PalInx:=0;
		for Plane:=2 downto 0 do
			begin
			PalInx:=PalInx shl 1;
			PlaneOff:=Plane * Info.BytesPerLine;
			if (RawImage[Inx+PlaneOff] and Bit)<>0 then PalInx:=PalInx or 1;
			end;
		Scan[X]:=Info.Palette1[PalInx];
		end;
	end;
end;



procedure HandlePlane4Bits1(var RawImage: TBodyData; BM: TBitmap; Info: TPCXInfo);
var X,Y, Width,Height,Inx,Bit: integer;
var Plane,PlaneOff: integer;
var PalInx: integer;
const Bits: array [0..7] of byte = ($80,$40,$20,$10,$08,$04,$02,$01);
var Scan: pRGBTripleArray;
begin
BM.PixelFormat:=pf24Bit;
Width:=Info.Xmax - Info.Xmin + 1;
Height:=Info.Ymax - Info.Ymin + 1;
BM.Width:=Width;
BM.Height:=Height;
for Y:=0 to Height-1 do
	begin
	Scan:=BM.ScanLine[Y];
	for X:=0 to Width-1 do
		begin
		Bit:=Bits[X Mod 8];
		Inx:=(X div 8)+(Y * 4 * Info.BytesPerLine);
		PalInx:=0;
		for Plane:=3 downto 0 do
			begin
			PalInx:=PalInx shl 1;
			PlaneOff:=Plane * Info.BytesPerLine;
			if (RawImage[Inx+PlaneOff] and Bit)<>0 then PalInx:=PalInx or 1;
			end;
		Scan[X]:=Info.Palette1[PalInx];
		end;
	end
end;



procedure HandlePlane1Bits8(var RawImage: TBodyData; BM: TBitmap; Info: TPCXInfo);
var X,Y, Width,Height,Inx: integer;
var Scan: pRGBTripleArray;
begin
BM.PixelFormat:=pf24Bit;
Width:=Info.Xmax - Info.Xmin + 1;
Height:=Info.Ymax - Info.Ymin + 1;
BM.Width:=Width;
BM.Height:=Height;
if Info.Pal2Tag=$0C then
	begin
	{Has a palette}
	for Y:=0 to Height-1 do
		begin
		Scan:=BM.ScanLine[Y];
		 for X:=0 to Width-1 do
			begin
			Inx:=X +(Y * Info.BytesPerLine);
			Scan[X]:=Info.Palette2[RawImage[Inx]];
			end;
		end;
	end
else
	begin
	{Has no palette so must be gray scale}
	for Y:=0 to Height-1 do
		begin
		Scan:=BM.ScanLine[Y];
		for X:=0 to Width-1 do
			begin
			Inx:=X +(Y * Info.BytesPerLine);
			Scan[X].rgbtRed:=RawImage[Inx];
			Scan[X].rgbtGreen:=RawImage[Inx];
			Scan[X].rgbtBlue:=RawImage[Inx];
			end;
		end;
	end
end;


procedure HandlePlane3Bits8(var RawImage: TBodyData; BM: TBitmap; Info: TPCXInfo);
var X,Y, Width,Height,Inx1,Inx2,Inx3: integer;
var Scan: pRGBTripleArray;
begin
BM.PixelFormat:=pf24Bit;
Width:=Info.Xmax - Info.Xmin + 1;
Height:=Info.Ymax - Info.Ymin + 1;
BM.Width:=Width;
BM.Height:=Height;
for Y:=0 to Height-1 do
	begin
	Scan:=BM.ScanLine[Y];
	for X:=0 to Width-1 do
		begin
		Inx1:=X +(3 * Y * Info.BytesPerLine);
		Inx2:=Inx1 + Info.BytesPerLine;
		Inx3:=Inx1 + Info.BytesPerLine * 2;
		Scan[X].rgbtRed:=RawImage[Inx1];
		Scan[X].rgbtGreen:=RawImage[Inx2];
		Scan[X].rgbtBlue:=RawImage[Inx3];
		end;
	end;
end;



procedure DecodeRawData(var RawImage: TBodyData; BM: TBitmap; Info: TPCXInfo);
begin
if (Info.Planes=1) and (Info.BitsPerPixel=1)  then HandlePlane1Bits1(RawImage,BM,Info)
else if (Info.Planes=1) and (Info.BitsPerPixel=2) then HandlePlane1Bits2(RawImage,BM,Info)
else if (Info.Planes=3) and (Info.BitsPerPixel=1) then HandlePlane3Bits1(RawImage,BM,Info)
else if (Info.Planes=4) and (Info.BitsPerPixel=1) then HandlePlane4Bits1(RawImage,BM,Info)
else if (Info.Planes=1) and (Info.BitsPerPixel=8) then HandlePlane1Bits8(RawImage,BM,Info)
else if (Info.Planes=3) and (Info.BitsPerPixel=8) then HandlePlane3Bits8(RawImage,BM,Info);

SetBMPPPM(BM,round(Info.HDPI * Meters2Inches),round(Info.VDPI * Meters2Inches));
AdvanceProgress(psRunning,0);
end;





{-------------------------------------------------------------------------------}

function GetHighest(var RawImage: TBodyData): integer;
var I: integer;
begin
Result:=0;
for I:=0 to high(RawImage) do
 if RawImage[I]>Result then Result:=RawImage[I];
end;


function LoadPCXStream(Stream: TStream; BM: TBitmap): TPCXInfo;
{Load PCX stream into specified bitmap }
var Info: TPCXInfo;
var RawImage: TBodyData;
begin
StreamEndian:=LittleEnd;
FillMemory(@Info,SizeOf(TPCXInfo),0);
AdvanceProgress(psStarting,4);
ReadHeader(Stream,Info);
ReadRawImage(Stream,Info,RawImage);
ReadSecondPalette(Stream,Info);
DecodeRawData(RawImage,BM,Info);
Info.PeakValue:=GetHighest(RawImage);
Result:=Info;
AdvanceProgress(psEnding,0);
end;




function ReadPCXFile(BM: TBitmap; FileName: string): TPCXInfo;
{Read PCX from file }
var Stream: TMemoryStream;
begin
Stream := TMemoryStream.Create;
Stream.LoadFromFile(Filename);
try
Result:=LoadPCXStream(Stream,BM);
finally Stream.Free; end;
end;


end.
