unit SplineLib;

interface

uses Vector;


procedure DoQuadraticBezier(const Source: T2DPolygon; var Destination: T2DPolygon);
procedure GetCardinalSpline(const Source: T2DPolygon; var Destination: T2DPolygon; Tension: double = 0.5);

procedure DoCubicSplineLine(Steps: Integer; L1,L2: T2DLine; ClearArray: boolean; var PG: T2DVectorArray);


implementation

{---------------- Recursive Bezier Quadratic Spline ---------------------------}

function IsZero(const A: double): Boolean;
const Epsilon = 1E-15 * 1000;
begin
Result := Abs(A) <= Epsilon;
end;


function GetEndPointTangent(EndPnt, Adj: T2DVector; tension: double): T2DVector;
{ Calculates Bezier points from cardinal spline endpoints.}
begin
{ tangent at endpoints is the line from the endpoint to the adjacent point}
Result:=VectorAdd2D(ScalarProduct2D(VectorSubtract2D(Adj, EndPnt), tension), EndPnt);
end;



procedure GetInteriorTangent(const pts: T2DPolygon; Tension: double; var P1,P2: T2DVector);
{ Calculate incoming and outgoing tangents.}
{Pts[0] = Previous point, Pts[1] = Current Point, Pts[2] = Next Point}
var Diff,TV: T2DVector;
begin
{ Tangent Vector = Next Point - Previous Point * Tension}
Diff:=VectorSubtract2D(pts[2],pts[0]);
TV:=ScalarProduct2D(Diff,Tension);

{ Add/Subtract tangent vector to get control points}
P1:=VectorSubtract2D(pts[1],TV);
P2:=VectorAdd2D(pts[1],TV);
end;


function VectorMidPoint(const P1,P2: T2DVector): T2DVector;
begin
Result:=ScalarDivide2D(VectorAdd2D(P1,P2),2);
end;


{Don't change item order}

type TBezierPoints = packed record
 BeginPoint,BeginControl,
 EndControl,EndPoint: T2DVector;
 end;


function ControlBetweenBeginEnd(BeginPoint,BeginControl,EndControl,EndPoint: double): boolean;
{ Are control points are between begin and end point}
begin
Result:=False;
if BeginControl < BeginPoint then
	begin
	if BeginControl < EndPoint then exit;
	end
else if BeginControl > EndPoint then exit;

if EndControl < BeginPoint then
	begin
	if EndControl < EndPoint then exit;
	end
else if EndControl > EndPoint then exit;
Result:=True;
end;


function RecursionDone(const Points: TBezierPoints): Boolean;
{ Function to check that recursion can be terminated}
{ Returns true if the recusion can be terminated }
const BezierPixel = 1;
var dx, dy: double;
begin
dx := Points.EndPoint.x - Points.BeginPoint.x;
dy := Points.EndPoint.y - Points.BeginPoint.y;
if Abs(dy) <= Abs(dx) then
	begin
	{ shallow line - check that control points are between begin and end}
	Result:=False;
	if not ControlBetweenBeginEnd(Points.BeginPoint.X,Points.BeginControl.X,Points.EndControl.X,Points.EndPoint.X) then exit;

	Result:=True;
	if IsZero(dx) then exit;

	if (Abs(Points.BeginControl.y - Points.BeginPoint.y - (dy / dx) * (Points.BeginControl.x - Points.BeginPoint.x)) > BezierPixel) or
	   (Abs(Points.EndControl.y -   Points.BeginPoint.y - (dy / dx) * (Points.EndControl.x -   Points.BeginPoint.x)) > BezierPixel) then
		begin
		Result := False;
		exit;
		end
	else
		begin
		Result := True;
		exit;
		end;
	end
else
	begin
	{ steep line - check that control points are between begin and end}
	Result:=False;
	if not ControlBetweenBeginEnd(Points.BeginPoint.Y,Points.BeginControl.Y,Points.EndControl.Y,Points.EndPoint.Y) then exit;
	Result:=True;
	if IsZero(dy) then exit;

	if (Abs(Points.BeginControl.x - Points.BeginPoint.x - (dx / dy) * (Points.BeginControl.y - Points.BeginPoint.y)) > BezierPixel) or
	   (Abs(Points.EndControl.x -   Points.BeginPoint.x - (dx / dy) * (Points.EndControl.y -   Points.BeginPoint.y)) > BezierPixel) then
		begin
		Result := False;
		exit;
		end
	else
		begin
		Result := True;
		exit;
		end;
	end;
end;



procedure BezierRecursion(var Points: TBezierPoints; var PtsOut: T2DPolygon; var Alloc, OutCount: Integer; level: Integer);
{Recursively subdivide the space between the two Bezier end-points}
var Points2: TBezierPoints; { for the second recursive call}
begin
{Out of memory?}
if OutCount = Alloc then
	begin
	{then double hte memory allocation}
	Alloc := Alloc * 2;
	SetLength(PtsOut, Alloc);
	end;

if (level = 0) or RecursionDone(Points) then { Recursion can be terminated}
	begin
	if OutCount = 0 then
		begin
		PtsOut[0] := Points.BeginPoint;
		OutCount := 1;
		end;
	PtsOut[OutCount] := Points.EndPoint;
	Inc(OutCount);
	end
else
	begin
	{Split Points into two halves}
	Points2.EndPoint:=Points.EndPoint;
	Points2.EndControl:=VectorMidPoint(Points.EndControl, Points.EndPoint);
	Points2.BeginPoint:=VectorMidPoint(Points.BeginControl, Points.EndControl);
	Points2.BeginControl:=VectorMidPoint(Points2.BeginPoint,Points2.EndControl);

	Points.BeginControl:=VectorMidPoint(Points.BeginPoint,  Points.BeginControl);
	Points.EndControl:=VectorMidPoint(Points.BeginControl, Points2.BeginPoint);
	Points.EndPoint:=VectorMidPoint(Points.EndControl, Points2.BeginControl);

	Points2.BeginPoint := Points.EndPoint;

	{ Do recursion on the two halves}
	BezierRecursion(Points, PtsOut, Alloc, OutCount, level - 1);
	BezierRecursion(Points2, PtsOut, Alloc, OutCount, level - 1);
	end;
end;


procedure DoQuadraticBezier(const Source: T2DPolygon; var Destination: T2DPolygon);
{Generate Bezier spline from Source polygon and store result in Destination }
{Source Format: P[0] = Start Point, P[1]= Control Point, P[2] = End Point }
var B, Alloc,OutCount: Integer;
var ptBuf: TBezierPoints;
begin
if (Length(Source) - 1) mod 3 <> 0 then exit;
OutCount := 0;
{Start with allocation of 150 to save allocation overhead}
Alloc := 150;
SetLength(Destination, Alloc);
for B:=0 to (Length(Source) - 1) div 3 - 1 do
	begin
	Move(Source[B * 3], ptBuf.BeginPoint, SizeOf(ptBuf));
	BezierRecursion(ptBuf, Destination, Alloc, OutCount, 8);
	end;
{Trim Destination to actual length}
SetLength(Destination,OutCount);
end;



procedure GetCardinalSpline(const Source: T2DPolygon; var Destination: T2DPolygon; Tension: double = 0.5);
{Generate cardinal spline from Source with result in Destination}
{Generate tangents to get the Cardinal Spline}
var i: Integer;
var pt: T2DPolygon;
var P1,P2: T2DVector;
begin
{We need at least 2 points}
if Length(Source) <= 1 then exit;

{ The points and tangents require count * 3 - 2 points.}
SetLength(pt, Length(Source) * 3 - 2);
tension := tension * 0.3;

{Calculate Tangents for each point and store results in new array}

{Do the first point}
pt[0]:=Source[0];
pt[1]:=GetEndPointTangent(Source[0], Source[1], tension);

{Do intermediates points}
for i := 0 to Length(Source) - 3 do
	begin
	GetInteriorTangent(T2DPolygon(@(Source[i])), tension, P1,P2);
	pt[3 * i + 2]:=P1;
	pt[3 * i + 3]:=Source[i + 1];
	pt[3 * i + 4]:=P2;
	end;
{Do last point}
pt[Length(Pt) - 1]:=Source[Length(Source) - 1];
pt[Length(Pt) - 2]:=GetEndPointTangent(Source[Length(Source) - 1], Source[Length(Source) - 2], Tension);

DoQuadraticBezier(pt, Destination);
end;


{------------------- Cubic Bezier Spline --------------------------------------}


procedure DoCubicSplineLine(Steps: Integer; L1,L2: T2DLine; ClearArray: boolean; var PG: T2DVectorArray);
{Do cubic Bezier spline between L1.P1 and L2.P1  }
{L1.P1 = Point-1, L1.P2 = Control1, L2.P1=Control2, L2.P2 = Point-2}
var P: Integer;
var V: T2DVector;
var T: double;
var A,B,C,D,E,F,G,H : double;
begin
if ClearArray then SetLength(PG,0);
A :=  L2.P2.X - (3  * L2.P1.X) + (3  * L1.P2.X) - L1.P1.X;
B := (3  * L2.P1.X) - (6  * L1.P2.X) + (3  * L1.P1.X);
C := (3  * L1.P2.X) - (3  * L1.P1.X);
D := L1.P1.X;

E := L2.P2.Y - (3  * L2.P1.Y) + (3  * L1.P2.Y) - L1.P1.Y;
F := (3  * L2.P1.Y) - (6  * L1.P2.Y) + (3  * L1.P1.Y);
G := (3  * L1.P2.Y) - (3  * L1.P1.Y);
H := L1.P1.Y;

for P:=0 to Steps-1 do
	begin
	T :=P / (Steps-1);
	V.X := (((A  * T) + B)  * T + C)  * T + D;
	V.Y := (((E  * T) + F)  * T + G)  * T + H;
	SetLength(PG,Length(PG)+1);
	PG[High(PG)]:=V;
	end;
end;



end.
