unit StreamLib;

interface

uses Windows,Classes,SysUtils,DSupport;

{------------------------------------------------------------------------------}


type TTempFileStream = class(THandleStream)
 private
  FFileName: string;
 public
  procedure SetSize(NewSize: Longint); override;
  procedure SetSize(const NewSize: Int64); override;
  procedure SaveToFile(const FileName: string);
  procedure LoadFromFile(const FileName: string);
  constructor Create(const Prefix: string);
  destructor Destroy; override;
  property FileName: string read FFileName;
 end;

{Specialized file stream that save data scrambled to protect it}


type TSafeStream = class(TTempFileStream)
 private
 protected
 public
  constructor Create; overload;
  function Read(var Buffer; Count: Longint): Longint; override;
  function Write(const Buffer; Count: Longint): Longint; override;
 end;

{------------------------------------------------------------------------------}


type TNumberType = (LittleEnd,BigEnd);

var StreamEndian: TNumberType=LittleEnd;	{Current file is big or little "endian"}

{---- Big/Little Endian number routines ----}

procedure ReverseByteOrder(var D; N: integer);

function GetStreamByte(Stream: TStream): Byte;
function GetStreamWord(Stream: TStream): Word;
function GetStreamDWord(Stream: TStream): DWord;
function GetStreamInt64(Stream: TStream): int64;
function GetStreamDouble(Stream: TStream): double;
function GetStreamString(Stream: TStream; Count: integer): AnsiString; overload;
function GetStreamString(Stream: TStream; Term: AnsiChar): AnsiString; overload;
procedure GetStreamByteArray(Stream: TStream; Count: integer; var BA: TDynByteArray);

function GetStreamSingle(Stream: TStream): single;
function PeekStreamByte(Stream: TStream): byte;
function PeekStreamWord(Stream: TStream): Word;

procedure WriteStreamByte(Stream: TStream; B: byte);
procedure WriteStreamByteBool(Stream: TStream; B: boolean);
procedure WriteStreamWord(Stream: TStream; V: word);
procedure WriteStreamDWord(Stream: TStream; V: DWord);
procedure WriteStreamDouble(Stream: TStream; V: double);

{---- Text and Character I/O -----}

type TCharSet = set of #0..#$7f;

procedure WriteStreamChar(Stream: TStream; B: AnsiChar);
procedure WriteStreamString(Stream: TStream; S: AnsiString);
procedure WriteStreamStringLn(Stream: TStream; S: AnsiString);
procedure WriteStreamNChars(Stream: TStream; C: AnsiChar; Count: integer);
procedure WriteStreamTruncStr(Stream: TStream; S: AnsiString; Len: integer);

function GetStreamChar(Stream: TStream): AnsiChar;
function StreamSkipToChar(Stream: TStream; Target: AnsiChar): boolean;
function PeekStreamChar(Stream: TStream): AnsiChar;
function LookBackStreamChar(Stream: TStream; BackCount: integer): AnsiChar;
function GetStreamPreviousChar(Stream: TStream): AnsiChar;
function StreamGetToChar(Stream: TStream; Target: AnsiChar): AnsiString;
function StreamReadString(Stream: TStream; Count: integer): AnsiString;
function StreamSkipToChars(Stream: TStream; Chars: TCharSet): boolean;
procedure StreamSkipWhiteSpace(Stream: TStream);

function StreamEOF(Stream: TStream): boolean;
function StreamRealIn(Stream: TStream): double;
function StreamIntegerIn(Stream: TStream): integer;


function FileToString(const FileName: TFileName; var S: AnsiString): boolean;
function StreamToString(Stream: TStream): AnsiString;

procedure AssignStream(var F: Textfile; S: TStream );

var LastChar: AnsiChar;	{Last char from some text stream operations}

implementation


procedure ReverseByteOrder(var D; N: integer);
{Reverse the order of bytes in any type of variable}
var Bytes: PByteArray;
var T: Byte;
var I,I2: integer;
begin
Bytes:=@D;
I2:=N-1;
for I:=0 to (N-1) div 2 do
	begin
        T:=Bytes^[I2]; Bytes^[I2]:=Bytes^[I]; Bytes^[I]:=T;
        Dec(I2);
        end;
end;



{------------------ Stream Reading Support --------------------------}




function GetStreamByte(Stream: TStream): Byte;
{Read byte from specified stream}
begin
Stream.Read(Result,1);
end;



function PeekStreamByte(Stream: TStream): byte;
{Peek one bye ahead from specified stream}
begin
Stream.Read(Result,1);
Stream.Position:=Stream.Position-1;
end;



function GetStreamWord(Stream: TStream): Word;
{Read word and handle big/little endian}
begin
Stream.Read(Result,2);
if StreamEndian=BigEnd then ReverseByteOrder(Result,2);
end;



function PeekStreamWord(Stream: TStream): Word;
{Read word and handle big/little endian}
begin
Result:=GetStreamWord(Stream);
Stream.Position:=Stream.Position-2;
end;


function GetStreamDWord(Stream: TStream): DWord;
{Read 32 bit word and handle big/little endian}
begin
Stream.Read(Result,4);
if StreamEndian=BigEnd then ReverseByteOrder(Result,4);
end;


function GetStreamInt64(Stream: TStream): int64;
{Read 64 bit integer and handle big/little endian}
begin
Stream.Read(Result,8);
if StreamEndian=BigEnd then ReverseByteOrder(Result,8);
end;


function GetStreamDouble(Stream: TStream): double;
{Read double and handle big/little endian}
begin
Stream.Read(Result,8);
if StreamEndian=BigEnd then ReverseByteOrder(Result,8);
end;


function GetStreamSingle(Stream: TStream): single;
{Read single and handle big/little endian}
begin
Stream.Read(Result,4);
if StreamEndian=BigEnd then ReverseByteOrder(Result,4);
end;


function GetStreamString(Stream: TStream; Count: integer): AnsiString;
var I: integer;
var C: AnsiChar;
begin
Result:='';
for I:=0 to Count-1 do
	begin
	Stream.Read(C,1);
	Result:=Result+C;
	end;
end;


procedure GetStreamByteArray(Stream: TStream; Count: integer; var BA: TDynByteArray);
var I: integer;
var B: byte;
begin
SetLength(BA,Count);
for I:=0 to Count-1 do
	begin
	Stream.Read(B,1);
	BA[I]:=B;
	end;
end;

function GetStreamString(Stream: TStream; Term: AnsiChar): AnsiString;
var C: AnsiChar;
begin
Result:='';
repeat
	begin
	Stream.Read(C,1);
	if C<>Term then Result:=Result+C;
	end
until C=Term;
end;



{----------------- Stream output routines ----------------------}


procedure WriteStreamByte(Stream: TStream; B: byte);
{Write byte to stream}
begin
Stream.Write(B,1);
end;




procedure WriteStreamByteBool(Stream: TStream; B: boolean);
{Write a byte boolean value}
begin
if B then WriteStreamByte(Stream,1)
else WriteStreamByte(Stream,0);
end;



procedure WriteStreamWord(Stream: TStream; V: word);
{Write word and handle big/little endian}
begin
if StreamEndian=BigEnd then ReverseByteOrder(V,2);
Stream.Write(V,2);
end;


procedure WriteStreamDWord(Stream: TStream; V: DWord);
{Write dword and handle big/little endian}
begin
if StreamEndian=BigEnd then ReverseByteOrder(V,4);
Stream.Write(V,4);
end;


procedure WriteStreamDouble(Stream: TStream; V: double);
{Write double and handle big/little endian}
begin
if StreamEndian=BigEnd then ReverseByteOrder(V,8);
Stream.Write(V,8);
end;




{------------ Non-Numeric Routines --------------------}


procedure WriteStreamChar(Stream: TStream; B: AnsiChar);
{Write byte to stream}
begin
Stream.Write(B,1);
end;



procedure WriteStreamString(Stream: TStream; S: AnsiString);
{Write delphi string to stream}
begin
Stream.Write(Pointer(S)^,Length(S));
end;



procedure WriteStreamTruncStr(Stream: TStream; S: AnsiString; Len: integer);
{Write string to stream, trunc to Len}
{ For strings shorter than Len, Pad with zeros}
var I,L: integer;
begin
L:=Length(S);
for I:=1 to Len-1 do
 if I>L then WriteStreamByte(Stream,0)
 else WriteStreamChar(Stream,S[I]);
WriteStreamByte(Stream,0);
end;




procedure WriteStreamStringLn(Stream: TStream; S: AnsiString);
{Write delphi string plus CRLF to stream}
begin
WriteStreamString(Stream,S+CRLF_Char);
end;



procedure WriteStreamNChars(Stream: TStream; C: AnsiChar; Count: integer);
{Write delphi string to stream}
var I: integer;
begin
for I:=1 to Count do WriteStreamChar(Stream,C);
end;



function GetStreamChar(Stream: TStream): AnsiChar;
{Read char from specified stream}
begin
Stream.Read(Result,1);
end;



function PeekStreamChar(Stream: TStream): AnsiChar;
{Peek one char ahead from specified stream}
begin
Stream.Read(Result,1);
Stream.Position:=Stream.Position-1;
end;



function LookBackStreamChar(Stream: TStream; BackCount: integer): AnsiChar;
{Peek one char back from specified stream}
var OldPos: int64;
begin
OldPos:=Stream.Position;
Stream.Position:=Stream.Position-BackCount;
Stream.Read(Result,1);
Stream.Position:=OldPos;
end;
        


function GetStreamPreviousChar(Stream: TStream): AnsiChar;
{Peek at previous char}
begin
if Stream.Position>0 then
	begin
	Stream.Position:=Stream.Position-1;
	Result:=GetStreamChar(Stream);
	end
else Result:=#0;
end;


function StreamSkipToChar(Stream: TStream; Target: AnsiChar): boolean;
{Skip to specified character or EOF}
{Returns true if specified character found}
begin
while (Stream.Position<Stream.Size) do
if GetStreamChar(Stream)=Target then break;
Result:=Stream.Position<Stream.Size;
end;



procedure StreamSkipWhiteSpace(Stream: TStream);
{Skip chars up to first non-white space char}
var C: AnsiChar;
begin
while (Stream.Position<Stream.Size) do
	begin
	C:=PeekStreamChar(Stream);
	if C>#$20 then break
	else GetStreamChar(Stream);
	end;
end;



function StreamSkipToChars(Stream: TStream; Chars: TCharSet): boolean;
{Skip to specified character or EOF}
{Returns true if specified character found}
begin
while (Stream.Position<Stream.Size) do
if GetStreamChar(Stream) in Chars then break;
Result:=Stream.Position<Stream.Size;
end;



function StreamGetToChar(Stream: TStream; Target: AnsiChar): AnsiString;
{Returns string up to specified character or EOF}
var C: AnsiChar;
begin
Result:='';
while (Stream.Position<Stream.Size) do
	begin
	C:=GetStreamChar(Stream);
	if C=Target then break
	else Result:=Result+C;
	end;
end;


function StreamReadString(Stream: TStream; Count: integer): AnsiString;
{Read string of specified number of characters}
var R: integer;
begin
SetLength(Result,Count);
R:=Stream.Read(Result[1],Count);
SetLength(Result,R);
end;


function StreamToString(Stream: TStream): AnsiString;
{Read whole stream as string}
var MS: TMemoryStream;
begin
Result:='';
MS:=TMemoryStream.Create;
try
MS.LoadFromStream(Stream);
SetString(Result, PChar(MS.memory), MS.Size);
finally ms.Free; end;
end;



function FileToString(const FileName: TFileName; var S: AnsiString): boolean;
var FS: TFileStream;
begin
Result:=True;
FS:=TFileStream.Create(FileName,fmOpenRead or fmShareDenyWrite);
try try
SetLength(S, FS.Size);
FS.Read(Pointer(S)^, FS.Size);
finally FS.Free; end;
except Result:=False; S:=''; end;
end;



function StreamEOF(Stream: TStream): boolean;
begin
Result:=Stream.Position>=Stream.Size;
end;



function StreamIntegerIn(Stream: TStream): integer;
{ Read ACII integer from stream}
var S: String[128];
begin
S:='';
Stream.Read(LastChar,1);
while (LastChar<'0') or (LastChar>'9') do
	begin
	if StreamEOF(Stream) then begin Result:=0; exit; end;
	if (LastChar = '-') or (LastChar = '+') then S:=S+LastChar
	else S:='';
	Stream.Read(LastChar,1);
	end;
while (LastChar>='0') and (LastChar<='9') do
	begin
	S:=S+LastChar;
	Stream.Read(LastChar,1);
	if StreamEOF(Stream) then break;
	end;
Result:=StrToInt(S);
end;



function StreamRealIn(Stream: TStream): double;
{ Read a ASCII real from the stream}
var S: String[128];
begin
S:='';
Stream.Read(LastChar,1);
while ((LastChar<'0') or (LastChar>'9')) and (LastChar<>'.') do
	begin
	if StreamEOF(Stream) then begin Result:=0; exit; end;
	if (LastChar = '-') or (LastChar = '+') then S:=S+LastChar
	else S:='';
	Stream.Read(LastChar,1);
	end;
while (LastChar>='0') and (LastChar<='9') or
	(LastChar='.') or (LastChar='E') or (LastChar='e') or
	(LastChar='+') or (LastChar='-') do
	begin
	S:=S+LastChar;
	Stream.Read(LastChar,1);
	if StreamEOF(Stream) then break;
	end;
Result:=StrToFloat(S);
end;



{-------------------------------------------------------------------------------}
{Routines to assign a stream to a Text file ------------------------------------}
{ Used like this:
 var FH: TextFile;
 var MS: TMemoryStream;

 MS:=TMemoryStream.Create;
 AssignStream(FH,MS);
 Rewrite(FH);
 WriteLn(FH,'Hello World', N);  }
{-------------------------------------------------------------------------------}

Function GetDevStream(var F: TTextRec): TStream;
{Get the stream reference stored in the textrec userdata area}
begin
Move( F.Userdata, Result, Sizeof( Result ));
Assert( Assigned( Result ));
end;



Function DevIn( var F: TTextRec ): Integer;
{Called by Read, ReadLn etc. to fill the textfiles buffer from the stream.}
begin
Result := 0;
with F do
 begin
 Bufend := GetDevStream(F).Read( BufPtr^, BufSize );
 BufPos := 0;
 end;
end;



Function DevFlushIn( var F: TTextRec ): Integer;
{A dummy method, flush on input does nothing.}
begin
Result := 0;
end;



Function DevOut( var F: TTextRec ): Integer;
{ Write the textfile buffers content to the stream. }
begin { DevOut }
Result := 0;
with F do
 If BufPos > 0 Then
	begin
	GetDevStream(F).WriteBuffer( BufPtr^, BufPos );
	BufPos := 0;
	end;
end;




Function DevClose( var F: TTextRec ): Integer;
{Does nothing here. }
begin
Result := 0;
end;


Function DevOpen( var F: TTextRec ): Integer;
{prepare the textfile for I/O}
begin
Result := 0;
with F do
 begin
 Case Mode Of
 fmInput:
 	begin { Reset }
 	InOutFunc := @DevIn;
 	FlushFunc := @DevFlushIn;
 	BufPos := 0;
 	Bufend := 0;
 	GetDevStream( F ).Position := 0;
 	end; { Case fmInput }
 fmOutput:
	begin { Rewrite }
	InOutFunc := @DevOut;
	FlushFunc := @DevOut;
	BufPos := 0;
	Bufend := 0;
	GetDevStream( F ).Position := 0;
	end; { Case fmOutput }
 fmInOut:
	begin { Append }
	Mode := fmOutput;
	DevOpen( F );
	GetDevStream(F).Seek( 0, soFromend );
	end; { Case fmInOut }
  end;
 end;
end;


Procedure AssignStream(var F: Textfile; S: TStream );
{ Attach a stream to a Textfile to allow I/O via WriteLn/ReadLn }
{ F is the textfile to attach the stream to }
{ S is the stream }
begin
Assert( Assigned( S ));
with TTextRec(F) do
 begin
 Mode := fmClosed;
 BufSize := SizeOf(Buffer);
 BufPtr := @Buffer;
 OpenFunc := @DevOpen;
 CloseFunc := @DevClose;
 Name[0] := #0;
 { Store stream reference into Userdata area }
 Move( S, Userdata, Sizeof(S));
 end;
end;


{------------------------------------------------------------------------------}

procedure ScrambleBits( const Source; var Dest; count : Integer );
{Move data from source to destination, scrambling bits on the way}
{ doesn't handle over lapping blocks}
{     ->EAX     Pointer to source       }
{       EDX     Pointer to destination  }
{       ECX     Count                   }
asm
        PUSH    ESI		{Save registers}
        PUSH    EDI
        MOV     ESI,EAX		{Point source and destination}
        MOV     EDI,EDX
	MOV	EDX,ECX		{Save Count}
        CLD

	SAR	ECX,2           {Copying four bytes at a time }
	JZ      @SMALL
@LOOP1:	LODSD			{Get next four bytes.}
	XOR     EAX,$55555555	{Scramble it}
	STOSD			{Save letter on screen.}
	LOOP	@LOOP1

@SMALL:	MOV     ECX,EDX		{Restore count}
	AND     ECX,03H
	JZ      @@exit

@LOOP2:	LODSB			{Get next character.}
	XOR     AL,$55		{Scramble it}
	STOSB			{Save letter on screen.}
	LOOP	@LOOP2
@@exit:
        POP     EDI
        POP     ESI
end;




function FileGetTempName(const Prefix: string): string;
var TempPath, TempFile: string;
var R: Cardinal;
begin
Result := '';
R := GetTempPath(0, nil);
SetLength(TempPath, R);
R := GetTempPath(R, PChar(TempPath));
if R <> 0 then
	begin
	SetLength(TempPath, StrLen(PChar(TempPath)));
	SetLength(TempFile, MAX_PATH);
	R := GetTempFileName(PChar(TempPath), PChar(Prefix), 0, PChar(TempFile));
	if R <> 0 then
		begin
		SetLength(TempFile, StrLen(PChar(TempFile)));
		Result := TempFile;
		end;
	end;
end;



function FileCreateTemp(var Prefix: string): THandle;
var TempName: string;
begin
Result := INVALID_HANDLE_VALUE;
TempName := FileGetTempName(Prefix);
if TempName <> '' then
	begin
	Result := CreateFile(PChar(TempName), GENERIC_READ or GENERIC_WRITE, 0, nil,
	OPEN_EXISTING, FILE_ATTRIBUTE_TEMPORARY or FILE_FLAG_DELETE_ON_CLOSE, 0);
	// In certain situations it's possible that CreateFile fails yet the file is
	// actually created, therefore explicitly delete it upon failure.
	if Result = INVALID_HANDLE_VALUE then
	DeleteFile(TempName);
	Prefix := TempName;
	end;
end;




procedure TTempFileStream.SaveToFile(const FileName: string);
var FS: TStream;
begin
FS:=TFileStream.Create(FileName, fmCreate);
try
Position:=0;
FS.CopyFrom(Self,Size)
finally FS.Free; end;
end;



procedure TTempFileStream.LoadFromFile(const FileName: string);
var FS: TStream;
begin
FS := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
try
Self.CopyFrom(FS,FS.Size);
finally FS.Free; end;
end;


constructor TTempFileStream.Create(const Prefix: string);
var FileHandle: THandle;
begin
FFileName := Prefix;
FileHandle := FileCreateTemp(FFileName);
if FileHandle = INVALID_HANDLE_VALUE then raise Exception.Create('Unable to create Temp File');
inherited Create(FileHandle);
end;



destructor TTempFileStream.Destroy;
begin
if THandle(Handle) <> INVALID_HANDLE_VALUE then CloseHandle(Handle);
inherited Destroy;
end;



procedure TTempFileStream.SetSize(NewSize: Integer);
begin
inherited SetSize(NewSize);
end;




procedure TTempFileStream.SetSize(const NewSize: Int64);
begin
inherited SetSize(NewSize);
end;



{ TSafeStream }

constructor TSafeStream.Create;
begin
inherited Create('');
end;



function TSafeStream.Read(var Buffer; Count: Integer): Longint;
var BA: PByte;
begin
GetMem(BA,Count);
Result:=Inherited Read(BA^,Count);
ScrambleBits(BA^, Buffer, Result);
FreeMem(BA);
end;



function TSafeStream.Write(const Buffer; Count: Integer): Longint;
var BA: PByte;
begin
GetMem(BA,Count);
ScrambleBits(Buffer, BA^, Count);
Result:=Inherited Write(BA^,Count);
FreeMem(BA);
end;

end.
