unit TIFRaw;

{ Unit to read common versions of 6.0 TIFF file formats }

interface

uses Windows,SysUtils,Classes,Graphics, TifGlobal, GeoTiff, StreamLib, Math,
     Huge2D, DSupport,BitmapLib;


{Type to hold the parameters from current TIF file}

const PalSize=256;

Type
TTIFFInfo = record
 Width,Depth,
 BytesPerRow,Bits: Word;
 PhotoMetric: Word;
 RowsPerStrip: Word;
 StripCount: Word;
 Samples: Word;
 PlanarConfig: Word;
 Compression: Word;
 BitsPerSample: Word;
 ByteCount: DWord;
 Offset: DWord;
 Palette: array [0..PalSize-1] of TRGBQuad;
 end;

var VerNumber: word;		{Identifies file as TIFF }

var TiffFilename: string;	{Current input filename }
var TIFInfo: TTIFFInfo;		{Contain all info about current file }

var LineBuffer: Pointer;	{Array to store each line of data}


{ Public procedures that can be called from external programs }

procedure ReadRawTIFFile(FileName: string; RawData: T2DIntArray);
procedure ReadRawTIFStream(Stream: TStream; RawData: T2DIntArray);
procedure RawDataToBMP(RawData: T2DIntArray; Info: TTIFFInfo; BM: TBitmap);


implementation




procedure UnhandledCompression(C: integer);
{Exception to deal with unhandled compression}
var S: string;
begin
S:='Unsupported Compression = '+IntToStr(C);
raise Exception.Create(S);
end;




{----------------- TIFF Routines ------------------------------------}


procedure SetDefaults(var FI: TTIFFInfo);
{Fill parameter structure with default values}
var I : Integer;
begin
FI.Width:=0;
FI.Depth:=0;
FI.Bits:=0;
FI.PlanarConfig:=1;
FI.Samples:=1;
FI.PhotoMetric:=1;
FI.Compression:=1;
FI.Offset:=0;
for I:=0 to PalSize-1 do
	begin
	FI.Palette[I].rgbRed:=I;
	FI.Palette[I].rgbGreen:=I;
	FI.Palette[I].rgbBlue:=I;
	FI.Palette[I].rgbReserved:=0;
	end;
end;


procedure SetBilevelPalette(OneIsBlack: boolean; var FI: TTIFFInfo);
{Set bilevel paletter so first two entries control black and white}
{if OneisBlack is true, a one in the data signifies a black pixel}
var B,W: integer;
begin
if OneIsBlack then begin B:=1; W:=0; end
else begin B:=0; W:=1; end;
FI.Palette[B].rgbRed:=0; FI.Palette[B].rgbGreen:=0; FI.Palette[B].rgbBlue:=0;
FI.Palette[W].rgbRed:=255; FI.Palette[W].rgbGreen:=255; FI.Palette[W].rgbBlue:=255;
end;



procedure GetColorMap(Stream: TStream;Offset: integer; var FI: TTIFFInfo);
{Read the TIF file color map into TIF parameters}
var I,Pos: integer;
begin
Pos:=Stream.Position;
Stream.Seek(Offset,soFromBeginning);
for I:=0 to (1 shl FI.BitsPerSample)-1 do
        begin
        if I >= 256 then break;
        FI.Palette[I].rgbRed:= GetStreamWord(Stream) shr 8;
        end;
for I:=0 to (1 shl FI.BitsPerSample)-1 do
        begin
        if I >= 256 then break;
        FI.Palette[I].rgbGreen:= GetStreamWord(Stream) shr 8;
        end;
for I:=0 to (1 shl FI.BitsPerSample)-1 do
        begin
        if I >= 256 then break;
        FI.Palette[I].rgbBlue:=GetStreamWord(Stream) shr 8;
        end;
Stream.Seek(Pos,soFromBeginning);
end;



procedure DecodeTag(Stream: TStream; var FI: TTIFFInfo);
{Read and decode a TIF tag }
var Length,Offset,Value : DWord;
var Pos: DWord;
var Tag,FieldType : Integer;
begin
Tag:=GetStreamWord(Stream);
FieldType:=GetStreamWord(Stream);
Length:=GetStreamDWord(Stream);
Offset:=GetStreamDWord(Stream);
if (FieldType = tifLong) or (TifNumberType=TifLittleEnd) then  Value:=Offset
else Value:=Offset shr 16;

case Tag of
 SubfileType: ;
 ImageWidth: FI.Width:=Value;
 ImageLength: FI.Depth:=Value;
 RowsPerStrip: FI.RowsPerStrip:=Value;
 StripOffsets:
	begin
	FI.Offset:=Value;
	FI.StripCount:=Length;
	end;
 StripByteCounts: FI.ByteCount:=Value;
 SamplesPerPixel: FI.Samples:=Value;
 BitsPerSample:
	begin
	if Length > 1 then
		begin
		Pos:=Stream.Position;
		Stream.Seek(Offset,soFromBeginning);
		FI.BitsPerSample:=GetStreamWord(Stream);
		Stream.Seek(Pos,soFromBeginning);
		end
	else FI.BitsPerSample:=Value;
	end;
 PlanarConfiguration: FI.PlanarConfig:=Value;
 Compression: FI.Compression:=Value;
 Group3Options: ;
 Group4Options: ;
 FillOrder: ;
 Threshholding: ;
 CellWidth: ;
 CellLength: ;
 MinSampleValue: ;
 MaxSampleValue: ;
 PhotometricInterp: FI.PhotoMetric:=Value;
 GrayResponseUnit: ;
 GrayResponseCurve: ;
 ColorResponseUnit: ;
 ColorResponseCurves: ;
 XResolution: ;
 YResolution: ;
 ResolutionUnit: ;
 Orientation: ;
 DocumentName: ;
 PageName: ;
 XPosition: ;
 YPosition: ;
 PageNumber: ;
 ImageDescription: ;
 Make: ;
 Model: ;
 FreeOffsets: ;
 FreeByteCounts: ;
 ColorMap: GetColorMap(Stream,Offset,FI);
 else DecodeGeoTiffTag(Stream,Tag,Length,Offset);
 end;
end;


{--------------- Handle Color Scheme Conversions ------------------------------}


type TCMYK = packed record
 Cyan,Magenta,Yellow,Black: byte;
 end;

type TCMYKArray = array [0..32768] of TCMYK;
type PCMYKArray = ^TCMYKArray;



function CMYKToRGB(CMYK: TCMYK): TRGBTriple;
{Convert CYMYK to RGB}
const MaxValue = 255;
begin
Result.rgbtRed:=MaxValue-EnsureRange(CMYK.Cyan+CMYK.Black,0,255);
Result.rgbtGreen:=MaxValue-EnsureRange(CMYK.Magenta+CMYK.Black,0,255);
Result.rgbtBlue:=MaxValue-EnsureRange(CMYK.Yellow+CMYK.Black,0,255);
end;




procedure Copy1BitData(Source: PByteArray; RawData: T2DIntArray;  FI: TTIFFInfo);
{Copy 1 bit data into separate array elemnets}
var MI,X: integer;
const BitMask: array [0..7] of byte = ($80,$40,$20,$10,$08,$04,$02,$01);
var White,Black: byte;
begin
if FI.PhotoMetric=1 then begin White:=1; Black:=0; end
else begin White:=0; Black:=1; end;
X:=0;
for X:=0 to RawData.Cols-1 do
	begin
	MI:=X mod 8;
        RawData[X,RawData.YIndex]:=PickInteger(IsBitSet(Source[X div 8],BitMask[MI]),White,Black);
	end;
RawData.YIndex:=RawData.YIndex + 1;
end;



procedure Copy4BitData(Source: PByteArray; RawData: T2DIntArray; FI: TTIFFInfo);
{Copy 4 bit data into separate array elemnets}
var I,X: integer;
begin
for X:=0 to RawData.Cols-1 do
	begin
	if (X mod 2)<>0 then RawData[X,RawData.YIndex]:=Source[X div 2] and $0F
	else RawData[X,RawData.YIndex]:=(Source[X div 2] and $F0) shr 4;
	end;
RawData.YIndex:=RawData.YIndex + 1;
end;


procedure Copy8BitData(Source: PByteArray; RawData: T2DIntArray; FI: TTIFFInfo);
{Copy 8 bit data into separate array elemnets}
var I: integer;
begin
for I:=0 to RawData.Cols-1 do RawData[I,RawData.YIndex]:=Source[I];
RawData.YIndex:=RawData.YIndex + 1;
end;




procedure ReorderRGBQuad(Source: PRGBQuadArray; RawData: T2DIntArray; FI: TTIFFInfo);
{Change BGR color order to RGB for Windows BMP}
var X: integer;
begin
for X:=0 to RawData.Cols-1 do
	begin
	RawData[X,RawData.YIndex]:=RGB(Source[X].rgbBlue,Source[X].rgbGreen,Source[X].rgbRed);
	end;
RawData.YIndex:=RawData.YIndex + 1;
end;



procedure Copy24BitData(Source: PRGBTripleArray; RawData: T2DIntArray; FI: TTIFFInfo);
{Change BGR color order to RGB for raw data array}
var X: integer;
begin
for X:=0 to RawData.Cols-1 do
	begin
	RawData[X,RawData.YIndex]:=RGB(Source[X].rgbtRed,Source[X].rgbtGreen,Source[X].rgbtBlue);
	end;
RawData.YIndex:=RawData.YIndex + 1;
end;



procedure DoCMYKToRGBQuad(Source: PCMYKArray; RawData: T2DIntArray; FI: TTIFFInfo);
{Change BGR color order to RGB for Windows BMP}
var X: integer;
var RGBTriple: TRGBTriple;
begin
for X:=0 to RawData.Cols-1 do                           
	begin
	RGBTriple:=CMYKToRGB(Source[X]);
	RawData[X,RawData.YIndex]:=RGB(RGBTriple.rgbtBlue,RGBTriple.rgbtGreen,RGBTriple.rgbtRed);
	end;
RawData.YIndex:=RawData.YIndex + 1;
end;




procedure InvertColorScheme(Source: PByteArray; RawData: T2DIntArray; FI: TTIFFInfo);
{Invert the color scheme of the TIFF data make it compatable with Windows}
var I: integer;
begin
for I:=0 to RawData.Cols-1 do RawData[I,RawData.YIndex]:=255 - Source[I];
RawData.YIndex:=RawData.YIndex + 1;
end;



procedure CopyGrayScale16(Source: PWordArray; RawData: T2DIntArray; FI: TTIFFInfo);
{Copy 16-bit gray scale data into array}
var X: integer;
var T: Word;
begin
for X:=0 to RawData.Cols-1 do
	begin
	RawData[X,RawData.YIndex]:=Source[X];
	end;
RawData.YIndex:=RawData.YIndex + 1;
end;



procedure ConvertColorScheme(var Buffer: Pointer; RawData: T2DIntArray;  FI: TTIFFInfo);
{Convert the color to either RBG or Gray scale}
begin
case FI.Bits of
 1: Copy1BitData(Buffer,RawData,FI);
 4: Copy4BitData(Buffer,RawData,FI);
 8:	begin
	if FI.PhotoMetric=0 then InvertColorScheme(Buffer,RawData,FI)
	else Copy8BitData(Buffer,RawData,FI);
	end;
 16: CopyGrayScale16(Buffer,RawData,FI);
 24: Copy24BitData(Buffer,RawData,FI);
 32:	begin
	if FI.PhotoMetric=Photometric_Separated then DoCMYKToRGBQuad(Buffer,RawData,FI)
	else ReorderRGBQuad(Buffer,RawData,FI);
	end;
 end;
end;

{------------------ LZW Decompression Routines ----------------------------}


var LZWWData,LZWBits,CodeWidth: integer;	{Varibale to decode LZW data}


const ClearCode = 256;		{LZW codes}
const EOICode = 257;
const NullCode = 4096;


function GetLZWCode(Stream:TStream; CodeSize: integer): integer;
{Read one variable-width code from the stream}
{CodeSize holds the number of bits wide the code is}
var CodeMask: integer;
var Atom: byte;
begin
Result:=EOICode;
CodeMask:=(1 shl CodeSize) - 1;
while LZWBits < CodeSize do
	begin
	if Stream.Read(Atom,1)<1 then exit;
	Inc(LZWWData, Cardinal(Atom) shl (24 - LZWBits));
	Inc(LZWBits, 8);
	end;

Result:=(LZWWData and ($FFFFFFFF - CodeMask)) shr (32 - CodeSize);
LZWWData:=LZWWData shl CodeSize;
Dec(LZWBits, CodeSize);
end;



function ReadLZWStrip(Stream: TStream; RawData: T2DIntArray; Lines: integer; FI: TTIFFInfo): boolean;
{Reads and decompresses a single 'Strip' from input file}
{Each 'Strip' can contain multiple scan lines of data }
{Each scan line must be stored in the Windows BMP dword aligned}
var L: integer;
var LZWIndex: DWord;
var OldCode: DWord;
var BackLink: array[0..4095] of DWord;
var DecodeValues: array [0..4095] of Byte;
var Stack: array [0..4095] of Byte;
var StackPointer: PByte;
var FirstChar: Byte;
var Strip: PByteArray;

	procedure InitializeLZWTable;
	{Initialize decode table}
	var I: integer;
	begin
	for I:=0 to ClearCode - 1 do
		begin
		BackLink[I]:=NullCode;
		DecodeValues[I]:=I;
		end;
	end;


	procedure ResetLZW;
	{Reset state of LZW decoder}
	begin
	CodeWidth:=9;
	LZWIndex:=ClearCode + 2;
	OldCode:=NullCode;
	end;



	function DecodeLZW(Code: DWord;var Buffer: PByte): Boolean;
	{Decode current LZW code }
	var InCode: DWord; { buffer for passed code}
	begin
	Result:=False;
	if Code > LZWIndex then raise Exception.Create('Invalid LZW Code');
	
	{ handling for the first LZW code: print and keep it}
	if OldCode = NullCode then
		begin
		FirstChar:=DecodeValues[Code];
		Buffer^:=FirstChar;
		Inc(Buffer);
		OldCode:=Code;
		Result:=True;
		Exit;
		end;

	{ keep the passed LZW code}
	InCode:=Code;

	{ the first LZW code is always smaller than FFirstCode}
	if Code = LZWIndex then
		begin
		StackPointer^:=FirstChar;
		Inc(StackPointer);
		Code:=OldCode;
		end;

	{ loop to put decoded bytes onto the stack}
	while Code > ClearCode do
		begin
		StackPointer^:=DecodeValues[Code];
		Inc(StackPointer);
		Code:=BackLink[Code];
		end;

	{ place new code into code table}
	FirstChar:=DecodeValues[Code];
	StackPointer^:=FirstChar;
	Inc(StackPointer);
	BackLink[LZWIndex]:=OldCode;
	DecodeValues[LZWIndex]:=FirstChar;
	if LZWIndex < 4096 then Inc(LZWIndex);

	{ put decoded bytes (from the stack) into the target buffer}
	OldCode:=InCode;
	repeat
		begin
		Dec(StackPointer);
		Buffer^:=StackPointer^;
		Inc(Buffer);
		end
	until DWord(StackPointer) <= DWord(@Stack);
	Result:=True;
	end;



	function ReadStrip(Stream: TStream; Buffer: Pointer): integer;
	{Read the strip into buffer }
	var Code: integer;
	var Target: Pointer;
	begin
	Target:=Buffer;
	while true do
		begin
		Code:=GetLZWCode(Stream,CodeWidth);
		case Code of
		 ClearCode: ResetLZW;
		 EOICode: break;
		 else DecodeLZW(Code,PByte(Target));
		 end;
		if LZWIndex>=(1 shl CodeWidth)-1 then Inc(CodeWidth);
		end;
	Result:=Cardinal(Target)-Cardinal(Buffer);
	end;

begin
Result:=False;
try
GetMem(Strip,FI.BytesPerRow*Lines);
LZWWData:=0; LZWBits:=0;
InitializeLZWTable;
ResetLZW;
StackPointer:=@Stack;
ReadStrip(Stream,Strip);
for L:=0 to Lines-1 do
	begin
	AdvanceProgress(psRunning,0);
	Move(Strip[L*Fi.BytesPerRow],LIneBuffer^,FI.BytesPerRow);
	ConvertColorScheme(LineBuffer,RawData,FI);
	end;
finally FreeMem(Strip); end;
Result:=True;
end;


{---------------- Decode PackBits Compression -------------------------}

function ReadPackBitsLine(Stream: TStream; Destination: Pointer;BytesOut: Integer): Integer;
{Read single scanline of packbits data into Destination}
var B: Byte;
var P: Pointer;
var RepCount: integer;
begin
P:=Destination;
repeat
	begin
	RepCount:=SignedByteToInteger(GetStreamByte(Stream));
	if RepCount>=0 then
		begin
		{0-128 = Copy N bytes from input}
		Stream.Read(P^,RepCount+1);
		Inc(Cardinal(P),RepCount+1);
		end
	else if RepCount>-128 then
		begin
		{-128 to -1 = Repeat next byte N times}
		B:=GetStreamByte(Stream);
		FillChar(P^,-RepCount+1,B);
		Inc(Cardinal(P),-RepCount+1);
		end
	else ;	{-128 is a NOP }
	Result:=Cardinal(P)-Cardinal(Destination);
	end
until Result >= BytesOut;
end;



function ReadPackBitsStrip(Stream: TStream; RawData: T2DIntArray; Lines: integer; FI: TTIFFInfo): boolean;
{Read all 'PackBits' scanlines in a 'strip' and copy into BMP DWord aligned}
var L: integer;
begin
Result:=False;
for L:=0 to Lines-1 do
	begin
	AdvanceProgress(psRunning,0);     
	if ReadPackBitsLine(Stream,LineBuffer,FI.BytesPerRow) <> FI.BytesPerRow then exit;
	ConvertColorScheme(LineBuffer,RawData,FI);
	end;
Result:=True;
end;

{------------------ Handle Uncompressed data ---------------------------------}



function ReadUncompressedStrip(Stream: TStream; RawData: T2DIntArray; Lines: integer; FI: TTIFFInfo): boolean;
{Read uncompressed scanlines in a 'strip' and copy into BMP DWord aligned}
var L: integer;
begin
Result:=False;
for L:=0 to Lines-1 do
	begin
	AdvanceProgress(psRunning,0);
	if Stream.Read(LineBuffer^,FI.BytesPerRow) <> FI.BytesPerRow then exit;
	ConvertColorScheme(LineBuffer,RawData,FI);
	end;
Result:=True;
end;

{------------------Rountines To Copy Data into Array ---------------------}




function ReadStrip(Stream: TStream; RawData: T2DIntArray; Lines: integer; FI: TTIFFInfo): boolean;
{Read a 'Strip' using appropriate Decompression }
begin
case FI.Compression of
 Compress_None: Result:=ReadUncompressedStrip(Stream,RawData,Lines,FI);
 Compress_LZW: Result:=ReadLZWStrip(Stream,RawData,Lines,FI);
 Compress_PackBits: Result:=ReadPackBitsStrip(Stream,RawData,Lines,FI);
 else UnhandledCompression(FI.Compression);
 end;
end;



function CopyData(Stream: TStream; RawData: T2DIntArray; FI: TTIFFInfo): boolean;
{Copy all 'Strips' into buffer }
var I,C: integer;
begin
Result:=False;
if FI.StripCount=1 then
	begin
	{StripCount=1, the Offset points to actual data}
	AdvanceProgress(psStarting,FI.Depth);
	Stream.Seek(FI.Offset,soFromBeginning);
	if not ReadStrip(Stream,RawData,FI.Depth,FI) then exit;
	end
else
	begin
	{Offset points to table of values }
	AdvanceProgress(psStarting,FI.Depth);
	for I:=0 to FI.StripCount-1 do
		begin
		Stream.Seek(FI.Offset+(I*sizeof(DWord)),soFromBeginning);
		Stream.Seek(GetStreamDWord(Stream),soFromBeginning);
		if (((I+1) * FI.RowsPerStrip)-1) >= FI.Depth then C:=FI.Depth-(I*FI.RowsPerStrip)
		else C:=FI.RowsPerStrip;
		if not ReadStrip(Stream,RawData,C,FI) then exit;
		end;
	end;
Result:=True;
end;


procedure TestValidCompress(Compression: integer);
{Test valid compression and raise exception if not supported}
begin
case Compression of
 Compress_None,Compress_LZW,Compress_PackBits: ;
 else UnhandledCompression(Compression);
 end;
end;



procedure StoreInArray(Stream: TStream; RawData: T2DIntArray; FI: TTIFFInfo);
{Copy Tiff data into Delphi TBitmap Object }
var I,Size,BitsPerPixel: integer;
var BMPBytesPerLine: integer;
begin
if (FI.Width<=0) or (FI.Depth<=0) then raise Exception.Create('Width or Height is Zero');
TestValidCompress(FI.Compression);
RawData.SetDimensions(FI.Width,FI.Depth);
RawData.Clear;

RawData.XIndex:=0; RawData.YIndex:=0;
BitsPerPixel:=FI.Bits;
BMPBytesPerLine:=BytesPerScanline(FI.Width,BitsPerPixel, 32);

GetMem(LineBuffer,BMPBytesPerLine);
try
CopyData(Stream,RawData,FI);
finally FreeMem(LineBuffer); end;
end;




procedure ReadRawTIFStream(Stream: TStream; RawData: T2DIntArray);
{Read TIF file from Stream and copy it into Raw Data array}
{Currently configured to read single images, Add a loop}
{between comments "*Loop*" to read multiple images}
var Loc : DWord;
var I,N : Integer;
begin
TifNumberType:=GetStreamWord(Stream);
if TifNumberType = TifBigEnd then StreamEndian:=BigEnd
else StreamEndian:=LittleEnd;

VerNumber:=GetStreamWord(Stream);
if (TifNumberType<>TifLittleEnd) and (TifNumberType<>TifBigEnd) and (VerNumber<>42) then
	raise Exception.Create('Invalid TIF File');

{ Only read the first image }
{ *LOOP* to read multiple images}
Loc:=GetStreamDWord(Stream);
if Loc=0 then exit;
Stream.Seek(Loc,soFromBeginning);

{ Read TIF information tags }
SetDefaults(TIFInfo);
N:=GetStreamWord(Stream);
for I:=0 to N-1 do DecodeTag(Stream,TIFInfo);

{ save the file position }
Loc:=Stream.Position;

{ figure the bits per pixel}
TIFInfo.Bits:=TIFInfo.Samples * TIFInfo.BitsPerSample;
if TIFInfo.Bits=1 then SetBilevelPalette(TIFInfo.PhotoMetric=0,TIFInfo);

{ figure the bytes }
if TIFInfo.Bits =1 then TIFInfo.BytesPerRow:=MulDiv(TIFInfo.Width,TIFInfo.Bits,8)+1
else if TIFInfo.Bits < 8 then TIFInfo.BytesPerRow:=MulDiv(TIFInfo.Width,TIFInfo.Bits,8)
else if TIFInfo.Bits=8 then TIFInfo.BytesPerRow:=TIFInfo.Width
else if TIFInfo.Bits=16 then TIFInfo.BytesPerRow:=TIFInfo.Width * 2
else if TIFInfo.Bits=24 then TIFInfo.BytesPerRow:=TIFInfo.Width * 3
else if TIFInfo.Bits=32 then TIFInfo.BytesPerRow:=TIFInfo.Width * 4
else raise Exception.Create('Unsupported Bits/Pixel: '+IntToStr(TIFInfo.Bits));;

StoreInArray(Stream,RawData,TIFInfo);
AdvanceProgress(psEnding,0);
{ Restore pointer to next entry }
Stream.Seek(Loc,soFromBeginning);
{ *LOOP* to read multiple images}

OnTIFProgress:=nil;
end;




procedure ReadRawTIFFile(FileName: string; RawData: T2DIntArray);
{Read TIF from file }
var Stream: TMemoryStream;
begin
Stream := TMemoryStream.Create;
Stream.LoadFromFile(Filename);
try
ReadRawTIFStream(Stream, RawData);
finally Stream.Free; end;
end;


{================== Routines to convert RawData to bitmaps ===================}


procedure Make1BitBitmap(RawData: T2DIntArray; Info: TTIFFInfo; BM: TBitmap);
{Convert 1-Bit TIFF data to Windows BMP}
var Line: PRGBQuadArray;
var T: TRGBQuad;
var X,Y: integer;
const White: TRGBQuad = (rgbBlue: 255; rgbGreen: 255; rgbRed: 255; rgbReserved: 0);
const Black: TRGBQuad = (rgbBlue: 0; rgbGreen: 0; rgbRed: 0; rgbReserved: 0);
begin
for Y:=0 to RawData.Rows-1 do
	begin
	Line:=BM.ScanLine[Y];
	for X:=0 to RawData.Cols-1 do
		begin
		if RawData[X,Y]<>0 then Line[X]:=White
		else Line[X]:=Black;
		end;
	end;
end;




procedure Make4BitBitmap(RawData: T2DIntArray; Info: TTIFFInfo; BM: TBitmap);
{Convert 4-Bit TIFF data to Windows BMP}
var Line: PRGBQuadArray;
var T: TRGBQuad;
var X,Y: integer;
begin
for Y:=0 to RawData.Rows-1 do
	begin
	Line:=BM.ScanLine[Y];
	for X:=0 to RawData.Cols-1 do
		begin
		Line[X]:=Info.Palette[RawData[X,Y]];
		end;
	end;
end;






procedure Make8BitBitmap(RawData: T2DIntArray; Info: TTIFFInfo; BM: TBitmap);
{Convert 8-Bit TIFF data to Windows BMP}
var Line: PRGBQuadArray;
var T: TRGBQuad;
var X,Y: integer;
begin
for Y:=0 to RawData.Rows-1 do
	begin
	Line:=BM.ScanLine[Y];
	for X:=0 to RawData.Cols-1 do
		begin
		Line[X]:=Info.Palette[RawData[X,Y]];
		end;
	end;
end;





procedure Make16BitBitmap(RawData: T2DIntArray; Info: TTIFFInfo; BM: TBitmap);
{Convert 16-Bit TIFF data to Windows BMP}
var Line: PRGBQuadArray;
var T: integer;
var X,Y: integer;
begin
for Y:=0 to RawData.Rows-1 do
	begin
	Line:=BM.ScanLine[Y];
	for X:=0 to RawData.Cols-1 do
		begin
		T:=RawData[X,Y] div 257;
		Line[X].rgbRed:=T;
		Line[X].rgbGreen:=T;
		Line[X].rgbBlue:=T;
		Line[X].rgbReserved:=0;
		end;
	end;
end;





procedure Make24BitBitmap(RawData: T2DIntArray; Info: TTIFFInfo; BM: TBitmap);
{Convert 24-Bit TIFF data to Windows BMP}
type TDWordArray = array [0..0] of DWord;
type PDWordArray = ^TDWordArray;
var Line: PDWordArray;
var T: integer;
var X,Y: integer;
begin
for Y:=0 to RawData.Rows-1 do
	begin
	Line:=BM.ScanLine[Y];
	for X:=0 to RawData.Cols-1 do
		begin
		Line[X]:=RawData[X,Y];
		end;
	end;
end;




procedure Make32BitBitmap(RawData: T2DIntArray; Info: TTIFFInfo; BM: TBitmap);
{Convert 24-Bit TIFF data to Windows BMP}
type TDWordArray = array [0..0] of DWord;
type PDWordArray = ^TDWordArray;
var Line: PDWordArray;
var T: integer;
var X,Y: integer;
begin
for Y:=0 to RawData.Rows-1 do
	begin
	Line:=BM.ScanLine[Y];
	for X:=0 to RawData.Cols-1 do
		begin
		Line[X]:=RawData[X,Y];
		end;
	end;
end;


procedure RawDataToBMP(RawData: T2DIntArray; Info: TTIFFInfo; BM: TBitmap);
{Convert RawTIFF data into Windows BMP file}
var X,Y: integer;
begin
BM.Width:=Info.Width;
BM.Height:=Info.Depth;
BM.PixelFormat:=pf32Bit;
case Info.Bits of
 1: Make1BitBitmap(RawData,Info,BM);
 4: Make4BitBitmap(RawData,Info,BM);
 8: Make8BitBitmap(RawData,Info,BM);
 16: Make16BitBitmap(RawData,Info,BM);
 24: Make24BitBitmap(RawData,Info,BM);
 32: Make32BitBitmap(RawData,Info,BM);
 end;
end;


end.
