unit XMLLibrary;

{
Routines for manually reading/writing XML documents
Note: TXMLDocument support routines are in XMLDocLibrary
}


interface

uses Classes,SysUtils,StreamLib,DSupport,HugeArry;

var XMLListParser: TStringList;		{General purpose list for parsing}

type TTagData = record
 Tag,Attributes: string;
 Empty: boolean
 end;


const MaxAttributes = 25;	{Maximum number of items in attribute array}

{attribute structure}

type TXMLAttribute = record
 Name,Value: shortstring;
 NewLine: boolean;
 end;

{Attribute array structure}

type TXMLAttArray = record
 Count: integer;
 Items: array [0..MaxAttributes-1] of TXMLAttribute;
 end;

{For validating XML files}

type TXMLErrorType = (etNoOpenTag,etNoCloseTag);

type TXMLErrorInfo = record
 Tag: string;
 ErrorType: TXMLErrorType;
 LineNumber: integer;
 BeginInx,EndInx: int64;
 end;

type TXMLErrorArray = array of TXMLErrorInfo;



function ExtractXMLTag(S: string; var First,Last: integer; var Tag: string): boolean;
function FindClosingTag(S, Tag: string; var First,Last: integer): boolean;
function GetXMLTagArgs(S: string; var TagStart,ArgStart,ArgEnd,TagEnd: integer; var Tag,Argument: string): boolean;

function ExtractElement(Stream: TStream; var Element: string): boolean;  overload;
function ExtractElement(Stream: TStream): string; overload;
function GetNextTag(Stream: TStream; var TagData: TTagData): boolean;
function GetNextTagNoAttributes(Stream: TStream; var TagData: TTagData): boolean;
procedure StringToIntegerArgs(S: string; var IA: THugeInteger);
procedure StringToDoubleArgs(S: string; DA: THugeDouble);
procedure GetIntegerElementArgs(Stream: TStream; IA: THugeInteger);
procedure GetDoubleElementArgs(Stream: TStream; DA: THugeDouble);
function ExtractXMLInteger(Stream: TStream): integer;
function ExtractXMLDouble(Stream: TStream): double;



procedure XMLNewLine(Stream: TStream);
procedure XMLOpenElementBlock(Stream: TStream; Element: string);
procedure XMLCloseElementBlock(Stream: TStream; Element: string);

procedure XMLWriteCompleteElementAttrib(Stream: TStream; Tag,Data: string);
procedure XMLWriteCompleteElement(Stream: TStream; Tag,Data: string);
procedure XMLWriteClosedElementAttrib(Stream: TStream; Tag: string);

procedure ResetXMLWriter;
procedure PutAttribute(Name,Value: string); overload;
procedure PutAttribute(Name,Value: string; NewLine: boolean); overload;
procedure XMLWriteLine(Stream: TStream; S: string);
procedure XMLWriteIndented(Stream: TStream; S: string);
procedure XMLWriteMultiLineIndent(Stream: TStream; S: string);
procedure XMLWriteIndentEOL(Stream: TStream; S: string);
procedure XMLWritePIDeclaration(Stream: TStream; Element: string);
procedure XMLWriteComment(Stream: TStream; Comment: string);

procedure XMLOpenInternalDocType(Stream: TStream; Root,DTDType,DTDName,DTDURL: string);
procedure XMLWriteInternalEntity(Stream: TStream; Name,Value: string);
procedure XMLCloseInternalDocType(Stream: TStream);

function AddEscapeChars(S: string): string;
function RemoveEscapeChars(S: string): string;


function XMLErrorToStr(ET: TXMLErrorType): string;
procedure ValidateXMLFile(Stream: TStream; var ErrorArray: TXMLErrorArray);


implementation



function AddEscapeChars(S: string): string;
var I: integer;
begin
Result:='';
for I:=1 to Length(S) do
	begin
	case S[I] of
	 CR_Char: Result:=Result + '&#10;';
	 LF_Char: Result:=Result + '&#13;';
	 '"': Result:=Result + '&quot;';
	 '''': Result:=Result + '&apos;';
	 '<': Result:=Result + '&lt;';
	 '>': Result:=Result + '&gt;';
	 '&': Result:=Result + '&amp;';
	 else Result:=Result + S[I];
	 end;
	end;
end;



function RemoveEscapeChars(S: string): string;
var I: integer;
var S1: string;
var Escape: boolean;
begin
Result:='';
Escape:=False;
for I:=1 to Length(S) do
	begin
	if S[I]='&' then
		begin
		Escape:=True;
		S1:='';
		end
	else if Escape then
		begin
		if S[I]=';' then
			begin
			Escape:=False;
			if S1='quot' then Result:=Result + '"'
			else if S1='apos' then Result:=Result + ''''
			else if S1='lt' then Result:=Result + '<'
			else if S1='gt' then Result:=Result + '>'
			else if S1='amp' then Result:=Result + '&'
			else if S1[1]='#' then
				begin
				S1[1]:='$';
				Result:=Result+Char(StrToIntDef(S1,$20));
				end;
			end
		else S1:=S1 + S[I];
		end
	else Result:=Result + S[I];
	end;
end;

{---------------- Routines to Parse XML tags ------------------------------------}

function ExtractXMLTag(S: string; var First,Last: integer; var Tag: string): boolean;
{Extract an XML starting at First. First and Last point to Each angle bracket }
begin
Result:=False;
First:=FindString('<',S,First);
if First=0 then exit;
Last:=FindString('>',S,First);
if Last=0 then exit;

Tag:=Copy(S,First+1,Last - First-1);
Result:=True;
end;



function FindClosingTag(S, Tag: string; var First,Last: integer): boolean;
{Find closing tag that matches specified Tag starting at First}
begin
Tag:='</'+Tag+'>';
First:=FindString(Tag,S,First);
Result:=First>0;
if Result then Last:=First+Length(Tag);
end;



function GetXMLTagArgs(S: string; var TagStart,ArgStart,ArgEnd,TagEnd: integer; var Tag,Argument: string): boolean;
{Extract tag and arguments starting TagStart}
begin
Result:=False;
if not ExtractXMLTag(S,TagStart,ArgStart,Tag) then exit;
if Tag[Length(Tag)]='\' then
	begin
	Result:=True;
	Argument:=Tag;
	exit;
	end;
Inc(ArgStart);
ArgEnd:=ArgStart;
Result:=FindClosingTag(S,Tag,ArgEnd,TagEnd);
Argument:=Copy(S,ArgStart,ArgEnd - ArgStart);
end;

{---------------- Routines to read XML Tag Data --------------------------}


function GetNextTagNoAttributes(Stream: TStream; var TagData: TTagData): boolean;
{Get tag, ignore attributes for speed}
var C: char;
begin
Result:=False; TagData.Empty:=False;
TagData.Tag:=''; TagData.Attributes:='';
{Look for tag start }
repeat if Stream.Read(C,1)<1 then exit until C='<';
{Read Tag name}
while (Stream.Read(C,1)>0) and not(C in ['>',' ']) do TagData.Tag:=TagData.Tag + C;
{skip attributes, read to final '>'}
if C=' ' then
	begin
	while (Stream.Read(C,1)>0) and (C<>'>') do;
	TagData.Empty:=LookBackStreamChar(Stream,2)='/';
	end;
Result:=True;
end;



function GetNextTag(Stream: TStream; var TagData: TTagData): boolean;
var C: char;
begin
Result:=False; TagData.Empty:=False;
TagData.Tag:=''; TagData.Attributes:='';
{Look for tag start }
repeat if Stream.Read(C,1)<1 then exit until C='<';
{Read Tag name}
while (Stream.Read(C,1)>0) and not(C in ['>',' ']) do TagData.Tag:=TagData.Tag + C;
{Read attribute strings, if any}
if C=' ' then
	begin
	while (Stream.Read(C,1)>0) do
		begin
		if (C='>') or 
		   ((C='/') and (PeekStreamChar(Stream)='>')) then break;
		TagData.Attributes:=TagData.Attributes + C;
		end;
	TagData.Empty:=C='/';
	end;
Result:=True;
end;



function ExtractElement(Stream: TStream; var Element: string): boolean;  overload;
{Extract element strings, assumes we have already passed starting tag}
var C: char;
begin
Result:=False;
Element:='';
{extract the the element string}
while (Stream.Read(C,1)>0) and (C<>'<') do Element:=Element + C;
{move past closing tag}
while (Stream.Read(C,1)>0) and (C<>'>') do;
Result:=True;
end;


function ExtractElement(Stream: TStream): string; overload;
begin
ExtractElement(Stream,Result);
end;


{----------------- Extract specific argument types ---------------------------}


procedure StringToIntegerArgs(S: string; var IA: THugeInteger);
{Extract integer arguments from string}
var I: integer;
begin
XMLListParser.Clear;
XMLListParser.CommaText:=S;
IA.Clear;
for I:=0 to XMLListParser.Count-1 do IA.Add(StrToInt(XMLListParser[I]));
end;




procedure StringToDoubleArgs(S: string; DA: THugeDouble);
{Extract integer arguments from string}
var I: integer;
begin
DA.Clear;
XMLListParser.Clear;
XMLListParser.CommaText:=S;
for I:=0 to XMLListParser.Count-1 do DA.Add(StrToFloat(XMLListParser[I]));
end;



procedure GetIntegerElementArgs(Stream: TStream; IA: THugeInteger);
var S: string;
begin
S:=ExtractElement(Stream);
StringToIntegerArgs(S,IA);
end;



procedure GetDoubleElementArgs(Stream: TStream; DA: THugeDouble);
var S: string;
begin
S:=ExtractElement(Stream);
StringToDoubleArgs(S,DA);
end;



function ExtractXMLInteger(Stream: TStream): integer;
begin
Result:=StrToInt(ExtractElement(Stream));
end;



function ExtractXMLDouble(Stream: TStream): double;
begin
Result:=StrToFloat(ExtractElement(Stream));
end;


{----------------- Routines to Write XML Files ------------------------------}

{Indent Parameters}

const TabIncrement = 4;		{Indent for file}
var CurrentIndent: integer;	{Current number of tabs we are indented}


{--------------- Low level I/o Routines ---------------------------------------}


procedure XMLNewLine(Stream: TStream);
{Write CRLF to XML stream}
begin
WriteStreamString(Stream,CRLF_Char);
end;


procedure XMLWriteLine(Stream: TStream; S: string);
{Write string and EOL to XML stream}
begin
WriteStreamString(Stream,S);
XMLNewLine(Stream);
end;


function GetIndentStr(Extra: integer): string;
{Return indent string plus any Extra}
var I: integer;
begin
Result:='';
for I:=1 to (CurrentIndent+Extra) * TabIncrement do Result:=' '+Result;
end;


procedure XMLIndent(Stream: TStream);
{Write indented to XML stream}
var S: string;
begin
S:=GetIndentStr(0);
WriteStreamString(Stream,S);
end;


procedure XMLWriteIndented(Stream: TStream; S: string);
{Write indented string to XML stream}
begin
XMLIndent(Stream);
WriteStreamString(Stream,S);
end;




procedure XMLWriteIndentEOL(Stream: TStream; S: string);
{Write indented string and EOL to XML stream}
begin
XMLWriteIndented(Stream,S);
XMLNewLine(Stream);
end;


procedure XMLWriteMultiLineIndent(Stream: TStream; S: string);
{Write each line in string indented}
var SL: TStringList;
var I: integer;
begin
SL:=TStringList.Create;
try
SL.Text:=S;
for I:=0 to SL.Count-1 do
  XMLWriteIndentEOL(Stream,SL[I]);
finally SL.Free; end;
end;



{----------------Routines to handle Attributes ---------------------------------}

var Attributes: TXMLAttArray;	{Global array of attributes}

function GetAttributeStr: string;
{Convert attribute list to array}
var I: integer;
begin
Result:='';
for I:=0 to Attributes.Count-1 do
	begin
	if Attributes.Items[I].NewLine then Result:=Result+CRLF_Char+GetIndentStr(1);
	Result:=Result + ' '+Attributes.Items[I].Name+'="'+Attributes.Items[I].Value+'"';
	end;
end;



procedure PutAttribute(Name,Value: string; NewLine: boolean);
{Add one attribute to list}
begin
if Attributes.Count>=MaxAttributes then exit;
Attributes.Items[Attributes.Count].Name:=Name;
Attributes.Items[Attributes.Count].Value:=Value;
Attributes.Items[Attributes.Count].NewLine:=NewLine;

Inc(Attributes.Count);
end;


procedure PutAttribute(Name,Value: string);
{Add one attribute to list}
begin
PutAttribute(Name,Value,False);
end;

{----------------------------------------------------------------------------}

procedure SetIndent(Increment: boolean);
begin
if Increment then Inc(CurrentIndent)
else Dec(CurrentIndent);
if CurrentIndent<0 then CurrentIndent:=0;
end;


procedure XMLOpenElementAttributes(Stream: TStream; Element: string; Closed,NewLine: boolean);
{Open element with attributes and increment indent}
var S: string;
begin
S:='<'+Element+GetAttributeStr;
if Closed then S:=S + '/>' else S:=S + '>';
XMLWriteIndented(Stream,S);
if NewLine then XMLNewLine(Stream);
Attributes.Count:=0;
end;



procedure XMLOpenElement(Stream: TStream; Element: string; Closed,NewLine: boolean);
{Open element and increment indent}
begin
Attributes.Count:=0;
XMLOpenElementAttributes(Stream,Element,Closed,NewLine);
end;



procedure XMLCloseElement(Stream: TStream; Element: string);
{Write ending bracket and decrement indent}
begin
WriteStreamString(Stream,'</'+Element+'>');
XMLNewLine(Stream);
end;



procedure XMLCloseElementIndent(Stream: TStream; Element: string);
{Write ending bracket and decrement indent}
begin
XMLIndent(Stream);
XMLCloseElement(Stream,Element);
end;



procedure XMLOpenElementBlock(Stream: TStream; Element: string);
begin
XMLOpenElementAttributes(Stream,Element,False,True);
SetIndent(True);
end;



procedure XMLCloseElementBlock(Stream: TStream; Element: string);
begin
SetIndent(False);
XMLCloseElementIndent(Stream,Element);
end;



procedure ResetXMLWriter;
begin
CurrentIndent:=0;
Attributes.Count:=0;
end;

{-----------------------------------------------------------------------------}



procedure XMLWriteClosedElementAttrib(Stream: TStream; Tag: string);
{Write closed element like: <tag attributes /> with attributes}
begin
XMLOpenElementAttributes(Stream,Tag,True,True);
end;



procedure XMLWriteCompleteElementAttrib(Stream: TStream; Tag,Data: string);
{Write complete element like: <tag>data</tag> with attributes}
begin
XMLOpenElementAttributes(Stream,Tag,False,False);
WriteStreamString(Stream,Data);
XMLCloseElement(Stream,Tag);
end;


procedure XMLWriteCompleteElement(Stream: TStream; Tag,Data: string);
{Write complete element like: <tag>data</tag>}
begin
XMLOpenElement(Stream,Tag,False,False);
WriteStreamString(Stream,Data);
XMLCloseElement(Stream,Tag);
end;


procedure XMLWritePIDeclaration(Stream: TStream; Element: string);
{Write Processing Instructions Declaration}
var S: string;
begin
S:='<?'+Element+GetAttributeStr+ '?>';
XMLWriteIndented(Stream,S);
Attributes.Count:=0;
XMLNewLine(Stream);
end;




procedure XMLWriteComment(Stream: TStream; Comment: string);
var S: string;
begin
S:='<!--'+Comment+'-->';
XMLWriteIndented(Stream,S);
XMLNewLine(Stream);
end;


procedure XMLOpenInternalDocType(Stream: TStream; Root,DTDType,DTDName,DTDURL: string);
{Open DOCTYPE Element}
var S: string;
begin
S:='<!DOCTYPE '+Root;
if DTDType<>'' then S:=S+' '+DTDType;
if DTDName<>'' then S:=S+' "'+DTDName+'"';
if DTDURL<>'' then S:=S+' "'+DTDURL+'"';
S:=S+' [';
XMLWriteIndented(Stream,S);
XMLNewLine(Stream);
SetIndent(True);
end;



procedure XMLCloseInternalDocType(Stream: TStream);
begin
XMLWriteIndented(Stream,']>');
XMLNewLine(Stream);
SetIndent(False);
end;



procedure XMLWriteInternalEntity(Stream: TStream; Name,Value: string);
{Write  Internal type !ENTITY Element}
var S: string;
begin
S:='<!ENTITY '+Name+' '+'"'+Value+'">';
XMLWriteIndented(Stream,S);
XMLNewLine(Stream);
end;


{------------------- Routines to validate XMPL -------------------------------}




function XMLErrorToStr(ET: TXMLErrorType): string;
begin
case ET of
 etNoOpenTag: Result:='No Open Tag';
 etNoCloseTag: Result:='No Close Tag';
 else Result:='Unknown';
 end;
end;



procedure ValidateXMLFile(Stream: TStream; var ErrorArray: TXMLErrorArray);
var TagData: TTagData;
var SL: TStringList;
var S: string;
var Inx1,Inx2: int64;
var LineCount: integer;



	function ReadChar(Stream: TStream; var C: char): boolean;
	begin
	Result:=Stream.Read(C,1)>0;
	if C=LF_Char then Inc(LineCount);
	end;


function IsTagTerminator(C: char): boolean;
begin

end;


	function FindNextTag(Stream: TStream; var TagData: TTagData; var Inx1,Inx2: int64): boolean;
	{Get tag, ignore attributes for speed}
	var C: char;
	const Terminator = [#0..#$20,'>'];
	begin
	Result:=False; TagData.Empty:=False;
	TagData.Tag:=''; TagData.Attributes:='';
	{Look for tag start }
	repeat if not ReadChar(Stream,C) then exit until C='<';
	Inx1:=Stream.Position-1;
	{Read Tag name}
	while ReadChar(Stream,C) and not(C in Terminator) do TagData.Tag:=TagData.Tag + C;
	{skip attributes, read to final '>'}
	if C<>'>' then
		begin
		while ReadChar(Stream,C) and (C<>'>') do;
		end;
		TagData.Empty:=LookBackStreamChar(Stream,2)='/';
	Inx2:=Stream.Position;
	Result:=True;
	end;




begin
LineCount:=1;
SetLength(ErrorArray,0);
SL:=TStringList.Create;
try
Stream.Seek(0,soBeginning);
while FindNextTag(Stream,TagData,Inx1,Inx2) do
	begin
	if TagData.Empty then continue;
	if TagData.Tag[1] in ['?','!'] then continue;
	if TagData.Tag[1]='/' then
		begin
		S:=Copy(TagData.Tag,2,Length(TagData.Tag)-1);
		if (SL.Count<1) or (S<>SL[SL.Count-1]) then
			begin
			SetLength(ErrorArray,Length(ErrorArray)+1);
			ErrorArray[High(ErrorArray)].LineNumber:=LineCount;
			ErrorArray[High(ErrorArray)].BeginInx:=Inx1;
			ErrorArray[High(ErrorArray)].EndInx:=Inx2;
			if (SL.Count>1) and (S=SL[SL.Count-2]) then
				begin
				ErrorArray[High(ErrorArray)].Tag:=SL[SL.Count-1];
				ErrorArray[High(ErrorArray)].ErrorType:=etNoCloseTag;
				if SL.Count>0 then SL.Delete(SL.Count-1);
				if SL.Count>0 then SL.Delete(SL.Count-1);
				end
			else
				begin
				ErrorArray[High(ErrorArray)].Tag:=TagData.Tag;
				ErrorArray[High(ErrorArray)].ErrorType:=etNoOpenTag;
				end;
			end
		else if SL.Count>0 then SL.Delete(SL.Count-1);
		end
	else SL.Add(TagData.Tag);
	end;
finally SL.Free; end;
end;






initialization
XMLListParser:=TStringList.Create;
finalization
XMLListParser.Free;
end.
