unit ChmHelper;

{ Object solves the problem that D6 doesn't understand CHM files. Normally, you would just
 overrider the handler, but D6 doesn't route all the help calls through the handler.
 As a result, we have to create our own Viewer and then trick Delphi into using it.
 To do that, we have to use the built in tester that is a part of the help system }

interface

uses Windows, Messages, SysUtils, Types, Classes, Forms, HelpIntfs, WinHelpViewer;

{HtmlHelp Commands }

const
  HH_DISPLAY_TOPIC        = $0000;
  HH_HELP_FINDER          = $0000;  { WinHelp equivalent}
  HH_DISPLAY_TOC          = $0001;
  HH_DISPLAY_INDEX        = $0002;
  HH_DISPLAY_SEARCH       = $0003;
  HH_SET_WIN_TYPE         = $0004;
  HH_GET_WIN_TYPE         = $0005;
  HH_GET_WIN_HANDLE       = $0006;
  HH_ENUM_INFO_TYPE       = $0007;  { Get Info type name, call repeatedly to enumerate, -1 at end}
  HH_SET_INFO_TYPE        = $0008;  { Add Info type to filter.}
  HH_SYNC                 = $0009;
  HH_RESERVED1            = $000A;
  HH_RESERVED2            = $000B;
  HH_RESERVED3            = $000C;
  HH_KEYWORD_LOOKUP       = $000D;
  HH_DISPLAY_TEXT_POPUP   = $000E;  { display string resource id or text in a popup window}
  HH_HELP_CONTEXT         = $000F;  { display mapped numeric value in dwData}
  HH_TP_HELP_CONTEXTMENU  = $0010;  { text popup help, same as WinHelp HELP_CONTEXTMENU}
  HH_TP_HELP_WM_HELP      = $0011;  { text popup help, same as WinHelp HELP_WM_HELP}
  HH_CLOSE_ALL            = $0012;  { close all windows opened directly or indirectly by the caller}
  HH_ALINK_LOOKUP         = $0013;  { ALink version of HH_KEYWORD_LOOKUP}
  HH_GET_LAST_ERROR       = $0014;  { not currently implemented // See HHERROR.h}
  HH_ENUM_CATEGORY        = $0015;	{ Get category name, call repeatedly to enumerate, -1 at end}
  HH_ENUM_CATEGORY_IT     = $0016;  { Get category info type members, call repeatedly to enumerate, -1 at end}
  HH_RESET_IT_FILTER      = $0017;  { Clear the info type filter of all info types.}
  HH_SET_INCLUSIVE_FILTER = $0018;  { set inclusive filtering method for untyped topics to be included in display}
  HH_SET_EXCLUSIVE_FILTER = $0019;  { set exclusive filtering method for untyped topics to be excluded from display}
  HH_INITIALIZE           = $001C;  { Initializes the help system.}
  HH_UNINITIALIZE         = $001D;  { Uninitializes the help system.}
  HH_PRETRANSLATEMESSAGE  = $00FD;  { Pumps messages. (NULL, NULL, MSG*).}
  HH_SET_GLOBAL_PROPERTY  = $00FC;  { Set a global property. (NULL, NULL, HH_GPROP)}

type
  HH_AKLINK = record
    cbStruct: Integer;
    fReserved: BOOL;
    pszKeywords: PChar;
    pszURL: PChar;
    pszMsgText: PChar;
    pszMsgTitle: PChar;
    pszWindow: PChar;
    fIndexOnFail: BOOL;
  end;

{$IF CompilerVersion<22}
type THelpEventData = integer;
{$IFEND}

{---------- Object that creates our own Help Viewer  --------------------------}

{$IF CompilerVersion < 17}
 THelpContext = integer;
{$IFEND}



type THTMLHelpViewer = class(TInterfacedObject, ICustomHelpViewer, IExtendedHelpViewer)
 private
  FViewerID: Integer;
  FHelpManager: IHelpManager;
  function HTMLHelpFileAvailable: Boolean;
  {$IFDEF VER140}
  procedure InternalShutDown;
  {$ENDIF}
 public
  { ICustomHelpViewer }
  function GetViewerName: string;
  function UnderstandsKeyword(const HelpString: String): Integer;
  function GetHelpStrings(const HelpString: String): TStringList;
  function CanShowTableOfContents: Boolean;
  procedure ShowHelp(const HelpString: String);
  procedure ShowTableOfContents;
  procedure NotifyID(const ViewerID: Integer);
  procedure SoftShutDown;
  procedure ShutDown;
  property HelpManager: IHelpManager read FHelpManager write FHelpManager;
  property ViewerID: Integer read FViewerID;
 public
  { IExtendedHelpViewer }
  function UnderstandsTopic(const Topic: String): Boolean;
  procedure DisplayTopic(const Topic: String);
  function UnderstandsContext(const ContextID: THelpContext; const HelpFileName: String): Boolean;
  procedure DisplayHelpByContext(const ContextID: THelpContext; const HelpFileName: String);
 end;



{---- Object to create our own tester. Allows our viewer to take presedence ---}


type  TWinHelpMakeBehave = class(TInterfacedObject, IWinHelpTester)
 public
  function CanShowALink(const ALink, FileName: string): Boolean;
  function CanShowTopic(const Topic, FileName: string): Boolean;
  function CanShowContext(const Context: THelpContext; const FileName: string): Boolean;
  function GetHelpStrings(const ALink: string): TStringList;
  function GetHelpPath: string;
  function GetDefaultHelpFile: string;
  end;


{--------------- Object encapsulate everything --------------------------------}


type TCHMHelper = class(TComponent)
 private
 public
  HelpViewer: THTMLHelpViewer;
  function HelpTranslator(Command: Word; Data: THelpEventData; var CallHelp: Boolean): Boolean;
  procedure LookupKeyword(Text: string);
  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;
 published
 end;


function HtmlHelp(hwndCaller: THandle; pszFile: PChar; uCommand: cardinal; dwData: longint): THandle; stdcall;


function HtmlHelp(hwndCaller: THandle; pszFile: PChar; uCommand: cardinal; dwData: longint): THandle; stdcall;
         external 'hhctrl.ocx' name 'HtmlHelpA';


procedure Register;


implementation



procedure Register;
begin
RegisterComponents('Fountain',[TCHMHelper]);
end;


{ THTMLHelpViewer }


function THTMLHelpViewer.CanShowTableOfContents: Boolean;
begin
Result:= HTMLHelpFileAvailable;
end;


function THTMLHelpViewer.GetHelpStrings(const HelpString: String): TStringList;
{ We cannot query the HTML Help API to get a list of topic titles.}
{ So we just return HelpString if a .chm file is available.}
begin
Result:= TStringList.Create;
if HTMLHelpFileAvailable then Result.Add(HelpString);
Assert(Result <> nil, 'ENSURE: GetHelpStrings must always return a valid string list');
end;


function THTMLHelpViewer.GetViewerName: String;
begin
Result:= 'HTMLHelp Viewer';
end;



function THTMLHelpViewer.HTMLHelpFileAvailable: Boolean;
begin
Result:= SameText(ExtractFileExt(FHelpManager.GetHelpFile), '.chm');
end;



procedure THTMLHelpViewer.NotifyID(const ViewerID: Integer);
begin
FViewerID:= ViewerID;
end;



procedure THTMLHelpViewer.ShowHelp(const HelpString: String);
var
AKLink: HH_AKLink;
begin
AKLink.cbStruct:= SizeOf(AKLink);
AKLink.fReserved:= False;
AKLink.pszKeywords:= PChar(HelpString);
AKLink.pszURL:= nil;
AKLink.pszMsgText:= nil;
AKLink.pszMsgTitle:= nil;
AKLink.pszWindow:= nil;
AKLink.fIndexOnFail:= True;
HTMLHelp(HelpManager.GetHandle, PChar(HelpManager.GetHelpFile), HH_KEYWORD_LOOKUP, Integer(@AKLink));
end;



procedure THTMLHelpViewer.ShowTableOfContents;
begin
HTMLHelp(HelpManager.GetHandle, PChar(HelpManager.GetHelpFile), HH_DISPLAY_TOC, 0);
end;



procedure THTMLHelpViewer.ShutDown;
begin
HTMLHelp(0, nil, HH_CLOSE_ALL, 0);
FHelpManager:= nil;
end;



procedure THTMLHelpViewer.SoftShutDown;
begin
HTMLHelp(0, nil, HH_CLOSE_ALL, 0);
end;



function THTMLHelpViewer.UnderstandsKeyword(const HelpString: String): Integer;
{ We can't check .chm files for keywords so we handle all requests unless there is no help installed.  we assume it does, if a .chm file is available at all}
begin
if HTMLHelpFileAvailable then Result:= 1 else Result:= 0
end;



procedure THTMLHelpViewer.DisplayHelpByContext(const ContextID: THelpContext; const HelpFileName: String);
begin
HTMLHelp(HelpManager.GetHandle, PChar(HelpFileName), HH_HELP_CONTEXT, ContextID);
end;



procedure THTMLHelpViewer.DisplayTopic(const Topic: String);
var  URL: string;
begin
if Topic = '' then URL:= ''
else if Topic[1] = '/' then URL:= '::' + Topic
else URL:= '::/' + Topic;
HTMLHelp(HelpManager.GetHandle, PChar(HelpManager.GetHelpFile + URL), HH_DISPLAY_TOPIC, 0);
end;



function THTMLHelpViewer.UnderstandsContext(const ContextID: THelpContext; const HelpFileName: String): Boolean;
{ It is not possible to check if the given context ID is mapped to a topic in the .chm file}
{ So we assume it does, if a .chm file is available at all}
begin
Result:= HTMLHelpFileAvailable;
end;



function THTMLHelpViewer.UnderstandsTopic(const Topic: String): Boolean;
{ It is not possible to check if .chm file contains a topic with the given file name}
{ So we assume it does, if a .chm file is available at all}
begin
Result:= HTMLHelpFileAvailable;
end;



{$IFDEF VER140}
procedure THTMLHelpViewer.InternalShutDown;
begin
SoftShutDown;
if Assigned(FHelpManager) then FHelpManager.Release(FViewerID);
end;
{$ENDIF}


{ TWinHelpMakeBehave }

function TWinHelpMakeBehave.CanShowALink(const ALink, FileName: string): Boolean;
begin
Result:= SameText(ExtractFileExt(FileName), '.hlp');
end;


function TWinHelpMakeBehave.CanShowContext(const Context: THelpContext; const FileName: string): Boolean;
begin
Result:= SameText(ExtractFileExt(FileName), '.hlp');
end;



function TWinHelpMakeBehave.CanShowTopic(const Topic, FileName: string): Boolean;
begin
Result:= SameText(ExtractFileExt(FileName), '.hlp');
end;



function TWinHelpMakeBehave.GetDefaultHelpFile: string;
begin
Result:= '';
end;



function TWinHelpMakeBehave.GetHelpPath: string;
begin
Result:= '';
end;



function TWinHelpMakeBehave.GetHelpStrings(const ALink: string): TStringList;
begin
Result:= TStringList.Create;
Assert(Result <> nil, 'ENSURE: GetHelpStrings must always return a valid string list');
end;




{============================================================================}

{ TCHMHelper }




function TCHMHelper.HelpTranslator(Command: Word; Data: THelpEventData; var CallHelp: Boolean): Boolean;
{translates WinHelp commands to HTMLHelp commands + calls API}
var HCommand: word;
begin
{make sure VCL doesn't activate WinHelp + function succeeds}
CallHelp:= false;
result:= true;
{translate WinHelp > HTMLHelp}
case Command of
 HELP_CONTENTS:
	begin
	HCommand:=  HH_DISPLAY_TOC;
	Data:= 0;
	end;
 HELP_CONTEXT: HCommand:= HH_HELP_CONTEXT;
 HELP_CONTEXTPOPUP: HCommand:= HH_HELP_CONTEXT;
 HELP_FINDER: HCommand:= HH_DISPLAY_TOPIC;
 HELP_KEY: HCommand:= HH_DISPLAY_INDEX;
 HELP_QUIT:
	begin
	HCommand:=  HH_CLOSE_ALL;
	Data:= 0;
	end;
 else
	begin
	HCommand:= HH_DISPLAY_TOPIC;
	Data:= 0;
	end;
 end;
{call HTML Help API}
HtmlHelp(Screen.ActiveControl.Handle, PChar(Application.HelpFile), HCommand, Data);
end;


procedure TCHMHelper.LookupKeyword(Text: string);
var Link: HH_AKLINK;
begin
link.cbStruct:=sizeof(HH_AKLINK);
link.fReserved:=False;
link.pszKeywords:=PChar(Text);
link.pszUrl:=nil;
link.pszMsgText:=nil;
link.pszMsgTitle:=nil;
link.pszWindow:=nil;
link.fIndexOnFail:=True;
HtmlHelp(0, PChar(Application.HelpFile), HH_KEYWORD_LOOKUP, Integer(@Link)) ;
end;




constructor TCHMHelper.Create(AOwner: TComponent);
begin
inherited;
HelpViewer:=THTMLHelpViewer.Create;
if not (csDesigning in ComponentState) then
	begin
	WinHelpTester:= TWinHelpMakeBehave.Create;
	HelpIntfs.RegisterViewer(HelpViewer, HelpViewer.FHelpManager);
	Application.OnHelp:=HelpTranslator;
	end;
end;



destructor TCHMHelper.Destroy;
begin
{$IFDEF VER140}
Application.OnHelp:= nil;
if Assigned(HelpViewer) then HelpViewer.InternalShutDown;
{$ENDIF}
inherited;
end;




end.

