unit ConsoleWindow;

interface

uses
  Forms,Windows,Types, Graphics, Messages, SysUtils, Classes, Controls, ExtCtrls,
  BitmapLib,DSupport,Math,Vector,IntVector,Contnrs,ImageLib, GLViewer;

{Mouse State of Console}

type TMouseState = record
 Pos: TPoint;
 Left,Right,Middle: boolean;
 end;


type TCopyMode = record
 Mode: integer;
 Alpha: byte;
 PerPixel: boolean;
 TransColor: TColor;
 end;

{Graphics Mode Structures, constants and variables }

type TCardType = (ctCGA,ctPCJR,ctEGA,ctVGA,ctVesa);

type TCharCell = (cc8x8,cc8x14,cc8x16,cc9x16);

type TModeType = (mtText,mtGraphics);

type TFontType = (ftNormal,ftCGAFont,ftEGAFont,ftVGAFont);

type TGraphicsMode = record
 Mode: Word;
 Card: TCardType;
 Res: TPoint;
 TextRes: TPoint;
 CharCell: TCharCell;
 Bits: integer;
 ModeType: TModeType;
 AspRatio: double;
 end;

type TWindowMode = (wmScroll,wmWrap,wmClip);


type TBitmapManager = class(TObject)
 private
  OldPen: TPen;
  OldBrush: TBrush;
  function GetBitmapByHandle(Handle: integer): TBitmap;
  procedure CopyTransparent(DestCanvas: TCanvas; DestRect: TRect; SourceCanvas: TCanvas; SourceRect: TRect; C: TColor);
  procedure RestoreContext(Canvas: TCanvas);
  procedure SaveContext(Canvas: TCanvas);
 protected
 public
  BitmapList: TObjectList;
  Screen: TBitmap;
  CopyMode: TCopyMode;
  constructor Create(Display: TBitmap);
  destructor Destroy; override;
  function Allocate: integer;
  procedure DeAllocate(Index: integer);
  procedure SetBitmapSize(Handle, Width, Height: integer);
  procedure GetBitmapSize(Handle: integer; var Width, Height: integer);
  procedure LoadBitmap(Handle: integer; Filename: string);
  procedure SaveBitmap(Handle: integer; Filename: string);
  procedure CopyImage(DestHandle, SourceHandle: integer; DestRect, SourceRect: TRect); overload;
  procedure CopyImage(DestBM,SourceBM: TBitmap; DestRect,SourceRect: TRect); overload;
  procedure DrawImage(DestHandle, SourceHandle: integer; X, Y: integer);
  procedure SetPixel(Handle, X, Y: integer; C: TColor);
  function GetPixel(Handle, X, Y: integer): TColor;
  procedure ClearImage(Handle: integer; Color: TColor);
  procedure DrawCircle(Handle, X, Y, Radius: integer);
  procedure DrawEllipse(Handle, X1, Y1, X2, Y2: integer);
  procedure DrawRectangle(Handle, X1, Y1, X2, Y2: integer);
  procedure Lineto(Handle, X, Y: integer);
  procedure MoveTo(Handle, X, Y: integer);
  procedure Polyline(Handle, Mode: integer; const Points: array of TPoint);
  function GetPenPosition(Handle: integer): TPoint;
  procedure SetFill(Handle, Mode: integer; C: TColor);
  procedure SetPen(Handle,Mode,Width,Pattern: integer; C: TColor);
  procedure DoTextOut(Handle, X, Y: integer; S: string);
  procedure SetFont(Handle: integer; Name: string);
  procedure SetFontStyle(Handle, Size: integer; Bold, Italic, Underline, Strike: boolean);
  procedure SetFontColor(Handle: integer; R, G, B: byte);
  procedure GetFontProperties(Handle: integer; S: string; var Width,Height: integer);
  procedure CopyFont(DestHandle, SourceHandle: integer);
 end;



type TVGAWindow = class(TImage)
 private
  FGutterX,FGutterY: integer;
  FColor: TColor;
  FColumns: integer;
  FRows: integer;
  FAttribute: byte;
  CharSize: TPoint;
  CursorTimer: TTimer;
  FShowCursor: boolean;
  FOnWindowChange: TNotifyEvent;
  FCharSpacing: TPoint;
  FExpandTabs: boolean;
  CursorDrawn: boolean;
  OldCursorLoc: TPoint;
  CursorBMP: TBitmap;
  FUseAttribute: boolean;
  OldPenPos: TPoint;
  CharRect: TRect;
  FBlockPaint: boolean;
  procedure SetColumns(const Value: integer);
  procedure SetRows(const Value: integer);
  procedure BlankLine(Line: integer);
  procedure ScrollWindow;
  function GetScreenHeight: integer;
  function GetScreenWidth: integer;
  procedure DrawChar(X, Y: integer; CA: word);
  procedure HandleFontChange(Sender: TObject);
  procedure ClearTextArray;
  function CursorToPixel(X, Y: integer): TPoint;
  procedure HandleTimer(Sender: TObject);
  procedure EnableCursor(Enable: boolean);
  procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
  procedure SetShowCursor(const Value: boolean);
  function FocusOwner: boolean;
  procedure SetCursorRangeCheck(X, Y: integer);
  procedure WindowChange;
  procedure DeleteChar;
  procedure SetCharSpacing(const Value: TPoint);
  procedure EraseBackground(var Msg: TWMEraseBkgnd); message WM_ERASEBKGND;
  procedure OutputCursor(X, Y: integer; C: TColor);
  procedure EraseCursor;
  procedure ClearTextBlock(Block: TRect);
  procedure RangeCheckCharPos(var X, Y: integer);
  procedure RangeCheckBlock(var Block: TRect);
  procedure SetColorPalette(P: array of TColor);
  function GetCharWidth(S: string): integer;
 protected
  procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X,Y: Integer); override;
  procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
  procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  procedure Paint; override;
 public
  CellSize: TPoint;
  TextArray: array of array of word;
  CursorX,CursorY: integer;
  BitmapManager: TBitmapManager;
  ColorPalette: array [0..255] of TColor;
  WinMode: TWindowMode;
  procedure SetPaletteItem(Inx: integer; C: TColor);
  procedure ClearBlock(Block: TRect);
  procedure DrawCursor(X, Y: integer; Show: boolean);
  procedure SetCharAttrib(X, Y: integer; CA: word);
  procedure StringOut(X, Y: integer; S: string; Attrib: byte); overload;
  procedure StringOut(X,Y: integer; S: string); overload;
  procedure StringOut(Str: String); overload;
  procedure GraphicsTextOut(S: string);
  procedure CharOut(Ch: Char); overload;
  procedure CharOut(Ch: Char; Attrib: byte); overload;
  procedure LineTo(X, Y: integer; Color: TColor);
  procedure MoveTo(X, Y: integer);
  procedure DrawPoint(X, Y: integer; Color: TColor);
  procedure SetImageSize(Cols, Rows: integer);
  procedure Clear;
  procedure Cursor(X,Y: integer);
  procedure MoveCursor(X, Y: Integer);
  procedure HighlightText(X1, Y1, X2, Y2: integer; Attrib: byte);
  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;
  property BlockPaint: boolean read FBlockPaint write FBlockPaint;
 published
  property ExpandTabs: boolean read FExpandTabs write FExpandTabs default True;
  property GutterX: integer read FGutterX write FGutterX;
  property GutterY: integer read FGutterY write FGutterY;
  property ScreenWidth: integer read GetScreenWidth;
  property ScreenHeight: integer read GetScreenHeight;
  property Rows: integer read FRows write SetRows default 25;
  property Columns: integer read FColumns write SetColumns default 80;
  property Attribute: byte read FAttribute write FAttribute;
  property UseAttribute: boolean read FUseAttribute write FUseAttribute;
  property ShowCursor: boolean read FShowCursor write SetShowCursor default True;
  property OnWindowChange: TNotifyEvent read FOnWindowChange write FOnWindowChange;
  property CharSpacing: TPoint read FCharSpacing write SetCharSpacing;
  property Font;
  property Color;
 end;


type TConsoleWindow = class(TScrollingWinControl)
 private
  Panel: TPanel;
  FVGABox: TVGAWindow;
  //* FGLViewer: TGLViewer;
  FOnControlC: TNotifyEvent;
  FFontType: TFontType;
  FAspectRatio: double;
  FDoubleSize: boolean;
  FGraphicsMode: TGraphicsMode;
  FRefreshTimer: TTimer;
  procedure HandleFontChange(Sender: TObject);
  procedure LoadVGAFont(ResName,FontName: string; Size: integer);
  procedure HandleVGAChange(Sender: TObject);
  function GetVGABorder: integer;
  procedure SetVGABorder(const Value: integer);
  function WaitForKeyPress: boolean;
  function WaitForBuffer: boolean;
  procedure SetFontType(const Value: TFontType);
  procedure SetAspectRatio(const Value: double);
  procedure SetDoubleSize(const Value: boolean);
  procedure EraseBackground(var Msg: TWMEraseBkgnd); message WM_ERASEBKGND;
  function WaitForKeyInBuffer: char;
  function ReadBufferKey(var C: char): boolean;
  procedure WMKeyDown(var Message: TWMKeyDown); message WM_KEYDOWN;
  function WaitScanCode(var Code: Word): boolean;
  procedure SetVGAMode(Mode: Word);
  function ReadLastKey: char;
  procedure HandleScanCodes(SC: byte);
  procedure HandleRefreshTimer(Sender: TObject);
//*     function GetGLMode: boolean;
//*     procedure SetGLMode(const Value: boolean);
 protected
  procedure CMWantSpecialKey(var Message: TCMWantSpecialKey); message CM_WANTSPECIALKEY;
  procedure KeyPress(var Key: Char); override;
  procedure HandleControlC;
 public
  EchoFlag: boolean;
  KeyPressed: boolean;
  CtrlCPressed: boolean;
  FLastKey: Char;
  ScanCode: integer;
  DOSScanCode: byte;
  NewScanCode: boolean;
  KeyInBuffer: string;
  MouseState: TMouseState;
  procedure Reset;
  function VGAIndexToColor(C: Integer): TColor;
  function ColorToVGAIndex(C: TColor): Integer;
  procedure ResetKeyInBuffer;
  function KeyBoardIn: Char;
  function BuffererKeyIn: Char;
  function GetScanCode(var Code: Word; Wait: boolean): boolean;
  procedure InjectKeystroke(Key: char);
  procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
  procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X,Y: Integer); override;
  procedure EnableRefresh(Enable: boolean);
  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;
  function GetLastKey(Clear: boolean): char;
  property LastKey: char read FLastKey;
  property RefreshTimer: TTimer read FRefreshTimer;
 published
  property VGAMode: Word read FGraphicsMode.Mode write SetVGAMode;
  property VGABorder: integer read GetVGABorder write SetVGABorder default 4;
  property FontType: TFontType read FFontType write SetFontType;
  property VGABox: TVGAWindow read FVGABox write FVGABox;
//*  property GLViewer: TGLViewer read FGLViewer write FGLViewer;
//*  property GLMode: boolean read  GetGLMode write SetGLMode;
  property OnControlC: TNotifyEvent read FOnControlC write FOnControlC;
  property AspectRatio: double read FAspectRatio write SetAspectRatio;
  property DoubleSize: boolean read FDoubleSize write SetDoubleSize;

  property OnKeyDown;
  property OnKeyPress;
  property Align;
  property Anchors;
  property AutoScroll;
  property AutoSize;
  property BevelEdges;
  property BevelInner;
  property BevelOuter;
  property BevelKind;
  property BevelWidth;
  property BiDiMode;
  property Constraints;
  property DockSite;
  property DragCursor;
  property DragKind;
  property DragMode;
  property Enabled;
  property Color nodefault;
  property Ctl3D;
  property Font;
  property ParentBiDiMode;
  property ParentColor;
  property ParentCtl3D;
  property ParentFont;
  property ParentShowHint;
  property PopupMenu;
  property ShowHint;
  property TabOrder;
  property TabStop;
  property Visible;
  property OnCanResize;
  property OnClick;
  property OnConstrainedResize;
  property OnContextPopup;
  property OnDblClick;
  property OnDockDrop;
  property OnDockOver;
  property OnDragDrop;
  property OnDragOver;
  property OnEndDock;
  property OnEndDrag;
  property OnEnter;
  property OnExit;
  property OnGetSiteInfo;
  property OnMouseDown;
  property OnMouseMove;
  property OnMouseUp;
  property OnMouseWheel;
  property OnMouseWheelDown;
  property OnMouseWheelUp;
  property OnResize;
  property OnStartDock;
  property OnStartDrag;
  property OnUnDock;
end;


procedure Register;

implementation

uses EBXMain;

{$R VGAFont.res}



procedure Register;
begin
RegisterComponents('Fountain', [TConsoleWindow]);
end;

{ TBitmapManager }



constructor TBitmapManager.Create(Display: TBitmap);
begin
inherited Create;
BitmapList:=TObjectList.Create(True);
Self.Screen:=Display;
CopyMode.Mode:=0;
CopyMode.Alpha:=255;
CopyMode.PerPixel:=False;
CopyMode.TransColor:=RGB(0,0,0);
OldPen:=TPen.Create;
OldBrush:=TBrush.Create;
end;



destructor TBitmapManager.Destroy;
begin
OldBrush.Free;
OldPen.Free;
Bitmaplist.Free;
inherited Destroy;
end;



function TBitmapManager.Allocate: integer;
var BM: TBitmap;
begin
BM:=TBitmap.Create;
Result:=BitmapList.Add(BM);
BM.PixelFormat:=pf32Bit;
end;



procedure TBitmapManager.DeAllocate(Index: integer);
begin
if (Index<0) or (Index>=BitmapList.Count) then exit;
BitmapList.Delete(Index);
BitmapList.Insert(Index,nil);
end;



function TBitmapManager.GetBitmapByHandle(Handle: integer): TBitmap;
begin
if Handle=-1 then Result:=Screen
else if (Handle>=0) and (Handle<BitmapList.Count) then Result:=TBitmap(BitmapList[Handle])
else raise Exception.Create('Invalid Handle: '+IntToStr(Handle));
if Result=nil then raise Exception.Create('Deallocated Handle: '+IntToStr(Handle));
end;



procedure TBitmapManager.SetBitmapSize(Handle,Width,Height: integer);
var BM: TBitmap;
begin
BM:=GetBitmapByHandle(Handle);
BM.Width:=Width;
BM.Height:=Height;
end;



procedure TBitmapManager.GetBitmapSize(Handle: integer; var Width,Height: integer);
var BM: TBitmap;
begin
BM:=GetBitmapByHandle(Handle);
Width:=BM.Width;
Height:=BM.Height;
end;



procedure TBitmapManager.LoadBitmap(Handle: integer; Filename: string);
var BM: TBitmap;
begin
BM:=GetBitmapByHandle(Handle);
ImageLib.LoadBitmapByExtension(BM,Filename,False);
end;


procedure TBitmapManager.SaveBitmap(Handle: integer; Filename: string);
var BM: TBitmap;
begin
BM:=GetBitmapByHandle(Handle);
ImageLib.SaveBitmapByExtension(BM,Filename,False,True,False);
end;



procedure TBitmapManager.CopyTransparent(DestCanvas: TCanvas; DestRect: TRect; SourceCanvas: TCanvas; SourceRect: TRect; C: TColor);
begin
TransparentBlt(DestCanvas.Handle, DestRect.Left, DestRect.Top, DestRect.Right - DestRect.Left,
    DestRect.Bottom - DestRect.Top, SourceCanvas.Handle, SourceRect.Left, SourceRect.Top,
    SourceRect.Right - SourceRect.Left, SourceRect.Bottom - SourceRect.Top, C);
end;


procedure TBitmapManager.CopyImage(DestBM,SourceBM: TBitmap; DestRect,SourceRect: TRect);
begin
DestBM.Canvas.Brush.Color:=clBlack;
case CopyMode.Mode of
 1: CopyTransparent(DestBM.Canvas,DestRect,SourceBM.Canvas,SourceRect,CopyMode.TransColor);
 2: AlphaBlendCanvas(DestBM.Canvas,DestRect,SourceBM.Canvas, SourceRect,CopyMode.Alpha,CopyMode.PerPixel);
 13: FlipBitmapHoriz(SourceBM,DestBM);
 14: FlipBitmapVertical(SourceBM,DestBM);
 15: FlipBitmap90(SourceBM,DestBM);
 16: FlipBitmap180(SourceBM,DestBM);
 17: FlipBitmap270(SourceBM,DestBM);
 else
	begin
	case CopyMode.Mode of
 	 0: DestBM.Canvas.CopyMode:=cmSrcCopy;
 	 3: DestBM.Canvas.CopyMode:=cmSrcAnd;
 	 4: DestBM.Canvas.CopyMode:=cmSrcPaint;
 	 5: DestBM.Canvas.CopyMode:=cmSrcInvert;
 	 6: DestBM.Canvas.CopyMode:=cmDstInvert;
 	 7: DestBM.Canvas.CopyMode:=cmNotSrcCopy;
 	 8: DestBM.Canvas.CopyMode:=cmMergePaint;
 	 9: DestBM.Canvas.CopyMode:=cmNotSrcErase;
 	 10: DestBM.Canvas.CopyMode:=cmSrcErase;
	 end;
	DestBM.Canvas.CopyRect(DestRect,SourceBM.Canvas,SourceRect);
	case CopyMode.Mode of
	 11: MakeMask(DestBM,CopyMode.TransColor);
	 12: SetAlphaByColor(DestBM,CopyMode.TransColor,CopyMode.Alpha,not CopyMode.PerPixel);
	 end;
	end
 end;
end;



procedure TBitmapManager.CopyImage(DestHandle,SourceHandle: integer; DestRect,SourceRect: TRect);
var SourceBM,DestBM: TBitmap;
begin
SourceBM:=GetBitmapByHandle(SourceHandle);
DestBM:=GetBitmapByHandle(DestHandle);
CopyImage(DestBM,SourceBM,DestRect,SourceRect);
end;



procedure TBitmapManager.DrawImage(DestHandle,SourceHandle: integer; X,Y: integer);
var SourceBM,DestBM: TBitmap;
begin
SourceBM:=GetBitmapByHandle(SourceHandle);
DestBM:=GetBitmapByHandle(DestHandle);
CopyImage(DestBM,SourceBM,Rect(X,Y,X+SourceBM.Width,Y+SourceBM.Height),Rect(0,0,SourceBM.Width,SourceBM.Height));
end;



procedure TBitmapManager.SetPixel(Handle: integer; X,Y: integer; C: TColor);
var BM: TBitmap;
var Row: pRGBQuadArray;
begin
BM:=GetBitmapByHandle(Handle);
Row:=BM.Scanline[Y];
Row[X]:=BitmapLib.ColorToQuad(C)
end;




function TBitmapManager.GetPixel(Handle: integer; X,Y: integer): TColor;
var BM: TBitmap;
var Row: pRGBQuadArray;
begin
BM:=GetBitmapByHandle(Handle);
Row:=BM.Scanline[Y];
Result:=QuadToColor(Row[X]);
end;




function TBitmapManager.GetPenPosition(Handle: integer): TPoint;
var BM: TBitmap;
begin
BM:=GetBitmapByHandle(Handle);
Result:=BM.Canvas.PenPos;
end;


procedure TBitmapManager.SaveContext(Canvas: TCanvas);
begin
OldPen.Assign(Canvas.Pen);
OldBrush.Assign(Canvas.Brush);
end;


procedure TBitmapManager.RestoreContext(Canvas: TCanvas);
begin
Canvas.Pen.Assign(OldPen);
Canvas.Brush.Assign(OldBrush);
end;


procedure TBitmapManager.ClearImage(Handle: integer; Color: TColor);
var BM: TBitmap;
begin
BM:=GetBitmapByHandle(Handle);
SaveContext(BM.Canvas);
BM.Canvas.Brush.Style:=bsSolid;
ClearBitmap(BM,Color);
RestoreContext(BM.Canvas);
end;



procedure TBitmapManager.SetPen(Handle,Mode,Width,Pattern: integer; C: TColor);
var BM: TBitmap;
begin
BM:=GetBitmapByHandle(Handle);
BM.Canvas.Pen.Width:=Width;
BM.Canvas.Pen.Style:=TPenStyle(Pattern);

BM.Canvas.Pen.Color:=C;
case Mode of
 0: BM.Canvas.Pen.Mode:=pmCopy;
 1: BM.Canvas.Pen.Mode:=pmXOR;
 else BM.Canvas.Pen.Mode:=pmCopy;
 end;
end;



procedure TBitmapManager.SetFill(Handle,Mode: integer; C: TColor);
var BM: TBitmap;
begin
BM:=GetBitmapByHandle(Handle);
BM.Canvas.Brush.Color:=C;
case Mode of
 0: BM.Canvas.Brush.Style:=bsSolid;
 1: BM.Canvas.Brush.Style:=bsClear;
 else BM.Canvas.Brush.Style:=bsSolid;
 end;
end;



procedure TBitmapManager.DrawRectangle(Handle,X1,Y1,X2,Y2: integer );
var BM: TBitmap;
begin
BM:=GetBitmapByHandle(Handle);
BM.Canvas.Rectangle(X1,Y1,X2,Y2);
end;



procedure TBitmapManager.DrawCircle(Handle,X,Y,Radius: integer);
var BM: TBitmap;
var R: TRect;
begin
BM:=GetBitmapByHandle(Handle);
R:=Rect(X,Y,X,Y);
InflateRect(R,Radius,Radius);
BM.Canvas.Ellipse(R);
end;


procedure TBitmapManager.DrawEllipse(Handle,X1,Y1,X2,Y2: integer );
var BM: TBitmap;
var R: TRect;
begin
BM:=GetBitmapByHandle(Handle);
R:=Rect(X1,Y1,X2,Y2);
BM.Canvas.Ellipse(R);
end;




procedure TBitmapManager.MoveTo(Handle,X,Y: integer);
var BM: TBitmap;
begin
BM:=GetBitmapByHandle(Handle);
BM.Canvas.MoveTo(X,Y);
end;



procedure TBitmapManager.Lineto(Handle,X,Y: integer);
var BM: TBitmap;
begin
BM:=GetBitmapByHandle(Handle);
BM.Canvas.LineTo(X,Y);
end;



procedure TBitmapManager.Polyline(Handle,Mode: integer; const Points: array of TPoint);
var BM: TBitmap;
begin
BM:=GetBitmapByHandle(Handle);
case Mode of
 0: BM.Canvas.Polygon(Points);
 1: BM.Canvas.Polyline(Points);
 2: BM.Canvas.PolyBezier(Points);
 else Raise Exception.Create('Invalid Mode');
 end;
end;


procedure TBitmapManager.DoTextOut(Handle,X,Y: integer; S: string);
var BM: TBitmap;
begin
BM:=GetBitmapByHandle(Handle);
BM.Canvas.TextOut(X,Y,S);
end;



procedure TBitmapManager.SetFont(Handle: integer; Name: string);
var BM: TBitmap;
begin
BM:=GetBitmapByHandle(Handle);
BM.Canvas.Font.Name:=Name;
end;


procedure TBitmapManager.SetFontStyle(Handle: integer; Size: integer; Bold,Italic,Underline,Strike: boolean);
var Styles: TFontStyles;
var BM: TBitmap;
begin
BM:=GetBitmapByHandle(Handle);
BM.Canvas.Font.Size:=Size;
Styles:=[];
if Bold then Include(Styles,fsBold);
if Italic then Include(Styles,fsItalic);
if Underline then Include(Styles,fsUnderline);
if Strike then Include(Styles,fsStrikeOut);
BM.Canvas.Font.Style:=Styles;
end;


procedure TBitmapManager.GetFontProperties(Handle: integer; S: string; var Width,Height: integer);
var Styles: TFontStyles;
var BM: TBitmap;
begin
BM:=GetBitmapByHandle(Handle);
Width:=BM.Canvas.TextWidth(S);
Height:=BM.Canvas.TextHeight(S);
end;


procedure TBitmapManager.SetFontColor(Handle: integer; R,G,B: byte);
var BM: TBitmap;
begin
BM:=GetBitmapByHandle(Handle);
BM.Canvas.Font.Color:=RGB(R,G,B);
end;





procedure TBitmapManager.CopyFont(DestHandle,SourceHandle: integer);
var DBM,SBM: TBitmap;
begin
DBM:=GetBitmapByHandle(DestHandle);
SBM:=GetBitmapByHandle(SourceHandle);
DBM.Canvas.Font.Assign(SBM.Canvas.Font);
end;


{---------------- TVGAWindow --------------------------------------------------}


{

Old CGA color set - used to define character attributes
on VGA screen.

0=black, 1=blue, 2=green, 3=cyan, 4=red, 5=magenta
6=brown, 7=white, 8=gray, 9=light blue, 10=light green
11=light cyan, 12=light red, 13=light magenta, 14=yellow
15=white

}

{Remember RGB colors are least to most }





const VGAPalette256: array [0..255] of TColor = (
	$000000, $AA0000, $00AA00, $AAAA00, $0000AA, $AA00AA, $0055AA, $AAAAAA,
	$555555, $FF5555, $55FF55, $FFFF55, $5555FF, $FF55FF, $55FFFF, $FFFFFF,
	$000000, $101010, $202020, $353535, $454545, $555555, $656565, $757575,
	$8A8A8A, $9A9A9A, $AAAAAA, $BABABA, $CACACA, $DFDFDF, $EFEFEF, $FFFFFF,
	$FF0000, $FF0041, $FF0082, $FF00BE, $FF00FF, $BE00FF, $8200FF, $4100FF,
	$0000FF, $0041FF, $0082FF, $00BEFF, $00FFFF, $00FFBE, $00FF82, $00FF41,
	$00FF00, $41FF00, $82FF00, $BEFF00, $FFFF00, $FFBE00, $FF8200, $FF4100,
	$FF8282, $FF829E, $FF82BE, $FF82DF, $FF82FF, $DF82FF, $BE82FF, $9E82FF,
	$8282FF, $829EFF, $82BEFF, $82DFFF, $82FFFF, $82FFDF, $82FFBE, $82FF9E,
	$82FF82, $9EFF82, $BEFF82, $DFFF82, $FFFF82, $FFDF82, $FFBE82, $FF9E82,
	$FFBABA, $FFBACA, $FFBADF, $FFBAEF, $FFBAFF, $EFBAFF, $DFBAFF, $CABAFF,
	$BABAFF, $BACAFF, $BADFFF, $BAEFFF, $BAFFFF, $BAFFEF, $BAFFDF, $BAFFCA,
	$BAFFBA, $CAFFBA, $DFFFBA, $EFFFBA, $FFFFBA, $FFEFBA, $FFDFBA, $FFCABA,
	$710000, $71001C, $710039, $710055, $710071, $550071, $390071, $1C0071,
	$000071, $001C71, $003971, $005571, $007171, $007155, $007139, $00711C,
	$007100, $1C7100, $397100, $557100, $717100, $715500, $713900, $711C00,
	$713939, $713945, $713955, $713961, $713971, $613971, $553971, $453971,
	$393971, $394571, $395571, $396171, $397171, $397161, $397155, $397145,
	$397139, $457139, $557139, $617139, $717139, $716139, $715539, $714539,
	$715151, $715159, $715161, $715169, $715171, $695171, $615171, $595171,
	$515171, $515971, $516171, $516971, $517171, $517169, $517161, $517159,
	$517151, $597151, $617151, $697151, $717151, $716951, $716151, $715951,
	$410000, $410010, $410020, $410031, $410041, $310041, $200041, $100041,
	$000041, $001041, $002041, $003141, $004141, $004131, $004120, $004110,
	$004100, $104100, $204100, $314100, $414100, $413100, $412000, $411000,
	$412020, $412028, $412031, $412039, $412041, $392041, $312041, $282041,
	$202041, $202841, $203141, $203941, $204141, $204139, $204131, $204128,
	$204120, $284120, $314120, $394120, $414120, $413920, $413120, $412820,
	$412D2D, $412D31, $412D35, $412D3D, $412D41, $3D2D41, $352D41, $312D41,
	$2D2D41, $2D3141, $2D3541, $2D3D41, $2D4141, $2D413D, $2D4135, $2D4131,
	$2D412D, $31412D, $35412D, $3D412D, $41412D, $413D2D, $41352D, $41312D,
	$000000, $000000, $000000, $000000, $000000, $000000, $000000, $000000);

const CGAPalette: array [0..15] of TColor =(
	$000000, $AA0000, $00AA00, $AAAA00, $0000AA, $AA00AA, $0055AA, $AAAAAA,
	$555555, $FF5555, $55FF55, $FFFF55, $5555FF, $FF55FF, $55FFFF, $FFFFFF);

const VGAPalette8: array [0..15] of TColor =(
	$000000, $AA0000, $00AA00, $AAAA00, $0000AA, $AA00AA, $0055AA, $AAAAAA,
	$000000, $AA0000, $00AA00, $AAAA00, $0000AA, $AA00AA, $0055AA, $AAAAAA);

const VGAPalette16: array [0..15] of TColor =(
	$000000, $AA0000, $00AA00, $AAAA00, $0000AA, $AA00AA, $0055AA, $AAAAAA,
	$555555, $FF5555, $55FF55, $FFFF55, $5555FF, $FF55FF, $55FFFF, $FFFFFF);

const VGAPalette4: array [0..1,0..3] of TColor =(
	($000000, $AAAA00, $AA00AA, $AAAAAA),
	($555555, $FFFF55, $FF55FF, $FFFFFF));

const VGAPalette1: array [0..1] of TColor =(
	$000000, $FFFFFF);



const GraphicsModes: array [0..60] of TGraphicsMode =(
	(Mode: $0; Card:ctVGA; Res:(X:320;Y:200); TextRes:(X:40;Y:25); CharCell:cc8x8; Bits:3; ModeType:mtText),
	(Mode: $1; Card:ctVGA; Res:(X:320;Y:200); TextRes:(X:40;Y:25); CharCell:cc8x8; Bits:3; ModeType:mtText),

//	(Mode: $2; Card:ctVGA; Res:(X:720;Y:400); TextRes:(X:80;Y:25); CharCell:cc9x16; Bits:3; ModeType:mtText),
//	(Mode: $3; Card:ctVGA; Res:(X:720;Y:400); TextRes:(X:80;Y:25); CharCell:cc9x16; Bits:3; ModeType:mtText),
	(Mode: $2; Card:ctVGA; Res:(X:640;Y:400); TextRes:(X:80;Y:25); CharCell:cc8x16; Bits:3; ModeType:mtText),
	(Mode: $3; Card:ctVGA; Res:(X:640;Y:400); TextRes:(X:80;Y:25); CharCell:cc8x16; Bits:3; ModeType:mtText),

	(Mode: $4; Card:ctCGA; Res:(X:320;Y:200); TextRes:(X:40;Y:25); CharCell:cc8x8; Bits:2; ModeType:mtGraphics),
	(Mode: $5; Card:ctCGA; Res:(X:320;Y:200); TextRes:(X:40;Y:25); CharCell:cc8x8; Bits:2; ModeType:mtGraphics),
	(Mode: $6; Card:ctCGA; Res:(X:640;Y:200); TextRes:(X:80;Y:25); CharCell:cc8x8; Bits:1; ModeType:mtGraphics),

//	(Mode: $7; Card:ctVGA; Res:(X:720;Y:400); TextRes:(X:80;Y:25); CharCell:cc9x16; Bits:1; ModeType:mtText),
	(Mode: $7; Card:ctVGA; Res:(X:640;Y:400); TextRes:(X:80;Y:25); CharCell:cc8x16; Bits:1; ModeType:mtText),

	(Mode: $8; Card:ctPCJR; Res:(X:160;Y:200); TextRes:(X:20;Y:25); CharCell:cc8x8; Bits:4; ModeType:mtGraphics),
	(Mode: $9; Card:ctPCJR; Res:(X:320;Y:200); TextRes:(X:40;Y:25); CharCell:cc8x8; Bits:4; ModeType:mtGraphics),
	(Mode: $A; Card:ctPCJR; Res:(X:640;Y:200); TextRes:(X:80;Y:25); CharCell:cc8x8; Bits:2; ModeType:mtGraphics),
	(Mode: $B; Card:ctEGA; Res:(X:320;Y:200); TextRes:(X:40;Y:25); CharCell:cc8x8; Bits:4; ModeType:mtGraphics),
	(Mode: $C; Card:ctEGA; Res:(X:640;Y:200); TextRes:(X:80;Y:25); CharCell:cc8x8; Bits:2; ModeType:mtGraphics),
	(Mode: $D; Card:ctEGA; Res:(X:320;Y:200); TextRes:(X:40;Y:25); CharCell:cc8x8; Bits:4; ModeType:mtGraphics),
	(Mode: $E; Card:ctEGA; Res:(X:640;Y:200); TextRes:(X:80;Y:25); CharCell:cc8x8; Bits:4; ModeType:mtGraphics),
//	(Mode: $F; Card:ctEGA; Res:(X:640;Y:350); TextRes:(X:80;Y:25); CharCell:cc8x14; Bits:1; ModeType:mtGraphics),
//	(Mode: $10; Card:ctEGA; Res:(X:640;Y:350); TextRes:(X:80;Y:25); CharCell:cc8x14; Bits:4; ModeType:mtGraphics),
	(Mode: $F; Card:ctEGA; Res:(X:640;Y:350); TextRes:(X:80;Y:25); CharCell:cc8x16; Bits:1; ModeType:mtGraphics),
	(Mode: $10; Card:ctEGA; Res:(X:640;Y:350); TextRes:(X:80;Y:25); CharCell:cc8x16; Bits:4; ModeType:mtGraphics),

	(Mode: $11; Card:ctVGA; Res:(X:640;Y:480); TextRes:(X:80;Y:30); CharCell:cc8x16; Bits:1; ModeType:mtGraphics),
	(Mode: $12; Card:ctVGA; Res:(X:640;Y:480); TextRes:(X:80;Y:30); CharCell:cc8x16; Bits:4; ModeType:mtGraphics),
	(Mode: $13; Card:ctVGA; Res:(X:320;Y:200); TextRes:(X:40;Y:25); CharCell:cc8x8; Bits:8; ModeType:mtGraphics),
	(Mode: $6A; Card:ctVesa; Res:(X:800;Y:600); TextRes:(X:100;Y:75); CharCell:cc8x16; Bits:4; ModeType:mtGraphics),
	(Mode: $100; Card:ctVesa; Res:(X:640;Y:400); TextRes:(X:80;Y:25); CharCell:cc8x16; Bits:8; ModeType:mtGraphics),
	(Mode: $101; Card:ctVesa; Res:(X:640;Y:480); TextRes:(X:80;Y:30); CharCell:cc8x16; Bits:8; ModeType:mtGraphics),
	(Mode: $102; Card:ctVesa; Res:(X:800;Y:600); TextRes:(X:100;Y:38); CharCell:cc8x16; Bits:4; ModeType:mtGraphics),
	(Mode: $103; Card:ctVesa; Res:(X:800;Y:600); TextRes:(X:100;Y:38); CharCell:cc8x16; Bits:8; ModeType:mtGraphics),
	(Mode: $104; Card:ctVesa; Res:(X:1024;Y:768); TextRes:(X:128;Y:48); CharCell:cc8x16; Bits:4; ModeType:mtGraphics),
	(Mode: $105; Card:ctVesa; Res:(X:1024;Y:768); TextRes:(X:128;Y:48); CharCell:cc8x16; Bits:8; ModeType:mtGraphics),
	(Mode: $106; Card:ctVesa; Res:(X:1280;Y:1024); TextRes:(X:160;Y:64); CharCell:cc8x16; Bits:4; ModeType:mtGraphics),
	(Mode: $107; Card:ctVesa; Res:(X:1280;Y:1024); TextRes:(X:160;Y:64); CharCell:cc8x16; Bits:8; ModeType:mtGraphics),
	(Mode: $108; Card:ctVesa; Res:(X:640;Y:960); TextRes:(X:80;Y:60); CharCell:cc8x16; Bits:4; ModeType:mtText),
	(Mode: $109; Card:ctVesa; Res:(X:1056;Y:400); TextRes:(X:132;Y:25); CharCell:cc8x16; Bits:4; ModeType:mtText),
	(Mode: $10A; Card:ctVesa; Res:(X:1056;Y:688); TextRes:(X:132;Y:43); CharCell:cc8x16; Bits:4; ModeType:mtText),
	(Mode: $10B; Card:ctVesa; Res:(X:1056;Y:800); TextRes:(X:132;Y:50); CharCell:cc8x16; Bits:4; ModeType:mtText),
	(Mode: $10C; Card:ctVesa; Res:(X:1056;Y:960); TextRes:(X:132;Y:60); CharCell:cc8x16; Bits:4; ModeType:mtText),
	(Mode: $10D; Card:ctVesa; Res:(X:320;Y:200); TextRes:(X:40;Y:25); CharCell:cc8x8; Bits:15; ModeType:mtGraphics),
	(Mode: $10E; Card:ctVesa; Res:(X:320;Y:200); TextRes:(X:40;Y:25); CharCell:cc8x8; Bits:16; ModeType:mtGraphics),
	(Mode: $10F; Card:ctVesa; Res:(X:320;Y:200); TextRes:(X:40;Y:25); CharCell:cc8x8; Bits:24; ModeType:mtGraphics),
	(Mode: $110; Card:ctVesa; Res:(X:640;Y:480); TextRes:(X:80;Y:30); CharCell:cc8x16; Bits:15; ModeType:mtGraphics),
	(Mode: $111; Card:ctVesa; Res:(X:640;Y:480); TextRes:(X:80;Y:30); CharCell:cc8x16; Bits:16; ModeType:mtGraphics),
	(Mode: $112; Card:ctVesa; Res:(X:640;Y:480); TextRes:(X:80;Y:30); CharCell:cc8x16; Bits:24; ModeType:mtGraphics),
	(Mode: $113; Card:ctVesa; Res:(X:800;Y:600); TextRes:(X:100;Y:38); CharCell:cc8x16; Bits:15; ModeType:mtGraphics),
	(Mode: $114; Card:ctVesa; Res:(X:800;Y:600); TextRes:(X:100;Y:38); CharCell:cc8x16; Bits:16; ModeType:mtGraphics),
	(Mode: $115; Card:ctVesa; Res:(X:800;Y:600); TextRes:(X:100;Y:38); CharCell:cc8x16; Bits:24; ModeType:mtGraphics),
	(Mode: $116; Card:ctVesa; Res:(X:1024;Y:768); TextRes:(X:128;Y:48); CharCell:cc8x16; Bits:15; ModeType:mtGraphics),
	(Mode: $117; Card:ctVesa; Res:(X:1024;Y:768); TextRes:(X:128;Y:48); CharCell:cc8x16; Bits:16; ModeType:mtGraphics),
	(Mode: $118; Card:ctVesa; Res:(X:1024;Y:768); TextRes:(X:128;Y:48); CharCell:cc8x16; Bits:24; ModeType:mtGraphics),
	(Mode: $119; Card:ctVesa; Res:(X:1280;Y:1024); TextRes:(X:160;Y:64); CharCell:cc8x16; Bits:15; ModeType:mtGraphics),
	(Mode: $11A; Card:ctVesa; Res:(X:1280;Y:1024); TextRes:(X:160;Y:64); CharCell:cc8x16; Bits:16; ModeType:mtGraphics),
	(Mode: $11B; Card:ctVesa; Res:(X:1280;Y:1024); TextRes:(X:160;Y:64); CharCell:cc8x16; Bits:24; ModeType:mtGraphics),
	(Mode: $130; Card:ctVesa; Res:(X:320;Y:200); TextRes:(X:40;Y:25); CharCell:cc8x8; Bits:8; ModeType:mtGraphics),
	(Mode: $131; Card:ctVesa; Res:(X:320;Y:400); TextRes:(X:40;Y:25); CharCell:cc8x16; Bits:8; ModeType:mtGraphics),
	(Mode: $132; Card:ctVesa; Res:(X:320;Y:400); TextRes:(X:40;Y:25); CharCell:cc8x16; Bits:16; ModeType:mtGraphics),
	(Mode: $133; Card:ctVesa; Res:(X:320;Y:400); TextRes:(X:40;Y:25); CharCell:cc8x16; Bits:32; ModeType:mtGraphics),
	(Mode: $134; Card:ctVesa; Res:(X:320;Y:240); TextRes:(X:40;Y:30); CharCell:cc8x8; Bits:8; ModeType:mtGraphics),
	(Mode: $135; Card:ctVesa; Res:(X:320;Y:240); TextRes:(X:40;Y:30); CharCell:cc8x8; Bits:16; ModeType:mtGraphics),
	(Mode: $136; Card:ctVesa; Res:(X:320;Y:240); TextRes:(X:40;Y:30); CharCell:cc8x8; Bits:32; ModeType:mtGraphics),
	(Mode: $13D; Card:ctVesa; Res:(X:640;Y:400); TextRes:(X:80;Y:25); CharCell:cc8x16; Bits:16; ModeType:mtGraphics),
	(Mode: $13E; Card:ctVesa; Res:(X:640;Y:400); TextRes:(X:80;Y:25); CharCell:cc8x16; Bits:32; ModeType:mtGraphics),
	(Mode: $145; Card:ctVesa; Res:(X:1600;Y:1200); TextRes:(X:200;Y:75); CharCell:cc8x16; Bits:8; ModeType:mtGraphics),
	(Mode: $146; Card:ctVesa; Res:(X:1600;Y:1200); TextRes:(X:200;Y:75); CharCell:cc8x16; Bits:16; ModeType:mtGraphics),
	(Mode: $14A; Card:ctVesa; Res:(X:1600;Y:1200); TextRes:(X:200;Y:75); CharCell:cc8x16; Bits:32; ModeType:mtGraphics)
	);



procedure TVGAWindow.SetColorPalette(P: array of TColor);
var I,L: integer;
begin
L:=Length(P);
for I:=0 to High(ColorPalette) do
ColorPalette[I]:=P[I mod L];
end;



procedure TVGAWindow.SetPaletteItem(Inx: integer; C: TColor);
begin
if Inx<=High(ColorPalette) then ColorPalette[Inx]:=C;
end;



procedure TVGAWindow.HandleFontChange(Sender: TObject);
begin
Canvas.Font:=Font;
CharSize.X:=Canvas.TextWidth('M');
CharSize.Y:=Canvas.TextHeight('M');
CellSize:=IntVectorAdd(CharSize,CharSpacing);
SetImageSize(FColumns,FRows);
Resize;
end;



procedure TVGAWindow.SetImageSize(Cols,Rows: integer);
{Set the image size in Rows and Columns of Text}
begin
CursorX:=0; CursorY:=0;

{Calculate the character size in pixels}
FColumns:=Cols;
FRows:=Rows;
SetLength(TextArray,Cols,Rows);
Picture.Bitmap.Canvas.Font:=Font;
Picture.Bitmap.Width:=(Cols * CellSize.X + GutterX * 2) - CharSpacing.X;
Picture.Bitmap.Height:=(Rows * CellSize.Y + GutterY * 2) - CharSpacing.Y;

NormalizeImageSize(Self);
WindowChange;
Clear;
end;



function TVGAWindow.CursorToPixel(X,Y: integer): TPoint;
begin
Result:=IntVectorMultiply(CellSize,Point(X,Y));
Result:=IntVectorAdd(Result,Point(GutterX,GutterY));
end;



procedure TVGAWindow.DrawChar(X,Y: integer; CA: word);
var C: char;
var A: byte;
var P: TPoint;
begin
C:=char(CA and $FF);
A:=Swap(CA);
Canvas.Font:=Font;
Canvas.Font.Color:=ColorPalette[$F and A];
Canvas.Brush.Color:=ColorPalette[$F and (A shr 4)];
P:=CursorToPixel(X,Y);
Canvas.TextOut(P.X,P.Y,C);
end;



procedure TVGAWindow.HighlightText(X1,Y1,X2,Y2: integer; Attrib: byte);
var X,Y: integer;
var C: char;
var A: word;
begin
X1:=EnsureRange(X1,0,High(TextArray));
Y1:=EnsureRange(Y1,0,High(TextArray[0]));
X2:=EnsureRange(X2,0,High(TextArray));
Y2:=EnsureRange(Y2,0,High(TextArray[0]));
for Y:=Y1 to Y2 do
 for X:=X1 to X2 do
	begin
        A:=TextArray[X,Y];
	A:=Swap(Attrib)+(A and $FF);
        TextArray[X,Y]:=A;
        DrawChar(X,Y,A);
	end;
end;




procedure TVGAWindow.ClearTextArray;
var X,Y: integer;
var A: Word;
begin
A:=Swap(Attribute)+$20;
for Y:=0 to High(TextArray[0])-1 do
 for X:=0 to High(TextArray) do TextArray[X,Y]:=A;
end;





procedure TVGAWindow.ClearTextBlock(Block: TRect);
{Clear block of text - Block is in character coordinates, not pixel coordinates}
var X,Y: integer;
var A: Word;
begin
A:=Swap(Attribute)+$20;
for Y:=0 to High(TextArray[0])-1 do
 for X:=0 to High(TextArray) do TextArray[X,Y]:=A;
end;


procedure TVGAWindow.BlankLine(Line: integer);
var A: Word;
var X: integer;
begin
A:=Swap(Attribute)+$20;
for X:=0 to High(TextArray) do TextArray[X,Line]:=A;
end;




procedure TVGAWindow.ClearBlock(Block: TRect);
{Clears block in character coordinates}
var BP: TRect;
begin
RangeCheckBlock(Block);
Cursor(Block.Left,Block.Top);
BP.TopLeft:=CursorToPixel(Block.Left,Block.Top);
BP.Top:=BP.Top+1;
BP.BottomRight:=CursorToPixel(Block.Right,Block.Bottom);
ClearBitmap(Picture.Bitmap,BP,ColorPalette[$7 and (Attribute shr 4)]);
ClearTextBlock(Block);
end;




procedure TVGAWindow.Clear;
{Clear the Screen with the specified Color}
begin
Cursor(0,0);
{Clear the Image }
ClearBitmap(Picture.Bitmap,ColorPalette[$7 and (Attribute shr 4)]);
{Clear the Text}
ClearTextArray;
end;





procedure TVGAWindow.ScrollWindow;
var X,Y: integer;
var R1,R2,R3: TRect;
begin
{Scroll underlying Text array}
for Y:=0 to High(TextArray[0])-1 do
 for X:=0 to High(TextArray) do TextArray[X,Y]:=TextArray[X,Y+1];
BlankLine(High(TextArray[0]));

{Scroll screen by moving pixels up by the height of one character}
R1:=Rect(0,0,Picture.Bitmap.Width,Picture.Bitmap.Height);
R2:=R1;
OffsetRect(R2,0,-CellSize.Y);
Canvas.CopyRect(R2,Canvas,R1);

{Blank the bottom line}
Canvas.Brush.Color:=Color;
Canvas.Brush.Style:=bsSolid;
Canvas.Pen.Color:=Color;
Canvas.Pen.Mode:=pmCopy;
Canvas.Pen.Style:=psSolid;
R3:=Rect(0,Picture.Bitmap.Height-CellSize.Y,Picture.Bitmap.Width,Picture.Bitmap.Height);
Canvas.Rectangle(R3);
end;


{------------- Blinking Cursor Handling ---------------------------------------}





procedure TVGAWindow.EraseCursor;
begin
if CursorDrawn then Canvas.Draw(OldCursorLoc.X,OldCursorLoc.Y,CursorBMP);
CursorDrawn:=False;
end;



procedure TVGAWindow.OutputCursor(X,Y: integer; C: TColor);
var P: TPoint;
begin
P:=CursorToPixel(X,Y);

CursorDrawn:=True;
OldCursorLoc:=P;
CursorBMP.Width:=CellSize.X;
CursorBMP.Height:=CellSize.Y;
CursorBMP.Canvas.CopyRect(Rect(0,0,CellSize.X,CellSize.Y),Canvas,Rect(P.X,P.Y,P.X+CellSize.X,P.Y+CellSize.Y));

Canvas.Pen.Color:=C;
Canvas.Pen.Width:=2;
Canvas.Pen.Mode:=pmXOR;
Canvas.MoveTo(P.X,P.Y+CellSize.Y-1);
Canvas.LineTo(P.X+CellSize.X-1,P.Y+CellSize.Y-1);
end;



procedure TVGAWindow.DrawCursor(X,Y: integer; Show: boolean);
begin
Canvas.Font.Color:=clWhite;
if Show then OutputCursor(X,Y,clSilver)
else EraseCursor;
end;


procedure TVGAWindow.HandleTimer(Sender: TObject);
begin
if CursorDrawn then DrawCursor(CursorX,CursorY,False)
else DrawCursor(CursorX,CursorY,True);
end;


procedure TVGAWindow.EnableCursor(Enable: boolean);
begin
if Enable then
	begin
	DrawCursor(CursorX,CursorY,True);
	CursorTimer.Enabled:=True
	end
else
	begin
	CursorTimer.Enabled:=False;
	DrawCursor(CursorX,CursorY,False);
	end;
end;

{------------------------------------------------------------------------------}


constructor TVGAWindow.Create(AOwner: TComponent);
begin
inherited Create(AOwner);
ControlStyle := ControlStyle + [csOpaque];
Picture.Bitmap.PixelFormat:=pf32bit;
BitmapManager:=TBitmapManager.Create(Picture.Bitmap);

CursorDrawn:=False;
OldCursorLoc:=Point(0,0);
CursorBMP:=TBitmap.Create;
CursorTimer:=TTimer.Create(Self);
CursorTimer.Enabled:=False;
CursorTimer.Interval:=200;
CursorTimer.OnTimer:=HandleTimer;

FGutterX:=5;
FGutterY:=5;
FCharSpacing:=Point(0,0);
Attribute:=$0F;
UseAttribute:=True;
SetImageSize(80,25);
Color:=clBlack;
Font.OnChange:=HandleFontChange;
Font.Name:='Terminal';
Font.Size:=9;
FShowCursor:=True;
Cursor(0,0);
ExpandTabs:=True;
if FShowCursor then EnableCursor(True);
OldPenPos:=Point(-1,-1);
BlockPaint:=False;
end;



destructor TVGAWindow.Destroy;
begin
CursorTimer.Free;
BitmapManager.Free;
CursorBMP.Free;
inherited;
end;



{------------------------------------------------------------------------------}

procedure TVGAWindow.SetCharAttrib(X,Y: integer; CA: word);
begin
SetCursorRangeCheck(X,Y);
DrawChar(CursorX,CursorY,CA);
TextArray[CursorX,CursorY]:=CA;
end;



procedure TVGAWindow.StringOut(X,Y: integer; S: string; Attrib: byte);
{Out the text string at the specified cursor position}
{Handles CA/LF characters and tabs}
var I: integer;
var LocalAtt: byte;


	procedure OutputChar(C: char);
	var CA: word;
	begin
	CA:=Swap(LocalAtt)+byte(C);
	DrawChar(CursorX,CursorY,CA);
	TextArray[CursorX,CursorY]:=CA;
	Inc(CursorX);
	end;


	procedure IncrementRow;
	begin
	Inc(CursorY);
	if CursorY>High(TextArray[0]) then
		begin
		case WinMode of
		 wmScroll: begin ScrollWindow; CursorY:=High(TextArray[0]); end;
		 wmWrap: CursorY:=0;
		 wmClip: CursorY:=High(TextArray[0]);
		 end;
		end;
	end;

	procedure HandleWrap;
	begin
	if CursorX>High(TextArray) then
		begin
		case WinMode of
		 wmScroll: begin CursorX:=0; IncrementRow; end;
		 wmWrap: begin CursorX:=0; IncrementRow; end;
		 wmClip: CursorX:=High(TextArray);
		 end;
		end;
	end;


	procedure HandleTab;
	begin
	repeat
		begin
		OutputChar(SP_Char);
		end
	until ((CursorX mod 8)=0) or (CursorX>High(TextArray));
	HandleWrap;
	end;


begin
EnableCursor(False);
Attribute:=Attrib;
if UseAttribute then LocalAtt:=Attribute
else LocalAtt:=$07;
SetCursorRangeCheck(X,Y);

{Output Text at the specified position}
for I:=1 to Length(S) do
	begin
	if (S[I]=TAB_Char) and ExpandTabs then HandleTab
	else case S[I] of
	 CR_Char: CursorX:=0;
	 LF_Char: IncrementRow;
	 else
		begin
		OutputChar(S[I]);
		HandleWrap;
		end;
	 end;
	end;

if FShowCursor then EnableCursor(True);
{Don't focus here or no other controls on the form}
{will work while you're doing continuous I/O}
//FocusOwner;
end;


procedure TVGAWindow.StringOut(X,Y: integer; S: string);
begin
StringOut(X,Y,S,Attribute);
end;



procedure TVGAWindow.StringOut(Str: String);
{XPL-Style Text routine simulator}
begin
StringOut(CursorX,CursorY,Str);
end;




procedure TVGAWindow.CharOut(Ch: Char);
{XPL-Style Chout routine}
begin
StringOut(CursorX,CursorY,Ch);
end;



procedure TVGAWindow.CharOut(Ch: Char; Attrib: byte);
begin
StringOut(CursorX,CursorY,Ch,Attrib);
end;


procedure TVGAWindow.DeleteChar;
var A: Word;
begin
A:=Swap(Attribute)+$20;
if CursorX>0 then
	begin
	MoveCursor(-1,0);
	DrawChar(CursorX,CursorY,A);
	TextArray[CursorX,CursorY]:=A;
	end;
end;



function TVGAWindow.GetCharWidth(S: string): integer;
var ABC: TABCFloat;
var A,B,C: integer;
begin
GetCharABCWidthsFloat(Canvas.Handle,ord(S[1]),ord(S[1]),ABC);
Result:=trunc(abs(ABC.abcfA)+abs(ABC.abcfB));
end;



procedure TVGAWindow.GraphicsTextOut(S: string);
var P: TPoint;
var I,H,CW: integer;
var OS: TBrushStyle;
var OC: TColor;
begin
OS:=Canvas.Brush.Style;
OC:=Canvas.Brush.Color;
H:=Canvas.TextHeight('M');
for I:=1 To Length(S) do
	begin
	P:=Canvas.PenPos;
	case S[I] of
	 CR_Char: Canvas.MoveTo(0,P.Y);
	 LF_Char: Canvas.MoveTo(P.X,P.Y+H);
	 else
		begin
		if (P.X + Canvas.TextWidth(S[I]))>Width then P:=Point(0,P.Y+H);
		if (fsItalic in Canvas.Font.Style) and (Canvas.Brush.Style<>bsClear) then
			begin
			CW:=GetCharWidth(S[I]);
			if P.X<>OldPenPos.X then CharRect:=Rect(P.X,P.Y,P.X+CW,P.Y+H)
			else CharRect:=Rect(CharRect.Right,P.Y,(P.X+CW),P.Y+H);
			Canvas.Brush.Style:=bsSolid;
			Canvas.FillRect(CharRect);
			Update;
			Canvas.Brush.Style:=bsClear;
			end;
		Canvas.TextOut(P.X,P.Y,S[I]);
		Canvas.Brush.Color:=OC;
		Canvas.Brush.Style:=OS;
		OldPenPos:=Canvas.PenPos;
		end;
	 end;
	end;
end;

{---------------------------------------------------------------------}


procedure TVGAWindow.RangeCheckCharPos(var X,Y: integer);
begin
X:=EnsureRange(X,0,High(TextArray));
Y:=EnsureRange(Y,0,High(TextArray[0]));
end;


procedure TVGAWindow.RangeCheckBlock(var Block: TRect);
begin
RangeCheckCharPos(Block.Left,Block.Top);
RangeCheckCharPos(Block.Right,Block.Bottom);
end;


procedure TVGAWindow.SetCursorRangeCheck(X,Y: integer);
begin
RangeCheckCharPos(X,Y);
CursorX:=X;
CursorY:=Y;
end;


procedure TVGAWindow.Cursor(X, Y: Integer);
{Set the cursor position}
begin
EnableCursor(False);
SetCursorRangeCheck(X,Y);
if FShowCursor then EnableCursor(True);
end;


procedure TVGAWindow.MoveCursor(X, Y: Integer);
{Move cursor position relative to current position}
begin
EnableCursor(False);
X:=CursorX+X;
Y:=CursorY+Y;
SetCursorRangeCheck(X,Y);
if FShowCursor then EnableCursor(True);
end;


{---------------------------------------------------------------------}


procedure TVGAWindow.MoveTo(X,Y: integer);
begin
Canvas.MoveTo(X,Y);
end;


procedure TVGAWindow.LineTo(X,Y: integer; Color: TColor);
begin
Canvas.Pen.Color:=Color;
Canvas.LineTo(X,Y);
end;


procedure TVGAWindow.DrawPoint(X,Y: integer; Color: TColor);
begin
Canvas.Pixels[X,Y]:=Color;
end;


{---------------- Property Handlers -----------------------------------}


procedure TVGAWindow.SetColumns(const Value: integer);
begin
if FColumns<>Value then
	begin
        FColumns := Value;
	SetImageSize(FColumns,FRows);
	end;
end;

procedure TVGAWindow.SetRows(const Value: integer);
begin
if FRows<>Value then
	begin
        FRows:=Value;
	SetImageSize(FColumns,FRows);
	end;
end;


function TVGAWindow.GetScreenHeight: integer;
begin
Result:=Picture.Bitmap.Height;
end;




function TVGAWindow.GetScreenWidth: integer;
begin
Result:=Picture.Bitmap.Width;
end;


function TVGAWindow.FocusOwner: boolean;
begin
Result:=not (csDesigning in ComponentState) and
	(Owner is TWinControl) and
	TWinControl(Owner).CanFocus;
if Result then TWinControl(Owner).SetFocus;
end;


procedure TVGAWindow.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
if FocusOwner then
	begin
	if Owner is TConsoleWindow then TConsoleWindow(Owner).MouseDown(Button,Shift,X,Y);
	end;
inherited MouseDown(Button, Shift, X, Y);
end;



procedure TVGAWindow.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
if not (csDesigning in ComponentState) and (Owner is TConsoleWindow) then
	TConsoleWindow(Owner).MouseMove(Shift,X,Y);
inherited;
end;



procedure TVGAWindow.MouseUp(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
begin
if not (csDesigning in ComponentState) and (Owner is TConsoleWindow) then
	TConsoleWindow(Owner).MouseUp(Button,Shift,X,Y);
inherited;
end;



procedure TVGAWindow.SetShowCursor(const Value: boolean);
begin
if FShowCursor<>Value then
	begin
	FShowCursor:=Value;
	EnableCursor(Value);
	end;
end;


procedure TVGAWindow.WindowChange;
begin
if Assigned(FOnWindowChange) then FOnWindowChange(Self);
end;


procedure TVGAWindow.CMColorChanged(var Message: TMessage);
begin
inherited;
WindowChange;
end;



procedure TVGAWindow.SetCharSpacing(const Value: TPoint);
begin
if (FCharSpacing.X<>Value.X) or (FCharSpacing.Y<>Value.Y) then
	begin
	FCharSpacing := Value;
	HandleFontChange(Self);
	end;
end;



procedure TVGAWindow.EraseBackground(var Msg: TWMEraseBkgnd);
begin
Msg.Result := 1;
end;



procedure TVGAWindow.Paint;
begin
if not BlockPaint then inherited Paint;
end;


{============ TConsoleWindow =================================================}

procedure TConsoleWindow.CMWantSpecialKey(var Message: TCMWantSpecialKey);
begin
{ We want to handle the left arrow ourselves so they don't shift the focus}
case Message.CharCode of
 VK_LEFT,VK_RIGHT,VK_UP,VK_DOWN: Message.Result := 1;
 end;
end;



procedure TConsoleWindow.HandleFontChange(Sender: TObject);
begin
{Pass font changes to underlying control}
VGABox.Font:=Font;
Resize;
end;




procedure TConsoleWindow.LoadVGAFont(ResName,FontName: string; Size: integer);
var RS: TResourceStream;
var Count: Cardinal;
var HFRS: Cardinal;
begin
HFRS := 0;
RS := TResourceStream.Create(hInstance,ResName,'RT_FONT');
try
if RS.Size > 14 then
	begin
	HFRS := AddFontMemResourceEx(RS.Memory,RS.Size,nil, @Count);
	if HFRS > 0 then
		begin
		Font.Handle:=CreateFont(0,0,0,0,FW_NORMAL,0,0,0,DEFAULT_CHARSET,OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS,DEFAULT_QUALITY,DEFAULT_PITCH, pchar(FontName));
		Font.Size:=Size;
		end;
	end;
finally RS.Free; end;
end;



procedure TConsoleWindow.HandleVGAChange(Sender: TObject);
var Pad: integer;
var ASX,ASY: double;
begin
if AspectRatio<1 then begin ASY:=1; ASX:=1/AspectRatio; end
else begin ASY:=AspectRatio; ASX:=1;  end;
Pad:=(Panel.BorderWidth + Panel.BevelWidth) * 2;
Panel.Left:=-HorzScrollBar.Position; Panel.Top:=-VertScrollBar.Position;
{Use underlying bitmap because VGABox is client aligned}
VGABox.Stretch:=True;
Panel.Width:=Round(VGABox.Picture.Bitmap.Width * ASX) + Pad;
Panel.Height:=Round(VGABox.Picture.Bitmap.Height * ASY) + Pad;
Panel.Color:=VGABox.Color;
Resize;
end;


procedure TConsoleWindow.HandleRefreshTimer(Sender: TObject);
begin
Self.VGABox.Repaint;
end;


procedure TConsoleWindow.EnableRefresh(Enable: boolean);
begin
VGABox.BlockPaint:=not Enable;
if not VGABox.BlockPaint then VGABox.Repaint;
end;



constructor TConsoleWindow.Create(AOwner: TComponent);
begin
inherited;
ControlStyle := ControlStyle + [csOpaque];
Panel:=TPanel.Create(Self);
Panel.Parent:=Self;
Panel.Visible:=True;
Panel.ControlStyle := Panel.ControlStyle + [csOpaque];
Panel.Name:='Panel';
FRefreshTimer:=TTimer.Create(Self);
FRefreshTimer.Enabled:=False;
FRefreshTimer.Interval:=33;
FRefreshTimer.OnTimer:=HandleRefreshTimer;

VGABox:=TVGAWindow.Create(Self);
VGABox.Parent:=Panel;
VGABox.Visible:=True;
VGABox.GutterX:=0;
VGABox.GutterY:=0;
VGABox.Align:=alClient;
VGABorder:=4;
VGABox.OnWindowChange:=HandleVGAChange;
VGABox.Name:='VGABox';

//* FGLViewer:=TGLViewer.Create(Self);
//* FGLViewer.Parent:=Panel;
//* FGLViewer.Visible:=False;
//* FGLViewer.Align:=alClient;
//* FGLViewer.Name:='GLViewer';

FAspectRatio:=1;
FDoubleSize:=True;
FontType:=ftVGAFont;
Font.OnChange:=HandleFontChange;
VGABox.Font:=Font;
TabStop:=True;
DoubleBuffered:=True;
Self.VertScrollBar.Tracking:=True;
Self.HorzScrollBar.Tracking:=True;
EchoFlag:=False;
VGAMode:=3;
end;



destructor TConsoleWindow.Destroy;
begin
//* FGLViewer.Free;
FVGABox.Free;
Panel.Free;
inherited;
end;



procedure TConsoleWindow.Reset;
begin
VGAMode:=3;
ResetKeyInBuffer;
VGABox.Picture.Bitmap.Canvas.Font.Name:='FixedSys';
VGABox.Picture.Bitmap.Canvas.Font.Size:=9;
VGABox.Picture.Bitmap.Canvas.Font.Style:=[];

VGABox.BitmapManager.CopyMode.Mode:=0;
VGABox.BitmapManager.CopyMode.Alpha:=255;
VGABox.BitmapManager.CopyMode.PerPixel:=False;
VGABox.BitmapManager.CopyMode.TransColor:=RGB(0,0,0);
VGABox.WinMode:=wmScroll;
VGABox.BitmapManager.BitmapList.Clear;
EnableRefresh(True);
//* GLMode:=False;
end;



procedure TConsoleWindow.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
if not (csDesigning in ComponentState) and
  (CanFocus or (GetParentForm(Self) = nil)) then SetFocus;
case Button of
 mbLeft: MouseState.Left:=True;
 mbRight: MouseState.Right:=True;
 mbMiddle: MouseState.Middle:=True;
 end;
inherited MouseDown(Button, Shift, X, Y);
end;



procedure TConsoleWindow.MouseMove(Shift: TShiftState; X, Y: Integer);
{Overridden and made public so it can be called by the VGA Box}
var P: TPoint;
begin
P:=Point(X,Y);
MouseState.Pos:=P;
inherited;
end;



procedure TConsoleWindow.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
{Overridden and made public so it can be called by the VGA Box}
begin
case Button of
 mbLeft: MouseState.Left:=False;
 mbRight: MouseState.Right:=False;
 mbMiddle: MouseState.Middle:=False;
 end;
inherited;
end;


procedure TConsoleWindow.SetFontType(const Value: TFontType);
begin
if FFontType<>Value then
	begin
	FFontType := Value;
	case FFontType of
	 ftNormal: VGABox.Font:=Font;
	 ftVGAFont: LoadVGAFont('EGA-VGAFONT','ASCII VGA',12);
	 ftEGAFont: LoadVGAFont('EGA-VGAFONT','ASCII EGA',12);
	 ftCGAFont: LoadVGAFont('CGAFont','PC8X8',10);
	 end;
	HandleFontChange(Self);
	end;
end;



function TConsoleWindow.GetVGABorder: integer;
begin
Result:=Panel.BorderWidth
end;



procedure TConsoleWindow.SetVGABorder(const Value: integer);
begin
Panel.BorderWidth:=Value;
end;



procedure TConsoleWindow.SetAspectRatio(const Value: double);
begin
if FAspectRatio<>Value then
	begin
	FAspectRatio:=Value;
	HandleVGAChange(Self);
	Resize;
	end;
end;





procedure TConsoleWindow.SetDoubleSize(const Value: boolean);
begin
if FDoubleSize<>Value then
	begin
	FDoubleSize := Value;
	HandleVGAChange(Self);
	end;
end;



{---------------- Console Input Routines --------------------------------------}

const scLeftShift = $2A; scRightShift = $36; scControl = $1D; scCapsLock = $3A;
      scScrollLock = $46; scPause = $45;

const StateKeys = [scLeftShift,scRightShift, scControl, scPause,scScrollLock, scCapsLock];

const NonASCIICodes = [$1D,$2A,$36, $37, $38, $3A..$49, $4B..$4D, $4F..$53,$57,$58];

const FKeys =[$3B..$44];


procedure TConsoleWindow.HandleScanCodes(SC: byte);
{Translate the code to the type of scan code produced under DOS}
begin
Form1.Caption:=IntToHex(SC,2);
DOSScanCode:=0;
{Ignore keyboard-state keys like Control, Shift etc.}
if SC in StateKeys then exit;
if (GetAsyncKeyState(VK_SHIFT) and $8000)<>0 then
	begin
	if SC in FKeys then DOSScanCode:=SC + $19
	else if SC=$0F then DOSScanCode:=SC;
	end
else if (GetAsyncKeyState(VK_Control) and $8000)<>0 then
	begin
	if SC in FKeys then DOSScanCode:=SC + $23
	else case SC of
	 $03: DOSScanCode:=SC;
	 $47: DOSScanCode:=$77;
	 $49: DOSScanCode:=$84;
	 $4B: DOSScanCode:=$73;
	 $4D: DOSScanCode:=$74;
	 $4F: DOSScanCode:=$75;
	 $51: DOSScanCode:=$76;
	 else DOSScanCode:=SC;
	 end;
	end
else if (GetAsyncKeyState(VK_MENU) and $8000)<>0 then
	begin
	if SC in FKeys then DOSScanCode:=SC + $3D
	else if SC in [$10..$19,$1E..$26,$2C..$32] then DOSScanCode:=SC
	else if SC in [$0C,$0D] then DOSScanCode:=SC + $76;
	end
else if SC in NonASCIICodes then DOSScanCode:=SC;
if DOSScanCode<>0 then
	begin
	KeyPressed:=True;
	FLastKey:=#0;
	end;
end;




procedure TConsoleWindow.WMKeyDown(var Message: TWMKeyDown);
begin
Scancode:=(Message.KeyData shr 16) and $FF;
NewScanCode:=True;
HandleScanCodes(ScanCode);
inherited;
end;



function TConsoleWindow.WaitScanCode(var Code: Word): boolean;
begin
Code:=0;
NewScanCode:=False;
VGABox.ShowCursor:=True;
try
repeat
	begin
	Application.ProcessMessages;
	Sleep(1);
	if CtrlCPressed or Application.Terminated then abort;
	end
until NewScanCode;
finally
 VGABox.ShowCursor:=False;
 Result:=NewScanCode;
 NewScanCode:=False;
 end;
Code:=ScanCode;
end;



function TConsoleWindow.GetScanCode(var Code: Word; Wait:boolean): boolean;
begin
if Wait then Result:=WaitScanCode(Code)
else
	begin
	Code:=ScanCode;
	Result:=NewScanCode;
	NewScanCode:=False;
	end;
end;


function TConsoleWindow.GetLastKey(Clear: boolean): char;
begin
Result:=FLastKey;
if Clear then KeyPressed:=False;
end;



{------------------------------------------------------------------------------}



procedure TConsoleWindow.HandleControlC;
begin
CtrlCPressed:=True;
if Assigned(FOnControlC) then FOnControlC(Self);
end;




procedure TConsoleWindow.KeyPress(var Key: Char);
begin
inherited;
FLastKey:=Key;
KeyPressed:=True;
if Key=CtrlH then
	begin
	if Length(KeyInBuffer)>0 then
		begin
		SetLength(KeyInBuffer,Length(KeyInBuffer)-1);
		VGABox.DeleteChar;
		end;
	end
else
	begin
	KeyInBuffer:=KeyInBuffer+Key;
	if EchoFlag then VGABox.CharOut(Key);
	end;
if Key=CtrlC then HandleControlC;
end;


function TConsoleWindow.ReadBufferKey(var C: char): boolean;
begin
Result:=Length(KeyInBuffer)>0;
if Result then
	begin
	C:=KeyInBuffer[1];
	Delete(KeyInBuffer,1,1);
	KeyPressed:=False;
	end
else C:=#0;
end;


function TConsoleWindow.WaitForKeyPress: boolean;
begin
VGABox.ShowCursor:=True;
try
repeat
	begin
	Application.ProcessMessages;
	Sleep(1);
	if CtrlCPressed or Application.Terminated then abort;
	end
until KeyPressed;
finally
 VGABox.ShowCursor:=False;
 Result:=KeyPressed;
 end
end;



function TConsoleWindow.ReadLastKey: char;
begin
Result:=FLastKey;
if Result=#0 then FLastKey:=Char(DOSScanCode)
else KeyPressed:=False;
end;


function TConsoleWindow.WaitForKeyInBuffer: char;
begin
Result:=#0;
if not KeyPressed then
	begin
	if WaitForKeyPress then Result:=ReadLastKey;
	end
else Result:=ReadLastKey;
end;




function TConsoleWindow.KeyBoardIn: Char;
begin
EchoFlag:=False;
if CanFocus then SetFocus;
Result:=WaitForKeyInBuffer;
end;



function TConsoleWindow.WaitForBuffer: boolean;
begin
Result:=True;
while not Application.Terminated and not CtrlCPressed do
	begin
	if (Length(KeyInBuffer)>0) and
	   (KeyInBuffer[Length(KeyInBuffer)]=CR_Char) then
		begin
		VGABox.CharOut(LF_Char);
		exit;
		end;
	KeyPressed:=False;
	WaitForKeyPress;
	KeyPressed:=False;
	end;
Result:=False;
end;



function TConsoleWindow.BuffererKeyIn: Char;
begin
if CanFocus then SetFocus;
EchoFlag:=True;
try
if not ReadBufferKey(Result) then
	begin
	WaitForBuffer;
	ReadBufferKey(Result);
	end;
finally EchoFlag:=False; end;
end;




procedure TConsoleWindow.ResetKeyInBuffer;
begin
CtrlCPressed:=False;
KeyPressed:=False;
KeyInBuffer:='';
FLastKey:=#0;
end;



procedure TConsoleWindow.InjectKeystroke(Key: char);
begin
KeyPress(Key);
end;

{------------------------------------------------------------------------------}


procedure TConsoleWindow.EraseBackground(var Msg: TWMEraseBkgnd);
var R: TRect;
var Canvas: TCanvas;
begin
Msg.Result:= 1;
if Msg.DC=0 then exit;
Canvas:=TCanvas.Create;
try
Canvas.Handle:=Msg.DC;
R.Left:=Panel.ClientRect.Right;
R.Top:=Panel.ClientRect.Top;
R.Right:=ClientRect.Right;
R.Bottom:=ClientRect.Bottom;
Canvas.Pen.Color:=Color;
Canvas.Brush.Color:=Color;
Canvas.Rectangle(R);

R.Left:=ClientRect.Left;
R.Top:=Panel.ClientRect.Bottom;
R.Right:=ClientRect.Right;
R.Bottom:=ClientRect.Bottom;
Canvas.Rectangle(R);
finally Canvas.Free; end;
end;



procedure TConsoleWindow.SetVGAMode(Mode: Word);
begin
//if FGraphicsMode.Mode=Mode then exit;
VGABox.Attribute:=$07;
case Mode of
 $0: FGraphicsMode:=GraphicsModes[0];
 $1: FGraphicsMode:=GraphicsModes[1];
 $2: FGraphicsMode:=GraphicsModes[2];
 $3: FGraphicsMode:=GraphicsModes[3];
 $4: FGraphicsMode:=GraphicsModes[4];
 $5: FGraphicsMode:=GraphicsModes[5];
 $6: FGraphicsMode:=GraphicsModes[6];
 $7: FGraphicsMode:=GraphicsModes[7];
 $8: FGraphicsMode:=GraphicsModes[8];
 $9: FGraphicsMode:=GraphicsModes[9];
 $A: FGraphicsMode:=GraphicsModes[10];
 $B: FGraphicsMode:=GraphicsModes[11];
 $C: FGraphicsMode:=GraphicsModes[12];
 $D: FGraphicsMode:=GraphicsModes[13];
 $E: FGraphicsMode:=GraphicsModes[14];
 $F: FGraphicsMode:=GraphicsModes[15];
 $10: FGraphicsMode:=GraphicsModes[16];
 $11: FGraphicsMode:=GraphicsModes[17];
 $12: FGraphicsMode:=GraphicsModes[18];
 $13: FGraphicsMode:=GraphicsModes[19];
 $6A: FGraphicsMode:=GraphicsModes[20];
 $100: FGraphicsMode:=GraphicsModes[21];
 $101: FGraphicsMode:=GraphicsModes[22];
 $102: FGraphicsMode:=GraphicsModes[23];
 $103: FGraphicsMode:=GraphicsModes[24];
 $104: FGraphicsMode:=GraphicsModes[25];
 $105: FGraphicsMode:=GraphicsModes[26];
 $106: FGraphicsMode:=GraphicsModes[27];
 $107: FGraphicsMode:=GraphicsModes[28];
 $108: FGraphicsMode:=GraphicsModes[29];
 $109: FGraphicsMode:=GraphicsModes[30];
 $10A: FGraphicsMode:=GraphicsModes[31];
 $10B: FGraphicsMode:=GraphicsModes[32];
 $10C: FGraphicsMode:=GraphicsModes[33];
 $10D: FGraphicsMode:=GraphicsModes[34];
 $10E: FGraphicsMode:=GraphicsModes[35];
 $10F: FGraphicsMode:=GraphicsModes[36];
 $110: FGraphicsMode:=GraphicsModes[37];
 $111: FGraphicsMode:=GraphicsModes[38];
 $112: FGraphicsMode:=GraphicsModes[39];
 $113: FGraphicsMode:=GraphicsModes[40];
 $114: FGraphicsMode:=GraphicsModes[41];
 $115: FGraphicsMode:=GraphicsModes[42];
 $116: FGraphicsMode:=GraphicsModes[43];
 $117: FGraphicsMode:=GraphicsModes[44];
 $118: FGraphicsMode:=GraphicsModes[45];
 $119: FGraphicsMode:=GraphicsModes[46];
 $11A: FGraphicsMode:=GraphicsModes[47];
 $11B: FGraphicsMode:=GraphicsModes[48];
 $130: FGraphicsMode:=GraphicsModes[49];
 $131: FGraphicsMode:=GraphicsModes[50];
 $132: FGraphicsMode:=GraphicsModes[51];
 $133: FGraphicsMode:=GraphicsModes[52];
 $134: FGraphicsMode:=GraphicsModes[53];
 $135: FGraphicsMode:=GraphicsModes[54];
 $136: FGraphicsMode:=GraphicsModes[55];
 $13D: FGraphicsMode:=GraphicsModes[56];
 $13E: FGraphicsMode:=GraphicsModes[57];
 $145: FGraphicsMode:=GraphicsModes[58];
 $146: FGraphicsMode:=GraphicsModes[59];
 $14A: FGraphicsMode:=GraphicsModes[60];
 else FGraphicsMode:=GraphicsModes[3];
 end;

case FGraphicsMode.Card of
 ctCGA,ctPCJR: FontType:=ftCGAFont;
 ctEGA: FontType:=ftEGAFont;
 ctVGA,ctVesa: FontType:=ftVGAFont;
 else  FontType:=ftVGAFont;
 end;

case FGraphicsMode.Bits of
 1: VGABox.SetColorPalette(VGAPalette1);
 2: VGABox.SetColorPalette(VGAPalette4[0]);
 3: VGABox.SetColorPalette(VGAPalette8);
 4: VGABox.SetColorPalette(VGAPalette16);
 8: VGABox.SetColorPalette(VGAPalette256);
 else VGABox.SetColorPalette(VGAPalette256);
 end;

AspectRatio:=(FGraphicsMode.Res.X/FGraphicsMode.Res.Y)/(4/3);
if AspectRatio<1.5 then AspectRatio:=1;
case FGraphicsMode.CharCell of
 cc8x8:
	begin
	FontType:=ftCGAFont;
	if DoubleSize and (FGraphicsMode.Res.X<640) and (FGraphicsMode.ModeType=mtText) then Font.Size:=11
	else Font.Size:=10;
	VGABox.CharSpacing:=Point(0,0);
	end;
 cc8x14:
	begin
	FontType:=ftVGAFont;
	Font.Size:=11;
	VGABox.CharSpacing:=Point(0,-1);
	end;
 cc8x16:
	begin
	FontType:=ftVGAFont;
	Font.Size:=12;
	VGABox.CharSpacing:=Point(-1,0);
	end;
 cc9x16:
	begin
	FontType:=ftVGAFont;
	Font.Size:=13;
	VGABox.CharSpacing:=Point(0,-1);
	end;
 end;
VGABox.SetImageSize(FGraphicsMode.TextRes.X,FGraphicsMode.TextRes.Y);
end;



function TConsoleWindow.VGAIndexToColor(C: Integer): TColor;
begin
VGABox.Canvas.Pen.Width:=1;
if FGraphicsMode.Bits<8 then
	begin
	if (C and $FF00)<>0 then VGABox.Canvas.Pen.Style:=psDot
	else VGABox.Canvas.Pen.Style:=psSolid;
	if (C and $80)<>0 then VGABox.Canvas.Pen.Mode:=pmXOR
	else VGABox.Canvas.Pen.Mode:=pmCopy;
	C:=C and $7F;
	end
else
	begin
	VGABox.Canvas.Pen.Style:=psSolid;
	VGABox.Canvas.Pen.Mode:=pmCopy;
	end;
case FGraphicsMode.Bits of
 1: Result:=VGABox.ColorPalette[C mod 2];
 2: Result:=VGABox.ColorPalette[C mod 4];
 3: Result:=VGABox.ColorPalette[C mod 16];
 4: Result:=VGABox.ColorPalette[C mod 16];
 8: Result:=VGABox.ColorPalette[C mod 256];
 15: Result:=Color15ToColor(C);
 16: Result:=Color16ToColor(C);
 24: Result:=C;
 32: Result:=C;
 end;
end;





function SearchPalette(Palette: array of TColor; C: TColor; Len: integer; var Index: integer): boolean;
var I: integer;
begin
Result:=True;
Index:=0;
for I:=0 to Len-1 do
 if Palette[I]=C then
	begin
	Index:=I;
	exit;
	end;
Result:=False;
end;


function TConsoleWindow.ColorToVGAIndex(C: TColor): Integer;
var Dots,XR: Word;
begin
if FGraphicsMode.Bits<8 then
	begin
	if VGABox.Canvas.Pen.Style=psSolid then Dots:=0
	else Dots:=$5500;
	if (C and $80)<>0 then
	if VGABox.Canvas.Pen.Mode=pmXOR then XR:=$80
	else XR:=0;
	end
else
	begin
	Dots:=0;
	XR:=0;
	end;
case FGraphicsMode.Bits of
 1: SearchPalette(VGABox.ColorPalette,2,C,Result);
 2: SearchPalette(VGABox.ColorPalette,4,C,Result);
 4: SearchPalette(VGABox.ColorPalette,16,C,Result);
 8: SearchPalette(VGABox.ColorPalette,256,C,Result);
 15: Color15ToColor(C);
 16: Color16ToColor(C);
 24: Result:=C;
 32: Result:=C;
 end;
Result:=Result or Dots or XR;
end;



//* function TConsoleWindow.GetGLMode: boolean;
//* begin
//* Result:=GLViewer.Visible;
//* end;

//* procedure TConsoleWindow.SetGLMode(const Value: boolean);
//* begin
//* VGABox.Visible:=not Value;
//* GLViewer.Visible:=Value;
//* if GLViewer.Visible then GLViewer.Repaint
//* else VGABox.Repaint;
//* end;

end.
