unit GPSReader;

{ Unit for reading Garming GPS25 Phase Data }

interface

uses Windows,Classes,Forms,SysUtils,MMSystem,CommPort;

{Status codes - indicates status of Position and Measurement data}

type TGPSStatusCode = (gsInvalidPort,gsTimeOut,gsNotFound,gsChecksum,gsValid,gsNotQueried);

{Structures to hold GPS data }

{Position Record - contains time and position information }

type
 TGPSPosition = packed record
 ID: byte;		{ID byte, always $28 for position record (needed for check sum)}
 Count: byte;		{Number of bytes in record (Needed for checksum calculation)}
 Alt: single;		{Ellipoid altitude (meters)}
 EstPosError: single;	{Estimated Position Error}
 HorzPosError: single;	{Horizontal Position Error}
 VertPosError: single;	{Vertical Position Error}
 FixType: word;		{Fix type: 0=No Fix, 1=No Fix, 2=2D, 3=3D, 4=2D Differential, 5=3D Differential}
 GPSWeekTime: double;	{GPS Time of week}
 Latitude: double;	{Latitude (Radians)}
 Longitude: double;	{Longitude}
 LongVel: single;	{Longitude Velocity (Meters/Sec)}
 LatVel: single;	{Latitude Velocity}
 AltVel: single;	{Altitude Velocity}
 CheckSum: byte;	{Checksum}
 end;


{Individual Satellite information item for array of satellites}

type
 TSatItem = Packed record
 Cycles: dword;		{Accumulated cycles}
 PseudoRange: double;	{Psuedo Range}
 Phase: word;		{Phase in 360/2048 of degree}
 CycleSlip: byte;	{Cycle slip. Non-zero = cycle slip}
 SignalStrength: byte;	{Signal strength}
 SatelliteID: byte;	{Satellite number}
 ValidInfo: byte;	{Valid information = non-zero}
 end;

{Array of satellite information}

type TSatArray = array [0..11] of TSatItem;

{Contains satellite measurement information}

type
 TGPSMeasure = packed record
 ID: byte;		{ID byte, always $28 for position record (needed for check sum)}
 Count: byte;		{Number of bytes in record (Needed for checksum calculation)}
 RecWeekTime: double;	{Receiver Time of week}
 RecWeekNum: word;	{Receiver week number}
 SatValues: TSatArray;	{Array of satellite information}
 CheckSum: byte;	{Checksum}
 end;

{Combined record containing all data GPS data}

type
 TGPSData = record
 PositionStatus: TGPSStatusCode;	{Status codes for each data item}
 MeasureStatus: TGPSStatusCode;
 TimeOffset: double;			{Time Offset in Secs - Compensate for data rate}
 Position: TGPSPosition;		{Position information}
 Measure: TGPSMeasure;			{Measurement information}
 end;

{ TGPSReader - This object reads GPS data from a Garmin GPS25 unit
  through the PC serial port. The object inherits serial access from
  the TCommPort parent object. Set the TCommPort properites to control
  the serial port. To read data, call 'RequestPositionData', 'RequestMeasureData',
  or 'RequestBoth.' The 'TimeOut' property controls how long it waits before
  it fails. Data is returned through the 'Data' parameter. PositionStatus
  and MeasureStatus members of the structure indicate the status of the data.}


type
 TGPSReader = class(TCommPort)
  private
    Buffer: string;
    FTimeOut: dword;
    SyncFlag: boolean;
    SyncStr: string[4];
    TargetID: AnsiChar;
    FReading: boolean;
    function EvaluateChecksum(var V; Count: integer): boolean;
    procedure DecodePositionData(var Data: TGPSData);
    procedure DecodeMeasureData(var Data: TGPSData);
    procedure ReceiveData(Sender: TObject; Count: Word);
    procedure FindSync(S: string);
    function ReadDataBlock(ID: AnsiChar; MinCount: integer): boolean;
    function CharsToTime(C: integer): double;
    procedure GetPositionData(var Data: TGPSData);
    procedure GetMeasureData(var Data: TGPSData);
  protected
  public
    constructor Create(AOwner: TComponent); override;
    procedure RequestMeasureData(var Data: TGPSData);
    procedure RequestPositionData(var Data: TGPSData);
    procedure RequestBoth(var Data: TGPSData);
  published
    property TimeOut: dword read FTimeout write FTimeout default 5000;
    property Reading: boolean read FReading write FReading default False;
  end;

function GetStatusCodeStr(SC: TGPSStatusCode): string;

procedure Register;

implementation

const MinMeasureCnt = 233;	{Minimum characters necessary for each message type}
const MinPositionCnt = 62;



function GetStatusCodeStr(SC: TGPSStatusCode): string;
{Return GPS status string}
begin
case SC of
 gsInvalidPort: Result:='Invalid Port';
 gsTimeOut: Result:='Timeout';
 gsNotFound: Result:='Not Found';
 gsChecksum: Result:='Checksum';
 gsValid: Result:='Valid';
 gsNotQueried: Result:='Not Queried';
 end;
end;



constructor TGPSReader.Create(AOwner: TComponent);
{Create GPS reader}
begin
inherited Create(AOwner);
OnReceive:=ReceiveData;
Events:=[tceRxChar];
FTimeOut:=5000;
FReading:=False;
end;




function TGPSReader.EvaluateChecksum(var V; Count: integer): boolean;
{Calculate checksum of a GPS structure}
var T: PByteArray;
var Sum,I: integer;
begin
T:=@V; Sum:=0;
for I:=0 to Count-1 do Sum:=Sum + T[I];
Result:=(Sum and $FF) = 0;
end;



function TGPSReader.CharsToTime(C: integer): double;
{Calculate the time delay for specified number of character}
{Used to find the time delay for the GPS data}
var CT: double;
begin
CT:=1/(GetBaudRateValue(BaudRate)/10);
Result:=C * CT;
end;



procedure TGPSReader.DecodePositionData(var Data: TGPSData);
{Find and decode GPS Position data}
var L: integer;
const PositionHead = #$03#$10#$28#$36;
begin
{Locate position information in buffer}
L:=POS(PositionHead,Buffer);
if L>0 then
	begin
	Move(Pointer(Integer(@Buffer[L])+2)^,Pointer(@Data.Position)^,SizeOf(TGPSPosition));
	if EvaluateCheckSum(Data.Position,SizeOf(TGPSPosition)) then Data.PositionStatus:=gsValid
	else Data.PositionStatus:=gsCheckSum;
	Data.TimeOffset:=CharsToTime(Length(Buffer));
	end
else Data.PositionStatus:=gsNotFound;
end;



procedure TGPSReader.DecodeMeasureData(var Data: TGPSData);
{Find and decode GPS Measurement data}
var L: integer;
const MeasureHead = #$03#$10#$29#$E2;
begin
{Locate measurement information in buffer}
L:=POS(MeasureHead,Buffer);
if L>0 then
	begin
	Move(Pointer(Integer(@Buffer[L])+2)^,Pointer(@Data.Measure)^,SizeOf(TGPSMeasure));
	if EvaluateCheckSum(Data.Measure,SizeOf(TGPSMeasure)) then Data.MeasureStatus:=gsValid
	else Data.MeasureStatus:=gsCheckSum;
	end
else Data.MeasureStatus:=gsNotFound;
end;



procedure TGPSReader.FindSync(S: string);
{Look for the four byte sync pattern }
var I,J: integer;
const SyncSize = 3;
begin
for I:=1 to Length(S) do
	begin
	if Length(SyncStr)>=SyncSize then Delete(SyncStr,1,1);
	SyncStr:=SyncStr+S[I];
	if (Length(SyncStr)>=SyncSize) and
	   (SyncStr[1]=#$03) and (SyncStr[2]=#$10) and
           (SyncStr[3]=TargetID) then
		begin
		SyncFlag:=True;
		Buffer:=SyncStr;
		for J:=I+1 to Length(S) do
	 	if S[J]<>#$10 then Buffer:=Buffer+S[J]
	 	else if Buffer[Length(Buffer)]<>#$10 then Buffer:=Buffer+S[J];
		break;
		end;
	end;
end;



procedure TGPSReader.ReceiveData(Sender: TObject; Count: Word);
{Event handler for received serial data}
var I: integer;
var S: string;
begin
ReadString(S,Count);
if SyncFlag then
	begin
	for I:=1 to Length(S) do
	 if (S[I]<>#$10) or (Length(Buffer)=0) then Buffer:=Buffer+S[I]
	 else if Buffer[Length(Buffer)]<>#$10 then Buffer:=Buffer+S[I];
	end
else FindSync(S);
end;



function TGPSReader.ReadDataBlock(ID: AnsiChar; MinCount: integer): boolean;
{Wait for data specified by ID, of length specified by MinCount}
var EndTime: dword;
var Cnt: integer;
begin
SyncFlag:=False;
SyncStr:=''; Buffer:='';
TargetID:=ID;
EndTime:=timeGetTime+FTimeOut;
FReading:=True;
repeat
	begin
	Sleep(20);
	Application.ProcessMessages;
	if Application.Terminated then abort;
	if timeGetTime>EndTime then break;
	end;
until Length(Buffer)>=MinCount;
FReading:=False;
Cnt:=Length(Buffer);
Result:=Cnt>=MinCount;
end;



procedure TGPSReader.GetMeasureData(var Data: TGPSData);
{Request Measurement Data and store result in Data}
begin
Data.MeasureStatus:=gsTimeOut;
if ReadDataBlock(#$29,MinMeasureCnt) then DecodeMeasureData(Data);
end;



procedure TGPSReader.GetPositionData(var Data: TGPSData);
{Request Measurement Data and store result in Data}
begin
Data.PositionStatus:=gsTimeOut;
if ReadDataBlock(#$28,MinPositionCnt) then DecodePositionData(Data);
end;



procedure TGPSReader.RequestMeasureData(var Data: TGPSData);
{Request Measurement Data and store result in Data}
begin
Data.PositionStatus:=gsNotQueried;
Data.MeasureStatus:=gsInvalidPort;
PortActive:=True;
GetMeasureData(Data);
end;



procedure TGPSReader.RequestPositionData(var Data: TGPSData);
{Request Measurement Data and store result in Data}
begin
Data.MeasureStatus:=gsNotQueried;
Data.PositionStatus:=gsInvalidPort;
PortActive:=True;
GetPositionData(Data);
end;



procedure TGPSReader.RequestBoth(var Data: TGPSData);
{Request both Measurement and Position data}
{Position is read last to avoid having a time >1 second old}
begin
Data.MeasureStatus:=gsInvalidPort;
Data.PositionStatus:=gsInvalidPort;
PortActive:=True;
GetMeasureData(Data);
GetPositionData(Data);
end;



procedure Register;
begin
RegisterComponents('Fountain',[TGPSReader]);
end;



end.
