unit ImageUndoer;

interface

uses Classes,SysUtils,ExtCtrls,Graphics, BitmapLib,CircularBuffer;




type TBitmapUndoer = class(TComponent)
 private
  function GetMaxUndo: integer;
  procedure SetMaxUndo(const Value: integer);
 protected
  FUndoer: TBitmapUndo;
  FRedoer: TBitmapUndo;
 public
  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;
  procedure StoreUndo(BM: TBitmap); dynamic;
  procedure Undo(BM: TBitmap); dynamic;
  procedure Redo(BM: TBitmap); dynamic;
  procedure Clear; dynamic;
  procedure DeleteLastUndo;
  property Undoer: TBitmapUndo read FUndoer;
  property Redoer: TBitmapUndo read FRedoer;
 published
  property MaxUndo: integer read GetMaxUndo write SetMaxUndo;
 end;


type TStoreQueryEvent = procedure(Sender: TObject; var CanStore: Boolean) of object;



type TImageUndoer = class(TBitmapUndoer)
 private
  FImage: TImage;
  FOnChange: TNotifyEvent;
  FOnStoreQuery: TStoreQueryEvent;
  FAutoUpdate: boolean;
  procedure SetAutoUpdate(const Value: boolean);
 protected
  procedure SetImage(V: TImage); virtual;
  function GetCount: integer; virtual;
  procedure HandleChangeEvent(Sender: TObject); virtual;
  procedure Changed; virtual;
  function  StoreQuery: boolean; virtual;
  procedure Notification(AComponent: TComponent; Operation: TOperation); override;
 public
  Enabled: boolean;
  constructor Create(AOwner: TComponent); override;
  procedure StoreUndo(Bitmap: TBitmap); overload;
  procedure StoreUndo; overload;
  procedure Undo;
  procedure Redo;
  procedure Clear; override;
 published
  property Image: TImage read FImage write SetImage;
  property OnStoreQuery: TStoreQueryEvent read FOnStoreQuery write FOnStoreQuery;
  property OnChange: TNotifyEvent read FOnChange write FOnChange;
  property Count: integer read GetCount;
  property AutoUpdate: boolean read FAutoUpdate write SetAutoUpdate;
 end;

procedure Register;

implementation


procedure Register;
begin
RegisterComponents('Fountain', [TImageUndoer]);
end;


{ TBitmapUndoer }


constructor TBitmapUndoer.Create(AOwner: TComponent);
begin
inherited;
FUndoer:=TBitmapUndo.Create(100);
FRedoer:=TBitmapUndo.Create(100);
end;



destructor TBitmapUndoer.Destroy;
begin
FUndoer.Free;
FRedoer.Free;
inherited;
end;


procedure TBitmapUndoer.Clear;
begin
FUndoer.Clear;
FRedoer.Clear;
end;


procedure TBitmapUndoer.StoreUndo(BM: TBitmap);
begin
FUndoer.AddUndoBitmap(BM);
FRedoer.Clear;
end;


procedure TBitmapUndoer.DeleteLastUndo;
begin
FUndoer.DeleteLastUndo;
end;


procedure TBitmapUndoer.Undo(BM: TBitmap);
begin
if FUndoer.Count<=0 then exit;
FRedoer.AddUndoBitmap(BM);
FUndoer.GetUndoBitmap(BM);
end;



procedure TBitmapUndoer.Redo(BM: TBitmap);
begin
if FRedoer.Count<=0 then exit;
FUndoer.AddUndoBitmap(BM);
FRedoer.GetUndoBitmap(BM);
end;



function TBitmapUndoer.GetMaxUndo: integer;
begin
Result:=FUndoer.Fileundo.MaxUndos;
end;

procedure TBitmapUndoer.SetMaxUndo(const Value: integer);
begin
FUndoer.Fileundo.MaxUndos:=Value;
FRedoer.Fileundo.MaxUndos:=Value;
end;


{ TImageUndoer }


procedure TImageUndoer.Changed;
begin
if Assigned(FOnChange) then FOnChange(Self);
end;



constructor TImageUndoer.Create(AOwner: TComponent);
begin
inherited;
FImage:=nil;
Enabled:=True;
end;


function TImageUndoer.GetCount: integer;
begin
Result:=FUndoer.Count;
end;

procedure TImageUndoer.HandleChangeEvent(Sender: TObject);
{Note the change event only captures drawing to canvas}
{Does not capture bitmap changing size - could maybe}
{make this work by creating a bitmap decendent and }
{overriding the SetWidth and SetHeight procedures}
begin
if not FAutoUpdate then exit;
StoreUndo;
end;



procedure TImageUndoer.SetImage(V: TImage);
begin
if FImage<>V then
	begin
	FImage:=V;
	end;
end;



procedure TImageUndoer.StoreUndo;
begin
if not Enabled then exit;
if FImage=nil then Raise Exception.Create('No Image Set');
if (FImage.Picture.Bitmap.Width=0) or (FImage.Picture.Bitmap.Height=0) then exit;
if not StoreQuery then exit;
inherited StoreUndo(FImage.Picture.Bitmap);
Changed;
end;



procedure TImageUndoer.StoreUndo(Bitmap: TBitmap);
begin
inherited StoreUndo(Bitmap);
Changed;
end;



function TImageUndoer.StoreQuery: boolean;
begin
Result:=True;
if Assigned(FOnStoreQuery) then FOnStoreQuery(Self,Result);
end;



procedure TImageUndoer.Undo;
begin
if FImage=nil then Raise Exception.Create('No Image Set');
inherited Undo(FImage.Picture.Bitmap);
Changed;
end;




procedure TImageUndoer.Redo;
begin
if FImage=nil then Raise Exception.Create('No Image Set');
inherited Redo(FImage.Picture.Bitmap);
Changed;
end;



procedure TImageUndoer.SetAutoUpdate(const Value: boolean);
begin
if FAutoUpdate<>Value then
	begin
	FAutoUpdate:= Value;
	if FAutoUpdate then
		begin
		{Assign event to bitmap canvas vs the image canvas, otherwise )
		{Image will create a bitmap of Width/Height size at startup}
		{Also there is a problem working with ImageEx where the undo images overwrite the backbitmap}
		if not (csDesigning in ComponentState) then
			Fimage.Picture.Bitmap.Canvas.OnChanging:=HandleChangeEvent;
		end
	else
		begin
		if not (csDesigning in ComponentState) then
			Fimage.Picture.Bitmap.Canvas.OnChanging:=nil;
		end;
	end;
end;



procedure TImageUndoer.Notification(AComponent: TComponent;
  Operation: TOperation);
{Handle the live deletion of an Image controlled by the object at designtime }
begin
inherited Notification(AComponent, Operation);
if (Operation = opRemove) and (AComponent = FImage) then FImage := nil;
end;



procedure TImageUndoer.Clear;
begin
inherited;
Changed;
end;


end.
