unit Magnifier;

interface

uses Windows,Forms,Types,Classes,Controls,Graphics,ExtCtrls,IntVector;

type TZoomFactor = 1..32;

type TMagUpdate = procedure(Sender: TObject; X, Y: Integer) of object;


type TMagnifier = class(TImage)
 private
  FZoom: TZoomFactor;
  FPosition: TPoint;
  ImageTimer: TTimer;
    FOnMagUpdate: TMagUpdate;
    FTrackMouse: boolean;
    FShowCurosr: boolean;
    FShowCursor: boolean;
    procedure DrawCross(Image: TImage; C: TPoint);
    procedure MagnifyScreen(Image: TImage; Pos: TPoint; Zoom: double);
    procedure DisplayCursor(Image: TImage);
    function ZoomRectange(R: TRect; Zoom: double): TRect;
 protected
  procedure TimerEvent(Sender: TObject);
  function GetEnabled: boolean;
  procedure SetEnabled(V: boolean);
 public
  CenterColor: TColor;
  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;
  property Position: TPoint read FPosition write FPosition;
 published
  property Zoom: TZoomFactor read FZoom write FZoom default 1;
  property Enabled: boolean read GetEnabled write SetEnabled;
  property OnMagUpdate: TMagUpdate read FOnMagUpdate write FOnMagUpdate;
  property TrackMouse: boolean read FTrackMouse write FTrackMouse;
  property ShowCursor: boolean read FShowCursor write FShowCursor;
 end;


procedure Register;

implementation


procedure Register;
begin
RegisterComponents('Fountain', [TMagnifier]);
end;


{------------------------------------------------------------------------------}

function TMagnifier.ZoomRectange(R: TRect; Zoom: double): TRect;
{Reduce rectangle size to zoom}
var Size: TPoint;
var ZX,ZY: integer;
begin
if (R.Bottom and 1)<>1 then Dec(R.Top);
Result:=R;
Size:=IntVectorSubtract(R.BottomRight,R.TopLeft);
ZX:=(Trunc(Size.X/Zoom) - Size.X) div 2;
ZY:=(Trunc(Size.Y/Zoom) - Size.Y) div 2;
InflateRect(Result,ZX,ZY);
end;



procedure TMagnifier.DrawCross(Image: TImage; C: TPoint);
const Size = 17;
var X,Y: integer;
begin
{Center Left-Right}
for X:=C.X-Size to C.X+Size-1 do
 if X<>C.X then
 	begin
 	if (X mod 2)=0 then Image.Canvas.Pixels[X,C.Y]:=clRed
 	else Image.Canvas.Pixels[X,C.Y]:=clBlue;
	end;
{Top Left-Right}
Image.Canvas.MoveTo(C.X-Size,C.Y-1);
Image.Canvas.LineTo(C.X-Size div 2,C.Y-1);
Image.Canvas.MoveTo(C.X+Size div 2,C.Y-1);
Image.Canvas.LineTo(C.X+Size,C.Y-1);

{Bottom Left-Right}
Image.Canvas.MoveTo(C.X-Size,C.Y+1);
Image.Canvas.LineTo(C.X-Size div 2,C.Y+1);
Image.Canvas.MoveTo(C.X+Size div 2,C.Y+1);
Image.Canvas.LineTo(C.X+Size,C.Y+1);

{Center Top-Bottom}
for Y:=C.Y-Size to C.Y+Size-1 do
 if Y<>C.Y then
 	begin
 	if (Y mod 2)=0 then Image.Canvas.Pixels[C.X,Y]:=clRed
 	else Image.Canvas.Pixels[C.X,Y]:=clBlue;
	end;

{Left Top-Bottom}
Image.Canvas.MoveTo(C.X-1,C.Y-Size div 2);
Image.Canvas.LineTo(C.X-1,C.Y-Size);
Image.Canvas.MoveTo(C.X-1,C.Y+Size div 2);
Image.Canvas.LineTo(C.X-1,C.Y+Size);

{Right Top-Bottom}
Image.Canvas.MoveTo(C.X+1,C.Y-Size div 2);
Image.Canvas.LineTo(C.X+1,C.Y-Size);
Image.Canvas.MoveTo(C.X+1,C.Y+Size div 2);
Image.Canvas.LineTo(C.X+1,C.Y+Size);
end;


procedure TMagnifier.DisplayCursor(Image: TImage);
{Draw cursor at center of image}
var C: TPoint;
begin
if not ShowCursor then exit;
C:=IntScalarDivide(Image.ClientRect.BottomRight,2);
Image.Canvas.Pen.Color:=clRed;
Image.Canvas.Brush.Color:=clBlack;
Image.Canvas.Pen.Width:=1;
Image.Canvas.Pen.Style:=psDot;
CenterColor:=Image.Canvas.Pixels[C.X,C.Y];
DrawCross(Image,C);
end;



procedure TMagnifier.MagnifyScreen(Image: TImage; Pos: TPoint; Zoom: double);
{Magnify desktop and display result in Image.
{Magnified image is taken at Pos and enlarge by Zoom factor}
var DeskTopDC : HDC;
var DeskTopHand: HWND;
var Canvas: TCanvas;
var ImageRect,ScreenRect: TRect;
var IC: TPoint;
begin
{Setup canvas for desktop}
DeskTopHand:=GetDeskTopWindow;
DeskTopDC:=GetDC(DeskTopHand);
Canvas:=TCanvas.Create;
Canvas.Handle:=DeskTopDC;

{Get Rectangles for Screen and Image}
ImageRect:=Image.ClientRect;
ScreenRect:=Rect(0,0,Screen.Width,Screen.Height);

{Use Image Rectangle and offset to specified position}
IC:=IntScalarDivide(ImageRect.BottomRight,2);
Pos:=IntVectorSubtract(Pos,IC);
OffsetRect(ImageRect,Pos.X,Pos.Y);

{Do magnification by shrinking rectangle}
ImageRect:=ZoomRectange(ImageRect,Zoom);

Image.Canvas.Pen.Color:=clBlue;
Image.Canvas.Brush.Color:=clBlue;
Image.Canvas.Brush.Style:=bsSolid;
Image.Canvas.Rectangle(Image.ClientRect);
Image.Canvas.CopyRect(Image.ClientRect,Canvas,ImageRect);
DisplayCursor(Image);
ReleaseDC(DeskTopHand, DeskTopDC);
Canvas.Free;
end;



{ TMagnifier }

constructor TMagnifier.Create(AOwner: TComponent);
begin
inherited;
ImageTimer:=TTimer.Create(Self);
ImageTimer.Enabled:=False;
ImageTimer.Interval:=50;
ImageTimer.OnTimer:=TimerEvent;
FZoom:=1;
TWinControl(AOwner).DoubleBuffered:=True;
TrackMouse:=True;
FShowCursor:=True;
end;



destructor TMagnifier.Destroy;
begin
ImageTimer.Free;
inherited;
end;



function TMagnifier.GetEnabled: boolean;
begin
Result:=ImageTimer.Enabled
end;



procedure TMagnifier.SetEnabled(V: boolean);
begin
if V<>ImageTimer.Enabled then
	begin
	ImageTimer.Enabled:=V;
	end;
end;


procedure TMagnifier.TimerEvent(Sender: TObject);
begin
if csDesigning in Self.ComponentState then exit;
if FTrackMouse then
	begin
	GetCursorPos(FPosition);
	Self.Picture.Bitmap.Width:=Self.Width;
	Self.Picture.Bitmap.Height:=Self.Height;
	end;
MagnifyScreen(Self,FPosition,FZoom);
if Assigned(OnMagUpdate) then OnMagUpdate(Self,FPosition.X,FPosition.Y);
end;




end.
