unit ScrollEditor;

interface

uses Windows,SysUtils,Messages,Classes,Controls,StdCtrls, Graphics,StrUtils, Clipbrd,
     Math, IntVector, DSupport, ScrollBarWindow, BItmapLib, Hash, ExtCtrls;


{=== Global variables and procedure declarations for all Editor Components ====}

const WildCard= CtrlW;
const NumericWildChar = CtrlN;


const CtrlBkSlash= #$1C;  CtrlRtBrkt= #$1D;
const CtrlCaret= #$1E;    CtrlUScore= #$1F;


function RemoveLineFeeds(S: string): string;
function AddLineFeeds(S: string): string;
function RemoveLineBreaks(S: string): string;

{============ Declarations for Highlight object ================================}
{A simple object is used to make the highlight shareable through the "Attach" feature}
{NonDelete Highlights don't delete the underlying text if the user types - Used for searching}

type THighLight = class(TObject)
 public
 NonDelete: boolean;
 Marker: TRect;
 end;


{============ Declarations for Row Mark object ================================}

type TRowMarks = class(TObject)
 private
 protected
  List: TStringList;
 public
  constructor Create;
  destructor Destroy; override;
  procedure Delete(Group: string; Inx: integer);
  function AddLine(Group: string; Line: integer; Color: TColor): integer;
  function IsMarkedLine(Line: integer; var Color: TColor; var Start: integer): boolean;
  procedure DeleteLine(Group: string; Line: integer);
  procedure ClearGroup(Group: string);
  procedure ClearAll;
  procedure DeleteGroup(Group: string);
 end;


{============ Declarations for Cursor Position object =========================}
{A simple object is used to make the Cursor shareable through the "Attach" feature}

type TCursorObject = class(TObject)
 public
  Position: TPoint;
  ColumnMarking: boolean;
 end;

{============ Declarations for Undo object ====================================}

const DefaultUndoSize = 256000;

{Defines all possible undo type commands}

type TUndoCommand=(ucMove,ucBackDelete,ucForwardDelete,ucBackInsert,ucForwardInsert,ucError,ucNull);

{TUndostate controls multi-step undo's.
 usStart = Start  of multi-step undo process
 usStep = Intermediate step of multi-step undo process
 usEnd = End of multi-step undo process
 usComplete = An undo that is completed in one step
}

type TUndoState=(usStart,usStep,usEnd,usComplete);

{Each TUndoItem holds all the information needed to undo a change }

type TUndoItem= Record
 Command: TUndoCommand;
 State: TUndoState;
 OldCursor: TPoint;
 UndoBuffer: string;
 end;

{Resizeable array of undo items}

type TUndoList = array of TUndoItem;

{Prototype for user supplied procedure to handle all undos}

type TUndoProcedure = procedure (UndoItem: TUndoItem; Done: boolean) of Object;

{Protottype for  undo state change event - mostly used for debug }

Type TStateEvent = procedure (UndoIndex: integer; Item: TUndoItem);

{Declaration for undo object}

type
 TUndoControl = class(TObject)
 private
  FUndoSize: integer;
  FUndoProcedure: TUndoProcedure;
  FOnStateChange: TStateEvent;
  FOldDeltaCursor: TPoint;
  FOnChange: TNotifyEvent;
  FConsolidStart: integer;
  procedure IncrementIndex;
  procedure SetUndoSize(V: integer);
  function CheckForNullPairs(UndoItem: TUndoItem): boolean;
 protected
  procedure UndoItem(ItemInx: integer; Done: boolean);
  procedure HandleChange;
 public
  UndoList: TUndoList;
  UndoIndex: integer;		{ Points to end of Undo list  }
  UndoStart : Integer;		{ Points to start of list  }
  SingleStep: boolean;
  constructor Create;
  destructor Destroy; override;
  function GetStateStr(C: TUndoState): string;
  function GetCommandStr(C: TUndoCommand): string;
  procedure InvertUndoItem(var UndoItem: TUndoItem);
  procedure Initialize;
  function StoreUndoItem(UndoItem: TUndoItem): boolean;
  procedure StoreChanges(Command: TUndoCommand; DeletedChars: string; LastCursor: TPoint; State: TUndoState);
  procedure DoDeltaCursorStore(CurrentCursor: TPoint);
  procedure MarkConsolidationStart;
  procedure ConsolidateUndos;
  procedure SetupDeltaUndo(CurrentCursor: TPoint);
  function DoUndo: boolean;
  property UndoSize: integer read FUndoSize write SetUndoSize default 16000;
  property UndoProcedure: TUndoProcedure read FUndoProcedure write FUndoProcedure;
  property OnChange: TNotifyEvent read FOnChange write FOnChange;
  property OnStateChange: TStateEvent read FOnStateChange write FOnStateChange;
end;


{Object for searching string lists for string that can cross lines}

type TListSearcher = class(TObject)
 private
  LineInx,LineLength: integer;
  CurLine: string;
  FStringList: TStringList;
  FTarget: string;
  procedure GetSpecifiedLine(L: integer);
  function GetCharAtCursor(Loc: TPoint): Char;
  function IncrementLoc(var Loc: TPoint): boolean;
  function DecrementLoc(var Loc: TPoint): boolean;
  function CompareChars(TargetChar,ListChar: Char): boolean;
  function MatchStringAtCursor(var Loc: TPoint; var Mark: TRect): boolean;
  function IsWholeWord(Mark: TRect): boolean;
 protected
  function EndOfList(Loc: TPoint): boolean;
  constructor Create(StringList: TStringList);
 public
  IgnoreCase: boolean;
  WholeWordMode: boolean;
  HyphenatedWords: boolean;
  function SearchForward(Target: string; var Loc: TPoint; var Mark: TRect): boolean;
  function SearchBackward(Target: string; var Loc: TPoint; var Mark: TRect): boolean;
 end;

{Edit Buffer Object - This is an elaborate version of the Delphi TStringList}
{Several routines are overridden to slightly change the behavoir of our buffer scheme}

type TCharType = (ctASCII,ctBigEnd,ctLittleEnd);

type TCharSet = set of char;

type TEditBuffer = class(TStringList)
 private
  function GetEmpty: boolean;
 protected
  procedure PutTextInList(const Value: string);
  function GetTextStr: string; override;
  procedure SetTextStr(const Value: string); override;
  procedure Changing; override;
  procedure Changed; override;
 public
  Busy: boolean;
  CharType: TCharType;
  BOM: Word;
  constructor Create;
  function CustomText(Delimter: string): string;
  property Empty: boolean read GetEmpty;
 end;


{=================== Scrolling Editor Object ==================================}

{
Notes:

1. Array Structure. The editor buffer consists of an array of
strings in a modified Delphi TStringList object. This modified
object overrides the several methods to make the text conform
with the editor's model.

2. Editor Model. As an array of "Lines", a TStringList normally
doesn't store Carriage Returns. When Carriage Return is
entered, the line is split. In "Multi-Line" mode, the editor
views the beginning of each line (except the first) as having
an implied Carriage Return.

In single-line mode, Carriage Returns are stored and then
displayed as a Control-M character. No lines are split.

3. Cursor. The cursor is indicated by a Delphi TPoint
structure. The line number is indicated by the "Y" component
of the point. This is a zero-based array so lines run from 0
to Count-1. The "Y" Component of the Cursor cannot exceed
Count-1;

The column is specified by the "X" component of the Point.
Since Delphi strings are indexed starting with 1, the first
character on the line is specified by 1. If the X is zero, X
is at the beginning of the line and is pointing at the
implied Carriage Return. The last character on a line equals
the length of the string and the "X" Component of the cursor
cannot exceed the line's Length.

4. Marker. A marker is used to mark a block of test and
usually highlight it. A marker is represented by a Delphi
TRect structure which consists of two points. The two points
represent the beginning and end of the highlighted test. The
highlighted text starts one character after the "TopLeft"
point and includes the character at the "BottomRight" point.
For example, the points (0,1) and (4,1) would mark characters
1, 2, 3, and 4 on the second line. If the two points are
equal, no text is marked.

}

const TabLimitLength =2048;		{Maximum number of characters before you start ignoring tabs}

{Defines single and multiline editing modes}

type TEditorStyle = (esSingleLine,esMultiLine);    

{Xed-style error conditions}

type TXEDError=(xeNone,xeNotFound,xeStart,xeEnd,xeNoUndo,xeNoRedo,xeCursor,xeLetter);

{Xed-style scoop register states}

type TScoopCommand = (scScoop,scUnScoop,scAbort);

{Keymapping styles for "Stand Alone" mode }

type TKeyMapping = (kmDefault,kmXED,kmXEDLite,kmNoMap);

{Xed-style scoop register states}

type TScoopEvent = procedure(Command: TScoopCommand; Mark: TRect) of object;

{Prototype for scale change events.}

Type TScaleEvent = procedure (Scale: double) of object;

{Actual Scrolling-Editor object declaration}

type
 PScrollEditor = ^TScrollEditor;
 TScrollEditor = class(TScrollBarWindow)
 private
  FScale: double;
  FCursorWidth: integer;
  ScreenCharHeight,ScreenCharWidth: integer;
  OldBlinkRate: integer;
  FShadowColor,FControlColor,
  FHighlightColor,FNonDeleteColor: TColor;
  HighFontColor,NoDelFontColor: TColor;
  FShadowLine: boolean;
  FTextMetric: TTextMetric;
  FShowAll: boolean;
  FLinearHScroll: boolean;
  FGutterSize: integer;
  CursorMarking: boolean;
  MarkStartPos: TPoint;
  MarkBasePos: TPoint;
  WordMarked: boolean;
  LeftMouse: boolean;
  DragStartPos: TPoint;
  FWorkColumn: integer;
  FOnChange: TNotifyEvent;
  FOnScoopEvent: TScoopEvent;
  FEditorStyle: TEditorStyle;
  FIsModified: boolean;
  FChangeCount: integer;
  FMoveMarking: boolean;
  FSpacesPerTab: integer;
  FStandAlone: boolean;
  FKeyMapping: TKeyMapping;
  FHighlightRow: integer;
  FReadOnly: boolean;
  FFilename: string;
  FOnUnFocus: TNotifyEvent;
  FOnFocus: TNotifyEvent;
    FOnPaint: TNotifyEvent;
    FUseTabStops: boolean;
    FOnScaleChange: TScaleEvent;
  procedure CreateTextCursor;
  procedure DisplayCursor;
  function CursorToColumn(Cursor: TPoint): integer;
  function ColumnToCursorX(Cursor: TPoint): integer;
  procedure GetColumnMarkers(Line, Width: integer; Marker: TRect; var M1,M2: integer);
  function RasPointInRect(M: TRect; P: TPoint; ExcludeWhite: boolean): boolean;
  function ColumnizeMarker(M: TRect): TRect;
  procedure GetScreenParameters;

  procedure WriteLine(X,Y, ColumnOffset, ClientWidth, MarkCol1,MarkCol2: integer; S: string; MarkedLine,CursorLine,Clear: boolean);
  procedure DisplayStringSegment(Canvas: TCanvas;var S: string; var PT: TPoint);
  procedure SetTextColors(Canvas:TCanvas; FontColor,BrushColor: TColor);
  procedure MarkCursorDifference;
  procedure ScrollToCursor(MarkDifference: boolean);

  procedure SetShowAll(V: boolean);
  procedure SetGutter(V: integer);
  procedure SetControlColor(V: TColor);
  procedure SetHighlightColor(V: TColor);
  procedure SetEditBuffer(V: TEditBuffer);
  procedure SetCursorWidth(V: integer);
  procedure SetIsModified(V: boolean);
  procedure SetShadowLine(V: boolean);
  procedure SetShadowColor(V: TColor);
  procedure SetMoveMarking(V: boolean);
  procedure SetSpacesPerTab(V: integer);

  procedure DoUndoTask(UndoItem: TUndoItem; CursorAfter,Done: boolean);
  procedure UndoCallBack(UndoItem: TUndoItem; Done: boolean);
  procedure RedoCallBack(UndoItem: TUndoItem; Done: boolean);
  procedure UndoChange(Sender: TObject);

  procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
  procedure WMSetFocus(var Message:TWMSetFocus); message WM_SETFOCUS;
  procedure WMKillFocus(var Message: TWMKillFocus); message WM_KillFocus;
  procedure WMSize(var Message: TWMSize); message WM_SIZE;
  procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
  procedure HandleBufferChange(Sender: TObject);
  procedure InitializeMoves;
  procedure FinalizeMoves(Success,ZeroColumn: boolean; ErrorCode: TXEDError);
  function GetText: string;
  procedure SetText(const Value: string);

  procedure HandleControlC(XEDStyle: boolean);
  procedure HandleControlX(Shift: boolean; XEDStyle: boolean);
  procedure SetHighlightRow(const Value: integer);
  procedure SetNonDeleteColor(const Value: TColor);
  procedure DrawRowMark(Canvas: TCanvas; X,Y,Width,Height: integer; BackColor,HiColor: TColor);
  procedure StandAloneInsertChar(C: char);
  procedure DoWordBackSearch;
  procedure DoWordFwdSearch;
  function GetTarget: string;
    procedure SetTUseTabStops(const Value: boolean);
    procedure ScaleScreen(Scale: double);
    procedure SetScale(const Value: double);
    function GetPreciseCharWidth: double;
 protected
  EditorAttached: boolean;
  procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
  procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  function DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): Boolean; override;
  function DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): Boolean; override;

  procedure DoStartDrag(var DragObject: TDragObject); override;
  procedure DragOver(Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean); override;
  procedure DisplayBuffer;
  procedure UpdateScrollBars(MaxWidth: integer);
  procedure DoHorzScroll(Sender: TObject; ScrollCode: TScrollCode; var ScrollPos: Integer); override;
  procedure DoVertChange(Sender: TObject); override;
  procedure Paint; override;
  procedure ScoopStart(Cursor: TPoint);
  procedure ScoopEnd(Cursor: TPoint);
  procedure KeyDown(var Key: Word; Shift: TShiftState); override;
  procedure KeyPress(var Key: Char); override;
 public
  UndoControl: TUndoControl;
  RedoControl: TUndoControl;
  EditBuffer: TEditBuffer;
  Searcher: TListSearcher;

  Error: TXEDError;

  AutoUpdate: boolean;
  OverWrite: Boolean;
  MarkHyphenated: boolean;
  AppendEOF: boolean;

  ScrollOffset: TPoint;
  WorkingCursor: TCursorObject;
  Highlight: THighLight;
  RowMarks: TRowMarks;

  property HighlightRow: integer read FHighlightRow write SetHighlightRow;
  function DoXEDControlKeys(Key: Char; Shift: boolean): boolean;
  function DefaultControlKeys(Key: char; Shift: boolean): boolean;
  function DoXEDLiteControlKeys(Key: Char; Shift: boolean): boolean;
  procedure FlipCase;
  procedure MatchIndentAbove;
  property MoveMarking: boolean read FMoveMarking write SetMoveMarking;
  property Canvas;
  procedure DragDrop(Source: TObject; X, Y: Integer); override;

  procedure DoChange(Optimize,DoEvent: boolean); overload;
  procedure DoChange(Optimize: boolean); overload;
  procedure DoClearBuffer;
  procedure ResetMarkers;
  procedure CenterVerticalCursor;
  procedure OptimizeScrolling;
  procedure ScrollVert(Sender: TObject; ScrollCode: TScrollCode; var ScrollPos: Integer);
  procedure CursorMoveMarking;

  function IsHighlighted: boolean;
  function IsValidCursorPosition(P: TPoint): boolean;
  function IsCursorAtEOL: boolean;
  function IsCursorAtStart: boolean;
  function IsCursorAtEnd: boolean;
  function IsEmpty: boolean;
  function RasComparePoints(P1,P2: TPoint): integer;
  function IsGreaterPoint(P1,P2: TPoint): boolean;
  function IsPointInsideMarker(X,Y: integer; M: TRect; ExcludeWhite: boolean): boolean;

    procedure ScaleCursorPos(var X, Y: integer); overload;
    procedure ScaleCursorPos(var P: TPoint); overload;
  function GetCurrentColumn: integer;
  function CursorToLinearPos(C: TPoint): integer;
  function CursorToLinearEnd(C: TPoint): integer;
  function GetLinearPos: integer;
  function GetLinearEnd: integer;
  function LinearToCursorPos(C: integer): TPoint;
  function MoveToLinear(C: integer): boolean;
  function CursorToPixelPos(X, Y: integer): TPoint;
  function GetCursorPixelPos: TPoint;
  function ScreenToCursor(X,Y: integer): TPoint;
  function ScreenToBufferPos(X,Y: integer): TPoint;
  function ExpandTabs(S: string): string;

  procedure KillScoop;
  function NormalizeMarker(M: TRect): TRect;

  function GetMarkBlock(Loc: TPoint; Delimiter: TCharSet): TRect;
  function FindNearestBlock(Delimiter: TCharSet): TRect;
  function GetAndMarkBlock(Delimiter: TCharSet): string;
  function GetNearestBlock(Delimiter: TCharSet): string;
  procedure MarkBlock(Delimiter: TCharSet);

  function MarkNearestWord(Loc: TPoint): TRect;
  function FindNearestWord: TRect;
  function GetNearestWord: string;
  procedure SetHighlight(X1,Y1,X2,Y2: integer; NonDelete: boolean);
  procedure MarkWord;
  procedure MarkParagraph(Loc: TPoint; var Marker: TRect);
  procedure MarkLine(Loc: TPoint; var Marker: TRect);
  procedure HighlightLine;
  procedure HighlightParagraph;
  procedure CursorToHighlightEnd;
  procedure CusorToHighlightStart;
  function IsForwardMarker: boolean;

  function GetCharBeforeLoc(Loc: TPoint; var C: char): boolean;
  function GetCharAfterLoc(Loc: TPoint; var C: char): boolean;
  function GetCharsAfterCursor(var S: string; Count: integer): boolean;
  function GetCharsBeforeCursor(var S: string; Count: integer): boolean;
  function GetChars(var S: string; Backward: boolean; Count: integer): boolean;

  function GetCurrentLine: string;
  function GetNextChar: char;
  function GetLastChar: char;
  function GetBeginningOfLine: string;
  function GetEndingOfLine(AddCR: boolean): string;
  function GetWholeLines(First,Last: integer; UseCRLF: boolean): string;
  function GetHighlightedText: string;

  procedure InsertNewLine;
  procedure InsertIntoLine(Str: string);
  procedure SplitLineAtCursor;

  function GetForwardLoc(Loc: TPoint): TPoint;
  function GetBackLoc(Loc: TPoint): TPoint;
  function GetCountBetweenCursors(Marker: TRect): integer;
  function GetLineStartPos(Loc: TPoint): TPoint;
  function DoBackLoc(var Loc: TPoint): boolean;
  function DoForwardLoc(var Loc: TPoint): boolean;
  function DoIncrementLoc(var Loc: TPoint; Delta: integer): boolean;
  function DoColumnMove(Column: integer): integer;
  function DoUpLine: boolean;
  function DoDownLine: boolean;
  function DoForwardLine: boolean;
  function DoBackLine: boolean;
  function DoBackChar: boolean;
  function DoBackOverWord: boolean;
  function DoBackOverSeparators: boolean;
  function DoMoveForwardOverWord: boolean;
  function DoMoveForwardOverSeparators: boolean;

  function DoBackWord(ToWordStart: boolean): boolean;
  function DoForwardWord(ToWordEnd: boolean): boolean;
  function DoPageUp: boolean;
  function DoPageDown: boolean;
  function DoForwardChar: boolean;
  function DoBackDelete: boolean;
  function DoForwardDelete: boolean;
  function DoDeleteForwardChars(Count: integer): boolean;
  function DoDeleteBackChars(Count: integer): boolean;
  function DoDeleteChars(Backward: boolean; Count: integer): boolean;
  function DoDelBackLine: boolean;
  function DoDelForwardLine: boolean;
  function DoBackAll: boolean;
  function DoForwardAll: boolean;
  procedure DoInsertChar(C: char);
  procedure DoInsertString(S: string); overload;
  procedure DoInsertString(S: string; Raw: boolean); overload;
  procedure DoForwardInsertString(S: string);
  procedure DeleteWholeLines(First,Last: integer);
  procedure DeleteMarkedText(Marker: TRect);
  function GetMarkedText(Marker: TRect): string;
  procedure GetSurroundChars(Marker: TRect; var C1, C2: char);
  procedure DoDeleteHighlightedText;
  procedure DoDelCurrentLine;

  function DoForwardSearchMove(Target: string; var Loc: TPoint; var Mark: TRect): boolean;
  function DoBackwardSearchMove(Target: string; var Loc: TPoint; var Mark: TRect): boolean;
  function DoForwardSearch(Target: string; Loc: TPoint; var Mark: TRect): boolean;
  function DoBackwardSearch(Target: string; Loc: TPoint; var Mark: TRect): boolean;

  procedure StartScoopAction;
  procedure EndScoopAction;

  function SetScrollPosition(P: TPoint): boolean;
  procedure SetVerticalScrollPosition(V: integer);
  function GetRelativeLine: integer;
  function SetRelativeLine(Line: integer): boolean;

  procedure DoScrollUpLine;
  procedure DoScrollDownLine;
  procedure DoScrollUpPage;
  procedure DoScrollDownPage;

  function IcOverwriteChar(C: char): boolean;
  procedure IcInsertOverwriteString(S: string; OverWriteCR: boolean);
  function IcDeleteHighlightedText(CompleteUndo: boolean): boolean;

  procedure BackDelete;
  procedure ForwardDelete;
  procedure DeleteHighlighted;
  procedure InsertChar(C: char);
  procedure InsertCmdChar(Key: Char);
  procedure InsertString(S: string; DoEvent,Raw: boolean); overload;
  procedure InsertString(S: string); overload;
  procedure NoEventInsert(S: string);

  procedure BlockInsert(StringList: TStringList); overload;
  procedure BlockInsert(S: string); overload;
  function GetTextBlock(Block: TRect): string;
  procedure DisposeOfBlock(Block: TRect; C: Char; Fill: boolean);

  procedure ForwardChar;
  procedure BackChar;
  procedure BackWord(ToWordStart: boolean);
  procedure ForwardWord(ToWordStart: boolean);
  procedure DownLine;
  procedure UpLine;
  procedure ForwardLine;
  procedure PageDown;
  procedure PageUp;
  procedure BackLine;
  procedure ForwardAll;
  procedure MoveToHighlightEnd;
  procedure BackAll;
  function GotoLine(L: integer): boolean;
  procedure SetPosition(P: TPoint);
  function SetCursorPosition(P: TPoint): boolean;
  procedure DelBackLine;
  procedure DelForwardLine;
  procedure DelCurrentLine;
  procedure ReplaceCurrentLine(S: string);
  procedure DeleteBackWord(ToWordStart: boolean);
  procedure DeleteForwardWord(ToWordEnd: boolean);

  function ForwardSearch(Target: string): boolean;
  function BackwardSearch(Target: string): boolean;

  procedure ScrollUpLine;
  procedure ScrollDownLine;
  procedure ScrollUpPage;
  procedure ScrollDownPage;

  function Undo: boolean;
  function Redo: boolean;

  function CutHighlighted: string;
  procedure SelectAll;
  procedure CutToClipboard;
  procedure CopyToClipBoard;
  procedure PasteFromClipboard;
  procedure ClearBuffer;
  procedure Clear;
  procedure AttachToEditor(Editor: TScrollEditor);
  procedure DetachEditor;

  procedure LoadFromFile(Filename: string);
  procedure SaveToFile(const FileName: string);
  procedure Add(S: string);

  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;
  property IsModified: boolean read FIsModified write SetIsModified default False;
  property ChangeCount: integer read FChangeCount;
  property Text: string read GetText write SetText;
  property Filename: string read FFilename write FFilename;
  property Scale: double read FScale write SetScale;
 published
  property StandAlone: boolean read FStandAlone write FStandAlone default false;
  property Strings: TEditBuffer read EditBuffer write SetEditBuffer;
  property ScrollBars;
  property ShowAll: boolean read FShowAll write SetShowall default false;
  property GutterSize: integer read FGutterSize write SetGutter default 5;
  property OnChange: TNotifyEvent read FOnChange write FOnChange;
  property OnScoopEvent: TScoopEvent read FOnScoopEvent write FOnScoopEvent;
  property OnFocus: TNotifyEvent read FOnFocus write FOnFocus;
  property OnScaleChange: TScaleEvent read FOnScaleChange write FOnScaleChange;
  property OnUnFocus: TNotifyEvent read FOnUnFocus write FOnUnFocus;
  property EditorStyle: TEditorStyle read FEditorStyle write FEditorStyle;
  property CursorWidth: integer read FCursorWidth write SetCursorWidth default 3;
  property LinearHScroll: boolean read FLinearHScroll write FLinearHScroll;
  property HighlightColor: TColor read FHighlightColor write SetHighlightColor default clNavy;
  property NonDeleteColor: TColor read FNonDeleteColor write SetNonDeleteColor default clBlack;
  property ControlColor: TColor read FControlColor write SetControlColor default clRed;
  property ShadowColor: TColor read FShadowColor write SetShadowColor default clSilver;
  property ShadowLine: boolean read FShadowLine write SetShadowLine default False;
  property UseTabStops: boolean read FUseTabStops write SetTUseTabStops default True;
  property SpacesPerTab: integer read FSpacesPerTab write SetSpacesPerTab default 8;
  property KeyMapping: TKeyMapping read FKeyMapping write FKeyMapping default kmDefault;
  property ReadOnly: boolean read FReadOnly write FReadOnly default false;
  property Align;
  property OnKeyPress;
  property OnKeyDown;
  property OnClick;
  property OnDblClick;
  property OnDragDrop;
  property OnDragOver;
  property OnEndDrag;
  property OnEnter;
  property OnExit;
  property OnMouseDown;
  property OnMouseMove;
  property OnMouseUp;
  property OnPaint: TNotifyEvent read FOnPaint write FOnPaint;
end;



procedure Register;

implementation


{==============================================================================}
{================== Global Support Subroutines ================================}
{==============================================================================}


procedure Register;
begin
RegisterComponents('Fountain', [TScrollEditor]);
end;




function GetLeadingWhiteSpace(S: string): string;
{Extract leadin tabs and spaces from string}
var I: integer;
begin
Result:='';
for I:=1 to Length(S) do
 if S[I] in [Tab_Char,SP_Char] then Result:=Result + S[I] else break;
end;




function DoToggleCase(var S: string): boolean;
{Toggle case of all chars in string}
var I: integer;
begin
Result:=False;
for I:=1 to Length(S) do
 if IsAlpha(S[I]) then
	begin
	if IsUpperCaseChar(S[I]) then S[I]:=LowerChar(S[I])
	else  S[I]:=UpperChar(S[I]);
	Result:=True;
	end;
end;


{==============================================================================}
{================== Row Marker Object Implementation ==========================}
{==============================================================================}

{ TRowMarks }


constructor TRowMarks.Create;
begin
List:=TStringList.Create;
List.Sorted:=True;
List.CaseSensitive:=True;
end;


destructor TRowMarks.Destroy;
begin
ClearAll;
List.Free;
inherited;
end;



function TRowMarks.AddLine(Group: string; Line: integer; Color: TColor): integer;
var Inx: integer;
var IH: TIntegerHashTable;
begin
if List.Find(Group,Inx) then IH:=TIntegerHashTable(List.Objects[Inx])
else
	begin
	IH:=TIntegerHashTable.Create(5);
	List.AddObject(Group,IH);
	end;
Result:=IH.InsertPointer(Line,Color);
end;



procedure TRowMarks.Delete(Group: string; Inx: integer);
var IH: TIntegerHashTable;
begin
if List.Find(Group,Inx) then
	begin
	IH:=TIntegerHashTable(List.Objects[Inx]);
	IH.Items.Delete(Inx);
	IH.Rehash;
	end;
end;



function TRowMarks.IsMarkedLine(Line: integer; var Color: TColor; var Start: integer): boolean;
var Inx: integer;
var IH: TIntegerHashTable;
begin
Result:=False;
while Start<List.Count do
	begin
	IH:=TIntegerHashTable(List.Objects[Start]);
	Inc(Start);
	Inx:=IH.Search(Line);
	Result:=Inx>-1;
	if Result then
		begin
		Color:=TColor(IH.Items.Pointers[Inx]);
		break;
		end;
	end;
end;


procedure TRowMarks.DeleteLine(Group: string; Line: integer);
var I,Inx: integer;
var IH: TIntegerHashTable;
begin
if List.Find(Group,Inx) then
	begin
	IH:=TIntegerHashTable(List.Objects[Inx]);
	I:=IH.Search(Line);
	if I>-1 then IH.Delete(I);
	IH.Rehash;
	end;
end;


procedure TRowMarks.ClearGroup(Group: string);
var IH: TIntegerHashTable;
var Inx: integer;
begin
if List.Find(Group,Inx) then
	begin
	IH:=TIntegerHashTable(List.Objects[Inx]);
	IH.Clear;
	end;
end;



procedure TRowMarks.ClearAll;
var I: integer;
var IH: TIntegerHashTable;
begin
for I:=List.Count-1 downto 0 do
	begin
	IH:=TIntegerHashTable(List.Objects[I]);
	IH.Clear;
	IH.Free;
	List.Delete(I);
	end;
end;


procedure TRowMarks.DeleteGroup(Group: string);
var IH: TIntegerHashTable;
var Inx: integer;
begin
if List.Find(Group,Inx) then
	begin
	IH:=TIntegerHashTable(List.Objects[Inx]);
	IH.Clear;
	IH.Free;
	List.Delete(Inx);
	end;
end;



{==============================================================================}
{================== Undoer Object Implementation ==============================}
{==============================================================================}



function TUndoControl.GetCommandStr(C: TUndoCommand): string;
{Get undo command string - for debugging}
begin
case C of
 ucBackDelete: Result:='BackDelete';
 ucForwardDelete: Result:='ForwardDelete';
 ucMove: Result:='Move';
 ucBackInsert: Result:='BackInsert';
 ucForwardInsert: Result:='ForwardInsert';
 ucNull: Result:='Null';
 ucError: Result:='Error';
 end;
end;



function TUndoControl.GetStateStr(C: TUndoState): string;
{Get undo State string - for debugging}
begin
case C of
 usStart: Result:='Start';
 usStep: Result:='Step';
 usEnd: Result:='End';
 usComplete: Result:='Complete';
 end;
end;



procedure TUndoControl.Initialize;
{ initialize Undo list  }
begin
UndoIndex:=0; UndoStart:=0;
if Assigned(FOnStateChange) then FOnStateChange(UndoIndex,UndoList[0]);
HandleChange;
end;



procedure TUndoControl.SetUndoSize(V: integer);
{Handle changing the Undo list size}
var I: integer;
begin
if V<>FUndoSize then
	begin
	FUndoSize:=V;
	SetLength(UndoList,FUndoSize);
	for I:=0 to FUndoSize-1 do UndoList[I].UndoBuffer:='';
	Initialize;
	end;
end;


constructor TUndoControl.Create;
begin
inherited Create;
UndoSize:=DefaultUndoSize;
SingleStep:=False;
end;


destructor TUndoControl.Destroy;
begin
inherited Destroy;
end;



procedure TUndoControl.HandleChange;
{Handle Undo Change event}
begin
if Assigned(FOnChange) then FOnChange(Self);
end;



procedure TUndoControl.InvertUndoItem(var UndoItem: TUndoItem);
{Convert an Undo Item to its inverse operation - for redoing}
begin
case UndoItem.State of
 usStart: UndoItem.State:=usEnd;
 usEnd: UndoItem.State:=usStart;
 end;

case UndoItem.Command of
 ucForwardDelete: UndoItem.Command:=ucForwardInsert;
 ucBackDelete: UndoItem.Command:=ucBackInsert;
 ucBackInsert: UndoItem.Command:=ucBackDelete;
 ucForwardInsert: UndoItem.Command:=ucForwardDelete;
 end;
end;



procedure TUndoControl.IncrementIndex;
{ Increment undo indexes }
begin
Inc(UndoIndex);
if UndoIndex>=UndoSize then UndoIndex:=0;
if UndoIndex=UndoStart then
	begin
	Inc(UndoStart);
	if  UndoStart>=UndoSize then UndoStart:=0;
	end;
if UndoIndex=FConsolidStart then
	begin
	Inc(FConsolidStart);
	if  FConsolidStart>=UndoSize then FConsolidStart:=0;
	end;
end;



function TUndoControl.CheckForNullPairs(UndoItem: TUndoItem): boolean;
{Check for a start and ending null commands in sequence}
{These can be thrown away}
var LastInx: integer;
var LastItem: TUndoItem;
begin
Result:=False;
if UndoIndex=UndoStart then exit;
if (UndoItem.Command<>ucNull) or (UndoItem.State<>usEnd) then exit;
if UndoIndex=0 then LastInx:= UndoSize-1 else LastInx:=UndoIndex-1;
LastItem:=UndoList[LastInx];
if (LastItem.Command<>ucNull) or (LastItem.State<>usStart) then exit;
UndoIndex:=LastInx;
Result:=True;
end;


procedure TUndoControl.MarkConsolidationStart;
{Mark the beginning of a sequence you want consolidated}
begin
FConsolidStart:=UndoIndex;
end;


procedure TUndoControl.ConsolidateUndos;
{Consolidate discrete undo items since last mark }
var I: integer;
begin
if (FConsolidStart<0) or (FConsolidStart>=UndoSize) then exit;
if FConsolidStart = (UndoIndex-1) then UndoList[UndoIndex-1].State:=usComplete
else
	begin
	UndoList[FConsolidStart].State:=usStart;
	if FConsolidStart>UndoIndex then
		begin
		{Handle consolidation that wrap around circular buffer}
		for I:=FConsolidStart+1 to UndoSize-1 do UndoList[I].State:=usStep;
		for I:=0 to UndoIndex-2 do UndoList[I].State:=usStep;
		end
	else for I:=FConsolidStart+1 to UndoIndex-2 do UndoList[I].State:=usStep;
	UndoList[UndoIndex-1].State:=usEnd;
	end;
end;







function TUndoControl.StoreUndoItem(UndoItem: TUndoItem): boolean;
{ Store undo information after a change}
var OldInx: integer;
begin
OldInx:=UndoIndex;
if not CheckForNullPairs(UndoItem) then
	begin
	UndoList[UndoIndex]:=UndoItem;
	IncrementIndex;
	end;
if Assigned(FOnStateChange) then FOnStateChange(OldInx,UndoItem);
Result:=True;
end;



procedure TUndoControl.StoreChanges(Command: TUndoCommand; DeletedChars: string; LastCursor: TPoint; State: TUndoState);
{ Store most recent changes }
var UndoItem: TUndoItem;
begin
UndoItem.Command:=Command;
UndoItem.UndoBuffer:=DeletedChars;
UndoItem.OldCursor:=LastCursor;
UndoItem.State:=State;
if StoreUndoItem(UndoItem) then
	begin
	HandleChange;
	end;
end;




procedure TUndoControl.SetupDeltaUndo(CurrentCursor: TPoint);
{Sets the start of conditionally storing an undo item}
{You must call the DoDeltaCursorStore after any changes}
begin
FOldDeltaCursor:=CurrentCursor;
end;



procedure TUndoControl.DoDeltaCursorStore(CurrentCursor: TPoint);
{Store undo information if cursor has moved since calling SetupDeltaUndo}
begin
if IntVectorDistance(FOldDeltaCursor,CurrentCursor)<>0 then
StoreChanges(ucMove,'',FOldDeltaCursor,usComplete);
end;



procedure TUndoControl.UndoItem(ItemInx: integer; Done: boolean);
{ Undo item specified by ItemInx by calling user supplied procedure }
begin
if Assigned(FUndoProcedure) then
	begin
	FUndoProcedure(UndoList[ItemInx],Done);
	end;
end;




function TUndoControl.DoUndo: boolean;
{ Undo last item by transmitting info to user supplied procedure }
var Done: boolean;
begin
Result:=False;
while UndoIndex<>UndoStart do
	begin
	Result:=True;
	if UndoIndex=0 then UndoIndex:= UndoSize-1 else Dec(UndoIndex);
	Done:=SingleStep or (UndoList[UndoIndex].State = usComplete) or (UndoList[UndoIndex].State = usStart);
	UndoItem(UndoIndex,Done);
	if Done then break;
	end;
end;



{==============================================================================}
{================== StringList Searcher Implementation ========================}
{==============================================================================}



constructor TListSearcher.Create(StringList: TStringList);
begin
inherited Create;
FStringList:=StringList;
IgnoreCase:=False;
WholeWordMode:=False;
HyphenatedWords:=True;
end;


procedure TListSearcher.GetSpecifiedLine(L: integer);
{Check to see if we have already loaded current line}
{Load it if it is not in the local variable CurLine}
begin
if LineInx<>L then
	begin
	CurLine:=FStringList[L];
	LineInx:=L;
	LineLength:=Length(CurLine);
	end;
end;


function TListSearcher.GetCharAtCursor(Loc: TPoint): Char;
{Get character at specified cursor location}
{optimized for speed by caching lines }
begin
GetSpecifiedLine(Loc.Y);
if (Loc.X=0) or (Loc.X>LineLength) then Result:=CR_Char
else Result:=CurLine[Loc.X];
if IgnoreCase then Result:=LowerChar(Result);
end;



function TListSearcher.EndOfList(Loc: TPoint): boolean;
{Test if we are at the end of the list}
{optimized for speed by caching lines }
begin
GetSpecifiedLine(Loc.Y);
Result:=(Loc.Y = FStringList.Count-1) and (Loc.X=LineLength);
end;


function TListSearcher.IncrementLoc(var Loc: TPoint): boolean;
{Increment the specified cursor location}
{optimized for speed by caching lines }
begin
Result:=False;
GetSpecifiedLine(Loc.Y);
if Loc.X>=LineLength then
	begin
	if Loc.Y>=FStringList.Count-1 then exit;
	Inc(Loc.Y);
	Loc.X:=0;
	end
else Inc(Loc.X);
Result:=True;
end;




function TListSearcher.DecrementLoc(var Loc: TPoint): boolean;
{Increment the specified cursor location}
{optimized for speed by caching lines }
begin
Result:=False;
if Loc.X>0 then Dec(Loc.X)
else
	begin
	if Loc.Y<=0 then exit;
	Dec(Loc.Y);
	GetSpecifiedLine(Loc.Y);
	Loc.X:=LineLength;
	end;
Result:=True;
end;



function TListSearcher.CompareChars(TargetChar,ListChar: Char): boolean;
{Compare two characters taking into account Wildcards, Wordbreakd, and Numerics}
begin
if IgnoreCase then ListChar:=LowerChar(ListChar);
case TargetChar of
 WildCard: Result:=True;
 NumericWildChar: Result:=IsDigit(ListChar);
 else Result:=TargetChar = ListChar;
end;
end;



function TListSearcher.IsWholeWord(Mark: TRect): boolean;
{Returns true if marked word is a whole word}
var C1,C2,C3,C4: char;
var P: TPoint;
begin
C2:=GetCharAtCursor(Mark.TopLeft);
P:=Mark.TopLeft;
if (Mark.Left=0) or not DecrementLoc(P) then C1:=' ' else C1:=GetCharAtCursor(P);

if not IncrementLoc(Mark.BottomRight) then C3:=' ' else C3:=GetCharAtCursor(Mark.BottomRight);
if not IncrementLoc(Mark.BottomRight) then C4:=' ' else C4:=GetCharAtCursor(Mark.BottomRight);
Result:=IsWordBreakEx(C2,C1,False) and IsWordBreakEx(C3,C4,False);
end;



function TListSearcher.MatchStringAtCursor(var Loc: TPoint; var Mark: TRect): boolean;
{Tests to see if FTarget can be found at cursor}
var I,Len: integer;
var L: TPoint;
begin
Result:=False;
L:=Loc;
Mark.TopLeft:=Loc;
Len:=Length(FTarget);
for I:=1 to Len do
	begin
	if not CompareChars(FTarget[I],GetCharAtCursor(L)) then exit;
	if I>=Len then break;
	if not IncrementLoc(L) then exit;
	end;
Loc:=L;
Mark.BottomRight:=Loc;
DecrementLoc(Mark.TopLeft);
Result:=True;
end;




function TListSearcher.SearchForward(Target: string; var Loc: TPoint; var Mark: TRect): boolean;
{Search forward from specified "Loc" }
{Returns True if string is found }
{"Mark" marks the found string in the text}
var FirstChar: Char;
begin
LineInx:=-1;
Result:=False;
if Target='' then exit;
Result:=True;
if IgnoreCase then FTarget:=LowerCase(Target) else FTarget:=Target;
FirstChar:=FTarget[1];
repeat
	begin
	if CompareChars(FirstChar,GetCharAtCursor(Loc)) then
	   if MatchStringAtCursor(Loc,Mark) then
	      if not WholeWordMode or IsWholeWord(Mark) then exit;
	end
until not IncrementLoc(Loc);
Result:=False;
end;




function TListSearcher.SearchBackward(Target: string; var Loc: TPoint; var Mark: TRect): boolean;
{Search Backward from specified "Loc" }
{Returns True if string is found }
{"Mark" marks the found string in the text}
var FirstChar: Char;
var OldLoc: TPoint;
begin
LineInx:=-1;
Result:=False;
if (Target='') or ((Loc.X=0) and (Loc.Y=0)) then exit;
Result:=True;
if IgnoreCase then FTarget:=LowerCase(Target) else FTarget:=Target;
FirstChar:=FTarget[1];
repeat
	begin
	if CompareChars(FirstChar,GetCharAtCursor(Loc)) then
		begin
		OldLoc:=Loc;
		if MatchStringAtCursor(Loc,Mark) then
	 	if not WholeWordMode or IsWholeWord(Mark) then exit;
		Loc:=OldLoc;
		end;
	end
until not DecrementLoc(Loc);
Result:=False;
end;


{==============================================================================}
{================== Specialized String List Object ============================}
{==============================================================================}



function TEditBuffer.GetEmpty: boolean;
begin
Result:=(Count=0) or ((Count=1) and (Strings[0]=''));
end;


procedure TEditBuffer.Changed;
{Overridden method to keep track of when the editor is busy}
begin
inherited Changed;
Busy:=False;
end;



procedure TEditBuffer.Changing;
{Overridden method to keep track of when the editor is busy}
begin
inherited Changing;
Busy:=True;
end;



constructor TEditBuffer.Create;
begin
inherited Create;
Busy:=False;
end;


function TEditBuffer.GetTextStr: string;
{Overridden routines to remove final CR/LF}
{Used to make Stringlist object compatible with the editor model}
var L: integer;
begin
Result:=inherited GetTextStr;
{Not sure why I did this???}
L:=Length(Result);
if (L>1) and (Result[L]=LF_Char) and (Result[L-1]=CR_Char) then SetLength(Result,L-2);
end;



procedure TEditBuffer.PutTextInList(const Value: string);
{Insert string into buffer, splitting it up into lines}
{Similar to the Overriden "SetTextStr" routine except allows #0}
var P, Start, LastPtr : PChar;
var L: integer;


	procedure LoadASCIIText(const Value: string);
	var S: string;
	begin
	while P<LastPtr do
		begin
		Start := P;
		while (P<LastPtr) and not (P^ in [#10, #13]) do Inc(P);
		SetString(S, Start, P - Start);
		Add(S);
		if P^ = #13 then Inc(P);
	        if P^ = #10 then Inc(P);
	        end;
	end;



	procedure LoadANSIText(const Value: string);
	var S: string;
	begin
	while P<LastPtr do
		begin
		S:='';
		while (P<LastPtr) and not (P^ in [#10, #13]) do
			begin
			if P^<>#0 then S:=S+P^;
			Inc(P);
			end;
		Add(S);
		if P^ = #13 then Inc(P);
		if P^ = #00 then Inc(P);
	        if P^ = #10 then Inc(P);
		if P^ = #00 then Inc(P);
	        end;
	end;


begin
P:=Pointer(Value);
if P = nil then exit;
LastPtr:=P + Length(Value);
BOM:=byte(Value[1]) or byte(Value[2]) shl 8;
case BOM of
 $FFFE: CharType:=ctLittleEnd;
 $FEFF: CharType:=ctBigEnd;
 else CharType:=ctASCII;
end;
Clear;
BeginUpdate;
try
if CharType<>ctASCII then Inc(P,2);
if CharType<>ctASCII then LoadANSIText(Value) else LoadASCIIText(Value);
L:=Length(Value);
if (L>1) and (Value[L]=LF_Char) and (Value[L-1]=CR_Char) then Add('');
finally EndUpdate; end;
end;



procedure TEditBuffer.SetTextStr(const Value: string);
{Overridden routines to insert final CR/LF}
var L: integer;
begin
L:=Length(Value);
if (L>0) and (Value[L]=CR_Char) then PutTextInList(Value + CR_Char)
else PutTextInList(Value);
end;



function TEditBuffer.CustomText(Delimter: string): string;
{Return StringList data as string with specified delimiter}
var I, L, Size, Count: Integer;
var P: PChar;
var S: string;
begin
Count := GetCount;
Size := 0;
for I := 0 to Count - 1 do Inc(Size, Length(Get(I)) + Length(Delimter));
SetString(Result, nil, Size);
P := Pointer(Result);
for I := 0 to Count - 1 do
	begin
	S := Get(I);
	L := Length(S);
	if L <> 0 then
		begin
		System.Move(Pointer(S)^, P^, L);
		Inc(P, L);
		end;
	L := Length(Delimter);
	if L <> 0 then
		begin
		System.Move(Pointer(Delimter)^, P^, L);
		Inc(P, L);
		end;
	end;
SetLength(Result,Length(Result)-Length(Delimiter));
end;


{==============================================================================}
{================== Scrolling Editor Implementation ===========================}
{==============================================================================}




procedure TScrollEditor.AttachToEditor(Editor: TScrollEditor);
{Attach this editor to another specified editor}
{We do this by destorying all the key components}
{ and using the other editors components as our own}
begin
if EditorAttached then exit;
EditBuffer.Free;
EditBuffer:=Editor.EditBuffer;
UndoControl.Free;
UndoControl:=Editor.UndoControl;
RedoControl.Free;
RedoControl:=Editor.RedoControl;
WorkingCursor.Free;
WorkingCursor:=Editor.WorkingCursor;
HighLight.Free;
HighLight:=Editor.Highlight;
Searcher.Free;
Searcher:=Editor.Searcher;
EditorAttached:=True;
end;



procedure TScrollEditor.DetachEditor;
begin
if not EditorAttached then exit;
EditBuffer:=TEditBuffer.Create;
EditBuffer.OnChange:=HandleBufferChange;

UndoControl:=TUndoControl.Create;
UndoControl.UndoProcedure:=UndoCallBack;
UndoControl.OnChange:=UndoChange;

RedoControl:=TUndoControl.Create;
RedoControl.UndoProcedure:=RedoCallBack;

WorkingCursor:=TCursorObject.Create;

HighLight:=THighLight.Create;

Searcher:=TListSearcher.Create(EditBuffer);

EditorAttached:=False;
end;


{------------------ Screen Scaling Routines ----------------------------------}


procedure TScrollEditor.ScaleCursorPos(var X,Y: integer);
begin
X:=Ceil(X / FScale);
Y:=Ceil(Y / FScale);
end;



procedure TScrollEditor.ScaleCursorPos(var P: TPoint);
begin
ScaleCursorPos(P.X,P.Y);
end;




procedure TScrollEditor.ScaleScreen(Scale: double);
var form: tagXFORM;
begin
{Text shivers when scaled, so at least stop it for scales of one}
if Scale = 1 then exit;
SetGraphicsMode(Canvas.Handle, GM_ADVANCED);
SetMapMode(Canvas.Handle, MM_ANISOTROPIC);
form.eM11 := 1/Scale;
form.eM12 := 0;
form.eM21 := 0;
form.eM22 := 1/Scale;
form.eDx := 0;
form.eDy := 0;
SetWorldTransform(Canvas.Handle, form);
end;



procedure TScrollEditor.SetScale(const Value: double);
begin
if Value = FScale then exit;
FScale:=Value;
if FScale<1 then FScale:=1;
Invalidate;
CreateTextCursor;
if Assigned(FOnScaleChange) then FOnScaleChange(FScale);
end;



{-----------------------------------------------------------------------------}


function TScrollEditor.CursorToColumn(Cursor: TPoint): integer;
{Count columns to cursor on current line}
var I,LS: integer;
var S: string;
begin
Result:=0;
S:=EditBuffer[Cursor.Y];
LS:=Length(S);
if LS=0 then exit;
if FShowAll or (LS>TabLimitLength) then Result:=Cursor.X
else
	begin
	if Cursor.X>LS then exit;
	for I:=1 to Cursor.X do
	if (S[I]=TAB_CHAR) then Result:=Result + (FSpacesPerTab - (Result mod FSpacesPerTab))
	else Inc(Result);
	end;
end;



function TScrollEditor.GetCurrentColumn: integer;
{Return column of current cursor}
begin
Result:=CursorToColumn(WorkingCursor.Position);
end;



function TScrollEditor.ColumnToCursorX(Cursor: TPoint): integer;
{ Convert column to Cursor X position, skipping tabs as necessary}
var Len,C: integer;
var S: string;
begin
Result:=Cursor.X;
S:=EditBuffer[Cursor.Y];
Len:=Length(S);
if FShowAll or (Len<1) or (Len>TabLimitLength) then exit;
C:=0; Result:=1;
while C<Cursor.X do
	begin
	if Result>Len then break;
	if S[Result]=TAB_CHAR then C:=C + (FSpacesPerTab - (C mod FSpacesPerTab))-1;
	Inc(Result); Inc(C);
	end;
Result:=Result-1;
end;


function TScrollEditor.GetPreciseCharWidth: double;
{Get precise character width by averaging the width of multiple characters}
begin
Result:=Canvas.TextWidth('MMMMMMMMMM')/10;
end;



function TScrollEditor.GetCursorPixelPos: TPoint;
{Convert Cursor Position to Pixel Position}
var W: double;
begin
W:=GetPreciseCharWidth;
GetTextMetrics(Canvas.Handle,FTextMetric);
Result.X:=Round(W * (CursorToColumn(WorkingCursor.Position)-ScrollOffset.X)+FGutterSize);
Result.Y:=FTextMetric.tmHeight * (WorkingCursor.Position.Y-ScrollOffset.Y);
end;



function TScrollEditor.ScreenToCursor(X,Y: integer): TPoint;
{Convert Screen X,Y to Cursor Position}
{Cursor can be outside areas where there is text}
var DX,DY,GR: integer;
begin
GetTextMetrics(Canvas.Handle,FTextMetric);
GR:=X - FGutterSize;
if GR<0 then GR:=0;
DX:=round((GR/FTextMetric.tmAveCharWidth)-0.5);
DY:=round((Y/FTextMetric.tmHeight)-0.5);
Result.X:=DX + ScrollOffSet.X;
Result.Y:=DY + ScrollOffset.Y;
end;



function TScrollEditor.ScreenToBufferPos(X,Y: integer): TPoint;
{Convert X,Y Pixel position to cursor position, forcing it to valid text area}
begin
Result:=ScreenToCursor(X,Y);
if Result.X<0 then Result.X:=0;
if Result.Y<0 then Result.Y:=0;
if Result.Y>=EditBuffer.Count then Result.Y:=EditBuffer.Count-1;
Result.X:=ColumnToCursorX(Result);
if Result.X>Length(EditBuffer[Result.Y]) then  Result.X:=Length(EditBuffer[Result.Y]);
end;




procedure TScrollEditor.GetScreenParameters;
{Get Screen size parameters}
begin
ScaleScreen(FScale);
GetTextMetrics(Canvas.Handle,FTextMetric);
ScreenCharHeight:=ClientHeight div FTextMetric.tmHeight;
ScreenCharWidth:=Round(ClientWidth / FTextMetric.tmAveCharWidth);
end;



function TScrollEditor.CursorToPixelPos(X,Y: integer): TPoint;
begin
GetScreenParameters;
if X<ScrollOffset.X then X:=ScrollOffset.X;
if Y<ScrollOffset.Y then Y:=ScrollOffset.Y;
if X>(ScrollOffset.X+ScreenCharWidth) then X:=ScrollOffset.X+ScreenCharWidth;
if Y>(ScrollOffset.Y+ScreenCharWidth) then Y:=ScrollOffset.Y+ScreenCharHeight;
Result.X:=(X-ScrollOffset.X) * FTextMetric.tmAveCharWidth;
Result.Y:=(Y-ScrollOffset.Y) * FTextMetric.tmHeight;
end;



procedure FitToWindow(Linear: boolean; WinSize,Position,GuardBand: integer; var Offset: integer);
{Set Offset so Position falls within WinSize minus GuardBand}
var M: integer;
begin
M:=Position-Offset;
if Linear or (M<GuardBand) or (M>(WinSize-GuardBand)) then
	begin
	if Position>(WinSize-GuardBand) then Offset:=Position-(WinSize-GuardBand)
	else if Linear then Offset:=0;
	if (M<GuardBand) and not Linear then OffSet:=Position-GuardBand;
	end;
if Offset<0 then Offset:=0;
end;



procedure TScrollEditor.CenterVerticalCursor;
{Scroll so cursor is in the center of the screen}
begin
ScrollOffset.Y:=WorkingCursor.Position.Y - (ScreenCharHeight div 2);
if ScrollOffset.Y<0 then ScrollOffset.Y:=0;
end;




procedure TScrollEditor.OptimizeScrolling;
{Optimize scrolling so cursor is in Window and not against sides}
var CurColumn: integer;
const GuardBandX = 4;
const GuardBandY = 1;
begin
GetScreenParameters;
CurColumn:=CursorToColumn(WorkingCursor.Position);
FitToWindow(FLinearHScroll,ScreenCharWidth,CurColumn,GuardBandX,ScrollOffset.X);
FitToWindow(False,ScreenCharHeight,WorkingCursor.Position.Y,GuardBandY,ScrollOffset.Y);
end;




procedure TScrollEditor.CreateTextCursor;
{Create the cursor we are using}
var P: TPoint;
begin
GetTextMetrics(Canvas.Handle,FTextMetric);
CreateCaret(Handle,0,Round(FCursorWidth * FScale),Round(FTextMetric.tmHeight * FScale));
P:=GetCursorPixelPos;
SetCaretPos(P.X,P.Y);
SetCaretBlinkTime(500);
ShowCaret(Handle);
end;



procedure TScrollEditor.SetCursorWidth(V: integer);
{Change the cursor width }
begin
if V<>FCursorWidth then
	begin
	DestroyCaret;
	FCursorWidth:=V;
	CreateTextCursor;
	end;
end;



procedure TScrollEditor.DisplayCursor;
{Display Windows cursor resource at the current position}
var P: TPoint;
begin
P:=GetCursorPixelPos;
SetCaretPos(P.X,P.Y);
ShowCaret(Handle);
end;



procedure TScrollEditor.WMGetDlgCode(var Message: TWMGetDlgCode);
{This block of code is required so Edit Screen recieves arrows and tab keys}
begin
inherited;
Message.Result:=Message.Result or
	DLGC_WANTALLKEYS or DLGC_WANTTAB or DLGC_WANTARROWS;
end;


procedure TScrollEditor.WMSetFocus(var Message: TWMSetFocus);
{When we receive the focus from Windows, construct and display flashing cursor}
begin
inherited;
OldBlinkRate:=GetCaretBlinkTime;
CreateTextCursor;
if Assigned(FOnFocus) then FOnFocus(Self);
end;


procedure TScrollEditor.WMKillFocus(var Message: TWMKillFocus);
{When we loose the focus destroy flashing cursor}
begin
inherited;
SetCaretBlinkTime(OldBlinkRate);
DestroyCaret;
if Assigned(FOnUnFocus) then FOnUnFocus(Self);
end;



procedure TScrollEditor.WMSize(var Message: TWMSize);
begin
inherited;
{Don't optiomize scrooling here! The program will scroll away
from where you are looking everytime you adjust the window}
//OptimizeScrolling;
end;




procedure TScrollEditor.CMFontChanged(var Message: TMessage);
begin
inherited;
if Canvas <> nil then Canvas.Font := Font;
end;


{---------------- XED style scoop handlers ------------------------------------}

var ScoopCursorPos: TPoint;


procedure TScrollEditor.KillScoop;
{Called to deactivate scoop }
begin
if Assigned(FOnScoopEvent) then FOnScoopEvent(scAbort,Rect(0,0,0,0));
end;


procedure TScrollEditor.ScoopStart(Cursor: TPoint);
{Called when scoop buffer is enabled}
begin
ScoopCursorPos:=Cursor;
end;



procedure TScrollEditor.ScoopEnd(Cursor: TPoint);
{Called after command has completed to handle cursor changes}
var Comm: TScoopCommand;
var R: integer;
begin
R:= RasComparePoints(ScoopCursorPos,Cursor);
if R=0 then exit;
if R<0 then Comm:=scUnScoop
else Comm:=scScoop;
if Assigned(FOnScoopEvent) then FOnScoopEvent(Comm,Rect(ScoopCursorPos,Cursor));
end;



procedure TScrollEditor.StartScoopAction;
{Initiate a scoop action, using current position}
begin
ScoopStart(WorkingCursor.Position);
end;



procedure TScrollEditor.EndScoopAction;
{Terminate a scoop action, using current position}
begin
ScoopEnd(WorkingCursor.Position);
end;




{---------------- Text Marking Routines ---------------------------------------}


procedure TScrollEditor.ResetMarkers;
{Reset markers so no text is marked}
begin
Highlight.NonDelete:=False;
Highlight.Marker:=Rect(0,0,0,0);
FHighlightRow:=-1;
WorkingCursor.ColumnMarking:=False;
end;



function TScrollEditor.IsHighlighted: boolean;
{Returns true if some text is highlighted}
begin
Result:=(Highlight.Marker.Left <> Highlight.Marker.Right) or
   (Highlight.Marker.Top <> Highlight.Marker.Bottom);
end;




function TScrollEditor.CursorToLinearPos(C: TPoint): integer;
{Find the equavalent linear location of the specfied cursor position}
var Y: integer;
begin
Result:=-1;
if C.Y>=EditBuffer.Count then exit;
if C.X>Length(EditBuffer[C.Y]) then exit;
Result:=0;
{Count backwards and watch the "Busy" flag to avoid problems with multitasking}
for Y:=C.Y-1 downto 0 do
 if EditBuffer.Busy then begin Result:=-1; exit; end
 else inc(Result,Length(EditBuffer[Y])+1);
Inc(Result,C.X);
end;




function TScrollEditor.CursorToLinearEnd(C: TPoint): integer;
{Find the linear distance to end of buffer}
var Y: integer;
begin
Result:=-1;
if C.Y>=EditBuffer.Count then exit;
if C.X>Length(EditBuffer[C.Y]) then exit;
Result:=Length(EditBuffer[C.Y]) - C.X;
{Count backwards and watch the "Busy" flag to avoid problems with multitasking}
for Y:=EditBuffer.Count-1 downto C.Y+1 do
 if EditBuffer.Busy then begin Result:=-1; exit; end
 else inc(Result,Length(EditBuffer[Y])+1);
end;



function TScrollEditor.GetLinearEnd: integer;
{Get the linear position of the cursor from end of buffer}
begin
Result:=CursorToLinearEnd(WorkingCursor.Position);
end;



function TScrollEditor.GetLinearPos: integer;
{Get the linear position of the cursor}
begin
Result:=CursorToLinearPos(WorkingCursor.Position);
end;



function TScrollEditor.LinearToCursorPos(C: integer): TPoint;
{Convert linear position to cursor position}
var Y,Y1,P1,P2,L: integer;
begin
Result:=Point(-1,-1);
Y1:=0; P1:=0; P2:=0;
for Y:=0 to EditBuffer.Count-1 do
	begin
	Inc(P2,Length(EditBuffer[Y])+1);
	Y1:=Y;
	if P2>C then break;
	P1:=P2;
	end;
if C>=P2 then exit;
Result.Y:=Y1;
Result.X:=C  - P1;
L:=Length(EditBuffer[Result.Y]);
if Result.X>L then Result.X:=L;
end;




function TScrollEditor.IsEmpty: boolean;
{Returns true if the buffer is empty}
begin
Result:=(EditBuffer.Count>0) and (EditBuffer.Count<2) and (Length(EditBuffer[0])<1);
end;




function TScrollEditor.RasComparePoints(P1,P2: TPoint): integer;
{Do raster compare by subtracting P1 - P2}
{Returns: P1>P2 = 1, P1=P2 = 0, P1<P2 = -1}
var P: TPoint;
begin
P:=IntVectorSubtract(P1,P2);
if P.Y=0 then Result:=Math.Sign(P.X)
else Result:=Math.Sign(P.Y);
end;



function TScrollEditor.IsGreaterPoint(P1,P2: TPoint): boolean;
{Compare point's raster magnitude}
begin
Result:=((P1.Y = P2.Y) and (P1.X >P2.X)) or (P1.Y > P2.Y);
end;



function TScrollEditor.RasPointInRect(M: TRect; P: TPoint; ExcludeWhite: boolean): boolean;
{Is specifed Raster point in Raster Rectangle (As opposed to the cursor point)}
begin
Result:=(RasComparePoints(P,M.TopLeft)>=0) and (RasComparePoints(P,M.BottomRight)<=0);
{Test to see if point is in unhighlighted, non-text part of rectangle}
if ExcludeWhite and (M.Top<>M.Bottom) then Result:=Result and (CursorToColumn(P)>=P.X);
end;



function TScrollEditor.IsForwardMarker: boolean;
{Is marker's bottom-right greater than the top-left}
begin
Result:=IsGreaterPoint(Highlight.Marker.BottomRight,Highlight.Marker.TopLeft);
end;



function TScrollEditor.NormalizeMarker(M: TRect): TRect;
{Normalize marker by making sure TopRight is <= BottomRight}
var P: TPoint;
begin
Result:=M;
if IsGreaterPoint(Result.TopLeft,Result.BottomRight) then
	begin
	P:=Result.TopLeft;
	Result.TopLeft:=Result.BottomRight;
	Result.BottomRight:=P;
	end;
end;



function TScrollEditor.ColumnizeMarker(M: TRect): TRect;
{Convert marker from array-index format to column format}
begin
Result:=M;
Result.Left:=CursorToColumn(Result.TopLeft);
Result.Right:=CursorToColumn(Result.BottomRight);
end;




function TScrollEditor.IsPointInsideMarker(X,Y: integer; M: TRect; ExcludeWhite: boolean): boolean;
{Return true if point is within marking rectangle}
var Cursor: TPoint;
begin
Result:=False;
if not IsHighlighted then exit;
M:=NormalizeMarker(M);
M:=ColumnizeMarker(M);
Cursor:=ScreenToCursor(X,Y);
Result:=RasPointInRect(M,Cursor,ExcludeWhite);
end;




function TScrollEditor.MarkNearestWord(Loc: TPoint): TRect;
{Return Mark for word nearst Loc}
var S: string;
var X1,X2,Len: integer;
var C1,C2,C3: char;
begin
Result:=Rect(Loc.X,Loc.Y,Loc.X,Loc.Y);
S:=EditBuffer[Loc.Y];
Len:=Length(S);
if Len<1 then exit;
X1:=0;
if Loc.X<>0 then
 for X1:=Loc.X downto 1 do
	begin
	C1:=S[X1];
	if X1-1>0 then C2:=S[X1-1] else C2:=' ';
	if X1+1<=Len then C3:=S[X1+1] else C3:=' ';
	{Handle single quote vs apostrophe}
	if IsWordBreakEx(C1,C2,MarkHyphenated) then break;
	if IsWordBreakEx(C1,C3,MarkHyphenated) then break;
	end;
for X2:=Loc.X to Len do
	begin
	if X2<Len then C1:=S[X2+1] else C1:=' ';
	if (Len>0) and (X2>0) then C3:=S[X2] else C3:=' ';
	if X2<Len-1 then C2:=S[X2+2] else C2:=' ';
	{Handle single quote vs apostrophe}
	if IsWordBreakEx(C1,C2,MarkHyphenated) then break;
	if IsWordBreakEx(C1,C3,MarkHyphenated) then break;
	end;
Result.Top:=Loc.Y;
Result.Bottom:=Loc.Y;
Result.Left:=X1;
Result.Right:=X2;
end;




function TScrollEditor.FindNearestWord: TRect;
{Return Mark for word nearst cursor}
begin
Result:=MarkNearestWord(WorkingCursor.Position);
end;



function TScrollEditor.GetNearestWord: string;
{Return the word nearest the cursor as a string}
begin
Result:=GetMarkedText(FindNearestWord);
end;



procedure TScrollEditor.MarkWord;
{Mark/Highlight word nearst cursor}
begin
ResetMarkers;
Highlight.Marker:=FindNearestWord;
WordMarked:=True;
end;



function TScrollEditor.GetMarkBlock(Loc: TPoint; Delimiter: TCharSet): TRect;
{Return Mark for block delineated by white space}
var S: string;
var X1,X2,Len: integer;
begin
Result:=Rect(0,0,0,0);
S:=EditBuffer[Loc.Y];
Len:=Length(S);
if Len<1 then exit;

X1:=0;
if Loc.X<>0 then
 for X1:=Loc.X downto 1 do
  if S[X1] in Delimiter then break;

for X2:=Loc.X to Len do
	begin
	if X2>=Len then break;
	if S[X2+1] in Delimiter then break;
	end;
Result.Top:=Loc.Y;
Result.Bottom:=Loc.Y;
Result.Left:=X1;
Result.Right:=X2;
end;



function TScrollEditor.FindNearestBlock(Delimiter: TCharSet): TRect;
{Return Mark for word nearst cursor}
begin
Result:=GetMarkBlock(WorkingCursor.Position,Delimiter);
end;



function TScrollEditor.GetNearestBlock(Delimiter: TCharSet): string;
{Return the word nearest the cursor as a string}
begin
Result:=GetMarkedText(FindNearestBlock(Delimiter));
end;



procedure TScrollEditor.MarkBlock(Delimiter: TCharSet);
{Mark/Highlight word nearst cursor}
begin
Highlight.Marker:=FindNearestBlock(Delimiter);
WordMarked:=True;
end;



function TScrollEditor.GetAndMarkBlock(Delimiter: TCharSet): string;
{Return the word nearest the cursor as a string}
begin
MarkBlock(Delimiter);
Result:=GetMarkedText(FindNearestBlock(Delimiter));
end;


{------------------------------------------------------------------------------}



function IsBlankLine(S: string): boolean;
{Test string to if it contains only white space}
var I: integer;
begin
Result:=False;
for I:=1 to Length(S) do
 if not (S[I] in [Tab_Char,SP_Char]) then exit;
Result:=True;
end;





procedure TScrollEditor.SetHighlight(X1,Y1,X2,Y2: integer; NonDelete: boolean);
{Set highlight marker, validating postions}
var R: TRect;
begin
R:=Rect(X1,Y1,X2,Y2);
NormalizeMarker(R);
R.Left:=EnsureRange(R.Left,0,High(Integer));
R.Right:=EnsureRange(R.Right,0,High(Integer));
R.Top:=EnsureRange(R.Top,0,PickInteger(EditBuffer.Count<1,0,EditBuffer.Count-1));
R.Bottom:=EnsureRange(R.Bottom,0,PickInteger(EditBuffer.Count<1,0,EditBuffer.Count-1));
Highlight.Marker:=R;
Highlight.NonDelete:=NonDelete;
end;


procedure TScrollEditor.HighlightLine;
{Highlight the line that currently has the cursor}
begin
MarkLine(WorkingCursor.Position,Highlight.Marker);
end;


procedure TScrollEditor.MarkParagraph(Loc: TPoint; var Marker: TRect);
{Return marker for paragraph nearst Loc}
var Y: integer;
begin
Marker.Left:=0; Marker.Top:=Loc.Y;
Marker.Right:=0; Marker.Bottom:=Loc.Y;
for Y:=Marker.Top downto 0 do
 if IsBlankLine(EditBuffer[Y]) then break else Marker.Top:=Y;

for Y:=Marker.Bottom to EditBuffer.Count-1 do
	begin
	if IsBlankLine(EditBuffer[Y]) then break;
        Marker.Bottom:=Y;
	end;
Marker.Right:=Length(EditBuffer[Marker.Bottom]);
end;



procedure TScrollEditor.MarkLine(Loc: TPoint; var Marker: TRect);
{Get marker for line specified by cursor}
begin
Marker.Left:=0;
Marker.Top:=Loc.Y;
Marker.Bottom:=Loc.Y;
Marker.Right:=Length(EditBuffer[Loc.Y]);
end;




procedure TScrollEditor.HighlightParagraph;
{Highlight paragraph nearst cursor}
begin
MarkParagraph(WorkingCursor.Position,Highlight.Marker);
end;





procedure TScrollEditor.MarkCursorDifference;
{Mark screen based on change in cursor position}
begin
if not CursorMarking then
	begin
	Highlight.Marker.TopLeft:=MarkStartPos;
	Highlight.Marker.BottomRight:=MarkStartPos;
	MarkBasePos:=MarkStartPos;
	end;
if not FMoveMarking then WorkingCursor.ColumnMarking:=((GetAsyncKeyState(VK_MENU) and $8000)<>0);
if IsGreaterPoint(WorkingCursor.Position,MarkBasePos) then
	begin
	Highlight.Marker.BottomRight:=WorkingCursor.Position;
        Highlight.Marker.TopLeft:=MarkBasePos;
	end
else
	begin
	Highlight.Marker.TopLeft:=WorkingCursor.Position;
        Highlight.Marker.BottomRight:=MarkBasePos;
	end;
CursorMarking:=True;
end;



procedure TScrollEditor.CursorMoveMarking;
{Handling marking for simple cursor movements}
begin
if ((GetAsyncKeyState(VK_SHIFT) and $8000)=0) and not FMoveMarking then
	begin
	ResetMarkers;
	CursorMarking:=False;
	end
else MarkCursorDifference;                           
end;



function IsLButtonDown: boolean;
{Returns true if left button is down}
begin
if boolean(GetSystemMetrics(SM_SWAPBUTTON)) then Result:=(GetAsyncKeyState(VK_RBUTTON) and $8000)<>0
else Result:=(GetAsyncKeyState(VK_LBUTTON) and $8000)<>0;
end;



procedure TScrollEditor.ScrollToCursor(MarkDifference: boolean);
{Scroll screen with delay until cursor is on screen}
{Note: Scrolling happens because of calls to DoChange which calls OptimiseScrolling}
var PT: TPoint;

	function InScrollZone(PT: TPoint): boolean;
	var OuterRect,InnerRect: TRect;
	begin
	OuterRect:=ClientRect;
        InflateRect(OuterRect,10,10);
	InnerRect:=ClientRect;
	if Mouse.IsDragging then Result:=not PtInRect(InnerRect,PT) and PtInRect(OuterRect,PT)
	else Result:=not PtInRect(InnerRect,PT);
	end;


	procedure DoVerticalScrollDelay(PT: TPoint);
	var T,DY: integer;
	begin
	if PT.Y<0 then DY:=abs(PT.Y)
	else DY:=PT.Y-ClientHeight;
	if DY>25 then T:=0 else T:=75;
	DoTimeDelay(T);
	end;


begin
while True do
	begin
	GetCursorPos(PT);
	PT:=ScreenToClient(PT);
	ScaleCursorPos(PT);
	WorkingCursor.Position:=ScreenToBufferPos(PT.X,PT.Y);
	if MarkDifference then MarkCursorDifference;
	DoChange(True,True);
	if not IsLButtonDown then break;
	if not InScrollZone(PT) then break;
	DoVerticalScrollDelay(PT);
	end;
DoChange(True,True);
end;



procedure TScrollEditor.DoChange(Optimize,DoEvent: boolean);
{Call this routine when edit buffer changes}
begin
if Optimize then OptimizeScrolling;
if AutoUpdate then
	begin
	Repaint;
	if DoEvent and Assigned(FOnChange) then FOnChange(Self);
	end;
end;



procedure TScrollEditor.DoChange(Optimize: boolean);
begin
DoChange(Optimize,True);
end;



procedure TScrollEditor.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
{Handle Mouse-Down event}
begin
inherited MouseDown (Button, Shift, X, Y);
UndoControl.SetupDeltaUndo(WorkingCursor.Position);
KillScoop;
ScaleCursorPos(X,Y);

if (ssLeft in Shift) then LeftMouse:=True;
SetFocus;
MarkStartPos:=WorkingCursor.Position;
WorkingCursor.Position:=ScreenToBufferPos(X,Y);
if not (ssRight in Shift) and (ssDouble in Shift) then
	begin
	if X<=FGutterSize then
		begin
		if (Highlight.Marker.Left=0) and
		   (Highlight.Marker.Top=WorkingCursor.Position.Y) and
		   (Highlight.Marker.Bottom=WorkingCursor.Position.Y) and
		   (Highlight.Marker.Right=Length(EditBuffer[WorkingCursor.Position.Y])) then HighlightParagraph
		else HighlightLine;
		end
	else MarkWord;
	end
else if ssShift in Shift then MarkCursorDifference
else if not (ssRight in Shift) and IsPointInsideMarker(X,Y,Highlight.Marker,True) and not WorkingCursor.ColumnMarking then BeginDrag(False,-1)
else if not (ssRight in Shift) or not IsPointInsideMarker(X,Y,Highlight.Marker,True) then
	begin
	CursorMarking:=False;
	ResetMarkers;
	MarkStartPos:=WorkingCursor.Position;
	end;
UndoControl.DoDeltaCursorStore(WorkingCursor.Position);

FWorkColumn:=-1;
DoChange(True,True);
end;



procedure TScrollEditor.MouseMove(Shift: TShiftState; X, Y: Integer);
{Handle mouse movement events - except for dragging}
begin
inherited MouseMove(Shift,X,Y);
ScaleCursorPos(X,Y);
if not LeftMouse or WordMarked then exit;
UndoControl.SetupDeltaUndo(WorkingCursor.Position);
ScoopStart(WorkingCursor.Position);
ScrollToCursor(True);

UndoControl.DoDeltaCursorStore(WorkingCursor.Position);
ScoopEnd(WorkingCursor.Position);
FWorkColumn:=-1;
end;



procedure TScrollEditor.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
{Handle mouse movement events - except for dragging}
begin
inherited MouseUp(Button,Shift,X,Y);
ScaleCursorPos(X,Y);
WordMarked:=False;
LeftMouse:=False;
FWorkColumn:=-1;
end;


procedure TScrollEditor.DoStartDrag(var DragObject: TDragObject);
{Handle the start of dragging by noting the Cursor Position}
begin
inherited DoStartDrag(DragObject);
DragStartPos:=WorkingCursor.Position;
end;



procedure TScrollEditor.DragOver(Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
{Scroll cursor in repsonse to dragging past the edge of the screen}
begin
inherited DragOver(Source, X, Y, State, Accept);
Accept:=Self is TScrollEditor;
if not Self.Focused then Self.SetFocus;
ScrollToCursor(False);
end;



procedure TScrollEditor.DragDrop(Source: TObject; X, Y: Integer);
{Handle drop event by copying highlighted text and dropping at destination}
{If control is not down highlighted source text is deleted}
var S: string;
var Change1: TPoint;
var DropPosition: TPoint;
var FromEditor,ToEditor: TScrollEditor;
var SameEditor: boolean;
var NewMark: TRect;



	function DoDelete: boolean;
	{Delete the highlighted block if we are moving}
	begin
	Result:=((GetAsyncKeyState(VK_Control) and $8000)=0);
	If Result then
		begin
		FromEditor.UndoControl.StoreChanges(ucBackDelete,S,FromEditor.WorkingCursor.Position,usStep);
		FromEditor.DoDeleteHighlightedText;
		end;
	end;


	procedure DoDrop;
	{Drop the highlighted text}
	begin
	ToEditor.UndoControl.StoreChanges(ucMove,'',ToEditor.WorkingCursor.Position,usStep);
	ToEditor.WorkingCursor.Position:=DropPosition;
	ToEditor.UndoControl.StoreChanges(ucBackInsert,S,ToEditor.WorkingCursor.Position,usStep);
	NewMark.TopLeft:=ToEditor.WorkingCursor.Position;
	ToEditor.DoInsertString(S);
	NewMark.BottomRight:=ToEditor.WorkingCursor.Position;
	end;


	procedure GetPositionChange(var Change: TPoint);
	{Workout change in cursor and highlight position after deleting text}
	begin
	Change1.Y:=ToEditor.Highlight.Marker.Bottom - ToEditor.Highlight.Marker.Top;
	if Highlight.Marker.Bottom=DropPosition.Y then Change1.X:=Highlight.Marker.Right - Highlight.Marker.Left
	else Change1.X:=0;
	end;



begin
inherited DragDrop(Source, X, Y);
ScaleCursorPos(X,Y);
if not (Source is TScrollEditor) or ReadOnly then exit;
FromEditor:=TScrollEditor(Source);
ToEditor:=TScrollEditor(Self);
SameEditor:=FromEditor = ToEditor;

if not IsPointInsideMarker(X,Y,Highlight.Marker,False) then
	begin
	DropPosition:=ScreenToBufferPos(X,Y);
	GetPositionChange(Change1);
	S:=FromEditor.GetHighlightedText;
	ToEditor.UndoControl.StoreChanges(ucMove,'',DragStartPos,usStart);
	FromEditor.CursorToHighlightEnd;
	{Determine Drop is before or after highlight}
	{Inserting and deleting}
	if RasComparePoints(Highlight.Marker.TopLeft,DropPosition)>0 then
		begin
		{Drop position is before highlight}
		DoDelete;
		DoDrop;
		end
	else
		begin
		{Drop position is after highlight}
		DoDrop;
		ToEditor.UndoControl.StoreChanges(ucMove,'',ToEditor.WorkingCursor.Position,usStep);
		ToEditor.WorkingCursor.Position:=ToEditor.Highlight.Marker.BottomRight;
		if DoDelete then
			begin
			DropPosition:=IntVectorSubtract(DropPosition,Change1);
			NewMark.TopLeft:=IntVectorSubtract(NewMark.TopLeft,Change1);
			NewMark.BottomRight:=IntVectorSubtract(NewMark.BottomRight,Change1);
			end;
		end;
	UndoControl.StoreChanges(ucMove,'',WorkingCursor.Position,usEnd);
	WorkingCursor.Position:=DropPosition;
	ToEditor.Highlight.Marker:=NewMark;
	end;
ToEditor.DoChange(True,True);
if not SameEditor then FromEditor.DoChange(False,True);
end;


{---------------- Buffer Display Routines -----------------------------------}




procedure TScrollEditor.SetTextColors(Canvas:TCanvas; FontColor,BrushColor: TColor);
{Set font and background (brush) color}
begin
Canvas.Font.Color:=$FFFFFF and FontColor;
Canvas.Brush.Color:=BrushColor;
end;



procedure TScrollEditor.DisplayStringSegment(Canvas: TCanvas; var S: string; var PT: TPoint);
{Display subseqment string keeping track of last pixel position}
begin
if S='' then exit;
Canvas.TextOut(PT.X,PT.Y,S);
Canvas.Refresh;
GetCurrentPositionEx(Canvas.Handle,@PT);
PT.X:=PT.X-FTextMetric.tmOverhang;
S:='';
end;





procedure TScrollEditor.WriteLine(X,Y, ColumnOffset, ClientWidth, MarkCol1,MarkCol2: integer; S: string; MarkedLine,CursorLine,Clear: boolean);
{Write string specified in S to window @ X,Y}
{CollumnOffset = number of columns scrolled of screen to left}
{ClientWidth = Pixel width of display window}
{Clear = True = clear to edge of window}
var I,CurrentColumn: integer;
var PixelPosition: TPoint;
var C: Char;
var OutStr: string;
var CharMode,OldCharMode: integer;
const cmNormal=0; cmControl=1; cmInvert=8; cmLineMark=16;
const cmNotNormal=cmNormal or cmInvert;
const cmNotControl=cmControl or cmInvert;
begin
ScaleScreen(FScale);
GetTextMetrics(Canvas.Handle,FTextMetric);
OutStr:='';
OldCharMode:=$7FFF;
PixelPosition.X:=X; PixelPosition.Y:=Y;
{Trim off characters that are off the screen to the left}
if ColumnOffset<>0 then
	begin
	Delete(S,1,ColumnOffset);
	PixelPosition.X:=FGutterSize;
	end
else PixelPosition.X:=PixelPosition.X + FGutterSize;
{Trim off characters off the screen to the right}
if Length(S)>ScreenCharWidth then SetLength(S,ScreenCharWidth+20);
Canvas.MoveTo(X,Y);
{Loop through all characters in string}
for I:=1 to Length(S) do
	begin
	C:=S[I];
	{Deal with control characters}
        if MarkedLine then CharMode:=cmLineMark
	else if C<#$20 then begin C:=char(byte(C)+$40); CharMode:=cmControl end
	else CharMode:=cmNormal;
	CurrentColumn:=(I-1)+ColumnOffset;
	{Deal with marked columns}
	if (CurrentColumn>=MarkCol1) and (CurrentColumn<MarkCol2) then CharMode:=CharMode or cmInvert;
	{if Mode has changed, dump all accumulated characters with the same attribute}
	if CharMode <> OldCharMode then
		begin
		{Output accumulated characters}
		DisplayStringSegment(Canvas,OutStr,PixelPosition);
	 	case CharMode of
		 cmLineMark: SetTextColors(Canvas,NoDelFontColor,FNonDeleteColor);
	  	 cmNormal: if CursorLine then SetTextColors(Canvas,Font.Color,FShadowColor) else SetTextColors(Canvas,Font.Color,Color);
	  	 cmControl: SetTextColors(Canvas,clWhite,FControlColor);
	  	 cmNotNormal: if Highlight.NonDelete then SetTextColors(Canvas,not FNonDeleteColor,FNonDeleteColor)
	  	 	      else SetTextColors(Canvas,HighFontColor,FHighlightColor);
	  	 cmNotControl: SetTextColors(Canvas,clWhite xor clWhite,FControlColor xor clWhite);
		 end;
		end;
	{Accumulate characters of the same type}
	OutStr:=OutStr+C;
	OldCharMode:=CharMode;
	end;
{Display accumulated chars}
DisplayStringSegment(Canvas,OutStr,PixelPosition);
{Clear rest of line to screen color}
Canvas.Brush.Color:=Color;
if Clear then Canvas.FillRect(Rect(PixelPosition.X, PixelPosition.Y,PixelPosition.X+ClientWidth,PixelPosition.Y+FTextMetric.tmHeight));
end;




function TScrollEditor.ExpandTabs(S: string): string;
{Insert tabs into specified string}
var I,Len: integer;
begin
Result:='';
Len:=Length(S);
for I:=1 to Len do
 if S[I]=Tab_Char then
	begin
	repeat Result:=Result+' '
	until (Length(Result) mod FSpacesPerTab)=0;
	end
else Result:=Result + S[I];
end;



procedure MarkSpaces(var S: string; Marker: Char; ShowReturnSymbol: boolean);
{Substitute specified 'Marker' character for spaces}
var I: integer;
begin
for I:=1 to Length(S) do if S[I]=SP_Char then S[I]:=Marker;
if ShowReturnSymbol then S:=S+CtrlM;
end;



procedure ReplaceTabs(var S: string; Marker: Char);
{Substitute specified character for Tbs}
var I: integer;
begin
for I:=1 to Length(S) do if S[I]=Tab_Char then S[I]:=Marker;
end;



procedure TScrollEditor.GetColumnMarkers(Line,Width: integer; Marker: TRect; var M1,M2: integer);
{Set column markers based on the relation between Line the block markers }
begin
{Convert marker coordinates character to column based}
Marker:=ColumnizeMarker(Marker);

M1:=0; M2:=0;
if (Line<Marker.Top) or (Line>Marker.Bottom) then exit;
M2:=CursorToColumn(Point(Width,Line));
if WorkingCursor.ColumnMarking or (Line=Marker.Top) then M1:=Marker.Left;
if WorkingCursor.ColumnMarking or (Line=Marker.Bottom) then M2:=Marker.Right;
end;



procedure GetMarkerCharacters(Canvas: TCanvas; var SpaceMarker: char);
{Work out the proper character to mark spaces/tabs with}
begin
case Canvas.Font.Charset of
 ANSI_CHARSET,DEFAULT_CHARSET: SpaceMarker:=#$B7;
 OEM_CHARSET: SpaceMarker:=#$F9;
 else SpaceMarker:='_';
 end;
end;




procedure TScrollEditor.DrawRowMark(Canvas: TCanvas; X,Y,Width,Height: integer; BackColor,HiColor: TColor);
begin
Canvas.Brush.Style:=bsSolid;
Canvas.Brush.Color:=HiColor;
Canvas.Pen.Color:=HiColor;
if GetRGBIntensity(BackColor)>0.5 then Canvas.Pen.Mode:=pmMergePenNot
else Canvas.Pen.Mode:=pmMaskPenNot;
Canvas.Rectangle(Rect(X,Y,X+Width,Y+Height));
end;


function ContrastColorX(C: TColor): TColor;
{Return most constrating color to C}
var R,G,B: byte;
var A: double;
const Threshold = (255 * 3) div 2;
begin
ColorToRGBBytes(C,R,G,B);
A:=R+G+B;
if A>Threshold then Result:=clBlack else Result:=clWhite;
end;


procedure TScrollEditor.DisplayBuffer;
{Display all characters in the Edit buffer}
var Line: integer;
var LongestLine: integer;
var XOff,YOff,M1,M2: integer;
var S: string;
var Start,Stop: integer;
var SpaceMarker: char;
var Marker: TRect;
var CursorLine: boolean;
var RowColor: TColor;
var First: integer;
begin
Canvas.Lock;
try
LongestLine:=0;
HighFontColor:=ContrastColor(FHighlightColor);
NoDelFontColor:=ContrastColor(FNonDeleteColor);

Marker:=NormalizeMarker(Highlight.Marker);
GetMarkerCharacters(Canvas,SpaceMarker);

{Hide cursor if we have focus}
if GetFocus=Handle then HideCaret(Handle);
{Setup Parameters}
GetScreenParameters;
Start:=ScrollOffset.Y;
Stop:=Start + ScreenCharHeight + 2;
if Stop>= EditBuffer.Count then Stop:=EditBuffer.Count-1;
{Loop through all specified lines}
for Line:=Start to Stop do
	begin
	S:=EditBuffer[Line];
	if Length(S)>LongestLine then LongestLine:=Length(S);
	XOff:=-ScrollOffset.X * FTextMetric.tmAveCharWidth;
	YOff:=(Line-ScrollOffset.Y) * FTextMetric.tmHeight;
	GetColumnMarkers(Line,Length(S),Marker,M1,M2);
        if FShowAll then MarkSpaces(S,SpaceMarker,Line <> EditBuffer.Count-1)
	else if FUseTabStops and (Length(S) <= TabLimitLength) then S:=ExpandTabs(EditBuffer[Line]);
	CursorLine:=(Line=WorkingCursor.Position.Y) and FShadowLine;
	WriteLine(XOff,YOff,ScrollOffset.X,ClientWidth,M1,M2,S,Line=HighlightRow,CursorLine,not DoubleBuffered);
	end;
for Line:=Start to Stop do
	begin
	First:=0;
	while RowMarks.IsMarkedLine(Line,RowColor,First) do
		begin
		YOff:=(Line-ScrollOffset.Y) * FTextMetric.tmHeight;
		DrawRowMark(Canvas,0,YOff,ClientWidth,FTextMetric.tmHeight,Color,RowColor);
		end;
	end;

UpdateScrollBars(LongestLine);
{Show cursor if we have focus}
if GetFocus=Handle then DisplayCursor;
finally Canvas.Unlock; end;
end;




procedure TScrollEditor.Paint;
{Handle Windows "Paint" event}
begin
DisplayBuffer;
if assigned(FOnPaint) then FOnPaint(Self);
end;




{========= Low level cursor movements, inserts and deletes ===================}


function TScrollEditor.IsValidCursorPosition(P: TPoint): boolean;
{Returns true if P represents a valid cursor position}
begin
Result:=(P.Y>=0) and (P.Y<EditBuffer.Count) and
        (P.X>=0) and (P.X<=Length(EditBuffer[P.Y]));
end;



function TScrollEditor.IsCursorAtStart: boolean;
{Returns true if cursor is a start of buffer}
begin
Result:=(WorkingCursor.Position.Y=0) and (WorkingCursor.Position.X=0);
end;




function TScrollEditor.IsCursorAtEnd: boolean;
{Returns true if cursor is a end of buffer}
begin
Result:=(WorkingCursor.Position.Y = EditBuffer.Count-1) and
	(WorkingCursor.Position.X=Length(EditBuffer[EditBuffer.Count-1]));
end;



function TScrollEditor.IsCursorAtEOL: boolean;
{Returns true if cursor is End of Line}
begin
Result:=WorkingCursor.Position.X>=Length(EditBuffer[WorkingCursor.Position.Y]);
end;



function TScrollEditor.DoForwardLoc(var Loc: TPoint): boolean;
{Move the X,Y cursor location forward one linear step}
begin
Result:=False;
if Loc.X>=Length(EditBuffer[Loc.Y]) then
	begin
	if Loc.Y>=EditBuffer.Count-1 then exit;
	Inc(Loc.Y);
	Loc.X:=0;
	end
else Inc(Loc.X);
Result:=True;
end;




function TScrollEditor.GetForwardLoc(Loc: TPoint): TPoint;
{Get Next cursor location}
begin
DoForwardLoc(Loc);
Result:=Loc;
end;



function TScrollEditor.DoBackLoc(var Loc: TPoint): boolean;
{Move the X,Y cursor location backward one linear step}
begin
Result:=False;
if Loc.X>0 then Dec(Loc.X)
else
	begin
	if Loc.Y<=0 then exit;
	Dec(Loc.Y);
	Loc.X:=Length(EditBuffer[Loc.Y]);
	end;
Result:=True;
end;



function TScrollEditor.GetBackLoc(Loc: TPoint): TPoint;
{Get previous cursor location}
begin
DoBackLoc(Loc);
Result:=Loc;
end;



function TScrollEditor.DoIncrementLoc(var Loc: TPoint; Delta: integer): boolean;
{Move the cursor Delta number of linear steps forward or backward}
var I: integer;
begin
Result:=True;
if Delta<0 then for I:=1 to abs(Delta) do Result:=DoBackLoc(Loc)
else for I:=1 to Delta do Result:=DoForwardLoc(Loc);
end;



function TScrollEditor.GetLineStartPos(Loc: TPoint): TPoint;
{Get start of line XED style}
begin
if (Loc.X=0) and (Loc.Y>0) then Dec(Loc.Y)
else Loc.X:=0;
Result:=Loc;
end;



function TScrollEditor.GetCountBetweenCursors(Marker: TRect): integer;
{Count the number of character between two locations specified by "Marker"}
var Y: integer;
begin
Marker:=NormalizeMarker(Marker);
if Marker.Top = Marker.Bottom then Result:=Marker.Right - Marker.Left
else
	begin
	Result:=Length(EditBuffer[Marker.Top]) - Marker.Left;
	for Y:=Marker.Top+1 to Marker.Bottom-1 do Inc(Result,Length(EditBuffer[Y])+1);
	Inc(Result,Marker.Right+1); 
	end;
end;



{-------------------- Routines to Extract characters -------------------------}


function TScrollEditor.GetCharBeforeLoc(Loc: TPoint; var C: char): boolean;
{Returns the character before specified location}
var S: string;
begin
Result:=False;
if (Loc.Y=0) and (Loc.X=0) then exit;
if Loc.X=0 then C:=CR_Char
else
	begin
	S:=EditBuffer[Loc.Y];
	if Loc.X>Length(S) then exit;
	C:=S[Loc.X];
	end;
Result:=True;
end;



function TScrollEditor.GetCharAfterLoc(Loc: TPoint; var C: char): boolean;
{Returns the character after specified location}
begin
Result:=False;
if not DoForwardLoc(Loc) then exit;
Result:=GetCharBeforeLoc(Loc,C);
end;




function TScrollEditor.GetCharsBeforeCursor(var S: string; Count: integer): boolean;
{Get "Count" number of characters before cursor}
var I: integer;
var Loc: TPoint;
var C: char;
begin
Result:=False;
S:='';
Loc:=WorkingCursor.Position;
for I:=1 to Count do
	begin
	if not GetCharBeforeLoc(Loc,C) then exit;
	S:=C+S;
	DoBackLoc(Loc);
	end;
Result:=True;
end;





function TScrollEditor.GetCharsAfterCursor(var S: string; Count: integer): boolean;
{Get "Count" number of characters after cursor}
var I: integer;
var Loc: TPoint;
var C: char;
begin
Result:=False;
S:='';
Loc:=WorkingCursor.Position;
for I:=1 to Count do
	begin
	if not GetCharAfterLoc(Loc,C) then exit;
	S:=S+C;
	DoIncrementLoc(Loc,1);
	end;
Result:=True;
end;




function TScrollEditor.GetChars(var S: string; Backward: boolean; Count: integer): boolean;
{Get specified number of characters before or after cursor}
begin
if Backward then Result:=GetCharsBeforeCursor(S,Count)
else Result:=GetCharsAfterCursor(S,Count);
end;




function TScrollEditor.GetNextChar: char;
{Get character after cursor}
begin
if not GetCharAfterLoc(WorkingCursor.Position, Result) then Result:=#0;
end;



function TScrollEditor.GetLastChar: char;
{Get character before cursor}
begin
if not GetCharBeforeLoc(WorkingCursor.Position, Result) then Result:=#0;
end;



function TScrollEditor.GetCurrentLine: string;
{Get the line at the current cursor position}
begin
Result:=EditBuffer[WorkingCursor.Position.Y];
end;



function TScrollEditor.GetBeginningOfLine: string;
{Get characters from the cursor to the beginning of the line}
{If you are at the beginning of the line, return previous line (XED style)}
begin
if WorkingCursor.Position.X = 0 then
	begin
	{Since this crosses the beginning of line, include implied CR}
	if WorkingCursor.Position.Y>0 then Result:=EditBuffer[WorkingCursor.Position.Y-1]+CR_Char
	else Result:='';
	end
else
	begin
	Result:=EditBuffer[WorkingCursor.Position.Y];
	Delete(Result,WorkingCursor.Position.X+1,Length(Result)-WorkingCursor.Position.X+1);
	end;
end;




function TScrollEditor.GetEndingOfLine(AddCR: boolean): string;
{Get characters after cursor to end of line}
begin
Result:=EditBuffer[WorkingCursor.Position.Y];
if AddCR and (WorkingCursor.Position.Y<(EditBuffer.Count-1)) then Result:=Result + CR_Char;
Delete(Result,1,WorkingCursor.Position.X);
end;




{-------------- Low Level Delete Routines ---------------------------------------}



procedure MergeTwoLines(SL: TStringList; Inx: integer);
{Merge lines specified by Inx and Inx+1, effectively deleting the CR}
var S,S1,S2: string;
begin
if Inx>=SL.Count-1 then exit;
S1:=SL.Strings[Inx];
S2:=SL.Strings[Inx+1];
S:=S1 + S2;
SL.Strings[Inx]:=S;
SL.Delete(Inx+1);
end;




function TScrollEditor.DoBackDelete: boolean;
{Low level delete back one character}
{Does no undoing or other fancy stuff}
var S: string;
begin
Result:=False;
if IsCursorAtStart then exit;
if WorkingCursor.Position.X=0 then
	begin
	DoBackChar;
	MergeTwoLines(EditBuffer,WorkingCursor.Position.Y);
	end
else
	begin
	S:=EditBuffer[WorkingCursor.Position.Y];
	Delete(S,WorkingCursor.Position.X,1);
	EditBuffer[WorkingCursor.Position.Y]:=S;
	DoBackChar;
	end;
FHighlightRow:=-1;
Result:=True;
end;



function TScrollEditor.DoForwardDelete: boolean;
{Low level delete forward one character}
{Does no undoing or other fancy stuff}
begin
Result:=False;
if IsCursorAtEnd then exit;
if IsCursorATEOL then
	begin
	MergeTwoLines(EditBuffer,WorkingCursor.Position.Y);
	end
else
	begin
	if not DoForwardChar then exit;
	DoBackDelete;
	end;
FHighlightRow:=-1;
Result:=True;
end;



function TScrollEditor.DoDeleteBackChars(Count: integer): boolean;
{Low level delete back Count number of characters}
{Does no undoing or other fancy stuff}
var S: string;
var L: integer;
begin
Result:=False;
while Count>0 do
	begin
	if IsCursorAtStart then exit;
	if WorkingCursor.Position.X = 0 then
		begin
		DoBackDelete;
		Dec(Count);
		end
	else
		begin
		L:=WorkingCursor.Position.X;
		if L>Count then L:=Count;
		S:=EditBuffer[WorkingCursor.Position.Y];
		Delete(S,WorkingCursor.Position.X-L+1,L);
		EditBuffer[WorkingCursor.Position.Y]:=S;
		WorkingCursor.Position.X:=WorkingCursor.Position.X -L;
		Dec(Count,L);
		end;
	end;
FHighlightRow:=-1;
Result:=True;
end;



function TScrollEditor.DoDeleteForwardChars(Count: integer): boolean;
{Low level delete forward Count number of characters}
{Does no undoing or other fancy stuff}
var S: string;
var L: integer;
begin
Result:=False;
while Count>0 do
	begin
	if IsCursorAtEnd then exit;
	if IsCursorAtEOL then
		begin
		DoForwardDelete;
		Dec(Count);
		end
	else
		begin
		S:=EditBuffer[WorkingCursor.Position.Y];
		L:=Length(S) - WorkingCursor.Position.X;
		if L>Count then L:=Count;
		Delete(S,WorkingCursor.Position.X+1,L);
		EditBuffer[WorkingCursor.Position.Y]:=S;
		Dec(Count,L);
		end;
	end;
FHighlightRow:=-1;
Result:=True;
end;




function TScrollEditor.DoDeleteChars(Backward: boolean; Count: integer): boolean;
{Low level delete back/forward Count number of characters}
{Does no undoing or other fancy stuff}
begin
if Backward then Result:=DoDeleteBackChars(Count)
else Result:=DoDeleteForwardChars(Count);
end;



function TScrollEditor.DoDelBackLine: boolean;
{Low level delete back one Line}
{Does no undoing or other fancy stuff}
var S: string;
begin
Result:=False;
if Self.IsCursorAtStart then exit;
if WorkingCursor.Position.X=0 then DoBackDelete;
S:=EditBuffer[WorkingCursor.Position.Y];
Delete(S,1,WorkingCursor.Position.X);
EditBuffer[WorkingCursor.Position.Y]:=S;
WorkingCursor.Position.X:=0;
FHighlightRow:=-1;
Result:=True;
end;



function TScrollEditor.DoDelForwardLine: boolean;
{Low level delete forward one line}
{Does no undoing or other fancy stuff}
var S: string;
begin
Result:=False;
if IsCursorAtEnd then exit;
if IsCursorAtEOL then Result:=DoForwardDelete
else
	begin
	S:=EditBuffer[WorkingCursor.Position.Y];
	Delete(S,WorkingCursor.Position.X+1,Length(S)-WorkingCursor.Position.X);
	EditBuffer[WorkingCursor.Position.Y]:=S;
	DoForwardDelete;
	Result:=True;
	end;
end;




procedure TScrollEditor.DoDelCurrentLine;
{Delete current line without removing CR}
begin
if WorkingCursor.Position.Y>EditBuffer.Count-1 then exit;
WorkingCursor.Position.X:=0;
EditBuffer[WorkingCursor.Position.Y]:='';
FHighlightRow:=-1;
end;



{------------------ Highlight Manipulating Routines --------------------------}



procedure TScrollEditor.CursorToHighlightEnd;
{Move cursor to end of highlighted area}
var Marker: TRect;
begin
Marker:=NormalizeMarker(Highlight.Marker);
WorkingCursor.Position:=Marker.BottomRight;
end;



procedure TScrollEditor.CusorToHighlightStart;
{Move cursor to beginning of highlighted area}
var Marker: TRect;
begin
Marker:=NormalizeMarker(Highlight.Marker);
WorkingCursor.Position:=Marker.TopLeft;
end;



procedure BlankColumns(var S: string; Column1,Column2: integer);
{Blank specified columns in the string}
var I: integer;
begin
if Column1>Length(S) then exit;
if Column2>Length(S) then Column2:=Length(S);
for I:=Column1 to Column2 do S[I]:=SP_Char;
end;



procedure TScrollEditor.DisposeOfBlock(Block: TRect; C: Char; Fill: boolean);
{ Fill A Block/Delete Of Text Along Row/column Lines }
var Y,C1,C2: integer;
var S: string;
var P: TPoint;
begin
KillScoop;
UndoControl.StoreChanges(ucNull,'',WorkingCursor.Position,usStart);
P:=WorkingCursor.Position;
for Y:=Block.Top to Block.Bottom do
	begin
	C1:=ColumnToCursorX(Point(Block.Left+1,Y));
	C2:=ColumnToCursorX(Point(Block.Right,Y));
	UndoControl.StoreChanges(ucMove,'',WorkingCursor.Position,usStep);
	WorkingCursor.Position:=Point(0,Y);
	S:=EditBuffer[Y];
	UndoControl.StoreChanges(ucForwardDelete,S,WorkingCursor.Position,usStep);
	if Fill then BlankColumns(S,C1,C2) else Delete(S,C1,C2-C1+1);
	UndoControl.StoreChanges(ucForwardInsert,S,WorkingCursor.Position,usStep);
        EDitBuffer[Y]:=S;
	end;
ResetMarkers;
UndoControl.StoreChanges(ucMove,'',WorkingCursor.Position,usStep);
WorkingCursor.Position:=P;
UndoControl.StoreChanges(ucNull,'',WorkingCursor.Position,usEnd);
DoChange(True,True);
end;



procedure TScrollEditor.DeleteWholeLines(First,Last: integer);
{Delete the specified whole lines - optimized for speed}
var I: Integer;
begin
for I:=Last downto First do EditBuffer.Delete(I);
end;



procedure TScrollEditor.DeleteMarkedText(Marker: TRect);
{Delete all marked text}
var S: string;
var Delta: integer;
begin
Marker:=NormalizeMarker(Marker);
if Marker.Top <> Marker.Bottom then
	begin
	{Delete bottom line of block first}
	S:=EditBuffer[Marker.Bottom];
	Delete(S,1,Marker.Right);
	EditBuffer[Marker.Bottom]:=S;
	{Delete all lines in between}
	DeleteWholeLines(Marker.Top + 1,Marker.Bottom -1);
	{Delete Top line of block }
	S:=EditBuffer[Marker.Top];
	Delete(S,Marker.Left+1,Length(S)-Marker.Left);
	EditBuffer[Marker.Top]:=S;
	MergeTwoLines(EditBuffer,Marker.Top);
	WorkingCursor.Position:=Marker.TopLeft
	end
else
	begin
	S:=EditBuffer[Marker.Top];
	Delta:=Marker.Right - Marker.Left;
	Delete(S,Marker.Left+1,Delta);
	EditBuffer[Marker.Top]:=S;
	WorkingCursor.Position:=Marker.TopLeft;
	end;
end;



procedure TScrollEditor.DoDeleteHighlightedText;
{Delete all highlighted text}
{Does no undoing or other fancy stuff}
begin
DeleteMarkedText(Highlight.Marker);
ResetMarkers;
end;



{------------------------------------------------------------------------------}


function TScrollEditor.GetWholeLines(First,Last: integer; UseCRLF: boolean): string;
{Return specified number of whole lines - optimized for speed}
var I, L, Size: Integer;
var P: PChar;
var S, LB: string;
begin
Size := 0;
if UseCRLF then LB := CRLF_Char
else LB:=CR_Char;
for I := First to Last do Inc(Size, Length(EditBuffer[I]) + Length(LB));
SetString(Result, nil, Size);
P := Pointer(Result);
for I := First to Last do
	begin
	S := EditBuffer[I];
	L := Length(S);
	if L <> 0 then
		begin
		System.Move(Pointer(S)^, P^, L);
		Inc(P, L);
		end;
	L := Length(LB);
	System.Move(Pointer(LB)^, P^, L);
	Inc(P, L);
	end;
end;



function TScrollEditor.GetMarkedText(Marker: TRect): string;
{Return highlighted text as string}
var S: string;
begin
Marker:=NormalizeMarker(Marker);
S:=EditBuffer[Marker.Top];
if Marker.Top <> Marker.Bottom then
	begin
	Result:=RightStr(S,Length(S)-Marker.Left)+CR_Char;
	Result:=Result+GetWholeLines(Marker.Top + 1,Marker.Bottom -1,False);
	S:=EditBuffer[Marker.Bottom];
	Result:=Result+LeftStr(S,Marker.Right);
	end
else Result:=Copy(S,Marker.Left+1,Marker.Right - Marker.Left);
end;



procedure TScrollEditor.GetSurroundChars(Marker: TRect; var C1,C2: char);
{Return char before and after highlighted text}
var S: string;
begin
Marker:=NormalizeMarker(Marker);
S:=EditBuffer[Marker.Top];
if Marker.Left=0 then C1:=#0 else C1:=S[Marker.Left];
S:=EditBuffer[Marker.Bottom];
if Marker.Right>=Length(S)-1 then C2:=#0 else C2:=S[Marker.Right+1];
end;



function TScrollEditor.GetHighlightedText: string;
{Return highlighted text as string}
begin
if WorkingCursor.ColumnMarking then Result:=GetTextBlock(Highlight.Marker)
else Result:=GetMarkedText(Highlight.Marker);
end;


{-------------- Low Leve Insert Routines --------------------------------------}


procedure TScrollEditor.InsertIntoLine(Str: string);
{Insert a string containing no CR's into line at cursor}
var S: string;
begin
S:=EditBuffer[WorkingCursor.Position.Y];
Insert(Str,S,WorkingCursor.Position.X+1);
EditBuffer[WorkingCursor.Position.Y]:=S;
Inc(WorkingCursor.Position.X,Length(Str));
end;



procedure TScrollEditor.SplitLineAtCursor;
{Since CRs aren't actually stored, inserting a CR is a special operation}
{That splits the line at the cursor creating two lines}
var S,S1,S2: string;
begin
S:=EditBuffer[WorkingCursor.Position.Y];
S1:=LeftStr(S,WorkingCursor.Position.X);
S2:=RightStr(S,Length(S)-WorkingCursor.Position.X);
EditBuffer[WorkingCursor.Position.Y]:=S1;
Inc(WorkingCursor.Position.Y);
WorkingCursor.Position.X:=0;
EditBuffer.Insert(WorkingCursor.Position.Y,S2);
end;




procedure TScrollEditor.InsertNewLine;
{Insert a new line in the buffer}
{In single line mode, a CR is inserted and treated like any other control character}
{In multi-line mode, a new lines is created by splitting the line in two at the cursor}
begin
if FEditorStyle = esSingleLine then InsertIntoLine(CR_Char)
else SplitLineAtCursor;
end;



procedure TScrollEditor.DoInsertChar(C: char);
{Low level insert one character}
{Does no undoing or other fancy stuff}
begin
FHighlightRow:=-1;
{Handle CR by Splitting Line}
if C=CR_Char then InsertNewLine
else InsertIntoLine(C);
end;



procedure TScrollEditor.DoInsertString(S: string; Raw: boolean);
{Low level insert String - Does no undoing or other fancy stuff}
{If Raw mode is enabled, CRs and LFs are inserted}
var S1: string;
var LastChar: Char;
var Inx: integer;
begin
if S='' then exit;
FHighlightRow:=-1;
Inx:=1;
if Raw then InsertIntoLine(S)
else repeat
	begin
	LastChar:=ExtractStringLine(S,S1,Inx);
	InsertIntoLine(S1);
	case LastChar of
	 CR_Char,LF_Char: InsertNewLine;
	 end;
	end
until Inx>Length(S);
end;


procedure TScrollEditor.DoInsertString(S: string);
{Low level insert String - Does no undoing or other fancy stuff}
begin
DoInsertString(S,False);
end;




procedure TScrollEditor.DoForwardInsertString(S: string);
{Low level forwared insert String}
{Does no undoing or other fancy stuff}
var P: TPoint;
begin
P:=WorkingCursor.Position;
DoInsertString(S);
WorkingCursor.Position:=P;
end;




{------------------------------------------------------------------------------}


function TScrollEditor.DoColumnMove(Column: integer): integer;
{ Move Cursor Forward To specified Column }
begin
Result:=0;
WorkingCursor.Position.X:=0;
while Result<Column do
	begin
	if IsCursorAtEOL then break;
	if Self.GetNextChar=Tab_Char then Result:=Result or 7;
	DoForwardChar;
	Inc(Result);
	end;
Inc(Result);
end;



function TScrollEditor.DoForwardChar: boolean;
{Low level move one character forward}
{Does no undoing or other fancy stuff}
begin
Result:=DoForwardLoc(WorkingCursor.Position);
end;



function TScrollEditor.DoBackChar: boolean;
{Low level move one character backward}
{Does no undoing or other fancy stuff}
begin
Result:=DoBackLoc(WorkingCursor.Position);
end;



function TScrollEditor.DoBackOverWord: boolean;
{move backward to first word separator}
begin
Result:=False;
if IsCursorAtStart then exit;
while not IsCursorAtStart do
	begin
	if IsWhiteSpace(GetLastChar) then break;
        DoBackChar;
	end;
Result:=True;
end;



function TScrollEditor.DoBackOverSeparators: boolean;
{move backward to first non-separator}
begin
Result:=False;
if IsCursorAtStart then exit;
while not IsCursorAtStart do
	begin
	if not IsWhiteSpace(GetLastChar) then break;
        DoBackChar;
	end;
Result:=True;
end;



function TScrollEditor.DoMoveForwardOverWord: boolean;
{move foreward to first word separator}
begin
Result:=False;
if IsCursorAtEnd then exit;
while not IsCursorAtEnd do
	begin
	if IsWhiteSpace(GetNextChar) then break;
	DoForwardChar;
	end;
Result:=True;
end;



function TScrollEditor.DoMoveForwardOverSeparators: boolean;
{move forward to first non-separator}
begin
Result:=False;
if IsCursorAtEnd then exit;
while DoForwardChar do
	begin
	if not IsWhiteSpace(GetNextChar) then break;
	end;
Result:=True;
end;



function TScrollEditor.DoForwardWord(ToWordEnd: boolean): boolean;
{do low level forward word - WordEnd controls}
{if cursor ends on the start of word or separator}
begin
if ToWordEnd then
	begin
	Result:=DoMoveForwardOverWord and DoMoveForwardOverSeparators;
	end
else
	begin
	Result:=DoMoveForwardOverSeparators and DoMoveForwardOverWord;
	end;
end;



function TScrollEditor.DoBackWord(ToWordStart: boolean): boolean;
{do low level back word - WordStart controls}
{if cursor ends on the start of word or separator}
begin
if ToWordStart then
	begin
	Result:=DoBackOverSeparators and DoBackOverWord;
	end
else
	begin
	Result:=DoBackOverWord and DoBackOverSeparators;
	end;
end;


function TScrollEditor.DoDownLine: boolean;
{Low level move down one line }
{Does no undoing or other fancy stuff}
begin
Result:=False;
if WorkingCursor.Position.Y>=EditBuffer.Count-1 then exit;
if FWorkColumn<0 then FWorkColumn:=CursorToColumn(WorkingCursor.Position);
Inc(WorkingCursor.Position.Y);
DoColumnMove(FWorkColumn);
Result:=True;
end;




function TScrollEditor.DoUpLine: boolean;
{Low level move up one line )}
{Does no undoing or other fancy stuff}
begin
Result:=False;
if WorkingCursor.Position.Y<=0 then exit;
if FWorkColumn<0 then FWorkColumn:=CursorToColumn(WorkingCursor.Position);
Dec(WorkingCursor.Position.Y);
DoColumnMove(FWorkColumn);
Result:=True;
end;



function TScrollEditor.DoBackLine: boolean;
{Low level move back one line (XED style)}
{Does no undoing or other fancy stuff}
begin
if (WorkingCursor.Position.X>0) or (WorkingCursor.Position.Y>0) then
	begin
	while DoBackChar do if WorkingCursor.Position.X=0 then break;
	Result:=True;
	end
else Result:=False;
end;





function TScrollEditor.DoForwardLine: boolean;
{Low level move forward one line (XED style)}
{Does no undoing or other fancy stuff}
begin
if (WorkingCursor.Position.Y<EditBuffer.Count-1) or not IsCursorAtEOL then
	begin
	while DoForwardChar do if WorkingCursor.Position.X=0 then break;
	Result:=True;
	end
else Result:=False;
end;




function TScrollEditor.DoPageUp: boolean;
{Low level move cursor up page}
{Does no undoing or other fancy stuff}
var I: Integer;
begin
Result:=False;
GetScreenParameters;
for I:=1 to ScreenCharHeight do
 if not DoBackLine then exit;
Result:=True;
end;



function TScrollEditor.DoPageDown: boolean;
{Low level move cursor down page}
{Does no undoing or other fancy stuff}
var I: Integer;
begin
Result:=False;
GetScreenParameters;
for I:=1 to ScreenCharHeight do
 if not DoForwardLine then exit;
Result:=True;
end;



function TScrollEditor.DoBackAll: boolean;
{Low level move cursor to beginning of the buffer}
{Does no undoing or other fancy stuff}
begin
WorkingCursor.Position.X:=0; WorkingCursor.Position.Y:=0;
Result:=True;
end;



function TScrollEditor.DoForwardAll: boolean;
{Low level move cursor to end of the buffer}
{Does no undoing or other fancy stuff}
begin
WorkingCursor.Position.Y:=EditBuffer.Count-1;
WorkingCursor.Position.X:=Length(EditBuffer[WorkingCursor.Position.Y]);
Result:=True;
end;


{--------------- Scrolling Routines ----------------------------------------}

{Note: scrolling does actually change the text only which parts are displayed}


procedure TScrollEditor.DoScrollUpLine;
{Low level scroll window up one line without moving cursor}
begin
if ScrollOffset.Y>0 then Dec(ScrollOffset.Y);
end;



procedure TScrollEditor.DoScrollDownLine;
{Low level scroll window down one line without moving cursor}
begin
if ScrollOffset.Y<(EditBuffer.Count-1) then Inc(ScrollOffset.Y);
end;


procedure TScrollEditor.DoScrollUpPage;
{Low level scroll window up one page without moving cursor}
var I: integer;
begin
GetScreenParameters;
for I:=1 to ScreenCharHeight do DoScrollUpLine;
end;


procedure TScrollEditor.DoScrollDownPage;
{Low level scroll window down one page without moving cursor}
var I: integer;
begin
GetScreenParameters;
for I:=1 to ScreenCharHeight do DoScrollDownLine;
end;


{========= Intermediate Level Commands with undo but not complete =============}


function TScrollEditor.IcDeleteHighlightedText(CompleteUndo: boolean): boolean;
{Intermediate-Level Delete highlighted text with undo, display etc.}
begin
Result:=False;
if not IsHighlighted then exit;
if Highlight.NonDelete then
	begin
	ResetMarkers;
	exit;
	end;
if WorkingCursor.ColumnMarking then
	begin
	DisposeOfBlock(Highlight.Marker,SP_Char,Overwrite);
	Result:=True;
	exit;
	end;
ScoopStart(Highlight.Marker.BottomRight);
ScoopEnd(Highlight.Marker.TopLeft);
if CompleteUndo then UndoControl.StoreChanges(ucNull,'',WorkingCursor.Position,usStart);
UndoControl.StoreChanges(ucMove,'',WorkingCursor.Position,usStep);
CursorToHighlightEnd;
UndoControl.StoreChanges(ucBackDelete,GetHighlightedText,WorkingCursor.Position,usStep);
DoDeleteHighlightedText;
if CompleteUndo then UndoControl.StoreChanges(ucNull,'',WorkingCursor.Position,usEnd);
FWorkColumn:=-1;
if CompleteUndo then DoChange(True,True);
Result:=True;
end;



procedure TScrollEditor.IcInsertOverwriteString(S: string; OverWriteCR: boolean);
{Mid-level string insert and overwrite with undo, display and optimising}
var S1: string;
begin
if not IcDeleteHighlightedText(False) and OverWrite then
	begin
	{Assume we deleting to EOL by using string from cursor to EOL}
	S1:=GetEndingOfLine(False);
	{if Overwriting CRs or the Insert is shorter than rest of line - use the insert string}
	if OverWriteCR or (Length(S1)>Length(S)) then GetChars(S1,False,Length(S));
	UndoControl.StoreChanges(ucForwardDelete,S1,WorkingCursor.Position,usStep);
	DoDeleteChars(False,Length(S1));
	end;
UndoControl.StoreChanges(ucBackInsert,S,WorkingCursor.Position,usStep);
DoInsertString(S);
FWorkColumn:=-1;
end;



function TScrollEditor.IcOverwriteChar(C: char): boolean;
{Intermedia-Level char insert and overwrite with undo, display and optimising}
begin
UndoControl.StoreChanges(ucForwardDelete,GetNextChar,WorkingCursor.Position,usStep);
Result:=DoForwardDelete;
if Result then
	begin
	UndoControl.StoreChanges(ucBackInsert,C,WorkingCursor.Position,usStep);
	DoInsertChar(C);
	end;
end;



{========= High level cursor movements, inserts and deletes ===================}



function TScrollEditor.MoveToLinear(C: integer): boolean;
{Move to linear position in buffer}
var P: TPoint;
begin
P:=LinearToCursorPos(C);
Result:=SetCursorPosition(P);
end;



procedure TScrollEditor.BackDelete;
{High-level back delete with undo, display and optimising}
var LastChar: char;
var Cursor: TPoint;
var Line: integer;
begin
if IcDeleteHighlightedText(True) then exit;
ScoopStart(WorkingCursor.Position);
ScoopEnd(GetBackLoc(WorkingCursor.Position));
Cursor:=WorkingCursor.Position;
Line:=GetRelativeLine;
LastChar:=GetLastChar;
if DoBackDelete then
	begin
	UndoControl.StoreChanges(ucBackDelete,LastChar,Cursor,usStart);
	if LastChar = CR_Char then SetRelativeLine(Line);
	if OverWrite and (LastChar<>CR_Char) then
		begin
		{Delete overwrites with spaces}
		UndoControl.StoreChanges(ucBackInsert,' ',WorkingCursor.Position,usStep);
		DoInsertChar(' ');
		UndoControl.StoreChanges(ucMove,'',WorkingCursor.Position,usStep);
		DoBackChar;
		end;
	UndoControl.StoreChanges(ucNull,'',WorkingCursor.Position,usEnd);
	FWorkColumn:=-1;
	DoChange(True,True);
	Error:=xeNone;
	end
else Error:=xeStart;
end;



procedure TScrollEditor.ForwardDelete;
{High-level forward delete with undo, display and optimising}
var NextChar: char;
var Cursor: TPoint;
begin
if IcDeleteHighlightedText(True) then exit;
Cursor:=WorkingCursor.Position;
NextChar:=GetNextChar;
if DoForwardDelete then
	begin
	UndoControl.StoreChanges(ucForwardDelete,NextChar,Cursor,usComplete);
	FWorkColumn:=-1;
	DoChange(True,True);
	Error:=xeNone;
	end
else Error:=xeEnd;
end;




procedure TScrollEditor.DeleteHighlighted;
{High-level delete highlighted with undo, display and optimising}
begin
Self.IcDeleteHighlightedText(True);
end;



procedure TScrollEditor.InsertChar(C: char);
{High-level character insert with undo, display and optimising}
begin
UndoControl.StoreChanges(ucNull,'',WorkingCursor.Position,usStart);
IcDeleteHighlightedText(False);
if OverWrite and (C<>CR_Char) and
   (GetNextChar<>CR_CHAR) and not IsCursorAtEnd then IcOverwriteChar(C)
else
	begin
	UndoControl.StoreChanges(ucBackInsert,C,WorkingCursor.Position,usStep);
	DoInsertChar(C);
	end;
UndoControl.StoreChanges(ucNull,'',WorkingCursor.Position,usEnd);
FWorkColumn:=-1;
DoChange(True,True);
Error:=xeNone;
end;



procedure TScrollEditor.InsertString(S: string; DoEvent,Raw: boolean);
{High-level string insert with undo, display and optimising}
begin
KillScoop;
UndoControl.StoreChanges(ucNull,'',WorkingCursor.Position,usStart);
IcDeleteHighlightedText(False);
UndoControl.StoreChanges(ucBackInsert,S,WorkingCursor.Position,usEnd);
DoInsertString(S,Raw);
FWorkColumn:=-1;
DoChange(True,DoEvent);
Error:=xeNone;
end;


procedure TScrollEditor.InsertString(S: string);
{High-level string insert with undo, display and optimising}
begin
InsertString(S,True,False);
end;


procedure TScrollEditor.NoEventInsert(S: string);
{High-level string insert with undo, display and optimising}
begin
InsertString(S,False,False);
end;




procedure TScrollEditor.BlockInsert(StringList: TStringList);
{High-level Insert Block of text so left side remain aligned}
var X,Y,WC: integer;
var S: string;
begin
KillScoop;
UndoControl.StoreChanges(ucNull,'',WorkingCursor.Position,usStart);
WC:=GetCurrentColumn;
for Y:=0 to StringList.Count-1 do
	begin
	{Insert spaces to achieve column alignment}
	S:=FillString(SP_Char,WC-GetCurrentColumn);
	UndoControl.StoreChanges(ucBackInsert,S,WorkingCursor.Position,usStep);
	DoInsertString(S);
 	X:=WorkingCursor.Position.X;
	{Insert block string}
	IcInsertOverwriteString(StringList[Y],False);
	if WorkingCursor.Position.Y=EditBuffer.Count-1 then
		begin
		UndoControl.StoreChanges(ucMove,'',WorkingCursor.Position,usStep);
		DoForwardAll;
		end;
	if IsCursorAtEnd then
		begin
		UndoControl.StoreChanges(ucBackInsert,CR_Char,WorkingCursor.Position,usStep);
		DoInsertChar(CR_Char);
		end
	else
		begin
		UndoControl.StoreChanges(ucMove,'',WorkingCursor.Position,usStep);
		WorkingCursor.Position.X:=X;
		end;
	UndoControl.StoreChanges(ucMove,'',WorkingCursor.Position,usStep);
	DoDownLine;
	end;
UndoControl.StoreChanges(ucNull,'',WorkingCursor.Position,usEnd);
DoChange(True,True);
end;





procedure TScrollEditor.BlockInsert(S: string);
{ Insert Block of text so left side remain aligned}
{Composite - built with atomic commands}
var SL: TStringList;
begin
SL:=TStringList.Create;
try
SL.Text:=S;
BlockInsert(SL);
finally SL.Free; end;
end;




function TScrollEditor.GetTextBlock(Block: TRect): string;
{ Copy A Block Of Text Along Row/column Lines }
var Y,I,L: integer;
begin
Result:='';
{Make sure the block of is based on columns not cursor position}
{otherwise blocks with combinations of tabs and spaces won't work}
Block.Left:=CursorToColumn(Block.TopLeft);
Block.Right:=CursorToColumn(Block.BottomRight);

for Y:=Block.Top to Block.Bottom do
	begin
	I:=ColumnToCursorX(Point(Block.Left+1,Y));
	L:=ColumnToCursorX(Point(Block.Right,Y))-I+1;
	Result:=Result + Copy(EditBuffer[Y],I,L) + CR_Char;
	end;
end;



{------------------ Highlevel Cursor Moves -------------------------------------}


procedure TScrollEditor.InitializeMoves;
{Setup various things before move}
begin
ScoopStart(WorkingCursor.Position);
UndoControl.SetupDeltaUndo(WorkingCursor.Position);
MarkStartPos:=WorkingCursor.Position;
end;



procedure TScrollEditor.FinalizeMoves(Success,ZeroColumn: boolean; ErrorCode: TXEDError);
{Handle post move itimes}
begin
if Success then
	begin
	CursorMoveMarking;
	UndoControl.DoDeltaCursorStore(WorkingCursor.Position);
	if ZeroColumn then FWorkColumn:=-1;
	Error:=xeNone;
	end
else Error:=ErrorCode;
DoChange(True,True);
ScoopEnd(WorkingCursor.Position);
end;





procedure TScrollEditor.ForwardChar;
{High-level move forward one char with undo, marking, display and optimising}
begin
InitializeMoves;
FinalizeMoves(DoForwardChar,True,xeEnd);
end;



procedure TScrollEditor.BackChar;
{High-level move back one char with undo, marking, display and optimising}
begin
InitializeMoves;
FinalizeMoves(DoBackChar,True,xeStart);
end;




procedure TScrollEditor.BackWord(ToWordStart: boolean);
{Move back a word according to specified style}
begin
InitializeMoves;
FinalizeMoves(DoBackWord(ToWordStart),True,xeStart);
end;



procedure TScrollEditor.ForwardWord(ToWordStart: boolean);
{Move forward a word according to specified style}
begin
InitializeMoves;
FinalizeMoves(DoForwardWord(ToWordStart),True,xeEnd);
end;



procedure TScrollEditor.DownLine;
{High-level move down one line with undo, marking, display and optimising}
begin
InitializeMoves;
FinalizeMoves(DoDownLine,False,xeEnd);
end;



procedure TScrollEditor.UpLine;
{High-level move up one line with undo, marking, display and optimising}
begin
InitializeMoves;
FinalizeMoves(DoUpLine,False,xeStart);
end;



procedure TScrollEditor.BackLine;
{XED-style High-level move back one line with undo, marking, display and optimising}
begin
InitializeMoves;
FinalizeMoves(DoBackLine,True,xeStart);
end;




procedure TScrollEditor.ForwardLine;
{XED-style High-level move forward one line with undo, marking, display and optimising}
begin
InitializeMoves;
FinalizeMoves(DoForwardLine,True,xeEnd);
end;



procedure TScrollEditor.PageDown;
{High-level move page down with undo, marking, display and optimising}
begin
InitializeMoves;
FinalizeMoves(DoPageDown,True,xeEnd);
end;



procedure TScrollEditor.PageUp;
{High-level move page up with undo, marking, display and optimising}
begin
InitializeMoves;
FinalizeMoves(DoPageUp,True,xeStart);
end;




procedure TScrollEditor.BackAll;
{High-level move to beginning with undo, marking, display and optimising}
begin
InitializeMoves;
FinalizeMoves(DoBackAll,True,xeStart);
end;



procedure TScrollEditor.ForwardAll;
{High-level move to beginning with undo, marking, display and optimising}
begin
InitializeMoves;
FinalizeMoves(DoForwardAll,True,xeEnd);
end;



procedure TScrollEditor.MoveToHighlightEnd;
{High-level move to beginning with undo, marking, display and optimising}
begin
InitializeMoves;
CursorToHighlightEnd;
FinalizeMoves(True,True,xeEnd);
end;



{---------------- Highlevel Deletes -------------------------------------------}


procedure TScrollEditor.DelBackLine;
{High-level Delete back one line (XED style) with undo, marking, display and optimising}
var L: integer;
begin
Error:=xeNone;
if IcDeleteHighlightedText(True) then exit;
if IsCursorAtStart then Error:=xeStart
else
	begin
	ScoopStart(WorkingCursor.Position);
	ScoopEnd(Self.GetLineStartPos(WorkingCursor.Position));
	UndoControl.StoreChanges(ucBackDelete,GetBeginningOfLine,WorkingCursor.Position,usComplete);
	L:=GetRelativeLine;
	DoDelBackLine;
	SetRelativeLine(L);
	DoChange(True,True);
	FWorkColumn:=-1;
	end;
DoChange(True,True);
end;



procedure TScrollEditor.DelForwardLine;
{High-level Delete forward one line (XED style) with undo, marking, display and optimising}
begin
Error:=xeNone;
if IcDeleteHighlightedText(True) then exit;
if IsCursorAtEnd then Error:=xeEnd
else
	begin
	UndoControl.StoreChanges(ucForwardDelete,GetEndingOfLine(True),WorkingCursor.Position,usComplete);
	DoDelForwardLine;
	DoChange(True,True);
	FWorkColumn:=-1;
	end;
DoChange(True,True);
end;





procedure TScrollEditor.DeleteForwardWord(ToWordEnd: boolean);
{Delete forward one word. Style scontrol by ToWordEnd}
var M: TRect;
var S: string;
begin
Error:=xeNone;
if IsCursorAtEnd then Error:=xeEnd
else
	begin
	ScoopStart(WorkingCursor.Position);
	UndoControl.StoreChanges(ucMove,'',WorkingCursor.Position,usStart);

	M.TopLeft:=WorkingCursor.Position;
	DoForwardWord(ToWordEnd);
	M.BottomRight:=WorkingCursor.Position;
	S:=GetMarkedText(M);
	ScoopEnd(WorkingCursor.Position);

	UndoControl.StoreChanges(ucBackDelete,S,WorkingCursor.Position,usEnd);
	DeleteMarkedText(M);
	FWorkColumn:=-1;
	end;
DoChange(True,True);
end;



procedure TScrollEditor.DeleteBackWord(ToWordStart: boolean);
{Delete back one word. Style scontrol by ToWordStart}
var M: TRect;
var S: string;
begin
Error:=xeNone;
if IsCursorAtStart then Error:=xeStart
else
	begin
	ScoopStart(WorkingCursor.Position);
	UndoControl.StoreChanges(ucMove,'',WorkingCursor.Position,usStart);
	M.BottomRight:=WorkingCursor.Position;
	DoBackWord(ToWordStart);
	M.TopLeft:=WorkingCursor.Position;
	S:=GetMarkedText(M);
	ScoopEnd(WorkingCursor.Position);
	UndoControl.StoreChanges(ucForwardDelete,S,WorkingCursor.Position,usEnd);
	DeleteMarkedText(M);
	FWorkColumn:=-1;
	end;
DoChange(True,True);
end;




procedure TScrollEditor.DelCurrentLine;
{Delete line at cursor without removing CR}
begin
UndoControl.StoreChanges(ucForwardDelete,GetCurrentLine,WorkingCursor.Position,usComplete);
DoDelCurrentLine;
DoChange(True,True);
FWorkColumn:=-1;
end;




procedure TScrollEditor.ReplaceCurrentLine(S: string);
{Replace line where cursor is}
begin
UndoControl.StoreChanges(ucForwardDelete,GetCurrentLine,WorkingCursor.Position,usStart);
DoDelCurrentLine;
UndoControl.StoreChanges(ucBackInsert,S,WorkingCursor.Position,usEnd);
DoInsertString(S);
DoChange(True,True);
FWorkColumn:=-1;
end;



function TScrollEditor.CutHighlighted: string;
{Return and delete highliged text}
begin
Result:=GetHighlightedText;
if WorkingCursor.ColumnMarking then DisposeOfBlock(Highlight.Marker,SP_Char,Overwrite)
else IcDeleteHighlightedText(True);
end;



procedure TScrollEditor.SelectAll;
{Set marker so all text is selected}
begin
Highlight.Marker.Left:=0;
Highlight.Marker.Top:=0;
Highlight.Marker.Bottom:=EditBuffer.Count-1;
Highlight.Marker.Right:=Length(EditBuffer[EditBuffer.Count-1]);
DoChange(True,True);
end;



function TScrollEditor.GotoLine(L: integer): boolean;
{Goto specified line if it is in range}
begin
Result:=False;
if L>=EditBuffer.Count then Error:=xeEnd
else if L<0 then Error:=xeStart
else
	begin
	ScoopStart(WorkingCursor.Position);
	UndoControl.StoreChanges(ucMove,'',WorkingCursor.Position,usComplete);
	WorkingCursor.Position.Y:=L;
	WorkingCursor.Position.X:=0;
	DoChange(True,True);
	Result:=True;
	Error:=xeNone;
	ScoopEnd(WorkingCursor.Position);
	end;
end;


procedure TScrollEditor.SetPosition(P: TPoint);
{Set cursor position - ignoring invalid X or Y positions}
var WP: TPoint;
begin
WP:=WorkingCursor.Position;
if (P.Y<0) or (P.Y>=EditBuffer.Count) then P.Y:=WP.Y;
if (P.X<0) or (P.X>Length(EditBuffer[P.Y])) then P.X:=WP.X;
ScoopStart(WorkingCursor.Position);
UndoControl.StoreChanges(ucMove,'',WorkingCursor.Position,usComplete);
WorkingCursor.Position:=P;
ScoopEnd(WorkingCursor.Position);
DoChange(True,True);
end;




function TScrollEditor.SetCursorPosition(P: TPoint): boolean;
{Validate and set cursor position}
begin
if (P.Y>-1) and (P.Y<EditBuffer.Count) and
   (P.X>-1) and (P.X<=Length(EditBuffer[P.Y])) then
	begin
	ScoopStart(WorkingCursor.Position);
	UndoControl.StoreChanges(ucMove,'',WorkingCursor.Position,usComplete);
	WorkingCursor.Position:=P;
	Result:=True;
	ScoopEnd(WorkingCursor.Position);
	end
else Result:=False;
if Result then Error:=xeNone else Error:=xeCursor;
DoChange(Result,True);
end;



function TScrollEditor.SetScrollPosition(P: TPoint): boolean;
{Set scrolloffset -  return true if sucessful}
begin
Result:=(P.Y>=0) and (P.Y<EditBuffer.Count) and (P.X>=0);
if Result then ScrollOffset:=P;
end;




function TScrollEditor.GetRelativeLine: Integer;
{Get Position of cursor-line relative to top  of screen}
begin
Result:=WorkingCursor.Position.Y-ScrollOffset.Y;
end;



function TScrollEditor.SetRelativeLine(Line: integer): boolean;
{Scrool cursor-line to specified position relative to top of screen}
var O: integer;
begin
O:=WorkingCursor.Position.Y-Line;
Result:=Self.SetScrollPosition(Point(ScrollOffset.X,O));
end;



procedure TScrollEditor.ScrollUpLine;
{Scroll window up line without moving cursor}
begin
DoScrollUpLine;
DoChange(False,True);
end;


procedure TScrollEditor.ScrollDownLine;
{Scroll window down line without moving cursor}
begin
DoScrollDownLine;
DoChange(False,True);
end;


procedure TScrollEditor.ScrollUpPage;
{Scroll window up a page without moving cursor}
begin
DoScrollUpPage;
DoChange(False,True);
end;


procedure TScrollEditor.ScrollDownPage;
{Scroll window down a page without moving cursor}
begin
DoScrollDownPage;
DoChange(False,True);
end;


procedure TScrollEditor.SetVerticalScrollPosition(V: integer);
{Set vertical position of cursor without moving cursor}
var LineLimit: integer;
begin
LineLimit:=EditBuffer.Count-1;
if V>LineLimit then V:=LineLimit;
ScrollOffset.Y:=V;
DoChange(False,True);
end;



procedure TScrollEditor.DoClearBuffer;
{Clear all text from buffer and reset all parameters}
begin
FFilename:='';
EditBuffer.Clear;
EditBuffer.Add('');
WorkingCursor.Position:=Point(0,0);
ScrollOffset:=Point(0,0);
ResetMarkers;
UndoControl.Initialize;
RedoControl.Initialize;
end;



procedure TScrollEditor.ClearBuffer;
{Clear all text from buffer and reset all parameters}
begin
KillScoop;
DoClearBuffer;
DoChange(True,True);
end;

procedure TScrollEditor.Clear;
begin
ClearBuffer;
end;
{------------------ Search Routines -------------------------------------------}

function TScrollEditor.DoForwardSearchMove(Target: string; var Loc: TPoint; var Mark: TRect): boolean;
{Do low level forward search and move cursor if found}
{(Only used to simplify code since it just calls underlying control)}
begin
Result:=Searcher.SearchForward(Target,Loc,Mark);
end;



function TScrollEditor.DoBackwardSearchMove(Target: string; var Loc: TPoint; var Mark: TRect): boolean;
{Do low level backward search and move cursor if found}
{Moves cursor back so it does not find overlapped targets}
var L,C: integer;
begin
L:=Length(Target);
C:=0;
DoIncrementLoc(Loc,-(L-C));
Result:=Searcher.SearchBackward(Target,Loc,Mark);
end;



function TScrollEditor.DoForwardSearch(Target: string; Loc: TPoint; var Mark: TRect): boolean;
{Do low level forward search - don't move cursor}
{(Only used to simplify code since it just calls underlying control)}
begin
Result:=Searcher.SearchForward(Target,Loc,Mark);
end;



function TScrollEditor.DoBackwardSearch(Target: string; Loc: TPoint; var Mark: TRect): boolean;
{Do low level backward search - don't move cursor}
{Moves cursor back so it does not find overlapped targets}
begin
DoIncrementLoc(Loc,-Length(Target));
Result:=Searcher.SearchBackward(Target,Loc,Mark);
end;


function TScrollEditor.ForwardSearch(Target: string): boolean;
{Highlevel search moving cursor and handling undo}
var Loc: TPoint;
var Mark: TRect;
var Line: integer;
var OS: boolean;
begin
Result:=False;
Error:=xeNotFound;
Loc:=WorkingCursor.Position;
if DoForwardLoc(Loc) then
 if DoForwardSearchMove(Target,Loc,Mark) then
	begin
	Error:=xeNone;
	ScoopStart(WorkingCursor.Position);
	Self.UndoControl.StoreChanges(ucMove,'',WorkingCursor.Position,usComplete);
	Highlight.Marker:=Mark;
	Highlight.NonDelete:=True;
	Line:=GetRelativeLine;
	WorkingCursor.Position:=Mark.BottomRight;
	SetRelativeLine(Line);
	ScoopEnd(WorkingCursor.Position);
	Result:=True;
	end;
{Turn on linear scrolling to force }
{ it to optimise the search position}
OS:=FLinearHScroll;
FLinearHScroll:=True;
DoChange(True,True);
FLinearHScroll:=OS;
end;



function TScrollEditor.BackwardSearch(Target: string): boolean;
{Highlevel search moving cursor and handling undo}
var Loc: TPoint;
var Mark: TRect;
var Line: integer;
begin
Result:=False;
Error:=xeNotFound;
Loc:=WorkingCursor.Position;
if DoBackwardSearchMove(Target,Loc,Mark) then
	begin
	Error:=xeNone;
	ScoopStart(WorkingCursor.Position);
	Self.UndoControl.StoreChanges(ucMove,'',WorkingCursor.Position,usComplete);
	Highlight.Marker:=Mark;
	Highlight.NonDelete:=True;
	Line:=GetRelativeLine;
	WorkingCursor.Position:=Mark.BottomRight;
	SetRelativeLine(Line);
	DoChange(True,True);
	ScoopEnd(WorkingCursor.Position);
	Result:=True;
	end;
DoChange(True,True);
end;



{------------------- Clipboard Routines ---------------------------------------}


function RemoveLineFeeds(S: string): string;
{Romve all line feeds from a block of text}
var I: integer;
begin
for I:=1 to Length(S) do
 if S[I]<>LF_Char then Result:=Result + S[I];
end;



function RemoveLineBreaks(S: string): string;
{Romve CR and LF from a block of text}
var I: integer;
begin
Result:='';
for I:=1 to Length(S) do
 if (S[I] in [CR_Char,LF_Char]) then
	begin
	if (Result<>'') and (Result[Length(Result)] <> Sp_Char) then Result:=Result + Sp_Char;
	end
 else Result:=Result + S[I];
end;



function AddLineFeeds(S: string): string;
{Append line feeds to all carriage returns}
var I: integer;
begin
for I:=1 to Length(S) do
	begin
	Result:=Result + S[I];
	if S[I]=CR_Char then Result:=Result + LF_Char;
	end;
end;




procedure TScrollEditor.CutToClipboard;
{Cut the highlighted text to the clipboard}
begin
KillScoop;
ClipBoard.Open;
ClipBoard.AsText:=AddLineFeeds(CutHighlighted);
ClipBoard.Close;
end;



procedure TScrollEditor.CopyToClipBoard;
{Copy highlighted text to the clipboard}
begin
SetClipboardOEMText(AddLineFeeds(GetHighlightedText));
end;




procedure TScrollEditor.PasteFromClipboard;
{Paste text from clibboard to edit buffer at the cursor position}
begin
KillScoop;
if (GetAsyncKeyState(VK_MENU) and $8000)<>0 then BlockInsert(RemoveLineFeeds(ReadRawClipboardStr))
else InsertString(RemoveLineFeeds(ReadRawClipboardStr));
end;




{------------------ Undo Handling Routines -------------------------------------}

procedure TScrollEditor.UndoChange(Sender: TObject);
{Zero redo whenever a new item is inserted into do}
begin
RedoControl.Initialize;
end;



procedure TScrollEditor.DoUndoTask(UndoItem: TUndoItem; CursorAfter,Done: boolean);
{Carry out Undo/Redo tasks - CursorAfter controls when cursor is set}
{If CursorAfter is true, cursor is set after change otherwise before}
begin
if not CursorAfter then WorkingCursor.Position:=UndoItem.OldCursor;
case UndoItem.Command of
 ucBackDelete:
	begin
	DoInsertString(UndoItem.UndoBuffer);
	UndoItem.Command:=ucBackInsert;
	end;
 ucForwardDelete:
	begin
	DoForwardInsertString(UndoItem.UndoBuffer);
	UndoItem.Command:=ucForwardInsert;
	end;
 ucBackInsert:
	begin
	DoDeleteBackChars(Length(UndoItem.UndoBuffer));
	UndoItem.Command:=ucBackDelete;
	end;
 ucForwardInsert:
	begin
	Self.DoDeleteForwardChars(Length(UndoItem.UndoBuffer));
	UndoItem.Command:=ucForwardDelete;
	end;
 ucMove:
	begin
	UndoItem.UndoBuffer:='';
	end;
 end;
if CursorAfter then WorkingCursor.Position:=UndoItem.OldCursor;
if Done then DoChange(True,True);
end;




procedure TScrollEditor.UndoCallBack(UndoItem: TUndoItem; Done: boolean);
{This routine is called by Undo object everytime there is an "undo" event occurs}
{It "inverts" and stores the undo event in the "redo" buffer}
var OldPos: TPoint;
begin
OldPos:=WorkingCursor.Position;
DoUndoTask(UndoItem,True,Done);
RedoControl.InvertUndoItem(UndoItem);
if UndoItem.Command=ucMove then UndoItem.OldCursor:=OldPos;
RedoControl.StoreUndoItem(UndoItem);
end;




procedure TScrollEditor.RedoCallBack(UndoItem: TUndoItem; Done: boolean);
{This routine is called by Redo object everytime there is an "redo" event occurs}
{It "inverts" and stores the redo event back in the "undo" buffer}
var OldPos: TPoint;
begin
OldPos:=WorkingCursor.Position;
DoUndoTask(UndoItem,False,Done);
UndoControl.InvertUndoItem(UndoItem);
if UndoItem.Command=ucMove then UndoItem.OldCursor:=OldPos;
UndoControl.StoreUndoItem(UndoItem);
end;

{-----------------------------------------------------------------------------}


function TScrollEditor.Undo: boolean;
{Execute an undo operation}
begin
KillScoop;
ResetMarkers;
Result:=UndoControl.DoUndo;
if Result then
	begin
	DoChange(True,True);
	Error:=xeNone;
	end
else
	begin
	Error:=xeNoUndo;
	DoChange(False,True);
	end;
end;



function TScrollEditor.Redo: boolean;
{Execute a redo operation}
begin
KillScoop;
ResetMarkers;
Result:=RedoControl.DoUndo;
if Result then
	begin
	DoChange(True,True);
	Error:=xeNone;
	end
else
	begin
	Error:=xeNoRedo;
	DoChange(False,True);
	end;
end;


{------------------- File handling routines  --------------------------------}



procedure RemoveFinalEOF(Buffer: TEditBuffer);
var S: string;
var Line,Len: integer;
begin
Line:=Buffer.Count-1;
S:=Buffer[Line];
Len:=Length(S);
if (Len>0) and (S[Len]=CtrlZ) then
	begin
	SetLength(S,Len-1);
	Buffer[Line]:=S;
	end;
end;




procedure ASCIIToUnicode(var S: string);
var I,Size,OldLen: integer;
begin
OldLen:=Length(S);
Size:=OldLen * 2 + 2;
SetLength(S,Size);
for I:=OldLen downto 1 do PWordArray(S)[I]:=byte(S[I]);
S[1]:=#$FF; S[2]:=#$FE;
end;




procedure TScrollEditor.SaveToFile(const FileName: string);
var Stream: TStream;
var C: PChar;
var S: string;
begin
FFilename:=Filename;
Stream := TFileStream.Create(FileName, fmCreate);
try
S:=EditBuffer.GetTextStr;
if EditBuffer.CharType<>ctASCII then ASCIIToUnicode(S);
Stream.WriteBuffer(Pointer(S)^, Length(S));
if AppendEOF then
	begin
	C:=CtrlZ;
	Stream.Write(C^,1);
	end;
finally Stream.Free; end;
end;




procedure TScrollEditor.LoadFromFile(Filename: string);
{Loads specified file into buffer}
begin
DoClearBuffer;
FFilename:=Filename;
EditBuffer.LoadFromFile(Filename);
if AppendEOF then RemoveFinalEOF(EditBuffer);
DoChange(True,True);
end;



{------------------- Overridden Setup Methods  ---------------------------------}


constructor TScrollEditor.Create(AOwner: TComponent);
begin
inherited Create(AOwner);
ControlStyle :=ControlStyle-[csSetCaption]+[csOpaque];
Color:=clWhite;
Font.Pitch:=fpFixed;
Font.Name:='FixedSys';
ScrollBars:=ssBoth;
KeyMapping:=kmDefault;
EditBuffer:=TEditBuffer.Create;
EditBuffer.OnChange:=HandleBufferChange;

Searcher:=TListSearcher.Create(EditBuffer);
UndoControl:=TUndoControl.Create;
RedoControl:=TUndoControl.Create;
RedoControl.UndoProcedure:=RedoCallBack;
UndoControl.UndoProcedure:=UndoCallBack;
UndoControl.OnChange:=UndoChange;
WorkingCursor:=TCursorObject.Create;
HighLight:=THighLight.Create;
RowMarks:=TRowMarks.Create;
DoClearBuffer;
DoubleBuffered:=True;
FCursorWidth:=3;
FControlColor:=clRed;
FHighlightColor:=clNavy;
FShadowColor:=RGB(230,240,250);
FShadowLine:=False;
FShowAll:= False;
FLinearHScroll:= False;
AutoUpdate:=True;
MarkHyphenated:=False;
FGutterSize:=5;
WordMarked:=False;
LeftMouse:=False;
FWorkColumn:=-1;
FEditorStyle:=esMultiLine;
Error:=xeNone;
OverWrite:=False;
AppendEOF:=False;
EditorAttached:=False;
FMoveMarking:=False;
CursorMarking:=False;
WorkingCursor.ColumnMarking:=False;
FSpacesPerTab:=8;
FUseTabStops:=True;
FStandAlone:=False;
FReadOnly:=False;
FScale:=1;
end;



destructor TScrollEditor.Destroy;
begin
if not EditorAttached then
	begin
	EditBuffer.Free;
	RedoControl.Free;
	UndoControl.Free;
	WorkingCursor.Free;
	HighLight.Free;
	Searcher.Free;
	RowMarks.Free;
	end;
inherited Destroy;
end;


{-------------------- Handle Scroll bars -------------------------------------}



procedure TScrollEditor.UpdateScrollBars(MaxWidth: integer);
{Update scroll bars in response to cursor changes in editor}
begin
VertScrollBar.Max:=EditBuffer.Count;
VertScrollBar.Position:=ScrollOffset.Y;
VertScrollBar.LargeChange:=40;

if MaxWidth<255 then MaxWidth:=255;
HorzScrollBar.Max:=MaxWidth;
HorzScrollBar.Position:=ScrollOffset.X;
HorzScrollBar.LargeChange:=40;
end;



procedure TScrollEditor.DoVertChange(Sender: TObject);
{Overriden method from ScrollBarWindow}
begin
SetVerticalScrollPosition(VertScrollBar.Position);
end;


procedure TScrollEditor.DoHorzScroll(Sender: TObject; ScrollCode: TScrollCode; var ScrollPos: Integer);
{Overrides underlying Horizontal Scroll method to respond change in scroll bar}
const ScreenCharWidth = 40;
const GuardBand = 4;
begin
case ScrollCode of
 scLineUp:
	begin
	Dec(ScrollOffset.X);
	if ScrollOffset.X<0 then ScrollOffset.X:=0;
	end;
 scLineDown: Inc(ScrollOffset.X);
 scPageUp:
	begin
	Dec(ScrollOffset.X,ScreenCharWidth-GuardBand);
	if ScrollOffset.X<0 then ScrollOffset.X:=0;
	end;
 scPageDown:
 	begin
	Inc(ScrollOffset.X,ScreenCharWidth-GuardBand);
	if ScrollOffset.X>HorzScrollBar.Max then ScrollOffset.X:=HorzScrollBar.Max;
	end;

 scTrack: ScrollOffset.X:=ScrollPos;
 end;

DoChange(False,True);
SetFocus;
end;



procedure TScrollEditor.ScrollVert(Sender: TObject; ScrollCode: TScrollCode; var ScrollPos: Integer);
{Handle wheel-mouse events}
begin
case ScrollCode of
 scLineUp: ScrollUpLine;
 scLineDown: ScrollDownLine;
 scPageUp:  ScrollUpPage;
 scPageDown: ScrollDownPage;
 scTrack: SetVerticalScrollPosition(ScrollPos);
 end;
ScrollPos:=ScrollOffset.Y;
end;




function TScrollEditor.DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): Boolean;
{Overridden wheel mouse routine from TControl}
var P: integer;
begin
inherited DoMouseWheelDown(Shift,MousePos);
if (ssCtrl in Shift) then Scale:=Scale - 0.1
else ScrollVert(Self,scLineDown,P);
Result:=True;
end;



function TScrollEditor.DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): Boolean;
{Overridden wheel mouse routine from TControl}
var P: integer;
begin
inherited DoMouseWheelUp(Shift,MousePos);
if (ssCtrl in Shift) then Scale:=Scale + 0.1
else ScrollVert(Self,scLineUp,P);
Result:=True;
end;


{------------------- Routines to read/write object properties ----------------}


procedure TScrollEditor.SetEditBuffer(V: TEditBuffer);
begin
EditBuffer.Assign(V);
Invalidate;
end;



procedure TScrollEditor.SetShadowLine(V: boolean);
begin
if V<>FShadowLine then
	begin
	FShadowLine:=V;
	Invalidate;
	end;
end;



procedure TScrollEditor.SetShadowColor(V: TColor);
begin
if V<>FShadowColor then
	begin
	FShadowColor:=V;
	Invalidate;
	end;
end;



procedure TScrollEditor.SetControlColor(V: TColor);
begin
if V<>FControlColor then
	begin
	FControlColor:=V;
	Invalidate;
	end;
end;



procedure TScrollEditor.SetHighlightColor(V: TColor);
begin
if V<>FHighlightColor then
	begin
	FHighlightColor:=V;
	Invalidate;
	end;
end;



procedure TScrollEditor.SetNonDeleteColor(const Value: TColor);
begin
if FNonDeleteColor<>Value then
	begin
	FNonDeleteColor:=Value;
	Invalidate;
	end;
end;


procedure TScrollEditor.SetShowAll(V: boolean);
begin
if V<>FShowAll then
	begin
	FShowAll:=V;
	DoChange(True,True);
	end;
end;



procedure TScrollEditor.SetGutter(V: integer);
begin
if V<>FGutterSize then
	begin
	FGutterSize:=V;
	DoChange(True,True);
	end;
end;



procedure TScrollEditor.SetIsModified(V: boolean);
begin
if V<>FIsModified then
	begin
	FIsModified:=V;
	end;
Inc(FChangeCount);
end;



procedure TScrollEditor.SetMoveMarking(V: boolean);
begin
if V<>FMoveMarking then
	begin
	FMoveMarking:=V;
	MarkBasePos:=WorkingCursor.Position;
	end;
end;



procedure TScrollEditor.SetSpacesPerTab(V: integer);
begin
if V<>FSpacesPerTab then
	begin
	FSpacesPerTab:=V;
	DoChange(True,True);
	end;
end;



procedure TScrollEditor.SetTUseTabStops(const Value: boolean);
begin
if FUseTabStops<>Value then
	begin
	FUseTabStops := Value;
	DoChange(True,True);
	end;
end;




procedure TScrollEditor.MatchIndentAbove;
{ Indent Current Line Same As The Preceding One }
var LS,S: string;
var C: TPoint;
begin
if WorkingCursor.Position.Y<1 then exit;
{Get the previous line and extract the left indent characters}
LS:=GetLeadingWhiteSpace(EditBuffer[WorkingCursor.Position.Y-1]);

{Get Current Line, strip off old indent and then add in previous line's}
S:=GetCurrentLine;
S:=LS+TrimLeft(S);
C:=WorkingCursor.Position;
C.X:=Length(LS);
ReplaceCurrentLine(S);
SetCursorPosition(C);
end;



{------------------ Routines to Do Windows Memo like Commands -----------------}

procedure TScrollEditor.Add(S: string);
{Insert string at end of buffer}
begin
ForwardAll;
InsertString(S+CRLF_Char);
end;


{------------------ Event Handlers --------------------------------------------}

procedure TScrollEditor.HandleBufferChange(Sender: TObject);
{Handle editor buffer change event}
begin
IsModified:=True;
end;


{------------------ Handle Keystrokes In stand alone mode ---------------------}

var CmdMetaFlag: boolean;	{Unconditional insert for command line}

function TScrollEditor.GetTarget: string;
begin
if not IsHighlighted then MarkWord;
Result:=GetHighlightedText;
end;



procedure TScrollEditor.DoWordBackSearch;
var Target: string;
begin
Target:=GetTarget;
UndoControl.SetupDeltaUndo(WorkingCursor.Position);
CusorToHighlightStart;
UndoControl.DoDeltaCursorStore(WorkingCursor.Position);
Searcher.IgnoreCase:=True;
Searcher.WholeWordMode:=True;
BackwardSearch(Target);
end;



procedure TScrollEditor.DoWordFwdSearch;
var Target: string;
begin
Target:=GetTarget;
UndoControl.SetupDeltaUndo(WorkingCursor.Position);
CusorToHighlightStart;
CursorToHighlightEnd;
UndoControl.DoDeltaCursorStore(WorkingCursor.Position);
Searcher.IgnoreCase:=True;
Searcher.WholeWordMode:=True;
ForwardSearch(Target);
end;





procedure TScrollEditor.HandleControlX(Shift: boolean; XEDStyle: boolean);
begin
if IsHighlighted then CopyToClipBoard;
if FReadOnly then exit;
if XEDStyle then
	begin
	if Shift then DelForwardLine
	else DelBackLine;
	end
else DeleteHighlighted;
end;



procedure TScrollEditor.HandleControlC(XEDStyle: boolean);
begin
if IsHighlighted then CopyToClipBoard
else if XEDStyle and not FReadOnly then
	begin
	ResetMarkers;
	DeleteBackWord(True);
	end;
end;


procedure TScrollEditor.StandAloneInsertChar(C: char);
{Insert char for stand-alone mode}
begin
if not FReadOnly then InsertChar(C);
end;


procedure TScrollEditor.FlipCase;
{ Flip Case Of Letter At Cursor And Move Forward }
var S: string;
var C: TPoint;
var Letter: boolean;
begin
if FReadOnly then exit;
if IsCursorAtEnd then begin Error:=xeEnd; exit; end;
KillScoop;
if not IsHighlighted then
	begin
	Highlight.Marker.TopLeft:=WorkingCursor.Position;
	Highlight.Marker.BottomRight:=GetForwardLoc(WorkingCursor.Position);
	end;
NormalizeMarker(Highlight.Marker);
C:=Highlight.Marker.BottomRight;
S:=GetHighlightedText;
Letter:=DoToggleCase(S);

UndoControl.StoreChanges(ucNull,'',WorkingCursor.Position,usStart);
IcDeleteHighlightedText(False);
UndoControl.StoreChanges(ucBackInsert,S,WorkingCursor.Position,usStep);
DoInsertString(S);
UndoControl.StoreChanges(ucMove,'',WorkingCursor.Position,usEnd);
WorkingCursor.Position:=C;
DoChange(True);
if not Letter then Error:=xeLetter;
end;



function TScrollEditor.DoXEDControlKeys(Key: Char; Shift: boolean): boolean;
begin
Result:=True;
if CmdMetaFlag then
	begin
	StandAloneInsertChar(Key);
	CmdMetaFlag:=False;
	end
else case Key of
 CtrlA: ForwardChar;
 CtrlB: begin ResetMarkers; BackWord(True); end;
 CtrlC: HandleControlC(True);
 CtrlD: ForwardAll;
 CtrlE: BackAll;
 CtrlH: if not FReadOnly then BackDelete;
 CtrlJ: PageDown;
 CtrlK: PageUp;
 CtrlN: begin ResetMarkers; ForwardWord(True); end;
 CtrlO: if Shift then Redo else Undo;
 CtrlQ: BackChar;
 CtrlS: ForwardLine;
 CtrlT: ShowAll:=not ShowAll;
 CtrlU: if not FReadOnly then MatchIndentAbove;
 CtrlV: if not FReadOnly then PasteFromClipboard;
 CtrlW: BackLine;
 CtrlX: HandleControlX(Shift,True);
 CtrlZ: FlipCase;
 CtrlCaret: CmdMetaFlag:=True;
 else Result:=False;
 end;
end;





function TScrollEditor.DoXEDLiteControlKeys(Key: Char; Shift: boolean): boolean;
begin
Result:=True;
if CmdMetaFlag then
	begin
	StandAloneInsertChar(Key);
	CmdMetaFlag:=False;
	end
else case Key of
 CtrlC: HandleControlC(True);
 CtrlH: if not FReadOnly then BackDelete;
 CtrlO: if Shift then Redo else Undo;
 CtrlV: if not FReadOnly then PasteFromClipboard;
 CtrlX: HandleControlX(Shift,True);
 CtrlCaret: CmdMetaFlag:=True;
 else Result:=False;
 end;
end;


function TScrollEditor.DefaultControlKeys(Key: char; Shift: boolean): boolean;
begin
Result:=True;
case Key of
 CtrlH: if not FReadOnly then BackDelete;
 CtrlX: HandleControlX(Shift,False);
 CtrlC: HandleControlC(False);
 CtrlV: if not FReadOnly then PasteFromClipboard;
 CtrlY: Redo;
 CtrlZ: Undo;
 else Result:=false;
 end;
end;




procedure TScrollEditor.InsertCmdChar(Key: Char);
var Shift: boolean;
begin
Shift:=(GetAsyncKeyState(VK_SHIFT) and $8000)<>0;
case FKeyMapping of
 kmDefault: if not DefaultControlKeys(Key,Shift) then StandAloneInsertChar(Key);
 kmXed: if not DoXEDControlKeys(Key,Shift) then StandAloneInsertChar(Key);
 kmXedLite: if not DoXEDLiteControlKeys(Key,Shift) then StandAloneInsertChar(Key);
 kmNoMap: StandAloneInsertChar(Key);
 else  if not DefaultControlKeys(Key,Shift) then StandAloneInsertChar(Key);
 end;
end;



procedure TScrollEditor.KeyDown(var Key: Word; Shift: TShiftState);
var Handled: boolean;
begin
inherited;
if not FStandAlone then exit;
Handled:=True;
if (ssShift in Shift) and (ssCtrl in Shift) then
	begin
	case Key of
	 byte('2') : InsertCmdChar(#0);
	 end;
	end
else if ssCtrl in Shift then
	begin
	case Key of
	 VK_F9: DoWordBackSearch;
	 VK_F10: DoWordFwdSearch;
	 end
	end
else case Key of
 VK_LEFT: BackChar;
 VK_RIGHT: ForwardChar;
 VK_DELETE: if not FReadOnly then ForwardDelete;
 VK_INSERT: begin OverWrite:=not OverWrite; DoChange(True,True); end;
 else if EditorStyle=esMultiline then
	begin
	case Key of
	 VK_UP: UpLine;
	 VK_DOWN: DownLine;
	 VK_NEXT: PageDown;
	 VK_PRIOR: PageUp;
	 VK_HOME: BackAll;
	 VK_END: ForwardAll;
	 else Handled:=False;
	 end
	end
 else Handled:=False;
 end;
if Handled then Key:=0;
end;



procedure TScrollEditor.KeyPress(var Key: Char);
begin
inherited;
if (Key=#0) or not FStandAlone then exit;
InsertCmdChar(Key);
end;






function TScrollEditor.GetText: string;
begin
Result:=EditBuffer.Text;
end;



procedure TScrollEditor.SetText(const Value: string);
begin
EditBuffer.Text:=Value;
end;



procedure TScrollEditor.SetHighlightRow(const Value: integer);
begin
if FHighlightRow<>Value then
	begin
	ResetMarkers;
	FHighlightRow:=Value;
	Invalidate;
	end;
end;


end.
