unit intspin;

interface

uses Windows,Messages,Forms,Types,Classes,Controls,StdCtrls,ExtCtrls,Graphics,SysUtils,
  SpinButtons,Harmonizer;


{================== TIntegerSpin - Integer Spin Edit ==========================}

type TIntegerIncEvent = procedure(Sender: TObject; Inc: boolean; var Value: Int64) of object;


type
  TIntegerSpin = class(TCustomSpin)
  private
    FMinValue: Int64;
    FMaxValue: Int64;
    FIncrement: Int64;
    FOwnerIncrement: boolean;
    FOnIncrement: TIntegerIncEvent;
    FLastValid: Int64;
    FHexMode: boolean;
    FHexDigits: integer;
    procedure HandleUserIncrement(Inc: boolean);
    procedure SetValue (NewValue: Int64);
    procedure SetNoEventValue(NewValue: Int64);
    function CheckValue (NewValue: Int64): Int64;
    function GetValue: Int64;
    procedure CMExit(var Message: TCMExit);   message CM_EXIT;
    procedure UpClick (Sender: TObject); virtual;
    procedure SetHexMode(const Value: boolean);
    procedure SetHexDigits(const Value: integer);
    procedure SetMaxValue(const Value: Int64);
    procedure SetMinValue(const Value: Int64);
  protected
    function IsValidChar(Key: Char): Boolean; virtual;
    procedure DownClick (Sender: TObject); virtual;
    procedure KeyPress(var Key: Char); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
  public
    procedure ChainedIncrement(ParentSpin: TIntegerSpin; Inc: boolean; var Value: int64);
    constructor Create(AOwner: TComponent); override;
    property NoEventValue: Int64 read GetValue write SetNoEventValue;
  published
    property Increment: Int64 read FIncrement write FIncrement ;
    property MaxValue: Int64 read FMaxValue write SetMaxValue;
    property MinValue: Int64 read FMinValue write SetMinValue;
    property Value: Int64 read GetValue write SetValue;
    property OwnerIncrement: boolean read FOwnerIncrement write FOwnerIncrement default False;
    property OnIncrement: TIntegerIncEvent read FOnIncrement write FOnIncrement;
    property HexMode: boolean read FHexMode write SetHexMode default False;
    property HexDigits: integer read FHexDigits write SetHexDigits default 4;
    property AutoSelect;
    property AutoSize;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property MaxLength;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;




procedure Register;


implementation

procedure Register;
begin
RegisterComponents('Fountain', [TIntegerSpin]);
end;


{================== TIntegerSpin ================================================}


procedure TIntegerSpin.ChainedIncrement(ParentSpin: TIntegerSpin; Inc: boolean; var Value: int64);
{Recursive incrementor for multi-stage integer spin display i.e. Minutes, seconds and milliseconds}
{The routine will increment each spinner in the chain and wrap around when needed}
{Each spinner must be a TIntegerSpin, and OwnerIncrement must be enabled }
{The Max and Min value must be set to the limit values you want - they are used to wrap or limit}
{If Wrap is enabled, the value wrap and the next spinner up the chain is advanced}
{If Wrap is not enabled, the value is clamped to the upper and lower limit values}
{This routine should be called from you increment event of each spinner including the last one  }
var T: Int64;

	procedure HandleOverflow;
	begin
	if ParentSpin<>nil then
		begin
		T:=ParentSpin.Value;
		if assigned(ParentSpin.OnIncrement) then ParentSpin.OnIncrement(Self,Inc,T);
		ParentSpin.Value:=T;
		end;
	end;

begin
if Inc then
	begin
	System.Inc(Value);
	if Value>=Self.MaxValue then
		begin
		if Self.Wrap then
			begin
			Value:=Self.MinValue;
			HandleOverflow;
			end
		else Value:=Self.MaxValue;
		end;
	end
else
	begin
	Dec(Value);
	if Value<Self.MinValue then
		begin
		if Self.Wrap then
			begin
			Value:=Self.MaxValue-1;
			HandleOverflow;
			end
		else Value:=Self.MinValue;
		end;
	end;
end;




constructor TIntegerSpin.Create(AOwner: TComponent);
begin
inherited Create(AOwner);
FHexMode:=False;
FHexDigits:=4;
FButton.OnUpClick := UpClick;
FButton.OnDownClick := DownClick;
FIncrement := 1;
FLastValid:=Value;
end;



procedure TIntegerSpin.HandleUserIncrement(Inc: boolean);
var VT: Int64;
begin
VT:=Value;
if Assigned(FOnIncrement) then FOnIncrement(Self,Inc,VT);
Value:=VT;
end;


procedure TIntegerSpin.UpClick (Sender: TObject);
var M: integer;
begin
if ((GetAsyncKeyState(VK_Control) and $8000)<>0) then M:=10 else M:=1;;
if ReadOnly then MessageBeep(0)
else if FOwnerIncrement then HandleUserIncrement(True)
else Value := Value + FIncrement * M;
end;


procedure TIntegerSpin.DownClick (Sender: TObject);
var M: integer;
begin
if ((GetAsyncKeyState(VK_Control) and $8000)<>0) then M:=10 else M:=1;;
if ReadOnly then MessageBeep(0)
else if FOwnerIncrement then HandleUserIncrement(False)
else Value := Value - FIncrement * M;
end;



procedure TIntegerSpin.CMExit(var Message: TCMExit);
begin
inherited;
if Text='' then Value:=FLastValid;
SetValue(Value);
end;


function TIntegerSpin.GetValue: Int64;
var S: string;
begin
S:=Text;
if HexMode then S:='$'+S;
if TryStrToInt64(S,Result) then
	begin
	Result:=CheckValue(Result);
	FLastValid:=Result;
	end
else Result:=FLastValid;
end;



procedure TIntegerSpin.SetValue(NewValue: Int64);
begin
FLastValid:=NewValue;
if HexMode then Text:=IntToHex(CheckValue(NewValue),HexDigits)
else Text:=IntToStr(CheckValue(NewValue));
end;



procedure TIntegerSpin.SetNoEventValue(NewValue: Int64);
var OldBlock: boolean;
begin
OldBlock:=BlockChangeEvent;
try
BlockChangeEvent:=True;
SetValue(NewValue);
finally BlockChangeEvent:=OldBlock; end;
end;



function TIntegerSpin.IsValidChar(Key: Char): Boolean;
begin
if HexMode then Result := (Key in ['0'..'9','A'..'F','a'..'f']) or ((Key < #32) and (Key <> Chr(VK_RETURN)))
else Result := (Key in [IndSystemData.DecimalSeparator, '+', '-', '0'..'9']) or ((Key < #32) and (Key <> Chr(VK_RETURN)));
if not FEditorEnabled and Result and ((Key >= #32) or (Key = Char(VK_BACK)) or (Key = Char(VK_DELETE))) then Result := False;
end;



procedure TIntegerSpin.KeyPress(var Key: Char);
begin
if not IsValidChar(Key) then
	begin
	Key := #0;
	MessageBeep(0)
	end;
if Key <> #0 then inherited KeyPress(Key);
end;



function TIntegerSpin.CheckValue(NewValue: Int64): Int64;
begin
Result := NewValue;
if FUseLimit then
	begin
	if NewValue >= FMaxValue then
		begin
		if Wrap and (FMinValue<>FMaxValue) then
			begin
			while Result>=FMaxValue do Result:=Result - FMaxValue;
			Result:=Result + FMinValue;
			Wrapped(wmExceededMax);
			end
		else Result := FMaxValue;
		end;
	if NewValue < FMinValue then
		begin
		if Wrap and (FMinValue<>FMaxValue)  then
			begin
			while Result<FMinValue do Result:=FMinValue-Result;
			Result:=FMaxValue-Result;
			Wrapped(wmExceededMin);
			end
		else Result:= FMinValue;
		end;
	end;
end;




procedure TIntegerSpin.KeyDown(var Key: Word; Shift: TShiftState);
begin
if Key = VK_UP then UpClick(Self)
else if Key = VK_DOWN then DownClick(Self);
inherited KeyDown(Key, Shift);
end;


procedure TIntegerSpin.SetHexMode(const Value: boolean);
var V: Int64;
begin
if FHexMode<>Value then
	begin
	V:=Self.Value;
        FHexMode:=Value;
	SetValue(V);
	end;
end;



procedure TIntegerSpin.SetHexDigits(const Value: integer);
var V: Int64;
begin
if FHexDigits<>Value then
	begin
	V:=Self.Value;
	FHexDigits:=Value;
	SetValue(V);
	end;
end;

procedure TIntegerSpin.SetMaxValue(const Value: Int64);
begin
if FMaxValue<>Value then
	begin
	{Value overflow must be handled first or else "Value" will come }
	{back already trimmed to MaxValue and the "if" will always fail }
	if Self.Value>Value then Self.Value:=Value;
	FMaxValue := Value;
	end;
end;



procedure TIntegerSpin.SetMinValue(const Value: Int64);
begin
if FMinValue<>Value then
	begin
	{Value overflow must be handled first or else "Value" will come }
	{back already trimmed to MaxValue and the "if" will always fail }
	if Self.Value>FMinValue then Self.Value:=FMinValue;
	FMinValue := Value;
	end;
end;



end.

